
/* $Id: t2_comm.h,v 1.7 2000/10/20 21:35:04 chialin Exp $ */

/* t2_comm.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 6/12/98
 * header file for inter-processor communication between back-end nodes
 */


#ifndef _T2_COMM_H_
#define _T2_COMM_H_


#include	"t2_header.h"
#include	"t2_acomm.h"
#include	"t2_commfunc.h"
#include	"t2_tset.h"
#include	"t2_tarray.h"
#include	"t2_tqueue.h"
#include	"t2_qenv.h"
#include	"t2_accmsg.h"
#include	"t2_work.h"


// forward declaration
struct T2_CommRequest;
class T2_QEnv;
class T2_WorkList;
class T2_CommManager;

////////////////////////////////////////////////////////////////////////
//
// T2_CommObj
//
// A t2 system communication object, which contains information about
// communication subsystem -- groups, ranks, etc.
//
////////////////////////////////////////////////////////////////////////
class T2_CommObj {
public:

  static int getNumberNodes();
  static int getMyNode();

protected:
private:
};



////////////////////////////////////////////////////////////////////////
//
// T2_CommControlBlock
//
// Some methods defined for issuing sends/recvs and retrieving
// information out of T2_CommControlBlock
//
////////////////////////////////////////////////////////////////////////

// issue a non-blocking send
T2_AsyncResult
t2_issueISend(T2_CommControlBlock& cb,
	      char* buf, size_t nbytes, int dest, int tag);

// issue a non-blocking recv
T2_AsyncResult
t2_issueIRecv(T2_CommControlBlock& cb,
	      char* buf, size_t nbytes, int src, int tag);

// issue a probe
bool
t2_issueIProbe(T2_CommControlBlock& cb, int src, int tag);

// check if a non-blocking operation completes
bool
t2_commTestCompletion(T2_CommControlBlock& cb);

// get the message size (in bytes)
int t2_getMsgSize(T2_CommControlBlock& cb);

// get the sender
int t2_getSource(T2_CommControlBlock& cb);

// get the message tag
int t2_getMsgTag(T2_CommControlBlock& cb);



////////////////////////////////////////////////////////////////////////
//
// T2_CommCB
//
// A control block for communication.
//
////////////////////////////////////////////////////////////////////////
struct T2_CommCB: public T2_CommControlBlock {
  T2_CommRequest* key;
};



////////////////////////////////////////////////////////////////////////
//
// T2_Comm
//
// The constructor takes an argument that specifies the max number of
// concurrently pending requests that T2_Comm can handle.
//
////////////////////////////////////////////////////////////////////////
class T2_Comm {
public:
  T2_Comm(u_int n);
  ~T2_Comm();

  // query methods
  u_int getNumberFreeCB() const { return freel.getNumberElements(); };
  u_int getNumberPendingCB() const { return pendingl.getNumberElements(); };
  u_int getNumberDoneCB() const { return donel.getNumberElements(); };

  // issue a non-blocking send
  T2_AsyncResult isend(char* buf, size_t nbytes, int dest, int tag,
		       T2_CommRequest* key);

  // issue a non-blocking recv
  T2_AsyncResult irecv(char* buf, size_t nbytes, int src, int tag,
		       T2_CommRequest* key);

  // issue a message probe
  bool iprobe(int& src, int& tag, int& nbytes);

  // poll for completed operations
  u_int poll();

  // get the T2_CommRequest pointer for the first completed op on donel
  T2_CommRequest* getDoneRequest(int& msgsize, int& source, int& tag);

  // print status
  void printStatus(ostream& os) const;

protected:
  T2_Array<T2_CommCB> cbs;

  T2_TSet<T2_CommCB*> freel,
		      pendingl,
		      donel;

private:
};



////////////////////////////////////////////////////////////////////////
//
// T2_CommRequest
//
// A non-blocking communication request.
//
////////////////////////////////////////////////////////////////////////
enum T2_CommOp { 
	t2_cUnknown = 0, 
	t2_cISend, 
	t2_cIRecv, 
	t2_cISendKillQuery
};

struct T2_CommRequest {
  T2_CommOp  op;
  u_int      qid;		// query id
  T2_ProcID  proc;		// processor id for either src or dest
  int        tag;		// tag

  T2_AccMsgSendBuffer* msgbuf;

  T2_CommRequest();
};


inline
T2_CommRequest::T2_CommRequest()
{
  op = t2_cUnknown;
  msgbuf = NULL;
}

//inline size_t
//T2_CommRequest::allocBuffer(size_t n)
//{
//  t2_Assert (buf == NULL);
//  t2_Assert (nbytes == 0);
//  if (n == 0)
//     return 0;
//
//  t2_Alloc (buf, n, char);
//  nbytes = n;
//  return n;
//}



////////////////////////////////////////////////////////////////////////
//
// T2_AccMsgHeader
//
// Sending an acc message during the global combine phase actually
// involves sending two messages : a header message and a data message.
// The header message contains the query id of the query that sends
// the message and the size of the actual data message (ie the second
// message, referred to as the data message above).
// T2_AccMsgHeader is used as the buffer for the header message.
//
////////////////////////////////////////////////////////////////////////
class T2_AccMsgHeader {
public:
  T2_AccMsgHeader() {};
  ~T2_AccMsgHeader() {};

  // access methods
  u_int getQueryID() const { return fields[0]; };
  u_int getDataMsgSize() const { return fields[1]; };

  // set the various values
  void setQueryID(u_int qid) { fields[0] = qid; };
  void setDataMsgSize(u_int n) { fields[1] = n; };

  // get access to the fields
  const u_int* getFields() const { return fields; };

protected:
  u_int fields[2];

private:
};



////////////////////////////////////////////////////////////////////////
//
// T2_SendManager
//
// The class that manages the non-blocking sends issued for a particular
// processor. It basically consists of a list of T2_CommRequests (ie
// sends) ready to be issued.
//
// The constructor takes an argument that specifies the max number of 
// requests T2_SendManager can handle.
//
// Not sure what's the right policy for arbitrating among requests
// from different queries (ie which send request gets issued next?),
// so I'll just use a queue for now.
//
////////////////////////////////////////////////////////////////////////
class T2_SendManager {
friend class T2_CommManager;
public:
  T2_SendManager(u_int nreq=0);
  T2_SendManager(const T2_SendManager& m);
  ~T2_SendManager() {};

  // operator =
  T2_SendManager& operator =(const T2_SendManager& m);

  // query methods
  bool isFull() const { return readyl.isFull(); };
  bool isEmpty() const { return readyl.isEmpty(); };

  // get the number and the max number of isends in the ready queue
  u_int getNumberReadyISends() const { return readyl.getNumberElements(); };
  u_int getMaxNumberReadyISends() const { return readyl.getMaxQueueSize(); };

  // change the max size of the ready queue
  void setMaxNumberReadyISends(u_int n);

  // get the number of issued pending isends
  u_int getNumberPendingISends() const { return npending; };

  // decrement the number of issued pending isends
  u_int decrementNumberPendingISends(u_int i=1)
		{ t2_Assert (npending >= i); npending -= i; return npending; };

  // get the total number of registered isends, which is the sum of the
  // number of pending isends and the number of ready isends
  u_int getNumberRegisteredISends() const;

  // register a non-blocking send request ready to be issued
  void registerISend(T2_CommRequest* req, T2_Comm* comm);

  // issue a ready request queued in the ready queue
  // (must be called when readyToIssueISend() is true)
  void issueReadyISend(T2_Comm *com);

  // ready to issue another isend?
  bool readyToIssueISend();

protected:
  T2_TQueue<T2_CommRequest*> readyl;
  u_int npending;			// number of issued pending isends

  // issue an isend according to the given request
  void issueISend(T2_CommRequest* req, T2_Comm* comm);

private:
};


inline
T2_SendManager::T2_SendManager(u_int nreq):
			readyl(nreq), npending(0)
{
}

inline
T2_SendManager::T2_SendManager(const T2_SendManager& m)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_SendManager copy constructor");
}

inline T2_SendManager&
T2_SendManager::operator =(const T2_SendManager& m)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_SendManager operator =");
  return *this;
}

inline void
T2_SendManager::setMaxNumberReadyISends(u_int n)
{
  t2_Assert (isEmpty() == true);
  readyl.setMaxQueueSize(n);
}

inline u_int
T2_SendManager::getNumberRegisteredISends() const
{
  return getNumberReadyISends() + getNumberPendingISends();
}


////////////////////////////////////////////////////////////////////////
//
// T2_CommManager
//
// The constructor takes the following arguments:
//	nprocs:		the total number of back-end processors
//	nreqs:		the maximum number of communication requests
//			to be registered with T2_CommManager
//	c:		a pointer to a T2_Comm
//
////////////////////////////////////////////////////////////////////////
class T2_CommManager {
public:
  T2_CommManager(u_int nprocs, u_int nreqs, T2_Comm* c=NULL);
  ~T2_CommManager();

  // query methods
  u_int getNumberPendingRequests() const;

  // access to T2_Comm
  T2_Comm* getComm() const { return comm; };
  void setComm(T2_Comm* c) { comm = c; };

  // init
  void initBeforeQueryExecution();

  // ready to issue a send?
  bool readyToIssueISend();

  // register non-blocking sends for a qenv
  void registerISends(T2_QEnv& qenv);

  // poll for completed operations
  void poll(T2_Array<T2_QEnv>& qenvs, T2_WorkList& wlist);

  // accessing the number of expected recvs
  u_int getNumberExpectedRecvs() const { return nRecvExpect; };
  u_int incNumberExpectedRecvs(u_int i=1);
  u_int decNumberExpectedRecvs(u_int i=1);

  // accessing the number of current sends
  u_int getNumberCurrentSends() const { return nSendsCurrent; };

  // accessing the number of recvs and sends issued
  T2_Array<u_int>& getNumberRecvsIssued() { return nRecvsIssued; };
  T2_Array<u_int>& getNumberSendsIssued() { return nSendsIssued; };
  u_int getNumberRecvsIssued(u_int p) const { return nRecvsIssued[p]; };
  u_int getNumberSendsIssued(u_int p) const { return nSendsIssued[p]; };
  void incNumberRecvsIssued(u_int p) { nRecvsIssued[p]++; };
  void incNumberSendsIssued(u_int p) { nSendsIssued[p]++; };
  void decNumberRecvsIssued(u_int p);
  void decNumberSendsIssued(u_int p);

  // broadcast a message to other nodes to kill a query
  void killRemoteQuery(const u_int qid, const u_int numnodes, 
		       const T2_ProcID mynode);

  // update the number of expected recv's when a query is invalidated
  void decNumberExpectedRecvsWhenKillQuery(T2_QStatus& qstatus);

  // recv nmsgs[p] msgs from processor p and just drop the msgs
  void drainMessages(T2_ProcID mynode, T2_Array<T2_QEnv>& qenvs);

protected:
  T2_Array<T2_SendManager> senders;	// a send manager for each processor

  T2_Array<T2_CommRequest> requests;

  T2_TSet<T2_CommRequest*>
		freeRequests;		// list of free T2_CommRequests
//		doneRequests;		// list of completed requests,
					// not needed as done requests are
					// being handled directly in poll
					// without staging them in doneRequests

  const u_int nPreRecvs;		// number of recvs to be posted
					// at the beginning
  u_int nRecvExpect;			// total # of expected receives
  u_int nRecvsPending;			// actual # of pending irecvs
  u_int nSendsCurrent;			// # of sends that are either
					// queued by a send manager,
					// pending in the pending queue of
					// T2_Comm, waiting in the done queue
					// of T2_Comm (ie # sends that have
					// not left the communication 
					// sub-system)
  T2_Array<u_int> nRecvsIssued,		// total # recvs posted per proc
		  nSendsIssued;		// total # sends posted per proc

  T2_Comm* comm;

  u_int max_number_irecvs;		// max # of pending irecvs at any time

  void postISends();			// post more isends
  void postIRecvs();			// post more irecvs
  void actionISendCompletes(T2_CommRequest* req, T2_Array<T2_QEnv>& qenvs,
			    T2_WorkList& wlist);
  void actionIRecvCompletes(T2_CommRequest* req, T2_Array<T2_QEnv>& qenvs,
			    T2_WorkList& wlist);

private:
};


inline u_int
T2_CommManager::getNumberPendingRequests() const
{
  return requests.getNumberElements() - freeRequests.getNumberElements();
}

inline u_int
T2_CommManager::incNumberExpectedRecvs(u_int i)
{
  nRecvExpect += i;
  postIRecvs();
  return nRecvExpect;
}

inline u_int
T2_CommManager::decNumberExpectedRecvs(u_int i)
{
  t2_Assert (nRecvExpect >= i);
  return (nRecvExpect -= i);
}

inline void
T2_CommManager::decNumberRecvsIssued(u_int p)
{
  t2_Assert (nRecvsIssued[p] > 0);
  nRecvsIssued[p]++;
}

inline void
T2_CommManager::decNumberSendsIssued(u_int p)
{
  t2_Assert (nSendsIssued[p] > 0);
  nSendsIssued[p]++;
}


#endif	/* _T2_COMM_H_ */
