
/* $Id: t2_outhandle.h,v 1.13 2000/03/16 00:43:05 chialin Exp $ */

/* t2_outhandle.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 6/30/98
 * classes for various ways of handling the final output of a T2 query
 */


#ifndef _T2_OUTHANDLE_H_
#define _T2_OUTHANDLE_H_


#include	"t2_header.h"
#include	"t2_bfile.h"
#include	"t2_system.h"
#include	"t2_acc.h"
#include	"t2_output.h"
#include	"t2_outbuf.h"
#include	"t2_lio.h"
#include	"t2_nbsocket.h"
#include	"t2_qspec.h"
#include	"t2_rts.h"
#include	"t2_queryinfo.h"


// forward declarations
class T2_QEnv;
class T2_OutputSock;

///////////////////////////////////////////////////////////////////////
//
// T2_OutputHandle
//
// This is the base class for T2 objects handling the final output from
// a given query. All output handling objects derived from this base 
// class provide the following methods.
//      isBlocking()			return true if the output handling
//					object is blocking and therefore
//					cannot proceed until all isends for
//					the global-combine phase have completed
//					(eg blocking Meta-Chaos)
//	initBeforeQueryExecution	invoked at the beginning of the
//					query execution service (ie
//					after query plan is constructed
//					but before execution is actually
//					carried out), and is only called
//					once for the given query;
//					should be used to set up some
//					thing for the entire query
//					execution (eg socket connection,
//					meta-chaos oob communication, etc);
//					returns true if succeeds and
//					query should be executed, false if
//					fails and query should not be
//					executed
//
//	handleOutput			invoked at the beginning of the
//					output handling phase of every
//					strip-mined iteration (ie invoked
//					once per iteration);
//					returns true if more output for
//					the current iteration needs to
//					be sent out (ie output-handling
//					phase has NOT completed), 
//					returns false if the entire
//					output for the current iteration
//					has been sent out and the output
//					handling phase has completed
//
//	issueNextOutput			invoked during the output-handling
//					phase to finish sending the output;
//					returns true if more output for
//					the current iteration needs to
//					be sent out (ie output-handling
//					phase has NOT completed), 
//					returns false if the entire
//					output for the current iteration
//					has been sent out and the output
//					handling phase has completed;
//					that is, within the output-
//					handling phase of an iteration,
//					handlOutput is invoked once at
//					the beginning of the output-
//					handling phase and issueNextOutput
//					is invoked for as many times as
//					needed for the remaining period
//					of the output-handling phase
//					until the entire output for the
//					current iteration has been sent
//					out
//
//	finalize			invoked at the end when the
//					entire query is completed;
//					that is, only invoked once for
//					the entire query; 
//					should be used for cleaning up
//					whatever is used for the query
//					(eg closing sockets, etc)
//
//	prematureErrorHandling		invoked at the beginning of the
//					query execution servce if the
//					query was invalidated during
//					the query planning service
//					(ie no query execution at all)
//	executionErrorHandling		method invoked if query was 
// 					invalidated during query execution 
//					and can't reach the output-handling 
//					phase for the the strip-mined iteration 
//					(ie tile) in which the error occurs;
//					by default, it does nothing
//
///////////////////////////////////////////////////////////////////////
class T2_OutputHandle {
public:
  //T2_OutputHandle(T2_QEnv* p, bool v=false);
  T2_OutputHandle(T2_QEnv* p, bool v, bool t);
  virtual ~T2_OutputHandle() {};

  // a valid output handle?
  const bool isValid() const { return valid; };
  void validate() { valid = true; };
  void invalidate() { valid = false; };

  // get qenv
  T2_QEnv* getQEnv() const { return qenv; };

  // T2 protocol enabled/disabled?
  bool useT2Protocol() const { return uset2protocol; };

  // is the output-handling object blocking (see comment for isBlocking())
  // and not all isends for the global-combine phase complete?
  bool blockingAndISendsNotComplete() const;

  // is the output-handling object blocking and therefore must wait
  // until all pending isends for the global-combine phase complete?
  // (eg blocking Meta-Chaos)
  virtual bool isBlocking() const = 0;

  // initialization at the beginning of query execution; returns true
  // if initialization succeeds, false if initialization fails and 
  // query should not be executed
  virtual bool initBeforeQueryExecution(T2_RTS& rts) = 0;

  // handle the output, returns true if the output handling phase is
  // needed, false if the output handling phase is not needed
  virtual bool handleOutput(T2_Output& output,
			    T2_Accumulator& acc, T2_Iteration i) = 0;

  // method invoked when a non-blocking write completes;
  // return true if another write is issued, false if all writes had been
  // issued
  virtual bool issueNextOutputWrite() = 0;

  // method invoked when the query is completely done
  virtual void finalize() = 0;

  // method invoked if query was invalidated during query planning;
  // by default, it does nothing
  virtual void prematureErrorHandling() {};

  // method invoked if query was invalidated during query execution and can't
  // reach the output-handling phase for the the strip-mined iteration 
  // (ie tile) in which the error occurs;
  // by default, it does nothing
  virtual void executionErrorHandling(const T2_QueryInfo& qinfo) {};

  static T2_LIO* lio;

protected:
  T2_QEnv *qenv;
  bool valid;
  bool uset2protocol;

  // do not use these constructors and operator =
  T2_OutputHandle();
  T2_OutputHandle(const T2_OutputHandle&);
  T2_OutputHandle& operator =(const T2_OutputHandle&);

  // copy a string from src to dest
  static const char* setString(char*& dest, const char* src);

private:
};


inline
T2_OutputHandle::T2_OutputHandle(T2_QEnv* p, bool v, bool tp):
				qenv(p), valid(v), uset2protocol(tp)
{
  t2_Assert (p != NULL);
}

inline
T2_OutputHandle::T2_OutputHandle():
				qenv(NULL), valid(false), uset2protocol(true)
{
  t2_err_quit(__FILE__, __LINE__,
  "T2_OutputHandle do not use this constructor");
}

inline
T2_OutputHandle::T2_OutputHandle(const T2_OutputHandle& h):
				qenv(h.qenv), valid(h.valid),
				uset2protocol(h.uset2protocol)
{
  t2_err_quit(__FILE__, __LINE__,
  "T2_OutputHandle do not use copy constructor");
}

inline T2_OutputHandle&
T2_OutputHandle::operator =(const T2_OutputHandle& h)
{
  t2_err_quit(__FILE__, __LINE__, "T2_OutputHandle do not use operator =");
  return *this;
}

inline const char*
T2_OutputHandle::setString(char*& dest, const char* src)
{
  delete [] dest;

  if (src == NULL) {
     dest = NULL;
     return NULL;
  }

  size_t len = strlen(src);
  t2_Alloc (dest, len+1, char);
  strcpy(dest, src);
  return dest;
}


///////////////////////////////////////////////////////////////////////
//
// T2_OHBStream
//
///////////////////////////////////////////////////////////////////////
class T2_OHBStream {
public:
  T2_OHBStream();
  ~T2_OHBStream();

  // access methods
  u_int getNumberPointers() const { return ptrs.getNumberPointers(); };
  u_int getCurrentPtrIndex() const { return cur_ptr; };
  const char* getCurrentDataPointer() const;
  size_t getCurrentDataSize() const { return ptrs.getDataSize(cur_ptr); };
//  const char* getDataPointer(u_int i) const { return ptrs.getDataPointer(i);};
//  size_t getDataSize(u_int i) const { return ptrs.getDataSize(i); };
  
  // increment cur_ptr by one
  u_int incCurrentPtrIndex() { return ++cur_ptr; };

  // skip the rest of the pointers and force the condition that makes
  // allWritesDone() return true; usually used when error occurs during
  // writes
  void skipRemainingPtrs();

  // are all writes done?
  bool allWritesDone() const;

  // flush the output into pointer list
  u_int flushOutput(const T2_QueryInfo& qinfo, T2_Output& output);

  // clean up
  void cleanup();

protected:
  T2_OBuf obuf;			// output buffer
  T2_OutputDataPtrList ptrs;	// list of pointers for writing the output
  u_int cur_ptr;

private:
};


inline
T2_OHBStream::~T2_OHBStream()
{
  t2_Assert (obuf.getBufferSize() == 0);
  t2_Assert (ptrs.getNumberPointers() == 0);
}

inline void
T2_OHBStream::cleanup()
{
  obuf.releaseBuffer();
  ptrs.removeAllPointers();
  cur_ptr = 0;
}

inline const char*
T2_OHBStream::getCurrentDataPointer() const
{
  return ptrs.getDataPointer(cur_ptr);
}

inline bool
T2_OHBStream::allWritesDone() const
{
  t2_Assert (cur_ptr <= ptrs.getNumberPointers());
  return (cur_ptr == ptrs.getNumberPointers())? true : false;
}

inline void
T2_OHBStream::skipRemainingPtrs()
{
  t2_Assert (cur_ptr <= ptrs.getNumberPointers());
  cur_ptr = ptrs.getNumberPointers();
}



///////////////////////////////////////////////////////////////////////
//
// T2_OHBinaryFile
//
///////////////////////////////////////////////////////////////////////
class T2_OHBinaryFile: public T2_OutputHandle, public T2_OHBStream {
public:
  T2_OHBinaryFile(T2_QEnv* p, T2_QSpecOutput& qspec);
  ~T2_OHBinaryFile() { delete [] prefix; };

  // is the output-handling object blocking and therefore must wait
  // until all pending isends for the global-combine phase complete?
  // (eg blocking Meta-Chaos)
  bool isBlocking() const { return false; };

  // initialization at the beginning of query execution; returns true
  // if initialization succeeds, false if initialization fails and 
  // query should not be executed
  virtual bool initBeforeQueryExecution(T2_RTS& rts) { return true; };

  // handle the output
  virtual bool handleOutput(T2_Output& output,
			    T2_Accumulator& acc, T2_Iteration i);

  // method invoked when a non-blocking write completes;
  // return true if another write is issued, false if all writes had been
  // issued
  virtual bool issueNextOutputWrite();

  // method invoked when the query is completely done
  virtual void finalize() {};

  // method invoked if query was invalidated during query planning;
  // just create an empty output file
  virtual void prematureErrorHandling();

  // method invoked if query was invalidated during query execution and can't
  // reach the output-handling phase for the the strip-mined iteration 
  // (ie tile) in which the error occurs;
  // by default, it does nothing
  virtual void executionErrorHandling(const T2_QueryInfo& qinfo);

  // are all writes done?
  bool allWritesDone() const { return T2_OHBStream::allWritesDone(); };

protected:
  char *prefix;
  size_t prefix_len,		// the length of the prefix string
	 char_len;		// the length of the array used to store the
				// prefix string
  off_t cur_offset;		// since POSIX lio_listio only supports
				// whence = SEEK_SET, needs to remember where
				// the last write ended

  T2_OBFile obfile;		// output binary file

  void registerOneWrite();	// register the cur_ptr-th write with lio

  // clean up after all the writes are done for an iteration
  void cleanupAfterAllWritesDoneForIteration();

  // create output filename into prefix[];
  // the filename would be in the form of prefix-<iteration #>.<node rank>
  void createOutputFileName(T2_Iteration i);

private:
  // do not use copy constructor and operator =
  T2_OHBinaryFile(const T2_OHBinaryFile& h);
  T2_OHBinaryFile& operator =(const T2_OHBinaryFile& h);

};


inline void
T2_OHBinaryFile::registerOneWrite()
{
  t2_Assert (getCurrentPtrIndex() < getNumberPointers());
  t2_Assert (lio != NULL);
  t2_Assert (lio->noMoreFreeCB() == false);
  T2_OutputHandle& oh = *this;

  t2_say(cout, "just about to register with lio a write for pointer "
		<< getCurrentPtrIndex() << " of " << getNumberPointers()
		<< " ptr=" << (u_int) getCurrentDataPointer()
		<< " nbytes=" << getCurrentDataSize()
		<< endl);

# ifdef T2_ASSERTION
     bool r = 
# endif
  lio->registerWrite(T2_LIOKey(t2_lioOutputWrite, (void *) &oh),
			      obfile.getfd(), SEEK_SET, cur_offset,
			      getCurrentDataPointer(), getCurrentDataSize());
  t2_Assert (r == true);
  cur_offset += getCurrentDataSize();

  t2_say2(cout, "just register with lio a write for pointer "
		<< getCurrentPtrIndex() << " of " << getNumberPointers()
		<< endl);

  incCurrentPtrIndex();
}

inline
T2_OHBinaryFile::T2_OHBinaryFile(const T2_OHBinaryFile& h):
		T2_OutputHandle(h.getQEnv(), false, true), T2_OHBStream(),
		prefix(NULL), prefix_len(0), char_len(0), obfile()
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_OHBinaryFile copy constructor");
}

inline T2_OHBinaryFile&
T2_OHBinaryFile::operator =(const T2_OHBinaryFile& h)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_OHBinaryFile operator =");
  return *this;
}




///////////////////////////////////////////////////////////////////////
//
// T2_OHSocket
//
///////////////////////////////////////////////////////////////////////
class T2_OHSocket: public T2_OutputHandle, public T2_OHBStream {
public:
  T2_OHSocket(T2_QEnv* p, T2_RTS& rts, T2_QSpecOutput& qspec);
  ~T2_OHSocket() { delete [] hostname; };

  // is the output-handling object blocking and therefore must wait
  // until all pending isends for the global-combine phase complete?
  // (eg blocking Meta-Chaos)
  bool isBlocking() const { return false; };

  // initialization at the beginning of query execution; returns true
  // if initialization succeeds, false if initialization fails and 
  // query should not be executed
  virtual bool initBeforeQueryExecution(T2_RTS& rts);

  // handle the output
  virtual bool handleOutput(T2_Output& output,
			    T2_Accumulator& acc, T2_Iteration i);

  // method invoked when a non-blocking write completes;
  // return true if another write is issued, false if all writes had been
  // issued
  virtual bool issueNextOutputWrite();

  // method invoked when the query is completely done
  virtual void finalize();

  // method invoked if query was invalidated during query planning;
  // just connect to client and close connection immediately
  virtual void prematureErrorHandling();

  // are all writes done?
  bool allWritesDone() const { return T2_OHBStream::allWritesDone(); };

  static T2_OutputSock *outsock;

  // for debugging purpose:
  T2_nbSocket& getSocket() { return sock; };

protected:
  char* hostname;
  int port;
  T2_nbSocket sock;
  u_int nconnects;		// # time connected to the client

  bool issueOneWrite();		// issue a write

  // clean up after all the writes are done for an iteration
  void cleanupAfterAllWritesDoneForIteration() { T2_OHBStream::cleanup(); };

private:
  // do not use copy constructor and operator =
  T2_OHSocket(const T2_OHSocket& h);
  T2_OHSocket& operator =(const T2_OHSocket& h);

};


inline
T2_OHSocket::T2_OHSocket(const T2_OHSocket& h):
		T2_OutputHandle(h.getQEnv(), false, true), T2_OHBStream(),
		hostname(NULL), port(0), sock(), nconnects(0)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_OHSocket copy constructor");
}

inline T2_OHSocket&
T2_OHSocket::operator =(const T2_OHSocket& h)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_OHSocket operator =");
  return *this;
}



///////////////////////////////////////////////////////////////////////
//
// T2_OHMetaChaos
//
///////////////////////////////////////////////////////////////////////
#ifdef T2_METACHAOS
  #include	"t2_ohmc.h"
#endif



#endif	/* _T2_OUTHANDLE_H_ */
