
/* $Id: outhandle.C,v 1.12 2000/05/09 16:56:20 chialin Exp $ */

/* outhandle.C */


#include	<string.h>
#include	"t2_outhandle.h"
#include	"t2_outsock.h"
#include	"t2_packets.h"
#include	"t2_qenv.h"


////////////////////////////////////////////////////////////////////////
//
// The constructors for classes derived from T2_OutputHandle must
// set the flag 'T2_OutputHandle::valid' to true when the constructor
// succeeds, flase otherwise.
//
// Also, virtual function handleOutput must initialize T2_QStatus
// appropriately.
//
////////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////////
//
// some static variables
//
////////////////////////////////////////////////////////////////////////
T2_LIO* T2_OutputHandle::lio = NULL;
T2_OutputSock* T2_OHSocket::outsock = NULL;


////////////////////////////////////////////////////////////////////////
//
// T2_OutputHandle methods
//
////////////////////////////////////////////////////////////////////////
bool
T2_OutputHandle::blockingAndISendsNotComplete() const
{
  t2_Assert (qenv != NULL);
  if (isBlocking() == true &&
      qenv->getQueryStatus().getNumberAccSendsYetComplete() > 0)
     return true;
  return false;
}


////////////////////////////////////////////////////////////////////////
//
// T2_OHBStream methods
//
////////////////////////////////////////////////////////////////////////
T2_OHBStream::T2_OHBStream(): obuf(), ptrs(), cur_ptr(0)
{
}


u_int
T2_OHBStream::flushOutput(const T2_QueryInfo& qinfo, T2_Output& output)
{
  t2_Assert (obuf.getBufferSize() == 0);
  t2_Assert (ptrs.getNumberPointers() == 0);

  // allocate output buffer
  size_t output_size = output.getOutputBufferSize(qinfo);
  obuf.allocBuffer(output_size);
  T2_OutputBuffer outbuf(obuf);

  // get the set of data pointers to be written to a file
  output.flushOutput(qinfo, outbuf, ptrs);


  t2_say4(cout, "output.flushOutput returns "
		<< ptrs.getNumberPointers()
		<< " pointers to write to stream" << endl);

  return ptrs.getNumberPointers();
}



////////////////////////////////////////////////////////////////////////
//
// T2_OHBinaryFile methods
//
////////////////////////////////////////////////////////////////////////
T2_OHBinaryFile::T2_OHBinaryFile(T2_QEnv* p, T2_QSpecOutput& qspec):
			T2_OutputHandle(p, true, qspec.useT2Protocol()),
			T2_OHBStream(),
			prefix(NULL), prefix_len(0), obfile()
{
  t2_Assert (qspec.getOutputHandleType() == t2_oBFile);
  t2_Assert (p != NULL);

  const char* pre = qspec.getFilePrefix();
  if (pre == NULL) {
     invalidate();
     return;
  }

  prefix_len = strlen(pre);
  char_len = prefix_len + 30;
  t2_Alloc (prefix, char_len, char);
  strcpy(prefix, pre);
}


bool
T2_OHBinaryFile::handleOutput(T2_Output& output,
			      T2_Accumulator& acc, T2_Iteration i)
{

  t2_say2(cout, "T2_OHBinaryFile::handleOutput for qenv " << (int) qenv
	       << " is invoked" << endl);

  t2_Assert (qenv != NULL);

  // make sure we start with an empty buffer
  t2_Assert (obuf.getBufferSize() == 0);
  t2_Assert (obuf.getDataSize() == 0);
  t2_Assert (!obfile);

  // create complete filename as prefix-iter.mynode in prefix[]
  createOutputFileName(i);

  t2_say2(cout, "about to open output file " << prefix << endl);

  // open the binary output file
  obfile.open(prefix);
  if (!obfile) {
     t2_err_ret(__FILE__, __LINE__, "cannot open output file %s", prefix);
     return false;
  }

  if (T2_OHBStream::flushOutput(qenv->getQueryInfo(), output) == 0) {
     // nothing to write
     cleanupAfterAllWritesDoneForIteration();
     qenv->initOutputHandle(0);
     return false;				// no need of output handling
  }

  // reset cur_offset to 0, the beginning of the file
  cur_offset = 0;

  // register first non-blocking write with lio
  t2_Assert (getCurrentPtrIndex() == 0);
  registerOneWrite();

  // update query status
  qenv->initOutputHandle(1);

  return true;
}


bool
T2_OHBinaryFile::issueNextOutputWrite()
{
  if (allWritesDone() == false) {		// more writes needed
     registerOneWrite();
     return true;
  }

  // all writes are done
  cleanupAfterAllWritesDoneForIteration();

  // update query status
  t2_Assert (qenv != NULL);
  qenv->decNumberOutputWritesYetComplete();
  t2_Assert(qenv->getQueryStatus().getNumberOutputWritesYetComplete() == 0);

  return false;
}


void
T2_OHBinaryFile::cleanupAfterAllWritesDoneForIteration()
{
  obfile.close();
  T2_OHBStream::cleanup();
}


void
T2_OHBinaryFile::prematureErrorHandling()
{
  // For error handling, I'll just create an empty output file.

  // get complete file name first
  createOutputFileName(0);

  t2_say2(cout, "about to create empty output file " << prefix 
		<< " as error handling" << endl);

  // open the binary output file
  obfile.open(prefix);
  if (!obfile) {
     t2_err_ret(__FILE__, __LINE__, "cannot open output file %s", prefix);
     return;
  }
  obfile.close();
}


void
T2_OHBinaryFile::executionErrorHandling(const T2_QueryInfo& qinfo)
{
  // if the error occurs in the first strip-mined iteration (ie tile)
  // and therefore the query never gets a chance to create any output
  // file, then just go ahead and create an empty file

  if (qinfo.getCurrentTileNumber() == 0)	// error in first tile
     prematureErrorHandling();
}


void
T2_OHBinaryFile::createOutputFileName(T2_Iteration i)
{
  // the filename would be in the form of prefix-<iteration #>.<node rank>
  t2_Assert (prefix != NULL);
  t2_Assert (qenv != NULL);
  sprintf(prefix+prefix_len, "-%d.%d",
                             (int) i, qenv->getQueryInfo().getMyNode());
}



////////////////////////////////////////////////////////////////////////
//
// T2_OHSocket methods
//
////////////////////////////////////////////////////////////////////////
T2_OHSocket::T2_OHSocket(T2_QEnv* p, T2_RTS& rts, T2_QSpecOutput& qspec):
			 T2_OutputHandle(p, true, qspec.useT2Protocol()),
			 T2_OHBStream(),
			 hostname(NULL), port(0), sock(), nconnects(0)
{
  t2_Assert (qspec.getOutputHandleType() == t2_oSocket);
  t2_Assert (p != NULL);
  if (setString(hostname, qspec.getHostName()) == NULL) {
     invalidate();
     return;
  }
  port = qspec.getPortNumber();
}


bool
T2_OHSocket::initBeforeQueryExecution(T2_RTS& rts)
{
  t2_Assert (isValid() == true);
  t2_Assert (hostname != NULL);
  t2_Assert (nconnects == 0);

  // let's just open the socket right now
  T2_Socket bsock;				// temporary blocking socket
  if (bsock.connectByName(hostname, port) == false) {	// connect fails
     // invalidate the query for now -- may want to retry some time later,
     // but note that connectByName() would try multiple times
     t2_err_ret(__FILE__, __LINE__,
		"cannot connect to client %s at port %d", hostname, port);
     return false;
  }


  t2_say2(cout, "socket connects to " << hostname << " @ "
	        << port << " succeeded" << endl);


  // send out the hello packet T2_BEClientHelloPacket if T2 protocol enabled
  if (useT2Protocol() == true) {
     T2_System& system = rts.getSystem();
     T2_BEClientHelloPacket hello(system.getNumberNodes(), system.getMyNode());
     hello.send(bsock);
  }

  // transfer from blocking socket to non-blocking socket
  sock.transfer(bsock);

  nconnects = 1;
  return true;
}


bool
T2_OHSocket::handleOutput(T2_Output& output,
			  T2_Accumulator& acc, T2_Iteration i)
{
  t2_Assert (qenv != NULL);

  if (!sock) {					// socket not connected
     t2_err_ret(__FILE__, __LINE__, "socket back to client not connected");
     return false;
  }

  if (T2_OHBStream::flushOutput(qenv->getQueryInfo(), output) == 0) {
     // nothing to write
     cleanupAfterAllWritesDoneForIteration();
     qenv->initOutputHandle(0);
     return false;				// no need of output handling
  }


  // issue first non-blocking write with socket
  t2_Assert (getCurrentPtrIndex() == 0);
  if (issueOneWrite() == false) {		// the entire write completes
     cleanupAfterAllWritesDoneForIteration();
     qenv->initOutputHandle(0);
     return false;				// no need of output handling
  }

  // update query status
  qenv->initOutputHandle(1);

  // register this object with outsock
  t2_Assert (outsock != NULL);
# ifdef T2_ASSERTION
     bool r = 
# endif
  outsock->registerOHSocket(*this);
  t2_Assert (r == true);

  return true;
}


bool
T2_OHSocket::issueNextOutputWrite()
{
  if (issueOneWrite() == true)			// more writes needed
     return true;
  t2_Assert (allWritesDone() == true);

  // all writes are done
  cleanupAfterAllWritesDoneForIteration();

  // update query status
  t2_Assert (qenv != NULL);
  qenv->decNumberOutputWritesYetComplete();
  t2_Assert(qenv->getQueryStatus().getNumberOutputWritesYetComplete() == 0);

  return false;
}


bool
T2_OHSocket::issueOneWrite()
{
  if (!sock) {				// not connected ?
     t2_err_ret(__FILE__, __LINE__, "socket back to client not connected");
     skipRemainingPtrs();		// skip remaining writes
     return false;			// fake that write is completed
  }


  ssize_t xxx;
  switch (xxx = sock.testNBWrite()) {
	case -1:			// error occurs
		if (sock.getErrorValue() == EINVAL)	// got no pending write
		   break;

		t2_err_ret(__FILE__, __LINE__, "sock.testNBWrite failed");

		skipRemainingPtrs();	// skip remaining writes
		return false;		// fake that write is completed

	case 0:				// pending write completes
		break;

	default:			// pending write does not complete
		t2_Assert (xxx > 0);
		return true;
  }				// switch

  // ready to issue more writes
  while (allWritesDone() == false) {
     const char* data = getCurrentDataPointer();
     size_t nbytes = getCurrentDataSize();

     ssize_t r = sock.nbWrite(data, nbytes, sizeof(char));
     incCurrentPtrIndex();

     switch (r) {
	case -1:			// error occurs
		if (sock.getErrorValue() == EFAULT) {	// null data pointer
		   if (nbytes == 0)			// size is 0, fine
		      break;
		}

		t2_err_ret(__FILE__, __LINE__, "sock.nbWrite failed");

		skipRemainingPtrs();	// skip remaining writes
		return false;		// fake that write is completed

	case 0:				// nbwrite completes
		break;

	default:			// nbwrite does not complete
		t2_Assert (r > 0);
		return true;
     }					// switch

  }					// while (allWritesDone() == false)

  return false;				// all writes done
}


void
T2_OHSocket::finalize()
{
  t2_Assert (sock);
  if (!sock)
     return;

  // send out bye packet if T2 protocol enabled
  if (useT2Protocol() == true) {
     T2_BEClientByePacket bye;
     bye.send(sock);
  }

  sock.closeSocket();
  nconnects = 0;
}


void
T2_OHSocket::prematureErrorHandling()
{
  T2_Socket bsock;
  if (bsock.connectByName(hostname, port) == false) {	// connect fails
     // note that connectByName() would try multiple times
     t2_err_ret(__FILE__, __LINE__,
		"cannot connect to client %s at port %d", hostname, port);
     return;
  }

  bsock.closeSocket();

  t2_say2(cout, "socket connects to " << hostname << " @ "
	        << port << " successfully and disconnects immediately" << endl);
}


////////////////////////////////////////////////////////////////////////
//
// T2_OHMetaChaos methods
//
////////////////////////////////////////////////////////////////////////
#ifdef T2_METACHAOS
  #include	"ohmc.C"
#endif

