
/* t2_nbsocket.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/10/98
 */


#ifndef _T2_NBSOCKET_H_
#define _T2_NBSOCKET_H_


#include	"t2_header.h"
#include	"t2_socket.h"


//////////////////////////////////////////////////////////////////////////
//
// T2_nbSocket
//
// The "client" version of a non-blocking TCP socket.
// T2_nbSocket allows a blocking connect or a non-blocking connect.
// If a blocking connect is called, the blocking connect method will set
// the socket to non-blocking if the connect succeeds. If a non-blocking
// connect is called, the socket is set to non-blocking immediately.
//
//////////////////////////////////////////////////////////////////////////
class T2_nbSocket : public T2_Sock {
public:
  T2_nbSocket();
  ~T2_nbSocket() {};

  // close a socket
  void closeSocket();

  // get/set the convertEndian flag for input or output socket;
  bool inputConvertEndian() const { return input_convertendian; };
  bool outputConvertEndian() const { return output_convertendian; };
  bool setInputConvertEndian(bool t) { return (input_convertendian = t); };
  bool setOutputConvertEndian(bool t) { return (output_convertendian = t); };

  //////
  // blocking connect; return true if connected successully, false otherwise
  // with errval set to the error number
  // 	errval		reason
  //	-------		---------
  //	EADDRNOTAVAIL	ipaddr or hostname is NULL
  //	EINVAL		there is a non-blocking request pending
  //	others		error occurred in some system call
  //////
  bool connectByAddress(const char* ipaddr, short port);
  bool connectByName(const char* hname, short port);

  //////
  // non-blocking connect;
  // these methods return true if
  // (1) connected completed successfully and immdiately, in which case
  //     connected() returns true; or
  // (2) connect initiated successfully but not completed, in which case
  //     connected() returns false
  // otherwise, these non-blocking connect methods return false;
  // in short, if connect was initiated successfully, these methods return
  // true (connected() tells whether connect completed successfully);
  // otherwise, they return false and errval tells the reason:
  // 	errval		reason
  //	-------		---------
  //	EADDRNOTAVAIL	ipaddr or hostname is NULL
  //	others		error occurred in some system call
  //////
  bool nbConnectByAddress(const char* ipaddr, short port);
  bool nbConnectByName(const char* hname, short port);

  //////
  // check for completion of a pending connect issued earlier;
  // a null argument to testConnect means a blocking test;
  // methods return true if
  // (1) connect completed successfully, in which case
  // connected() returns true; or
  // (2) connect completed but unsucessfully, in which case
  // connected() returns false and errval contains the error number;
  // otherwise, this method returns false;
  // in summary, testConnect returns true if no connect request is pending;
  // note: in the case where testConnect is invoked when no pending connect
  // was issued, testConnect returns true and set errval to EINVAL
  //////
  bool testConnect(struct timeval* tval = NULL);
  bool testConnect(int nsec, int nusec);

  // accessing peer's sockaddr_in
  // getPeerAddress returns true if succeeds, false otherwise with errval
  // set to errno:
  // 	errval		reason
  //	-------		---------
  //	ENOTCONN	socket is not connected
  //	EADDRNOTAVAIL	buf is NULL
  //	others		error occurred in some system call
  // getPeerPort returns -1 if it fails and set errval to ENOTCONN
  const struct sockaddr_in& getPeerSockAddress() const { return remote; };
  bool getPeerAddress(char* buf, size_t len);	// return false if fails
						// (check errval for reason)
  short getPeerPort();				// return -1 if fails

  ///////
  //
  // blocking read and write
  //
  // Behave the same way as the blocking read and write of T2_Socket does.
  // In the case where a blocking read/write is issued when a pending
  // nonblocking write or read exists, the methods return -1 and have errval
  // set to EAGAIN.
  //
  ///////
  ssize_t bwrite(const void *buf, int count, size_t size);
  ssize_t bread(void *buf, int count, size_t size);

  //////
  // non-blocking binary read and write
  // Method nbWrite and nbRead are used to write and read a buffer into and from
  // the socket. The methods write or read as many bytes as possible, subject
  // to the size of the buffer and RETURN THE NUMBER OF BYTES THAT THEY FAILED
  // TO WRITE OR READ (as opposed to the number of bytes written or read).
  // They return -1 if error occurs, 0 if the write or read completes, and
  // a positive number if some number of bytes were written or read.
  // In case when a write or read is not completely satisfied (ie nbWrite or
  // nbRead returns a value > 0), it remembers the request and would try
  // to complete the request each time testNBWrite or testNBRead is called.
  // Like nbWrite and nbRead, both testNBWrite and testNBRead RETURN THE
  // NUMBER OF BYTES THAT ARE LEFT TO BE WRITTEN OR READ (not the number of
  // bytes written or read).
  // T2_nbSocket is able to have at most one pending write and one pending
  // read.
  // Here are the meaning for some of the possible errval returned by these
  // four methods below.
  // 	errval		reason (method)
  //	-------		---------------
  //	ENOTCONN	socket is not connected (all)
  //	EAGAIN		issuing another read/write with a read/write pending
  //			(nbWrite, nbRead)
  //	EFAULT		buf to nbWrite/nbRead is NULL (nbWrite, nbRead)
  //	EINVAL		call testNBWrite/testNBRead with no read/write pending
  //			(testNBWrite, testNBRead)
  //	others		error occurred in some system call
  //////
  ssize_t nbWrite(const void *buf, int count, size_t size);
  ssize_t nbRead(void *buf, int count, size_t size);
  ssize_t testNBWrite();
  ssize_t testNBRead();
  size_t getNumberBytesYetToWrite() const { return wlen; };
  size_t getNumberBytesYetToRead() const { return rlen; };

  // transfer a socket from a blocking socket s and detach s from that socket;
  // this is a way to turn a blocking socket into a non-blocking socket
  void transfer(T2_Socket& s);

protected:
  struct sockaddr_in remote;
  bool connectPending;		// is connect still pending ?

  void *rbuf;			// pointer to buffer for pending read
  const void *wbuf;		// pointer to buffer for pending write
  size_t rlen, wlen;		// # bytes left for pending read and write

  void *read_buf;		// pointer to begin of buffer for pending read
  const void *write_buf;	// pointer to begin of buffer for pending write
  size_t read_size, write_size;	// request size of pending read/write
  int read_count, write_count;	// request count of pending read/write

  bool input_convertendian,	// need to convert endian for input socket?
       output_convertendian;	// need to convert endian for output socket?

  // open a socket
  bool openSocket() { return T2_Sock::openSocket(AF_INET, SOCK_STREAM, 0); };

  // blocking and non-blocking connect by sockaddr_in
  bool connectBySockAddr(struct sockaddr_in& serv_addr);
  bool nbConnectBySockAddr(struct sockaddr_in& serv_addr);

private:
  // don't use copy constructor and operator =
  T2_nbSocket(const T2_nbSocket& s);
  T2_nbSocket& operator =(const T2_nbSocket& s);

};


inline
T2_nbSocket::T2_nbSocket():
		T2_Sock(), connectPending(false),
		rbuf(NULL), wbuf(NULL), rlen(0), wlen(0),
		input_convertendian(false), output_convertendian(false)
{
}

inline void
T2_nbSocket::closeSocket()
{
  rbuf = NULL;
  wbuf = NULL;
  rlen = wlen = 0;
  T2_Sock::closeSocket();
}

inline bool
T2_nbSocket::testConnect(int nsec, int nusec)
{
  struct timeval tval;
  tval.tv_sec = nsec;
  tval.tv_usec = nusec;
  return testConnect(&tval);
}


inline
T2_nbSocket::T2_nbSocket(const T2_nbSocket& s): T2_Sock(s)
{
  t2_err_quit(__FILE__,__LINE__,"don't use T2_nbSocket copy constructor");
}

inline T2_nbSocket&
T2_nbSocket::operator =(const T2_nbSocket& s)
{
  t2_err_quit(__FILE__,__LINE__,"don't use T2_nbSocket operator =");
  return *this;
}



#endif	/* ! _T2_NBSOCKET_H_ */
