
/* t2_qspec.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 4/22/98
 * T2 query specification
 */


#ifndef _T2_QSPEC_H_
#define _T2_QSPEC_H_


#include	<iostream.h>
#include	<string.h>
#include	<strings.h>
#include	"t2_header.h"
#include	"t2_bstream.h"
#include	"t2_tarray.h"
#include	"t2_shape.h"
#include	"t2_usrarg.h"



////////////////////////////////////////////////////////////////
//
// T2_QSpecDataset
//
////////////////////////////////////////////////////////////////
class T2_QSpecDataset {
public:
  T2_QSpecDataset();
  T2_QSpecDataset(const T2_QSpecDataset& d);

  // operator =
  T2_QSpecDataset& operator =(const T2_QSpecDataset& d);

  // const access methods
  T2_DSID getDatasetID() const { return ds_id; };
  T2_IteratorID getIteratorID() const { return ds_iter; };
  u_int getIndexID() const { return index_id; };
  u_int getProjID() const { return prj_id; };
  const T2_UsrArg& getIndexConstructorArg() const { return index_consarg; };
  const T2_UsrArg& getProjConstructorArg() const { return prj_consarg; };
  const T2_Box& getQueryBox() const { return qregion; };

  // non-const access methods
  T2_DSID& getDatasetID() { return ds_id; };
  T2_IteratorID& getIteratorID() { return ds_iter; };
  T2_IndexID& getIndexID() { return index_id; };
  u_int& getProjID() { return prj_id; };
  T2_UsrArg& getIndexConstructorArg() { return index_consarg; };
  T2_UsrArg& getProjConstructorArg() { return prj_consarg; };
  T2_Box& getQueryBox() { return qregion; };

  // write to ostream
  ostream& write(ostream& os) const;

#ifdef _T2_BSTREAM_H_
  // write to and read from a binary file
  T2_OBStream& bwrite(T2_OBStream& obf) const;
  T2_IBStream& bread(T2_IBStream& ibf);
#endif

protected:
  T2_DSID ds_id;		// dataset id number
  T2_IteratorID ds_iter;	// dataset iterator id number

  T2_IndexID index_id;		// index id number
  T2_UsrArg index_consarg;	// argument for index constructor

  u_int prj_id;			// projection function object id
  T2_UsrArg prj_consarg;	// argument for projection function constructor

  T2_Box qregion;		// query region

private:
};


inline
T2_QSpecDataset::T2_QSpecDataset()
{
  ds_id = 0;
  ds_iter = 0;
  index_id = 0;
  prj_id = 0;
}

inline T2_QSpecDataset&
T2_QSpecDataset::operator =(const T2_QSpecDataset& d)
{
  ds_id = d.ds_id;
  ds_iter = d.ds_iter;
  index_id = d.index_id;
  index_consarg = d.index_consarg;
  prj_id = d.prj_id;
  prj_consarg = d.prj_consarg;
  qregion.setNumberDimensions(d.qregion.getNumberDimensions());
  qregion = d.qregion;
  return *this;
}

inline
T2_QSpecDataset::T2_QSpecDataset(const T2_QSpecDataset& d)
{
  (*this) = d;
}

inline ostream&
operator <<(ostream& os, const T2_QSpecDataset& d)
{
  return d.write(os);
}


#ifdef _T2_BSTREAM_H_

inline T2_OBStream& 
operator <<(T2_OBStream& obf, const T2_QSpecDataset& d)
{
  return d.bwrite(obf);
}

inline T2_IBStream& 
operator >>(T2_IBStream& ibf, T2_QSpecDataset& d)
{
  return d.bread(ibf);
}

#endif	/* _T2_BSTREAM_H_ */



////////////////////////////////////////////////////////////////
//
// T2_QSpecOutput
//
// T2_OutputHandleType is a data type that is used to specify
// the desired type for output handling. It is declared as an
// enum with the following values:
//	t2_oUnknown	not known what to do
//	t2_oBFile	output to a binary file
//	t2_oSocket	output to a socket
//	t2_oMC		output through meta-chaos
//
////////////////////////////////////////////////////////////////
enum T2_OutputHandleType {
	t2_oUnknown = 0,
	t2_oBFile,
	t2_oSocket,
	t2_oMC
};

class T2_QSpecOutput {
public:
  T2_QSpecOutput();
  T2_QSpecOutput(const T2_QSpecOutput& p);
  ~T2_QSpecOutput() { delete [] hostname; delete [] prefix; delete [] client; };

  // accessing output handle type
  T2_OutputHandleType getOutputHandleType() const { return type; };
  T2_OutputHandleType setOutputHandleType(const T2_OutputHandleType& t);

  // enabling/disabling T2 protocol (default is enabled)
  void enableT2Protocol() { uset2protocol = true; };
  void disableT2Protocol() { uset2protocol = false; };
  bool useT2Protocol() const { return uset2protocol; };

  // accessing output endian-ness
  u_char getOutputEndian() const { return endian; };
  void useBigEndianOutput() { endian = T2_BigEndian; };
  void useLittleEndianOutput() { endian = T2_LittleEndian; };

  // access methods for outputing to a socket
  const char* getHostName() const { return hostname; };
  int getPortNumber() const { return port; };
  const char* setHostName(const char* h);
  int setPortNumber(int p) { return (port = p); };
  int& getPortNumber() { return port; };

  // access methods for outputing to a binary file
  const char* getFilePrefix() const { return prefix; };
  const char* setFilePrefix(const char* p);

  // access methods for outputing through meta-chaos
  const char* getClient() const { return client; };
  u_int getNumberClients() const { return nclients; };
  int getOffBandTag() const { return ofbtag; };
  int getDataTag() const { return datatag; };
  const char* setClient(const char* c);
  u_int setNumberClients(u_int n) { return (nclients = n); };
  int setOffBandTag(int t) { return (ofbtag = t); };
  int setDataTag(int t) { return (datatag = t); };

  // write output to ostream
  ostream& write(ostream& os) const;

#ifdef _T2_BSTREAM_H_
  // write to and read from a binary file
  T2_OBStream& bwrite(T2_OBStream& obf) const;
  T2_IBStream& bread(T2_IBStream& ibf);
#endif

protected:
  T2_OutputHandleType type;
  bool uset2protocol;			// use t2 protocol?
  u_char endian;			// what endian to use for the output
					// (note that T2_QSpec::endian specifies
					// the endian that the query submit-er
					// uses for the usr arg in the
					// query specification)

  char* hostname;			// hostname used by output sockets
  int port;

  char* prefix;				// prefix for output files

  char* client;				// name of client program for meta-chaos
  u_int nclients;			// number of nodes in client program
  int ofbtag,				// msg tag for off-band communication
      datatag;				// msg tag for communication data arrays

  // don't use operator =
  T2_QSpecOutput& operator =(const T2_QSpecOutput& p);

private:
  static const char* setString(char*& dest, const char* src);
};


inline const char*
T2_QSpecOutput::setString(char*& dest, const char* src)
{
  delete [] dest;

  if (src == NULL) {
     dest = NULL;
     return NULL;
  }

  size_t len = strlen(src);
  t2_Alloc (dest, len+1, char);
  strcpy(dest, src);
  return dest;
}

inline const char*
T2_QSpecOutput::setHostName(const char* h)
{
  return setString(hostname, h);
}

inline T2_OutputHandleType
T2_QSpecOutput::setOutputHandleType(const T2_OutputHandleType& t)
{
  return (type = t);
}

inline const char*
T2_QSpecOutput::setFilePrefix(const char* h)
{
  return setString(prefix, h);
}

inline const char*
T2_QSpecOutput::setClient(const char* c)
{
  return setString(client, c);
}

inline
T2_QSpecOutput::T2_QSpecOutput():
	type(t2_oUnknown), uset2protocol(true), endian(T2_BigEndian),
	hostname(NULL), port(0), 
	prefix(NULL),
	client(NULL), nclients(0), ofbtag(-1), datatag(-1)
{
}

inline
T2_QSpecOutput::T2_QSpecOutput(const T2_QSpecOutput& p)
{
  type = p.type;
  uset2protocol = p.uset2protocol;
  endian = p.endian;
  setHostName(p.hostname);
  port = p.port;
  setFilePrefix(p.prefix);
  setClient(p.client);
  setNumberClients(p.nclients);
  setOffBandTag(p.ofbtag);
  setDataTag(p.datatag);
}

inline T2_QSpecOutput&
T2_QSpecOutput::operator =(const T2_QSpecOutput& p)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_QSpecOutput operator =");
  return *this;
}

inline ostream&
operator <<(ostream& os, const T2_QSpecOutput& p)
{
  return p.write(os);
}


#ifdef _T2_BSTREAM_H_

inline T2_OBStream&
operator <<(T2_OBStream& obf, const T2_QSpecOutput& p)
{
  return p.bwrite(obf);
}

inline T2_IBStream&
operator >>(T2_IBStream& ibf, T2_QSpecOutput& p)
{
  return p.bread(ibf);
}

#endif	/* _T2_BSTREAM_H_ */




////////////////////////////////////////////////////////////////
//
// T2_QSpec
//
////////////////////////////////////////////////////////////////
class T2_QSpec {
public:
  T2_QSpec(const u_int nds = 0);
  ~T2_QSpec();

  // do not use copy constructor and operator =
  T2_QSpec(const T2_QSpec& q);
  T2_QSpec& operator =(const T2_QSpec& q);

  // const access methods
  const T2_Box& getOutputBox() const { return output_box; };
  u_int getAccID() const { return acc_id; };
  u_int getAccNavigatorID() const { return acc_navid; };
  u_int getAggrID() const { return aggr_id; };
  u_int getNumberDatasets() const { return dataset.getNumberElements(); };
  const T2_QSpecDataset& getDatasetSpec(u_int i) const { return dataset[i]; };
  const T2_UsrArg& getAccConstructorArg() const { return acc_consarg; };
  const T2_UsrArg& getAggrConstructorArg() const { return aggr_consarg; };
  const T2_QSpecOutput& getOutputSpec() const { return output; };

  // non-const access methods
  T2_Box& getOutputBox() { return output_box; };
  u_int& getAccID() { return acc_id; };
  u_int& getAccNavigatorID() { return acc_navid; };
  u_int& getAggrID() { return aggr_id; };
  T2_QSpecDataset& getDatasetSpec(u_int i) { return dataset[i]; };
  T2_UsrArg& getAccConstructorArg() { return acc_consarg; };
  T2_UsrArg& getAggrConstructorArg() { return aggr_consarg; };
  T2_QSpecOutput& getOutputSpec() { return output; };

  // accessing qspec endian-ness
  u_char getQSpecEndian() const { return endian; };
  void useBigEndianUserArg() { endian = T2_BigEndian; };
  void useLittleEndianUserArg() { endian = T2_LittleEndian; };

  // change the number of datasets
  void setNumberDatasets(u_int i);

  // write output to ostream
  ostream& write(ostream& os) const;

#ifdef _T2_BSTREAM_H_
  // write to and read from a binary file
  T2_OBStream& bwrite(T2_OBStream& obf) const;
  T2_IBStream& bread(T2_IBStream& ibf);
#endif

protected:
  T2_Box output_box;		// output query box
  u_int acc_id,			// accumulator id number
        acc_navid;		// accumulator navigation function id number
  T2_UsrArg acc_consarg;	// argument for acc constructor

  u_int aggr_id;		// aggregation function id number
  T2_UsrArg aggr_consarg;	// argument for aggregation function constructor

  T2_Array<T2_QSpecDataset> dataset;

  T2_QSpecOutput output;

  u_char endian;		// the endian that the user arg in the qspec is
				// (note that T2_SpecOutput::endian specifies
				// the endian to be used for the output)

private:
};

inline
T2_QSpec::T2_QSpec(const u_int nds)
	: output_box(0), acc_id(0), acc_navid(0), acc_consarg(),
	  aggr_id(0), aggr_consarg(), dataset(nds), output(),
	  endian(T2_BigEndian)
{
}

inline
T2_QSpec::~T2_QSpec()
{
}

inline
T2_QSpec::T2_QSpec(const T2_QSpec& q): output_box(0), dataset(0)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_QSpec copy constructor");
}

inline T2_QSpec&
T2_QSpec::operator =(const T2_QSpec& q)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_QSpec operator =");
  return *this;
}


inline void
T2_QSpec::setNumberDatasets(u_int i)
{
  dataset.clear(i);
}


#ifdef _T2_BSTREAM_H_

inline T2_OBStream&
operator <<(T2_OBStream& obf, const T2_QSpec& q)
{
  return q.bwrite(obf);
}

inline T2_IBStream&
operator >>(T2_IBStream& ibf, T2_QSpec& q)
{
  return q.bread(ibf);
}

#endif	/* _T2_BSTREAM_H_ */


inline ostream&
operator <<(ostream& os, const T2_QSpec& s)
{
  return s.write(os); 
}




////////////////////////////////////////////////////////////////
//
// T2_QBatch
//
// A class for a batch of query specifications.
//
////////////////////////////////////////////////////////////////
class T2_QBatch {
public:
  T2_QBatch(u_int n = 0): batch(n), valid(true) {};
  ~T2_QBatch() {};

  // valid?
  bool isValid() const { return valid; };

  // access methods
  u_int getNumberQueries() const { return batch.getNumberElements(); };
  const T2_QSpec& getQuerySpec(u_int i) const { return batch[i]; };
  T2_QSpec& getQuerySpec(u_int i) { return batch[i]; };

  // resize the batch and throw away all the existing query specs
  void setNumberQueries(u_int n);

  // write to ostream
  ostream& write(ostream& os) const;

#ifdef _T2_BSTREAM_H_
  // write to and read from a binary file
  T2_OBStream& bwrite(T2_OBStream& obf) const;
  T2_IBStream& bread(T2_IBStream& ibf);
#endif

protected:
  T2_Array<T2_QSpec> batch;
  bool valid;

private:
};

inline void
T2_QBatch::setNumberQueries(u_int n)
{
  batch.clear(n);
}

inline ostream&
T2_QBatch::write(ostream& os) const
{
  os << "T2_QBatch has " << getNumberQueries() << " queries, as follows";
  for (u_int i=0; i<getNumberQueries(); i++)
     os << endl << batch[i];
  return os;
}

inline ostream&
operator <<(ostream& os, const T2_QBatch& b)
{
  return b.write(os);
}


#ifdef _T2_BSTREAM_H_

inline T2_OBStream&
operator <<(T2_OBStream& obf, const T2_QBatch& b)
{
  return b.bwrite(obf);
}

inline T2_IBStream&
operator >>(T2_IBStream& ibf, T2_QBatch& b)
{
  return b.bread(ibf);
}

#endif	/* _T2_BSTREAM_H_ */



////////////////////////////////////////////////////////////////
//
// T2_QPtrBatch
//
// A class for a batch of query specification pointers.
//
// Note that T2_QPtrBatch does not allocate memory for any
// T2_QSpec pointer stored in T2_QPtrBatch, and therefore
// does not free the memory pointed to by any of its
// T2_QSpec pointers.
//
////////////////////////////////////////////////////////////////
class T2_QPtrBatch {
public:
  T2_QPtrBatch(u_int n = 0);
  ~T2_QPtrBatch();

  // valid?
  bool isValid() const { return valid; };

  // access methods
  u_int getNumberQueries() const { return batch.getNumberElements(); };
  u_int getNumberNonNullQueries() const;
  const T2_QSpec* getQuerySpec(u_int i) const { return batch[i]; };
  T2_QSpec*& getQuerySpec(u_int i) { return batch[i]; };

  // resize the batch and throw away all the existing query specs
  void setNumberQueries(u_int n);

  // add a query pointer to the list; returns true if succeeds, false otherwise
  // if where is not specified, would try to find an empty spot in the list
  bool addQuerySpec(T2_QSpec& q, int where=-1);

  // write to ostream
  ostream& write(ostream& os) const;

#ifdef _T2_BSTREAM_H_
  // write to a binary file, should behave exactly the same as T2_QSpec::bwrite
  T2_OBStream& bwrite(T2_OBStream& obf) const;
  // read from a binary file, should behave exactly the same as T2_QSpec::bread
  T2_IBStream& bread(T2_IBStream& ibf);
#endif

protected:
  T2_Array<T2_QSpec*> batch;
  bool valid;

private:
};


inline
T2_QPtrBatch::T2_QPtrBatch(u_int n): batch(n), valid(true)
{
  for (u_int i=0; i<n; i++)
    batch[i] = NULL;
}

inline u_int
T2_QPtrBatch::getNumberNonNullQueries() const
{
  const u_int n = getNumberQueries();
  u_int cnt = 0;
  for (u_int i=0; i<n; i++)
    if (batch[i] != NULL)
       cnt++;
  return cnt;
}

inline void
T2_QPtrBatch::setNumberQueries(u_int n)
{
  batch.clear(n);
  for (u_int i=0; i<n; i++)
    batch[i] = NULL;
}

inline ostream&
T2_QPtrBatch::write(ostream& os) const
{
  os << "T2_QPtrBatch has " << getNumberQueries() << " queries, as follows";
  for (u_int i=0; i<getNumberQueries(); i++) {
     if (batch[i] == NULL)
        os << endl << "NULL query spec";
     else
        os << endl << *(batch[i]);
  }
  return os;
}

inline ostream&
operator << (ostream& os, const T2_QPtrBatch& q)
{
  return q.write(os);
}


#ifdef _T2_BSTREAM_H_
inline T2_OBStream&
operator << (T2_OBStream& obf, const T2_QPtrBatch& q)
{
  return q.bwrite(obf);
}

inline T2_IBStream&
operator >>(T2_IBStream& ibf, T2_QPtrBatch& b)
{
  return b.bread(ibf);
}
#endif




#endif	/* _T2_QSPEC_H_ */
