
/* $Id: t2_tarray.h,v 1.8 2000/01/21 00:56:09 chialin Exp $ */

/* t2_tarray.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 12/23/97
 * C++ templates used by T2 for array of objects.
 * Adapted from the part of Titan code written by Carter T. Shock.
 */

#ifndef _T2_TARRAY_H_
#define _T2_TARRAY_H_

#include	"t2_header.h"
#include	"t2_template.h"


///////////////////////////////////////////////////
//
// template arrays
//
///////////////////////////////////////////////////
template <class T>
class T2_Array {
public:
  // this constructor would simply use ar[] without making a copy
  T2_Array(u_int sz = 0, T ar[] = 0L);

  T2_Array(const T2_Array<T>& t);
  ~T2_Array() { delete [] a; };

  // query methods
  u_int getNumberElements() const { return sz; };
  const T* getElements() const { return a; };
  T*& getElements() { return a; };

  // access methods
  const T& operator [](u_int i) const { t2_Assert(a && (i<sz)); return a[i]; };
  T& operator [](u_int i) { t2_Assert(a && (i<sz)); return a[i]; };

  // operator =
  T2_Array<T>& operator =(const T2_Array<T>& t);

  // resize the array and keep the old elements whenever possible;
  // if the new array is shorter, elements from the old array are truncated
  void resize(u_int newsz, T ar[] = 0L);

  // clear the array, ie resize the array and discard all elements
  u_int clear(u_int newsz, T ar[] = 0L);

protected:
  u_int sz;
  T *a;
private:
};



template <class T>
inline
T2_Array<T>::T2_Array(u_int sz, T ar[])
{
  if (ar == NULL) {
     if (sz > 0)
        t2_Alloc(a, sz, T);
     else
        a = NULL;
  }
  else
     a = ar;
  this->sz = sz;
}


template <class T>
inline
T2_Array<T>::T2_Array(const T2_Array<T>& t)
{
  sz = t.sz;
  if (sz == 0) {
     a = NULL;
     return;
  }
  t2_Alloc(a, sz, T);
  for (u_int i=0; i<sz; i++)
     a[i] = t.a[i];
}


template <class T>
inline T2_Array<T>&
T2_Array<T>::operator =(const T2_Array<T>& t)
{
  if (t.sz == 0) {
     delete [] a;
     a = NULL;
     sz = 0;
     return *this;
  }
  if (sz != t.sz) {
     delete [] a;
     t2_Alloc(a, t.sz, T);
     sz = t.sz;
  }
  for (u_int i=0; i<sz; i++)
     a[i] = t.a[i];
  return *this;
}


// resize the array and keep the old elements whenever possible;
// if the new array is shorter, elements from the old array are truncated
template <class T>
inline void
T2_Array<T>::resize(u_int newsz, T ar[])
{
  if (sz == newsz)			// no change of size
     return;
  if (newsz == 0) {
     delete [] a;
     a = NULL;
     sz = 0;
     return;
  }

  T *d = a;
  if (ar == NULL) {
     t2_Alloc(a, newsz, T);
  }
  else {
     a = ar;
  }
  u_int s = t2_min(sz, newsz);
  for (u_int i=0; i<s; i++)
     a[i] = d[i];
  sz = newsz;
  delete [] d;
}


// clear the array, ie resize the array and discard all elements
template <class T>
inline u_int
T2_Array<T>::clear(u_int newsz, T ar[])
{
  if (sz == newsz)			// no change of size
     return sz;
  delete [] a;
  if (ar == NULL) {
     if (newsz > 0) {
        t2_Alloc(a, newsz, T);
     }
     else
        a = NULL;
  }
  else
     a = ar;
  sz = newsz;
  return sz;
}



///////////////////////////////////////////////////
//
// template variable-length arrays
// Each element must have a constructor that takes
// no arguments, an assignment operator (ie
// operator =). New elements are allowed to be
// appended to the array, which will grow its
// size accordingly.
//
///////////////////////////////////////////////////
template <class Type>
class T2_VArray {
public:
  T2_VArray(u_int l);
  virtual ~T2_VArray();

  // query methods
  u_int segmentSize() const { return segsize; };
  u_int getNumberElements() const { return ndata; };
  bool isFull() const { return (ndata==(nsegments*segsize))? true:false; };

  // access methods
  Type& operator [] (int i);
  const Type& operator [] (int i) const;

  // append new data to array
  T2_VArray<Type>& operator << (const Type& e);

  // append a "blank" data (ie the one constructed with the constructor that
  // takes no arguments) to the array and return that blank data;
  // this is done by simply incrementing the number of elements in the
  // array (and grow the array if necessary, of course)
  Type& appendElement();

  // remove all elements from array and free all buffer space
  void clear();

  // the following set of methods allow for efficient sequential access;
  // the model is call getFirstSegment() first, then iterator through
  // elements in the segment and call getNextSegment() to get to the
  // next segment; call currentIsAtLastSegment() to see if current is pointing
  // at the last segment, or call currentIsNull() to test if current has run
  // passed the last segment;
  // for each segment, use getNumberElementsInCurrentSegment() to find out the
  // the number of elements in the current segment and call method
  // getCurrentSegmentElement(i) to access the i-th element in the current
  // segment
  void getFirstSegment() { cur = head; };
  void getNextSegment() { if (cur != NULL) cur = cur->next; };
  bool currentIsNull() const { return (cur == NULL)? true:false; };
  bool currentIsAtLastSegment() const { return (cur==tail)? true:false; };
  u_int getNumberElementsInCurrentSegment() const;
  const Type& getCurrentSegmentElement(u_int i) const;
  Type& getCurrentSegmentElement(u_int i);

  // dump the internal for debugging purpose
  ostream& dump(ostream& os) const;

protected:
  struct segment_t {
	 Type *data;
	 segment_t *next;

     segment_t(u_int l) : next(NULL) { t2_Alloc(data, l, Type); };
     ~segment_t() { delete [] data; };

     Type& operator [](u_int i) { return data[i]; };
     const Type& operator [](u_int i) const { return data[i]; };
  };

  segment_t *head, *tail, *cur;
  const u_int segsize;
  u_int ndata,			// number of data elements
        nsegments;		// number of segments

  // method to grow the array
  void grow();

  // method to initialize elements after grow is called; do nothing by default
  virtual void initElement(Type&) {};

private:
  // do not use copy constructor or operator =
  T2_VArray(const T2_VArray<Type>& t);
  T2_VArray<Type>& operator =(const T2_VArray<Type>& t);

};


template <class Type>
inline
T2_VArray<Type>::T2_VArray(u_int l) : segsize(l), ndata(0), nsegments(1)
{
  head = tail = cur = new segment_t(l);
}

template <class Type>
inline
T2_VArray<Type>::T2_VArray(const T2_VArray<Type>& t)
	: segsize(t.segsize), ndata(t.ndata), nsegments(t.nsegments)
{
  t2_err_quit(__FILE__, __LINE__,
	      "do not use T2_VArray copy constructor");
}

template <class Type>
inline
T2_VArray<Type>::~T2_VArray()
{
  segment_t *p = head;
  while (p != NULL) {
     segment_t *q = p->next;
     delete p;
     p = q;
  }
}

template <class Type>
inline Type&
T2_VArray<Type>::operator [] (int i)
{
  t2_Assert((u_int) i < ndata);
  t2_Assert(i >= 0);
  const int k = i / segsize;
  segment_t *p = head;
  for (int jxj=0; jxj<k; jxj++) {
     t2_Assert(p != NULL);
     p = p->next;
  }
  return (*p)[i % segsize];
}

template <class Type>
inline const Type&
T2_VArray<Type>::operator [] (int i) const
{
  t2_Assert((u_int) i < ndata);
  t2_Assert(i >= 0);
  const int k = i / segsize;
  segment_t *p = head;
  for (int jxj=0; jxj<k; jxj++) {
     t2_Assert(p != NULL);
     p = p->next;
  }
  return (*p)[i % segsize];
}


template <class Type>
inline T2_VArray<Type>&
T2_VArray<Type>::operator =(const T2_VArray<Type>&)
{ t2_err_quit(__FILE__, __LINE__,
	      "do not use T2_VArray operator=");
  return *this;
}

// append new data to array
template <class Type>
inline T2_VArray<Type>&
T2_VArray<Type>::operator << (const Type& e)
{
  if (isFull()) {
     grow();
     Type& t = (*tail)[0];
     initElement(t);
     t = e;
  }
  else	{
     Type& t = (*tail)[ndata % segsize];
     initElement(t);
     t = e;
  }
  ndata++;
  return *this;
}

// append a "blank" data element to array
template <class Type>
inline Type&
T2_VArray<Type>::appendElement()
{
  if (isFull()) {
     grow();
     Type& t = (*tail)[0];
     initElement(t);
     ndata++;
     return t;
  }
  Type& t = (*tail)[ndata % segsize];
  initElement(t);
  ndata++;
  return t;
}

template <class Type>
inline void
T2_VArray<Type>::grow()
{
  segment_t *p = new segment_t(segsize);
  if (head == NULL)
     head = p;
  else {
     t2_Assert(tail != NULL);
     tail->next = p;
  }
  tail = p;
  nsegments++;
}

template <class Type>
inline u_int
T2_VArray<Type>::getNumberElementsInCurrentSegment() const
{
  if (currentIsAtLastSegment() == false)
     return segsize;
  u_int n = ndata % segsize;
  if (n == 0) {			// this happens to be a full segment
     return segsize;
  }
  return n;
}

template <class Type>
inline const Type&
T2_VArray<Type>::getCurrentSegmentElement(u_int i) const
{
  t2_Assert (i < getNumberElementsInCurrentSegment());
  return (*cur)[i];
}

template <class Type>
inline Type&
T2_VArray<Type>::getCurrentSegmentElement(u_int i)
{
  t2_Assert (i < getNumberElementsInCurrentSegment());
  return (*cur)[i];
}

template <class Type>
inline void
T2_VArray<Type>::clear()
{
  segment_t *p = head;
  t2_Assert (p != NULL);
  while (p != NULL) {
     segment_t *q = p->next;
     delete p;
     p = q;
  }
  head = tail = cur = new segment_t(segsize);
  ndata = 0;
  nsegments = 1;
}

template <class Type>
inline ostream&
T2_VArray<Type>::dump(ostream& os) const
{
  os << "segsize = " << segsize
     << ", ndata = " << ndata
     << ", # segments = " << nsegments << endl;
  os << "head = " << (int) head << ", tail = " << (int) tail
     << ", cur = " << (int) cur << endl;
  os << "segments:";
  segment_t* p = head;
  while (p != NULL) {
    os << " " << (int) p;
    p = p->next;
  }
  return os;
}

template <class Type>
inline ostream&
dump(ostream& os, const T2_VArray<Type>& v)
{
  return v.dump(os);
}



#endif	/* _T2_TARRY_H_ */

