
/* $Id: t2_usrarg.h,v 1.20 2000/04/10 15:55:03 chialin Exp $ */

/* t2_usrarg.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 4/28/98
 */


#ifndef _T2_USRARG_H_
#define _T2_USRARG_H_


#include	<iostream.h>
#include 	<string.h>
#include	"t2_header.h"
#include	"t2_bstr.h"
#include	"t2_bstream.h"


// forward declaration
class T2_UsrArgWriter;
class T2_UsrArgReader;
class T2_CommManager;
class T2_SendManager;


/////////////////////////////////////////////////////////////////////////
//
// T2_UsrArg
//
// A class to store arguments provided by users.
//
/////////////////////////////////////////////////////////////////////////
class T2_UsrArg {
friend class T2_UsrArgWriter;
friend class T2_UsrArgReader;
friend class T2_CommManager;
friend class T2_SendManager;
public:
  T2_UsrArg();
  T2_UsrArg(const T2_UsrArg& a);
  ~T2_UsrArg();

  // give it a fixed-size buffer
  void setUsrArg(char* d, size_t l);

const void* getThis() const { return this; };

  // const access methods
  bool isFull() const { return (getDataSize()==getBufferSize())? true:false; };
  size_t getDataSize() const { return ndata; };
  size_t getBufferSize() const { return len; };
  const char* getStartOfData() const { return data; };
  const char* getStartOfEmptyBuffer() const
	{ return (data == NULL)? 0:data+ndata; };

  // non-const access methods
  size_t setDataSize(size_t s) { return (ndata = s); };
  size_t setBufferSize(size_t l) { return (len = l); };
  char* getStartOfData() { return data; };
  char* getStartOfEmptyBuffer() { return (data == NULL)? 0:data+ndata; };

  // operator =, which copies data buf in a into this object
  T2_UsrArg& operator =(const T2_UsrArg& a);

  // method steal() would delete the buffer that's already in the local object,
  // copy the pointer to the data buffer in argument a and the data size
  // to the local object, and set the buffer pointer in a to NULL, data size
  // and buffer size in a to 0;
  T2_UsrArg& steal(T2_UsrArg& a);

  // throw away the old data and allocate new space of s bytes
  size_t allocBuffer(size_t s);

  // remove data buffer
  void releaseBuffer() { delete [] data; data = NULL; len = ndata = 0; };

#ifdef _T2_BSTREAM_H_
  // write to and read from a binary file
  T2_OBStream& bwrite(T2_OBStream& obf) const;
  T2_IBStream& bread(T2_IBStream& obf);
#endif

  // write to a ostream
  ostream& write(ostream& os) const;

protected:
  char* data;
  size_t len,			// length of array data[]
         ndata;			// number of bytes stored in data

private:
};


inline
T2_UsrArg::T2_UsrArg()
{
  data = NULL;
  len = ndata = 0;
}

inline
T2_UsrArg::~T2_UsrArg()
{
  delete [] data;
}

inline void
T2_UsrArg::setUsrArg(char* d, size_t l)
{
  delete [] data;
  if (d == NULL) {
      data = NULL;
     len = ndata = 0;
  }
  else {
     data = d;
     len = ndata = l;
  }
}

inline T2_UsrArg&
T2_UsrArg:: operator =(const T2_UsrArg& a)
{
  delete [] data;
  len = a.len;
  ndata = a.ndata;
  if (len == 0) {
     data = NULL;
  }
  else {
     t2_Alloc (data, len, char);
     bcopy(a.data, data, ndata);
  }
  return *this;
}

inline T2_UsrArg&
T2_UsrArg:: steal(T2_UsrArg& a)
{
  delete [] data;
  len = a.len;
  ndata = a.ndata;
  if (len == 0) {
     data = NULL;
     t2_Assert (a.data == NULL);
  }
  else {
     data = a.data;
     a.data = NULL;
     a.len = a.ndata = 0;
  }
  return *this;
}

inline
T2_UsrArg::T2_UsrArg(const T2_UsrArg& a)
{
  (*this) = a;
}


#ifdef _T2_BSTREAM_H_

inline T2_OBStream&
T2_UsrArg::bwrite(T2_OBStream& obf) const
{
  obf << ndata;
  if (ndata) {
#    ifdef T2_ASSERTION
        ssize_t r = 
#    endif
     obf.bwrite(data, ndata, sizeof(char));
     t2_Assert (r == (ssize_t) ndata);
  }
  return obf;
}

inline T2_IBStream&
T2_UsrArg::bread(T2_IBStream& ibf)
{
  t2_Assert (ibf);
  size_t s;
  ibf >> s;
  allocBuffer(s);
  if (s) {
#    ifdef T2_ASSERTION
        ssize_t r = 
#    endif
     ibf.bread(data, s, sizeof(char));
     t2_Assert (r == (ssize_t) s);
  }
  ndata = s;
  return ibf;
}

inline T2_OBStream&
operator <<(T2_OBStream& obf, const T2_UsrArg& a)
{
  return a.bwrite(obf);
}

inline T2_IBStream&
operator >>(T2_IBStream& ibf, T2_UsrArg& a)
{
  return a.bread(ibf);
}

#endif	/* _T2_BSTREAM_H_ */

inline ostream&
operator << (ostream& os, const T2_UsrArg& a)
{
  return a.write(os);
}



/////////////////////////////////////////////////////////////////////////
//
// T2_UsrArgWriter
//
// A class to provide write capabilities for user-defined arguments.
//
/////////////////////////////////////////////////////////////////////////
class T2_UsrArgWriter {
public:
  T2_UsrArgWriter();
  T2_UsrArgWriter(T2_UsrArg& a);
  ~T2_UsrArgWriter();

  // attach the writer to a new T2_UsrArg obj
  void open(T2_UsrArg& a);
  void close();

  // query methods
  bool isFull() const { return bstr.isFull(); };
  size_t getBufferSize() const { return bstr.getBufferSize(); };
  size_t getNumberBytesWritten() const { return bstr.getNumberBytesWritten(); };

  // need to convert endian?
  bool convertEndian() const { return convert_endian; };
  void setConvertEndian(bool t) { convert_endian = t; };

  // write n bytes into T2_UsrArg
  T2_UsrArgWriter & write(const void* c, size_t n)
		{ bstr.bwrite(c,n,sizeof(char)); sync(); return *this; };

  // write a char string ended with a '\0' into T2_UsrArg (will write '\0'
  // into T2_UsrArg as well)
  T2_UsrArgWriter & writeString(const char* s)
		{ if (s == NULL) bstr << '\0';
		  else		 bstr.bwrite(s, strlen(s)+1, sizeof(char));
		  sync(); return *this; };

  T2_UsrArgWriter & operator << (bool b);

  T2_UsrArgWriter & operator << (char c);
  T2_UsrArgWriter & operator << (signed char c);
  T2_UsrArgWriter & operator << (unsigned char c);

  T2_UsrArgWriter & operator << (short s);
  T2_UsrArgWriter & operator << (unsigned short s);

  T2_UsrArgWriter & operator << (int i);
  T2_UsrArgWriter & operator << (unsigned int i);

  T2_UsrArgWriter & operator << (long l);
  T2_UsrArgWriter & operator << (unsigned long l);

  T2_UsrArgWriter & operator << (float f);
  T2_UsrArgWriter & operator << (double d);

  // do not use copy constructor and operator =
  T2_UsrArgWriter(const T2_UsrArgWriter& w);
  T2_UsrArgWriter& operator =(const T2_UsrArgWriter& w);

  // print all the bytes that have been written into the argument buffer
  // (for debugging purpose)
  ostream& writeBytesInBuffer(ostream& os) const
				{ return bstr.writeBytesInBuffer(os); };

protected:
  T2_UsrArg* arg;
  T2_FOBStr bstr;
  size_t arg_oldlen;		// arg.getDataSize() when open
  bool convert_endian;		// need to convert endian?

  // throw away the old data of arg and allocate new space of s bytes in arg
  size_t allocBuffer(size_t s);

  void sync() 	 		// update arg->ndata
	{ if (arg != NULL)
	     arg->setDataSize(arg_oldlen + bstr.getNumberBytesWritten()); };

private:
};


inline void
T2_UsrArgWriter::close()
{
  if (arg != NULL) {
     t2_AssertOrDo (arg->getDataSize() ==
		    arg_oldlen + bstr.getNumberBytesWritten() ||
		    arg->getBufferSize() == 0,
		cout << "arg->getDataSize = " << arg->getDataSize()
		     << ", arg->getBufSize = " << arg->getBufferSize()
		     << ", oldlen = " << arg_oldlen
		     << ", # bytes written = " << bstr.getNumberBytesWritten());
     arg = NULL;
  }
  bstr.close();
  arg_oldlen = 0;
  convert_endian = false;
}

inline
T2_UsrArgWriter::T2_UsrArgWriter()
		: arg(NULL), bstr(), arg_oldlen(0), convert_endian(false)
{
}

inline
T2_UsrArgWriter::T2_UsrArgWriter(T2_UsrArg& a)
		: arg(&a), bstr(), convert_endian(false)
{
  // make bstr point to the unwritten part of the buffer in a
  bstr.open(a.getStartOfEmptyBuffer(), a.getBufferSize() - a.getDataSize());
  arg_oldlen = a.getDataSize();
}

inline
T2_UsrArgWriter::~T2_UsrArgWriter()
{
  close();				// close the current T2_UsrArg
}

inline void
T2_UsrArgWriter::open(T2_UsrArg& a)
{
  close();				// close the current T2_UsrArg

  // now attach to the new T2_UsrArg
  arg = &a;
  bstr.open(a.getStartOfEmptyBuffer(), a.getBufferSize() - a.getDataSize());
  arg_oldlen = a.getDataSize();
}

inline size_t
T2_UsrArgWriter::allocBuffer(size_t s)
{
  if (arg == NULL)
     return 0;

  size_t ss = arg->allocBuffer(s);
  bstr.open(arg->getStartOfEmptyBuffer(), arg->getBufferSize());
  arg_oldlen = arg->getDataSize();
  return ss;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (bool b)
{
  if (convertEndian() == true) t2_swapBytes(&b, sizeof(bool));
  bstr << b; sync();
  if (convertEndian() == true) t2_swapBytes(&b, sizeof(bool));
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (char c)
{
  // no need to convert endian at all
  bstr << c; sync(); 
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (signed char c)
{
  // no need to convert endian at all
  bstr << c; sync();
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (unsigned char c)
{
  // no need to convert endian at all
  bstr << c; sync();
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (short s)
{
  if (convertEndian() == true) t2_swapBytes(&s, sizeof(short));
  bstr << s; sync(); 
  if (convertEndian() == true) t2_swapBytes(&s, sizeof(short));
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (unsigned short s)
{
  if (convertEndian() == true) t2_swapBytes(&s, sizeof(unsigned short));
  bstr << s; sync(); 
  if (convertEndian() == true) t2_swapBytes(&s, sizeof(unsigned short));
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (int i)
{
  if (convertEndian() == true) t2_swapBytes(&i, sizeof(int));
  bstr << i; sync(); 
  if (convertEndian() == true) t2_swapBytes(&i, sizeof(int));
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (unsigned int i)
{
  if (convertEndian() == true) t2_swapBytes(&i, sizeof(unsigned int));
  bstr << i; sync(); 
  if (convertEndian() == true) t2_swapBytes(&i, sizeof(unsigned int));
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (long l)
{
  if (convertEndian() == true) t2_swapBytes(&l, sizeof(long));
  bstr << l; sync(); 
  if (convertEndian() == true) t2_swapBytes(&l, sizeof(long));
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (unsigned long l)
{
  if (convertEndian() == true) t2_swapBytes(&l, sizeof(unsigned long));
  bstr << l; sync(); 
  if (convertEndian() == true) t2_swapBytes(&l, sizeof(unsigned long));
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (float f)
{
  if (convertEndian() == true) t2_swapBytes(&f, sizeof(float));
  bstr << f; sync(); 
  if (convertEndian() == true) t2_swapBytes(&f, sizeof(float));
  return *this;
}

inline T2_UsrArgWriter &
T2_UsrArgWriter::operator << (double d)
{
  if (convertEndian() == true) t2_swapBytes(&d, sizeof(double));
  bstr << d; sync(); 
  if (convertEndian() == true) t2_swapBytes(&d, sizeof(double));
  return *this;
}

inline
T2_UsrArgWriter::T2_UsrArgWriter(const T2_UsrArgWriter& w): arg(w.arg), bstr()
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_UsrArgWriter copy constructor");
}

inline T2_UsrArgWriter&
T2_UsrArgWriter::operator =(const T2_UsrArgWriter&)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_UsrArgWriter operator =");
  return *this;
}



/////////////////////////////////////////////////////////////////////////
//
// T2_UsrArgReader
//
// A class to provide read capabilities for user-defined arguments.
//
/////////////////////////////////////////////////////////////////////////
class T2_UsrArgReader {
public:
  T2_UsrArgReader();
  T2_UsrArgReader(const T2_UsrArg& a, u_int offset = 0);
  ~T2_UsrArgReader();

  // attach the writer to a new T2_UsrArg obj
  void open(const T2_UsrArg& a, u_int offset = 0);

  // how many bytes of data in total are available to read ?
//  size_t getDataSize() const { return arg->getDataSize(); };
  size_t getDataSize() const { return bstr.getBufferSize(); };

  // how many bytes of data yet to be read ?
  size_t getDataSizeYetToRead() const { return bstr.getNumberBytesUnread(); };

  // rewind to set pointer back to the beginning of the buffer
  void rewind() { bstr.rewind(); };
  operator const void*() const { return (bstr)? this : 0; };

  // read and copy n bytes from T2_UsrArg into c[]
  T2_UsrArgReader& read(void* c, size_t n) { bstr.bread(c,n,1); return *this; };

  // return at most n bytes from T2_UsrArg without copying; 'actual' is set to
  // the actual number of bytes returned (the pointer into T2_UsrArg
  // will advance 'actual' bytes)
  const void* read(size_t n, size_t& actual);

  // read a string that was written into T2_UsrArg with
  // T2_UsrArgWriter::writeString and copy it into s[] (will end s[] with '\0')
  T2_UsrArgReader& readString(char* s);

  // read a string that was written into T2_UsrArg with
  // T2_UsrArgWriter::writeString without copy (the pointer into T2_UsrArg
  // will advance to skip the string); the returned string ends with '\0'
  const char* readString();

  T2_UsrArgReader & operator >> (bool& b) { bstr >> b; return *this; };

  T2_UsrArgReader & operator >> (char& c) { bstr >> c; return *this; };
  T2_UsrArgReader & operator >> (signed char& c) { bstr >> c; return *this; };
  T2_UsrArgReader & operator >> (unsigned char& c) { bstr >> c; return *this; };

  T2_UsrArgReader & operator >> (short& s) { bstr >> s; return *this; };
  T2_UsrArgReader & operator >> (unsigned short& s) { bstr >> s; return *this;};

  T2_UsrArgReader & operator >> (int& i) { bstr >> i; return *this; };
  T2_UsrArgReader & operator >> (unsigned int& i) { bstr >> i; return *this; };

  T2_UsrArgReader & operator >> (long& l) { bstr >> l; return *this; };
  T2_UsrArgReader & operator >> (unsigned long& l) { bstr >> l; return *this; };

  T2_UsrArgReader & operator >> (float& f) { bstr >> f; return *this; };
  T2_UsrArgReader & operator >> (double& d) { bstr >> d; return *this; };

  // do not use copy constructor and operator =
  T2_UsrArgReader(const T2_UsrArgReader& w);
  T2_UsrArgReader& operator =(const T2_UsrArgReader& w);

  // print all the bytes in the binary string
  // (for debugging purpose)
  ostream& writeBytesInBuffer(ostream& os) const
				{ return bstr.writeBytesInBuffer(os); };

protected:
  const T2_UsrArg* arg;
  T2_FIBStr bstr;

private:
};


inline
T2_UsrArgReader::T2_UsrArgReader(): arg(NULL), bstr()
{
}

inline
T2_UsrArgReader::T2_UsrArgReader(const T2_UsrArg& a, u_int offset):
				arg(&a), bstr()
{
  // make bstr point to the offset into the data buffer
  if (a.getDataSize() <= offset) {
     // offset goes beyond the data length
     bstr.open(a.getStartOfData(), 0);
  }
  else {
     bstr.open(a.getStartOfData()+offset, a.getDataSize()-offset);
  }
}

inline
T2_UsrArgReader::~T2_UsrArgReader()
{
  bstr.close();
}

inline void
T2_UsrArgReader::open(const T2_UsrArg& a, u_int offset)
{
  // close the current T2_UsrArg
  bstr.close();

  // attach to the new T2_UsrArg
  arg = &a;
  // make bstr point to the offset into the data buffer
  if (a.getDataSize() <= offset) {
     // offset goes beyond the data length
     bstr.open(a.getStartOfData(), 0);
  }
  else {
     bstr.open(a.getStartOfData()+offset, a.getDataSize()-offset);
  }
}

inline
T2_UsrArgReader::T2_UsrArgReader(const T2_UsrArgReader& w): arg(w.arg), bstr()
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_UsrArgReader copy constructor");
}

inline T2_UsrArgReader&
T2_UsrArgReader::operator =(const T2_UsrArgReader&)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use T2_UsrArgReader operator =");
  return *this;
}


inline const void*
T2_UsrArgReader::read(size_t n, size_t& actual)
{
  const void* p = (const void*) bstr.getCurrentPosition();
  actual = bstr.skip(n);
  return p;
}

inline T2_UsrArgReader&
T2_UsrArgReader::readString(char* s)
{
  while (bstr.getNumberBytesUnread() > 0) {
     *this >> *s;
     if (*s == '\0')
	return *this;
     s++;
  }
  *s = '\0';
  return *this;
}

inline const char*
T2_UsrArgReader::readString()
{
  const char* p = bstr.getCurrentPosition();
  while (bstr.getNumberBytesUnread() > 0) {
     char c;
     *this >> c;
     if (c == '\0')
	break;
  }
  return p;
}



#endif	/* _T2_USRARG_H_ */
