
/* $Id: dsreg.C,v 1.3 2000/03/16 00:48:11 chialin Exp $ */

/* dsreg.C */

#include	<stdlib.h>
#include	<ctype.h>
#include	"t2_dsreg.h"


/////////////////////////////////////////////////////////////////////////////
//
// T2_DSRegistrationEntry methods
//
/////////////////////////////////////////////////////////////////////////////
bool
T2_DSRegistrationEntry::read(T2_TxtFile& file)
{
  if (!file)
     return false;

  t2_Assert (file.getFileName() != NULL);

  // The format of an entry in the dataset registration file is as follows:
  //   char[] dataset_name
  //   char[] dataset_constructor_name
  //   char[] an optional filename for an associated object (blob) that
  //	      the application code is responsible for the correct interpretation
  //   char[] one_line_description
  //   <iterator names>
  //   char input_method_A	-- could be f (file redirect),
  //					    h (assignment by hostname)
  //					    r (assignment by processor rank)
  //   <entry part A>
  //   <record for index 1>
  //   <record for index 2>
  // 		:
  //   <end_of_record mark>
  //
  // where <iterator names> have the following format:
  //   i iterator_name_1		# i for iterator identifier
  //   i iterator_name_2
  //		:
  // and <end_of_record mark> is currently defined as "===", as defined by
  // T2_DSRegistrationFileReader::check_eor_mark()
  //
  // The format for the <entry part A> depends on the input_method_A:
  //
  // if input_method_A is 'f'
  //	char[] dataset_details_filename		-- read data from another file
  //
  // if input_method_A is 'd'
  //	<data file entry 1>
  //	<data file entry 2>
  //		:
  //	<aux file entry 1>
  //	<aux file entry 2>
  //		:
  //
  // where dataset file entries have the following format:
  //		d				-- label for dataset file
  //		char[] filename			-- filename for data file
  //		u_int logical_disk_id		-- logical id of assigned disk
  // and aux file entries have the following format:
  //		char[] filename				-- replicated aux file
  //		a				-- label for aux file
  //		char[] filename			-- filename for aux file
  //						   (replicated on all procs)
  //
  // The format for an index record is defined as follows, as defined in
  // T2_IndexRegistrationEntry::read():
  //   char [] index_name
  //   char [] index_constructor_name
  //   char [] one_line_description
  //   char input_method_B	-- could be f (file redirect),
  //					    d (listed in the remaining of this
  //					       file)
  //   <entry part B>
  //
  // The format for the <entry part B> depends on the input_method_B:
  // if input_method_B is 'f'
  //	char[] index_details_filename		-- read data from another file
  //
  // if input_method_B is 'd'
  //    <index file entry 1>
  //    <index file entry 2>
  //		:
  // where <index file entry> has the following format:
  //		i				-- label for index file
  //		char[] filename			-- filename for index file
  //		<logical_file_id_range>		-- ranges of logical data file
  //						   id's for supported data
  //						   files
  // where <logical_file_id_range> can have one of the following form:
  // 


  const char* string;

  // read dataset name
  while (1) {
     if ((string = file.getNextUncommentString()) == NULL)	// eof
        return false;

     if (T2_DSRegistrationFileReader::check_eor_mark(string) == false)
	break;

     // so this is an end-of-record marker; skip it and read next string
  }
  str_alloc_cpy(dataset_name, string);

  const u_int lineno = file.getCurrentLineNumber();
  
  // read dataset constructor name
  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(NULL, 0,
		"ERROR: no dataset constructor name is given for the entry at line %d in file '%s'",
		lineno, file.getFileName());
     clear();
     return false;
  }
  str_alloc_cpy(dsconstructor_name, string);

  // read blob filename
  file.getNextLine();
  if (file.eof() == true) {
     t2_err_ret(NULL, 0,
		"ERROR: incomplete record for the dataset entry at line %d in file '%s'.",
		lineno, file.getFileName());
     clear();
     return false;
  }
  if ((string = file.getStartOfNextUncommentString(file.getLine())) != NULL) {
     // blob filename is not null
     char* p = (char *) string;
     while (*p != '\0' && *p != file.getCommentMark() && !isspace((int)*p))
        p++;
     *p = '\0';
     str_alloc_cpy(blob_fname, string);
  }
  file.getNextLine();

  // read description for dataset
  if (file.eof() == true) {
     t2_err_ret(NULL, 0,
		"ERROR: incomplete record for the dataset entry at line %d in file '%s'.",
		lineno, file.getFileName());
     clear();
     return false;
  }
  if ((string = file.getStartOfNextUncommentString(file.getLine())) != NULL) {
     // description is not null, so find start of comment, if any, and ignore it
     char* p = (char *) string;
     while (*p != '\0' && *p != file.getCommentMark())
        p++;
     *p = '\0';
     str_alloc_cpy(description, string);
  }
  file.getNextLine();

  // read iterators, data files, and aux files
  if (read_iterator_datafile_auxfile_part(file, string, false) == false) {
     // error
     clear();
     return false;
  }

  // read all index entries until end-of-record mark is read or eof
  // note that string already contains the 1st entry for the index part
  while (string != NULL && 
	 T2_DSRegistrationFileReader::check_eor_mark(string) == false)
  {
     T2_IndexRegistrationEntry& idx = indexes.appendElement();
     const u_int idx_lineno = file.getCurrentLineNumber();

     // string now points to an index name
     str_alloc_cpy(idx.index_name, string);

     // read index constructor name
     if ((string = file.getNextUncommentString()) == NULL) {
	t2_err_ret(NULL, 0,
		"ERROR: no constructor name is given for the index entry at line '%d' in file '%s'.",
		idx_lineno, file.getFileName());
	clear();
	return false;
     }
     str_alloc_cpy(idx.index_constructor_name, string);

     // read description for index
     file.getNextLine();
     if (file.eof() == true) {
	t2_err_ret(NULL, 0,
		"ERROR: incomplete record for the dataset entry at line %d in file '%s'.",
		lineno, file.getFileName());
	clear();
	return false;
     }
     if ((string=file.getStartOfNextUncommentString(file.getLine())) != NULL) {
	// description is not null, so find start of comment, if any,
	// and ignore it
	char* p = (char *) string;
	while (*p != '\0' && *p != file.getCommentMark())
	   p++;
	*p = '\0';
	str_alloc_cpy(idx.description, string);
     }
     file.getNextLine();

     if (read_index_details(file, idx.idxfiles, string, false) == false) {
	clear();
	return false;
     }

  }		// while (string != NULL && check_eor_mark(string) == false)

  return true;
}


bool
T2_DSRegistrationEntry::
read_iterator_datafile_auxfile_part(T2_TxtFile& dfile, const char*& string,
				    bool disable_file_redirect)
{
  t2_Assert (dfile.getFileName() != NULL);

  // read iterators, data files, and aux files
  while ((string = dfile.getNextUncommentString()) != NULL) {
    if (string[0] == '\0' || string[1] != '\0') {	// not a label, done
	return true;
    }

    u_int item_lineno = dfile.getCurrentLineNumber();
    switch (string[0]) {
//	case 'I':
//	case 'i': {	// label for an iterator
//			// get iterator name
//			if ((string = dfile.getNextUncommentString()) == NULL) {
//			   t2_err_ret(NULL, 0,
//				"ERROR: no iterator name given after the iterator label '%s' at line %d in file '%s'.",
//				string, item_lineno, dfile.getFileName());
//			   return false;
//     			}
//			char*& iname = iters.appendElement();
//			str_alloc_cpy(iname, string);
//		  }
//		  break;

	case 'D':
	case 'd':	// label for a data file
		  if (read_data_file_line(dfile, item_lineno) == false) {
		     // error
		     return false;
		  }
		  break;

	case 'A':
	case 'a':	// label for an aux file
		  {
			if ((string = dfile.getNextUncommentString()) == NULL) {
			   t2_err_ret(NULL, 0,
			        "ERROR: no filename is given after aux file label at line %d in file '%s'",
			        item_lineno, dfile.getFileName());
			   return false;
			}
			char *& auxname = auxfiles.appendElement();
			str_alloc_cpy(auxname, string);
		  }
		  
		  break;

	case 'F':
	case 'f':	// label for a dataset detail file
		  if (disable_file_redirect == false) {
			if ((string = dfile.getNextUncommentString()) == NULL) {
			   t2_err_ret(NULL, 0,
				"ERROR: no filename given after the file-redirection label at line %d in file '%s'.",
				item_lineno, dfile.getFileName());
			   return false;
			}
			if (read_dataset_details_file(string) == false) {
			   // error
			   return false;
			}
		  }
		  else {
			t2_err_ret(NULL, 0,
				"ERROR: file redirection label at line %d in file '%s' is not allowed recursively inside a redirection file.",
				item_lineno, dfile.getFileName());
			return false;
		  }
		  break;

	default:	// not a label
		  return true;

    }		// switch (string[0])
  }		// while ((string = dfile.getNextUncommentString()) != NULL)
  return true;	// eof
}



bool
T2_DSRegistrationEntry::read_dataset_details_file(const char* fname)
{
  t2_Assert (fname != NULL);
  T2_TxtFile detailf('#', 2048);
  detailf.open(fname, true);
  if (!detailf) {
     t2_err_ret(NULL, 0,
		"ERROR: cannot open dataset details file '%s'", fname);
     return false;
  }

  // parse detail file, but disable file redirection
  const char* string;
  return read_iterator_datafile_auxfile_part(detailf, string, true);
}


bool
T2_DSRegistrationEntry::read_data_file_line(T2_TxtFile& dfile, u_int lineno)
{
  t2_Assert ((const void *) dfile != NULL);
  t2_Assert (dfile.getFileName() != NULL);

  // this method assumes that the label 'd' has already been read
  const char* string = dfile.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(NULL, 0,
		"ERROR: no data file is given after the data file label at line %d in file '%s'.",
		lineno, dfile.getFileName());
     return false;
  }

  // get data filename
  T2_NameIDPair& nd = datafiles.appendElement();
  str_alloc_cpy(nd.name, string);
  lineno = dfile.getCurrentLineNumber();

  // get logical disk id
  if ((string = dfile.getNextUncommentString()) == NULL) {
     t2_err_ret(NULL, 0,
		"ERROR: data file '%s' at line %d in file '%s' has no assigned logical disk id.",
		nd.name, lineno, dfile.getFileName());
     return false;
  }
  int atoi_err;
  nd.id = t2_atoi(string, &atoi_err);
  if (atoi_err != 0) {
     t2_err_ret(NULL, 0,
		"ERROR: invalid logical disk id '%s' at line %d in file '%s' for data file '%s'.",
		string, dfile.getCurrentLineNumber(), dfile.getFileName(),
		nd.name);
     return false;
  }
  return true;
}


bool
T2_DSRegistrationEntry::
read_index_details_file(const char* fname, T2_VArray<T2_NameIDRangePair>& files)
{
  t2_Assert (fname != NULL);
  T2_TxtFile detailf('#', 2048);
  detailf.open(fname, true);
  if (!detailf) {
     t2_err_ret(NULL, 0,
		"ERROR: cannot open index details file '%s'", fname);
     return false;
  }
  const char* string;
  return read_index_details(detailf, files, string, true);
}


bool
T2_DSRegistrationEntry::
read_index_details(T2_TxtFile& ifile, T2_VArray<T2_NameIDRangePair>& files,
		   const char*& string, bool disable_file_redirect)
{
  t2_Assert ((const void *) ifile != NULL);
  t2_Assert (ifile.getFileName() != NULL);

  while ((string = ifile.getNextUncommentString()) != NULL) {
     if (string[0] == '\0' || string[1] != '\0') {	// not a label
	return true;
     }

     u_int lineno = ifile.getCurrentLineNumber();

     switch (string[0]) {
	case 'I':
	case 'i':	// an index file label
		  {
			T2_NameIDRangePair& ni = files.appendElement();

			// get an index file name
			if ((string = ifile.getNextUncommentString()) == NULL) {
			   ni.name = NULL;
			   t2_err_ret(NULL, 0,
				"ERROR: no index filename is given after index file label at line %d",
				lineno);
			   return false;
			}
			str_alloc_cpy(ni.name, string);

			// read data file id ranges
			int min_fid, max_fid;
			while ((string=ifile.getNextUncommentString()) != NULL)
			{
			   if (T2_IDRange::stringToRange((char *) string,
							  min_fid, max_fid)
				== false)
			   {
			      t2_err_ret(NULL, 0,
				   "ERROR: invalid logical data file id '%s' at line %d in file '%s'.",
				   string, ifile.getCurrentLineNumber(),
				   ifile.getFileName());
			      return false;		// error
			   }

			   if (min_fid < 0)		// end of record
			      break;

			   T2_IDRange& idr = ni.ranges.appendElement();
			   idr.id_low = min_fid;
			   idr.id_high = max_fid;
			}
		  }
		  break;

	case 'F':
	case 'f':	// a file redirection label
		  if (disable_file_redirect == false) {
		     if ((string = ifile.getNextUncommentString()) == NULL) {
		        t2_err_ret(NULL, 0,
				"ERROR: no filename given after the file-redirection label at line %d in file '%s'.",
				lineno,ifile.getFileName());
		        return false;
		     }
		     if (read_index_details_file(string, files) == false)
		        return false;		// error
		  }
		  else {
		     t2_err_ret(NULL, 0,
				"ERROR: file redirection label at line %d in file '%s' is not allowed recursively inside a redirection file",
				lineno, ifile.getFileName());
		     return false;
		  }
		  break;

	default:	// not a label
		  return true;
     }

  }		// while ((string = ifile.getNextUncommentString()) != NULL) 

  return true;		// eof
}


bool
T2_DSRegistrationEntry::check_content_correct(u_int min_invalid_log_did) const
{
  u_int i;
  bool r = true;

  // check names
  if (dataset_name == NULL) {
     t2_err_ret(NULL, 0, "ERROR: dataset name is NULL");
     r = false;
  }
  if (dsconstructor_name == NULL) {
     t2_err_ret(NULL, 0, "ERROR: dataset constructor name is NULL");
     r = false;
  }

  // check data files
  for (i=0; i<datafiles.getNumberElements(); i++) {
     if (datafiles[i].name == NULL) {
	t2_err_ret(NULL, 0, "ERROR: %d-th data filename is NULL.", i);
	r = false;
     }
     if (datafiles[i].id >= min_invalid_log_did) {
	t2_err_ret(NULL, 0,
		"ERROR: %d-th data file is assigned to invalid logical disk id %d; valid logical disk id's are in the range of 0-%d.",
		i, datafiles[i].id, min_invalid_log_did-1);
	r = false;
     }
  }

  // check for auxfiles
  for (i=0; i<auxfiles.getNumberElements(); i++) {
     const char* p = auxfiles[i];
     if (p == NULL) {
	t2_err_ret(NULL, 0, "ERROR: %d-th aux filename is NULL.", i);
	r = false;
     }
  }

//  // check for iterators
//  for (i=0; i<iters.getNumberElements(); i++) {
//     const char* p = iters[i];
//     if (p == NULL) {
//	t2_err_ret(NULL, 0,
//		   "ERROR: %d-th dataset iterator name is NULL.", i);
//	r = false;
//     }
//  }

  // check index entries
  for (i=0; i<indexes.getNumberElements(); i++) {
     const T2_IndexRegistrationEntry& ie = indexes[i];
     if (ie.index_name == NULL) {
	t2_err_ret(NULL, 0, "ERROR: %d-th index name is NULL.", i);
	r = false;
     }
     if (ie.index_constructor_name == NULL) {
	t2_err_ret(NULL, 0,
		   "ERROR: %d-th index constructor name is NULL.", i);
	r = false;
     }
     for (u_int j=0; j<ie.idxfiles.getNumberElements(); j++) {
	const T2_NameIDRangePair& nir = ie.idxfiles[j];
	if (nir.name == NULL) {
	   t2_err_ret(NULL, 0,
		   "ERROR: %d-th index %d-th index filename is NULL.", i, j);
	   r = false;
	}
	for (u_int k=0; k<nir.ranges.getNumberElements(); k++) {
	   const T2_IDRange& idr = nir.ranges[k];
	   if ((u_int) idr.id_low >= datafiles.getNumberElements()) {
	      if (idr.id_low == idr.id_high) {
	         t2_err_ret(NULL, 0,
		     "ERROR: %d-th index %d-th index file supports invalid data file logical id %d; valid  data file logical id's are 0-%d.",
		     i, j, idr.id_low, datafiles.getNumberElements()-1);
	      }
	      else {
	         t2_err_ret(NULL, 0,
		     "ERROR: %d-th index %d-th index supports invalid data file logical id range %d-%d; valid data file logical id's are 0-%d.",
		     i, j, idr.id_low, idr.id_high,
		     datafiles.getNumberElements()-1);
	      }
	      r = false;
	   }
	   else if ((u_int) idr.id_high >= datafiles.getNumberElements()) {
	      t2_err_ret(NULL, 0,
		     "ERROR: %d-th index %d-th index supports invalid data file logical id range %d-%d; valid data file logical id's are 0-%d.",
		     i, j, idr.id_low, idr.id_high,
		     datafiles.getNumberElements()-1);
	      r = false;
	   }
	}
     }
  }

  return r;
}


ostream&
T2_DSRegistrationEntry::write(ostream& os) const
{
  u_int i;
  os << "dataset-name: " << t2_stringNullWrap(dataset_name) << endl
     << "constructor: " << t2_stringNullWrap(dsconstructor_name) << endl
     << "description: " << t2_stringNullWrap(description) << endl
     << "iterator names:" << endl;
//  for (i=0; i<iters.getNumberElements(); i++)
//     os << "	" << (char *) iters[i] << endl;
  os << "data-files: " << endl;
  for (i=0; i<datafiles.getNumberElements(); i++)
     os << "	" << t2_stringNullWrap(datafiles[i].name)
	<< "	" << datafiles[i].id << endl;
  os << "aux files: " << endl;
  for (i=0; i<auxfiles.getNumberElements(); i++)
     os << "	" << t2_stringNullWrap(auxfiles[i]) << endl;
  os << "indexes:" << endl;
  for (i=0; i<indexes.getNumberElements(); i++) {
     const T2_IndexRegistrationEntry& ie = indexes[i];
     os << "	index-name: " << t2_stringNullWrap(ie.index_name) << endl
	<< "	constructor: " << t2_stringNullWrap(ie.index_constructor_name)
	<< endl
	<< "	description: " << t2_stringNullWrap(ie.description)
	<< endl
	<< "	index-files:" << endl;
     for (u_int j=0; j<ie.idxfiles.getNumberElements(); j++) {
	const T2_NameIDRangePair& idr = ie.idxfiles[j];
	os << "		" << t2_stringNullWrap(idr.name) << "	";

	for (u_int k=0; k<idr.ranges.getNumberElements(); k++) {
	   os << " " << idr.ranges[k].id_low;
	   if (idr.ranges[k].id_low < idr.ranges[k].id_high)
	      os << "-" << idr.ranges[k].id_high;
	}
	os << endl;
     }
  }

  return os;
}



/////////////////////////////////////////////////////////////////////////////
//
// T2_DSRegistrationFileReader methods
//
/////////////////////////////////////////////////////////////////////////////
T2_DSRegistrationFileReader::T2_DSRegistrationFileReader(const char* fname)
				: file('#', 2048)
{
  if (fname != NULL)
     open(fname);
}

void
T2_DSRegistrationFileReader::open(const char* fname)
{
  close();
  if (fname == NULL)
     return;

  file.open(fname, true);	// true makes file remember its filename
}


bool
T2_DSRegistrationFileReader::getNextEntry(T2_DSRegistrationEntry& e)
{
  if (!file)
     return false;

  e.clear();
  return e.read(file);
}


bool
T2_DSRegistrationFileReader::check_eor_mark(const char* s)
{
  if (s != NULL && s[0] == '=' && s[1] == '=' && s[2] == '=')
     return true;
  return false;
}
