
/* $Id: nbsocket.C,v 1.11 2000/03/16 00:48:11 chialin Exp $ */

/* nbsocket.C */


#include	<string.h>
#include	"t2_nbsocket.h"


////////////////////////////////////////////////////////////////////////
//
// T2_nbSocket methods
//
////////////////////////////////////////////////////////////////////////

//////
//
// blocking connection
//
//////
bool
T2_nbSocket::connectByName(const char* hname, short port)
{
  if (connectPending == true) {		// alread has a pending request
     t2_Assert (connected() == false);
     errval = EINVAL;
     return false;
  }

  closeSocket();
  if (hname == NULL) {
     errval = EADDRNOTAVAIL;		// Can't assign requested address
     return false;
  }

  struct hostent *host;
  if ((host = gethostbyname(hname)) == NULL) {
     errval = h_errno;
     t2_err_ret(__FILE__, __LINE__,
		    "gethostbyname error for host %s (%s)",
		    hname, hstrerror(h_errno));
     return false;
  }
  t2_Assert (host->h_addr_list != NULL);
  char **addr = host->h_addr_list;

  bzero((void *) &remote, sizeof(remote));
  remote.sin_family = AF_INET;
  remote.sin_port   = htons(port);
  memcpy(&remote.sin_addr, *addr, sizeof(struct in_addr));

  if (connectBySockAddr(remote) == false) {
     t2_err_ret(__FILE__, __LINE__, "cannot connect to server %s", hname);
     return false;
  }

  return true;
}


bool
T2_nbSocket::connectByAddress(const char* ipaddr, short port)
{
  if (connectPending == true) {		// alread has a pending request
     t2_Assert (connected() == false);
     errval = EINVAL;
     return false;
  }

  closeSocket();
  if (ipaddr == NULL) {
     errval = EADDRNOTAVAIL;		// Can't assign requested address
     return false;
  }

  bzero((void *) &remote, sizeof(remote));
  remote.sin_family = AF_INET;
  remote.sin_port   = htons(port);
  if (!inet_pton(AF_INET, ipaddr, &remote.sin_addr)) {
     errval = errno;
     t2_err_ret(__FILE__, __LINE__,
		"T2_nbSocket::connectByAddress encounters inet_pton error");
     return false;
  }

  if (connectBySockAddr(remote) == false) {
     t2_err_ret(__FILE__, __LINE__, "cannot connect to server %s", ipaddr);
     return false;
  }

  return true;
}


bool
T2_nbSocket::connectBySockAddr(struct sockaddr_in& serv_addr)
{
  t2_Assert (connected() == false);
  if (openSocket() == false)
     return false;

  const int num_attempts = 10;
  int na;
  for (na=0; na<num_attempts; na++) {
     if (connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0)
     {
        errval = errno;

	if (errval != EINTR) {
           t2_err_ret(__FILE__, __LINE__, "cannot connect to server");
           closeSocket();
           return false;
	}
	usleep(100000);
     }
     else
	break;				// connect succeeded !
  }

  if (na == num_attempts) {
     t2_err_ret(__FILE__, __LINE__, "cannot connect to server");
     closeSocket();
     return false;
  }

  if (setFcntlFlag(sockfd, O_NONBLOCK) < 0) {	// set non-blocking fails
     errval = errno;
     t2_err_ret(__FILE__, __LINE__, "cannot set socket to non-blocking");
     closeSocket();
     return false;
  }

  isconnect = true;
  errval = 0;
  return true;
}


bool
T2_nbSocket::getPeerAddress(char* buf, size_t len)
{
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     return false;
  }

  if (buf == NULL) {
     errval = EADDRNOTAVAIL;		// Can't assign requested address
     return false;
  }

  if (inet_ntop(AF_INET, &(remote.sin_addr), buf, len) == NULL) {
     errval = errno;
     t2_err_ret(__FILE__, __LINE__,
		"T2_nbSocket::getPeerAddress encounters inet_ntop error");
     return false;
  }

  errval = 0;
  return true;
}


short
T2_nbSocket::getPeerPort()
{
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     return -1;
  }
  return ntohs(remote.sin_port);
}


//////
//
// blocking read/write
//
//////
ssize_t
T2_nbSocket::bwrite(const void *buf, int count, size_t size)
{
  if (wbuf != NULL) {			// a nbwrite not completed yet
     errval = EAGAIN;
     t2_Assert (wlen > 0);
     return -1;				// ignore the new write request
  }

  ssize_t nbytes_wrote;
  return bWrite(buf, count, size, nbytes_wrote);
}

ssize_t
T2_nbSocket::bread(void *buf, int count, size_t size)
{
  if (rbuf != NULL) {			// a nbread not completed yet
     errval = EAGAIN;
     t2_Assert (rlen > 0);
     return -1;				// ignore the new read request
  }

  ssize_t nbytes_read;
  return bRead(buf, count, size, nbytes_read);
}


//////
//
// non-blocking read/write
//
//////
ssize_t
T2_nbSocket::nbWrite(const void *buf, int count, size_t size)
{

  t2_say4(cout, "entered nbWrite for fd = " << sockfd << endl);

  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     t2_Assert (wbuf == NULL);
     t2_Assert (wlen == 0);

     t2_say4(cout,
	     "     left nbWrite fd = " << sockfd
	     << " due to ENOTCONN" << endl);

     return -1;
  }

  if (wbuf != NULL) {			// previous write not completed yet
     errval = EAGAIN;
     t2_Assert (wlen > 0);

     t2_say4(cout,
	    "     left nbWrite fd = " << sockfd
	    << " due to EAGAIN" << endl);

     return -1;				// ignore the new write request
  }

  t2_Assert (wbuf == NULL);
  t2_Assert (wlen == 0);
  wlen = count * size;
  if (buf == NULL && wlen > 0) {	// null pointer !!
     errval = EFAULT;
     wlen = 0;

     t2_say4(cout, 
	    "     left nbWrite fd = " << sockfd
	    << " due to EFAULT" << endl);

     return -1;				// ignore the new write request
  }


  t2_say4(cout,
	"     in nbWrite about to call write for fd = " << sockfd
	<< " with " << wlen << " bytes" << endl);

  // remember the write request, so that later can convert endian if needed
  if (outputConvertEndian() == true) {
     write_buf = buf;
     write_count = count;
     write_size = size;

     char* bb = (char *) write_buf;
     t2_swapBytesArray(bb, write_count, write_size);
  }

  ssize_t r;
  if ((r = write(sockfd, buf, wlen)) < 0) {
     if (errno != EWOULDBLOCK) {	// write error
	errval = errno;
	t2_err_ret(__FILE__, __LINE__,
		   "T2_nbSocket::nbWrite encounters write error");
	wlen = 0;

	t2_say4(cout,
	      "     left nbWrite fd = " << sockfd
		<< " with write returning " << r 
		<< " and errval = " << errval << endl);

	return r;
     }

     errval = 0;			// completely out of buffer
     wbuf = buf;

     t2_say4(cout, "     left nbWrite fd = " << sockfd
		   << " with write returning EWOULDBLOCK" << endl);

     return wlen;			// wlen remains unchanged
  }


  t2_say4(cout, "     left nbWrite fd = " << sockfd
	        << " with write returning " << r << endl);


  t2_Assert ((u_int) r <= wlen);	// r bytes were written
  if ((wlen -= r) == 0) {		// nbWrite completes
     t2_Assert (wbuf == NULL);
  }
  else
     wbuf = (const char *) buf + r;
  errval = 0;
  return wlen;
}


ssize_t
T2_nbSocket::testNBWrite()
{

  t2_say4(cout, "entered testNBWrite for fd = " << sockfd << endl);

  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     t2_Assert (wbuf == NULL);
     t2_Assert (wlen == 0);

     t2_say4(cout,
	    "     left testNBWrite fd = " << sockfd
	    << " due to ENOTCONN" << endl);

     return -1;
  }

  if (wbuf == NULL) {			// no pending write
     errval = EINVAL;
     t2_Assert (wlen == 0);

     t2_say4(cout,
	    "     left testNBWrite fd = " << sockfd 
	    << " due to EINVAL" << endl);

     return -1;				// ignore the new write request
  }


{
  struct timeval tval;
  tval.tv_sec = 0;
  tval.tv_usec = 0;
  fd_set wset;
  FD_ZERO(&wset);
  FD_SET(sockfd, &wset);
  int r;
  if ((r = select(sockfd+1, NULL, &wset, NULL, &tval)) < 0) {
     errval = errno;
     t2_err_ret(__FILE__, __LINE__, "select returns error %d", errval);
     wlen = 0;
     wbuf = NULL;

     t2_say4(cout, "     left testNBWrite fd = " << sockfd
		  << " due to select error errval = " << errval
		  << " and r = " << r << endl);

     return r;
  }
  if (r == 0) {
     errval = ETIMEDOUT;
     return wlen;
  }
  if (!FD_ISSET(sockfd, &wset)) {
     t2_err_ret(__FILE__, __LINE__, "FD_ISSET returns 0");

     t2_say4(cout, "     left testNBWrite fd = " << sockfd 
		  << " due to FD_ISSET returns 0" << endl);

     return wlen;
  }


  t2_say4(cout, "     in testNBWrite fd = " << sockfd 
	       << " about to call getsockopt" << endl);


  int error = 0;
  socklen_t len = sizeof(error);
  if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (char *) &error, &len) < 0) {
	errval = errno;			// Solaris pending error
	t2_err_ret(__FILE__, __LINE__,
		"T2_nbSocket::testNBWrite encounters getsockopt error");
	closeSocket();
	wlen = 0;
	wbuf = NULL;
//	exit(-1);
	return 0;
  }
  if (error) {			// just in case
	errno = error;
	t2_err_ret(__FILE__, __LINE__,
	    "T2_nbSocket::testNBWrite getsockopt illustrates pending error");
	closeSocket();
	wlen = 0;
	wbuf = NULL;
//	exit(-1);
	return 0;
  }
}
  

  t2_say4(cout,
         "     in testNBWrite fd = " << sockfd 
	 << " calling write now" << endl);


  t2_Assert (wbuf != NULL);
  t2_Assert (wlen > 0);
  ssize_t r;
  if ((r = write(sockfd, wbuf, wlen)) < 0) {
     if (errno != EWOULDBLOCK) {	// write error
	errval = errno;
	t2_err_ret(__FILE__, __LINE__,
		  "T2_nbSocket::testNBWrite encounters write error");
	wlen = 0;
	wbuf = NULL;
	return r;
     }

     errval = 0;			// completely out of buffer
     return wlen;			// wbuf and wlen remain unchanged
  }


  t2_say4(cout, "     in testNBWrite writes to fd = " << sockfd
		<< " " << wlen << " bytes and returns "
		<< r << endl);


  t2_Assert ((u_int) r <= wlen);	// r bytes were written
  if ((wlen -= r) == 0) {		// nbWrite completes
     wbuf = NULL;

     // need to convert endian back to what it was before?
     if (outputConvertEndian() == true) {
	char* bb = (char *) write_buf;
        t2_swapBytesArray(bb, write_count, write_size);
     }
  }
  else
     wbuf = (const char *) wbuf + r;
  errval = 0;
  return wlen;
}


ssize_t
T2_nbSocket::nbRead(void *buf, int count, size_t size)
{
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     t2_Assert (rbuf == NULL);
     t2_Assert (rlen == 0);
     return -1;
  }

  if (rbuf != NULL) {			// previous read not completed yet
     errval = EAGAIN;
     t2_Assert (rlen > 0);
     return -1;				// ignore the new read request
  }

  t2_Assert (rbuf == NULL);
  t2_Assert (rlen == 0);
  rlen = count * size;
  if (buf == NULL && rlen > 0) {	// null pointer !!
     errval = EFAULT;
     rlen = 0;
     return -1;				// ignore the new read request
  }

  // remember the read request, so that later can convert endian if needed
  read_buf = buf;
  read_count = count;
  read_size = size;

  ssize_t r;
  if ((r = read(sockfd, buf, rlen)) < 0) {
     if (errno != EWOULDBLOCK) {	// read error
	errval = errno;
	t2_err_ret(__FILE__, __LINE__,
		   "T2_nbSocket::nbRead encounters read error");
	rlen = 0;
	return r;
     }

     errval = 0;			// completely out of buffer
     rbuf = buf;
     return rlen;			// rlen remains unchanged
  }

  t2_Assert ((u_int) r <= rlen);	// r bytes were read
  if ((rlen -= r) == 0) {		// nbRead completes
     t2_Assert (rbuf == NULL);
  }
  else
     rbuf = (char *) buf + r;
  errval = 0;
  return rlen;
}


ssize_t
T2_nbSocket::testNBRead()
{
  if (connected() == false) {
     errval = ENOTCONN;			// Socket is not connected
     t2_Assert (rbuf == NULL);
     t2_Assert (rlen == 0);
     return -1;
  }

  if (rbuf == NULL) {			// no pending read
     errval = EINVAL;
     t2_Assert (rlen == 0);
     return -1;				// ignore the new read request
  }

  t2_Assert (rbuf != NULL);
  t2_Assert (rlen > 0);
  ssize_t r;
  if ((r = read(sockfd, rbuf, rlen)) < 0) {
     if (errno != EWOULDBLOCK) {	// read error
	errval = errno;
	t2_err_ret(__FILE__, __LINE__,
		   "T2_nbSocket::testNBRead encounters read error");
	rlen = 0;
	rbuf = NULL;
	return r;
     }

     errval = 0;			// completely out of buffer
     return rlen;			// rbuf and rlen remain unchanged
  }

  t2_Assert ((u_int) r <= rlen);	// r bytes were read
  if ((rlen -= r) == 0) {		// nbRead completes
     rbuf = NULL;

     // need to convert endian?
     if (inputConvertEndian() == true) {
        char* bb = (char *) read_buf;
        t2_swapBytesArray(bb, read_count, read_size);
     }
  }
  else
     rbuf = (char *) rbuf + r;
  errval = 0;
  return rlen;
}



//////
//
// non-blocking connection
//
//////
bool
T2_nbSocket::nbConnectByName(const char* hname, short port)
{
  closeSocket();
  if (hname == NULL) {
     errval = EADDRNOTAVAIL;		// Can't assign requested address
     return false;
  }

  struct hostent *host;
  if ((host = gethostbyname(hname)) == NULL) {
     errval = h_errno;
     t2_err_ret(__FILE__, __LINE__,
		    "gethostbyname error for host %s (%s)",
		    hname, hstrerror(h_errno));
     return false;
  }
  t2_Assert (host->h_addr_list != NULL);
  char **addr = host->h_addr_list;

  bzero((void *) &remote, sizeof(remote));
  remote.sin_family = AF_INET;
  remote.sin_port   = htons(port);
  memcpy(&remote.sin_addr, *addr, sizeof(struct in_addr));

  return nbConnectBySockAddr(remote);
}


bool
T2_nbSocket::nbConnectByAddress(const char* ipaddr, short port)
{
  closeSocket();
  if (ipaddr == NULL) {
     errval = EADDRNOTAVAIL;		// Can't assign requested address
     return false;
  }

  bzero((void *) &remote, sizeof(remote));
  remote.sin_family = AF_INET;
  remote.sin_port   = htons(port);
  if (!inet_pton(AF_INET, ipaddr, &remote.sin_addr)) {
     errval = errno;
     t2_err_ret(__FILE__, __LINE__,
		"T2_nbSocket::nbConnectByAddress encounters inet_pton error");
     return false;
  }

  if (nbConnectBySockAddr(remote) == false) {
     t2_err_ret(__FILE__, __LINE__, "cannot connect to server %s", ipaddr);
     return false;
  }

  return true;
}


bool
T2_nbSocket::nbConnectBySockAddr(struct sockaddr_in& serv_addr)
{
  t2_Assert (connected() == false);
  t2_Assert (connectPending == false);
  if (openSocket() == false)
     return false;

  if (setFcntlFlag(sockfd, O_NONBLOCK) < 0) {	// set non-blocking fails
     errval = errno;
     t2_err_ret(__FILE__, __LINE__,
			"cannot set socket to non-blocking");
     closeSocket();
     return false;
  }

  int n;
  if ((n = connect(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)))
       < 0)
  {
     if (errno == EINPROGRESS) {
	t2_Assert (connected() == false);
	connectPending = true;
	errval = 0;
	return true;
     }

     // connect error
     errval = errno;
     t2_err_ret(__FILE__, __LINE__, "non-blocking connect error");
     closeSocket();
     t2_Assert (connected() == false);
     t2_Assert (connectPending == false);
     return false;
  }

  t2_Assert (isconnect == false);
  t2_Assert (connectPending == false);
  if (n == 0)			// wow, connect completed immediately !!
     isconnect = true;
  else				// becomes a pending connect request
     connectPending = true;
  errval = 0;
  return true;
}


bool
T2_nbSocket::testConnect(struct timeval* tval)
{
  if (sockfd < 0 || connectPending == false) {	// no pending connect request
     t2_Assert (sockfd >= 0 || connectPending == false);
     closeSocket();
     errval = EINVAL;
     return true;
  }
  t2_Assert (isconnect == false);

  // check if connect completed
  fd_set rset, wset;
  FD_ZERO(&rset);
  FD_SET(sockfd, &rset);
  wset = rset;

  int n, error = 0;
  if ((n = select(sockfd+1, &rset, &wset, NULL, tval)) == 0)
     return false;			// still in progress

  if (FD_ISSET(sockfd, &rset) || FD_ISSET(sockfd, &wset)) {
     socklen_t len = sizeof(error);
     if (getsockopt(sockfd, SOL_SOCKET, SO_ERROR, (char *) &error, &len) < 0) {
	errval = errno;			// Solaris pending error
	t2_err_ret(__FILE__, __LINE__,
		"T2_nbSocket::testConnect encounters getsockopt error");
	connectPending = false;
	closeSocket();
	return true;
     }
  }
  else {			// Stevens says it's because sockfd not set
     errval = errno;
     t2_err_ret(__FILE__, __LINE__, "select error");
     connectPending = false;
     t2_Assert (connected() == false);
     closeSocket();
     return true;
  }

  connectPending = false;	// no pending connect anymore

  if (error) {			// just in case
     closeSocket();
     errval = error;
     return true;
  }

  isconnect = true;		// connect succeeded
  errval = 0;
  return true;
}



//////
//
// transfer connection
//
//////
void
T2_nbSocket::transfer(T2_Socket& s)
{
  closeSocket();

  // get data from s
  memcpy((void *) &remote, (const void *) &s.remote, sizeof(remote));
  connectPending = false;
  rbuf = NULL;
  wbuf = NULL;
  rlen = wlen = 0;
  T2_Sock::operator = (s);
  setInputConvertEndian(s.inputConvertEndian());
  setOutputConvertEndian(s.outputConvertEndian());
  if (connected())				// set non-blocking
     setFcntlFlag(sockfd, O_NONBLOCK);

  // clear s
  bzero((void *) &s.remote, sizeof(s.remote));
  s.nbytes_read = 0;
  s.nbytes_wrote = 0;
  s.sockfd = -1;
  s.errval = 0;
  s.isconnect = false;
}
