
/* $Id: packets.C,v 1.16 2000/03/16 00:48:11 chialin Exp $ */

/* packets.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/26/98
 * methods dealing with packets exchanged between front-end and back-end
 */


#include	<unistd.h>
#include	"t2_bfile.h"
#include	"t2_packets.h"


///////////////////////////////////////////////////////////////////////////
//
// T2_FEBEPacket methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_FEBEPacket::send(T2_Socket& s) const
{
  t2_Assert (s);

  s << getTag();
  if (s.getNumberOfBytesLastWritten() == sizeof(tag))
     return true;
  return false;
}

u_char
T2_FEBEPacket::recv(T2_Socket& s)
{
  t2_Assert (s);

  u_char t;
  s >> t;

  if (s.getNumberOfBytesLastRead() == sizeof(t))
     return t;
  return T2_ErrorFEBEPacketTag;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_ConnectPacket methods
//
// The protocol for a T2 back-end node to connect to the T2 front-end
// is defined as follows.
//	T2-FE			T2-BE
//				connect
//	accept
//				send T2_ConnectPacketTag, be-endian
//	read T2_ConnectPacketTag,
//	     be-endian
//	send fe-endian
//				read fe-endian
//				send numnodes, mynode
//	read numnodes, mynode
//
///////////////////////////////////////////////////////////////////////////
bool
T2_ConnectPacket::send(T2_Socket& s) const
{
  if (!s)
     return false;

  if (T2_FEBEPacket::send(s) == false)
     return false;

  // send T2 back-end node endian
  s << my_endian;
  if (s.getNumberOfBytesLastWritten() != (ssize_t) sizeof(my_endian)) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: socket write error when writing connection packet.");
     return false;
  }

  // read T2 front-end endian
  u_char t2fe_endian;
  s >> t2fe_endian;
  if (s.getNumberOfBytesLastRead() < (ssize_t) sizeof(t2fe_endian)) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: socket read error when reading T2 front-end endian.");
     return false;
  }
  t2_Assert ((int) t2fe_endian == T2_BigEndian ||
	     (int) t2fe_endian == T2_LittleEndian);
  t2_say(cout, "T2 front-end is " << endian2string(t2fe_endian) << endl);
  
  // make output socket do endian conversion, if necessary
  if (t2fe_endian != my_endian) {
     t2_Assert (s.outputConvertEndian() == false);
     s.setOutputConvertEndian(true);
  }

  // send number of back-end nodes and my node rank
  s << numnodes << mynode;
  if (s.getNumberOfBytesLastWritten() < (ssize_t) sizeof(mynode)) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: socket write error when writing back-end node information.");
     return false;
  }

  return true;
}


u_char
T2_ConnectPacket::recv(T2_Socket& s, u_char fe_endian)
{
  if (!s)
     return T2_ErrorFEBEPacketTag;

  // read tag
  u_char t = T2_FEBEPacket::recv(s);
  if (t != T2_ConnectPacketTag)
     return t;

  // read back-end endian
  s >> my_endian;
  if (s.getNumberOfBytesLastRead() != sizeof(my_endian))
     return T2_ErrorFEBEPacketTag;

  // send T2 front-end endian
  s << fe_endian;
  if (s.getNumberOfBytesLastWritten() != (ssize_t) sizeof(fe_endian)) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: socket write error when writing front-end endian.");
     return false;
  }

  // read number of back-end nodes and the back-end node rank of the 
  // connecting node
  s >> numnodes;
  if (s.getNumberOfBytesLastRead() != sizeof(numnodes))
     return T2_ErrorFEBEPacketTag;

  s >> mynode;
  if (s.getNumberOfBytesLastRead() != sizeof(mynode))
     return T2_ErrorFEBEPacketTag;

  return T2_ConnectPacketTag;		// every thing goes right ...
}



///////////////////////////////////////////////////////////////////////////
//
// T2_AcceptPacket methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_AcceptPacket::send(T2_Socket& s) const
{
  if (!s)
     return false;

  if (T2_FEBEPacket::send(s) == false)
     return false;

  return true;
}


u_char
T2_AcceptPacket::recv(T2_Socket& s)
{
  if (!s)
     return T2_ErrorFEBEPacketTag;

  u_char t = T2_FEBEPacket::recv(s);
  if (t != T2_AcceptPacketTag) {
     t2_err_ret(__FILE__, __LINE__,
		"Expecting AcceptPacketTag %d but recvs tag = %d",
		T2_AcceptPacketTag, (int) t);
     return t;
  }

  return T2_AcceptPacketTag;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_ReadyPacket methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_ReadyPacket::send(T2_Socket& s) const
{
  if (!s)
     return false;

  if (T2_FEBEPacket::send(s) == false)
     return false;

  return true;
}


u_char
T2_ReadyPacket::recv(T2_Socket& s)
{
  if (!s)
     return T2_ErrorFEBEPacketTag;

  u_char t = T2_FEBEPacket::recv(s);
  if (t != T2_ReadyPacketTag) {
     t2_err_ret(__FILE__, __LINE__,
		"Expecting ReadyPacketTag %d but recvs tag = %d",
		T2_ReadyPacketTag, (int) t);
     return t;
  }

  return T2_ReadyPacketTag;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_ShutdownPacket methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_ShutdownPacket::send(T2_Socket& s) const
{
  if (!s)
     return false;

  if (T2_FEBEPacket::send(s) == false)
     return false;

  return true;
}


//u_char
//T2_ShutdownPacket::recv(T2_Socket& s)
//{
//  if (!s)
//     return T2_ErrorFEBEPacketTag;
//
//  int t = T2_FEBEPacket::recv(s);
//  if (t != T2_ShutdownPacketTag) {
//     t2_err_ret(__FILE__, __LINE__,
//		"Expecting ShutdownPacketTag %d but recvs tag = %d",
//		T2_ShutdownPacketTag, t);
//     return t;
//  }
//
//  return T2_ShutdownPacketTag;
//}



///////////////////////////////////////////////////////////////////////////
//
// T2_QueryAckPacket methods
//
///////////////////////////////////////////////////////////////////////////
static u_int cnt = 0;

bool
T2_QueryAckPacket::send(T2_Socket& s) const
{
  if (!s)
     return false;

  if (T2_FEBEPacket::send(s) == false)
     return false;

# ifdef T2_ASSERTION
     t2_say2(cout, "sending queryack cnt = " << cnt << endl);
     s << cnt++;
#endif

  return true;
}


u_char
T2_QueryAckPacket::recv(T2_Socket& s)
{
  if (!s)
     return T2_ErrorFEBEPacketTag;

  u_char t = T2_FEBEPacket::recv(s);
  if (t != T2_QueryAckTag) {
     t2_err_ret(__FILE__, __LINE__,
		"Expecting QueryAckTag %d but recvs tag = %d",
		T2_QueryAckTag, (int) t);
     return t;
  }

  return T2_QueryAckTag;
}




///////////////////////////////////////////////////////////////////////////
//
// T2_BEClientHelloPacket methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_BEClientHelloPacket::send(T2_Socket& s) const
{
  if (!s)
     return false;

  s << numnodes << mynode;
  return true;
}


bool
T2_BEClientHelloPacket::recv(T2_Socket& s)
{
  if (!s)
     return false;

  s >> numnodes >> mynode;
  return true;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_BEClientByePacket methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_BEClientByePacket::send(T2_Socket& s) const
{
  if (!s)
     return false;

  s << T2_BEClientByeTag;
  return true;
}


bool
T2_BEClientByePacket::send(T2_nbSocket& s) const
{
  if (!s)
     return false;

  u_char tag = T2_BEClientByeTag;
  s.bwrite((const void *) &tag, 1, sizeof(tag));
  return true;
}


bool
T2_BEClientByePacket::recv(T2_Socket& s)
{
  if (!s)
     return false;

  u_char tag;
  s >> tag;
  if (tag == T2_BEClientByeTag)
     return true;
  return false;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_BEConnection methods
//
///////////////////////////////////////////////////////////////////////////
T2_BEConnection::T2_BEConnection(): sock()
{
}


T2_BEConnection::~T2_BEConnection()
{
  shutdown();
}


void
T2_BEConnection::shutdown()
{
  if (sock) {
     T2_ShutdownPacket p;
#    ifdef T2_ASSERTION
        bool r = 
#    endif
     sendShutdownPacket(p);
     t2_Assert (r == true);
     sock.closeSocket();
  }
}


void
T2_BEConnection::transfer(T2_BEConnection& c)
{
  sock.transfer(c.sock);
}


bool
T2_BEConnection::sendQBatch(T2_QBatch& q)
{
  if (!sock) {
     t2_err_ret(__FILE__, __LINE__, "socket not connected");
     return false;
  }

  sock << T2_QueryPacketTag;		// send tag first

  q.bwrite(sock);			// write qbatch to a blocking socket
  if (sock.getErrorValue() > 0) {
     t2_err_ret(__FILE__, __LINE__, "write qbatch to socket error");
     return false;
  }
  return true;
}


bool
T2_BEConnection::sendQBatch(T2_QPtrBatch& q)
{
  if (!sock) {
     t2_err_ret(__FILE__, __LINE__, "socket not connected");
     return false;
  }

  t2_say3(cout,
	 "sending out to backend node query packet tag " 
	 << T2_QueryPacketTag << endl);

  sock << T2_QueryPacketTag;		// send tag first

  q.bwrite(sock);			// write qbatch to a blocking socket
  if (sock.getErrorValue() > 0) {
     t2_err_ret(__FILE__, __LINE__, "write qbatch to socket error");
     return false;
  }

  return true;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_FEConnection methods
//
///////////////////////////////////////////////////////////////////////////
T2_FEConnection::T2_FEConnection(): sock()
{
}


T2_FEConnection::~T2_FEConnection()
{
  sock.closeSocket();
}


void
T2_FEConnection::transfer(T2_FEConnection& c)
{
  sock.transfer(c.sock);
}


void
T2_FEConnection::shutdown()
{
  sock.closeSocket();
}


bool
T2_FEConnection::connectFrontEnd(const char* hname, short port,
				 int nprocs, int myproc, u_char endian)
{
  const u_int num_tries = 10;
  u_int i;
  for (i=0; i<num_tries; i++) {
     bool r = sock.connectByName(hname, port);		// connect socket
     if (r == true)					// succeeded !
	break;
     t2_err_ret(__FILE__, __LINE__, "sock.connectByName error, will retry");
     usleep(1000 + myproc*100);
  }

  if (i == num_tries) {
     t2_err_ret(__FILE__, __LINE__,
		"sock.connectByName failed %d times", num_tries);
     return false;
  }

  t2_Assert (endian == T2_BigEndian || endian == T2_LittleEndian);

  // send connect packet
  T2_ConnectPacket p(nprocs, myproc, endian);
  bool r = sendConnectPacket(p);
  if (r == false) {
     t2_err_ret(__FILE__, __LINE__, "sendConnectPacket error");
     sock.closeSocket();
     return false;
  }

  // expect accept packet
  T2_AcceptPacket a;
  u_char t = recvAcceptPacket(a);
  if (t != T2_AcceptPacketTag) {
     t2_err_ret(__FILE__, __LINE__, "recvAcceptPacket error");
     sock.closeSocket();
     return false;
  }

  // protocol is done
  return true;
}


u_char
T2_FEConnection::recvPacketTag()
{
  if (!sock) {
     t2_err_ret(__FILE__, __LINE__, "socket not connected");
     return T2_ErrorFEBEPacketTag;
  }

  u_char tag = T2_ErrorFEBEPacketTag;
  sock >> tag;
  return tag;
}


bool
T2_FEConnection::recvQBatch(T2_QBatch& q)
{
  if (!sock) {
     t2_err_ret(__FILE__, __LINE__, "socket not connected");
     return false;
  }

  q.bread(sock);		// read from blocking socket
  if (sock.getErrorValue() > 0) {
     t2_err_ret(__FILE__, __LINE__, "read qbatch from socket error");
     return false;
  }
  return true;
}


bool
T2_FEConnection::sendQBatchAck(const u_int mynode)
{
  if (mynode != 0) 		// for now, only node 0 sends back the ack
     return true;

  if (!sock) {
     t2_err_ret(__FILE__, __LINE__, "socket not connected");
     return false;
  }

  T2_QueryAckPacket ack;
  return ack.send(sock);
}
