
/* $Id: svm-front-end.C,v 1.5 2000/10/20 21:31:11 chialin Exp $ */

/* svm-front-end.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 1/22/2000
 * An example of a sequential application front-end process (a version that
 * uses standard T2_FrontEnd interface) for the SVM example.
 */


#include	<unistd.h>
#include	<sys/types.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	"t2_bfile.h"
#include	"t2_fe.h"
#include	"t2_frontend.h"
#include	"svm_front_end.h"


/* define _CONVERT_ENDIAN if this program will be running in
   little_endian machine: clients are assumed to be java clients,
   so they are big endian
*/

#if (!defined T2_LOCAL_MACHINE_BIG_ENDIAN && !defined T2_LOCAL_MACHINE_LITTLE_ENDIAN)
#   error Either T2_LOCAL_MACHINE_BIG_ENDIAN or T2_LOCAL_MACHINE_LITTLE_ENDIAN must be defined according to the endian-ness of the local machine.
#endif

#ifdef T2_LOCAL_MACHINE_LITTLE_ENDIAN
#   define _CONVERT_ENDIAN   
#endif



/////////////////////////////////////////////////////////////////////////////
//
// some help function prototypes and variables
//
/////////////////////////////////////////////////////////////////////////////
enum usageError { eUNKNOWN, eBADARG };
static void printusage(const char*);
static void badusage(usageError, const char*, char);

const char default_t2fe_host[]	= "condor.cs.umd.edu";
const u_int default_t2fe_port	= 9987,
	    default_t2_nqry	= 10;
const int default_sec		= 5;
const u_int default_client_port = 9986;

void use_interactive_query(T2_FrontEnd& fe);
bool
svm_inquire_functions(T2_FrontEnd& fe, const char* keyword,
		      const char* thumbnail_dir,
		      u_int& pid, u_int& accid, u_int& aggid),
svm_inquire_datasets(T2_FrontEnd& fe, const char* keyword,
		     u_int pid, u_int accid, u_int aggid,
		     const char* thumbnail_dir,
		     svm_ImageInfoCollection& allimages);
void AppFrontEnd(T2_FrontEnd& fe, svm_ImageInfoCollection& allimages, u_int client_port, const char* dump_qfile);


/////////////////////////////////////////////////////////////////////////////
//
// main
//
/////////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
 int t2fe_port				// port that T2 front-end listens to
     = default_t2fe_port,
     t2_nqry 				// # query batches to send to T2
     = default_t2_nqry,
     client_port
     = default_client_port;
 
 
 const char *hostname = default_t2fe_host,
     *qfile = NULL,
     *dump_qfile = NULL;		// dump query spec into given file
 bool  interactive_query = false;	// generate queries interactively?
 bool  app_front_end = false;		// act as app_front_end

 int sec = default_sec;

     // deal with the command-line arguments
 int i;
 for (i=0; i<argc; i++)
     {
     if (argv[i][0] == '-')
         {
         switch(argv[i][1])
             {
             case 'h':
                 printusage(argv[0]); return 0; break;

             case 'f':
                 i++;
                 if (i >= argc)
                     { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
                 hostname = argv[i];
                 i++;
                 if (i >= argc)
                     { badusage(eBADARG, argv[0], argv[i-2][1]); return -1; }
                 t2fe_port = atoi(argv[i]);
                 if (t2fe_port < 0)
                     { badusage(eBADARG, argv[0], argv[i-2][1]); return -1; }
                 break;

             case 'n':
                 i++;
                 if (i >= argc)
                     { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
                 t2_nqry = atoi(argv[i]);
                 if (t2_nqry < 0)
                     { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
                 break;

             case 'd':
                 i++;
                 if (i >= argc)
                     { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
                 dump_qfile = argv[i];
                 break;

             case 'q':
                 i++;
                 if (i >= argc)
                     { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
                 qfile = argv[i];
                 interactive_query = false;
                 break;

             case 'i':
                 i++;
                 qfile = NULL;
                 interactive_query = true;
                 break;

             case 'a':
                 i++;
                 if (i >= argc)
                     { badusage(eBADARG, argv[0], argv[i-2][1]); return -1; }
                 client_port = atoi(argv[i]);
                 if (client_port < 0)
                     { badusage(eBADARG, argv[0], argv[i-2][1]); return -1; }

                 qfile = NULL;
                 interactive_query = false;
                 app_front_end = true;
                 break;

             case 's':
                 i++;
                 if (i >= argc)
                     { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
                 sec = atoi(argv[i]);
                 if (t2_nqry < 0)
                     { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
                 break;

             default:
                 badusage(eUNKNOWN, argv[0], argv[i][1]); return -1; break;
             }
         }
     }

 cout << "T2 front-end host = " << hostname
      << ", port = " << t2fe_port << endl;
 if (qfile == NULL)
     if (interactive_query == true) 
	cout << "Will allow menu-driven interactive queries from users"
	     << endl;
     else if (app_front_end == true)
	cout << "Will wait for application client"
	     << endl;
     else
        cout << "Will issue " << t2_nqry << " empty query batches to back-end"
             << endl;
 else
     {
     cout << "Will read query batch file " << qfile << " and send the query"
          << endl
          << "batch in the file to the back-end " << t2_nqry << " times"
          << endl;
     }

     // only one of interactive_query and qfile can be set
 t2_Assert (interactive_query == false || qfile == NULL);

     // connecting to T2 front-end
 T2_FrontEnd fe;
 if (fe.connectT2FrontEndByHostname(hostname, t2fe_port) == false)
     {
         // cannot connect to T2 front-end
     if (fe.getErrorVal() == t2_feT2Busy)
         cout << "T2 front-end is busy" << endl;
     else
         t2_err_ret(__FILE__, __LINE__,
                    "fe.connectT2FrontEndByHostname error: %s",
                    fe.errorValToString());
     return -1;
     }

 cout << "Connected to T2 front-end" << endl;
 cout << "T2 currently has " << fe.getNumberBackEndNodes()
      << " back-end nodes." << endl
      << "I am in " << endian2string(fe.getMyEndianess())
      << " while T2 back-end is in " 
      << endian2string(fe.getT2BackEndEndianess())
      << endl;


     // inquire T2 front-end about constructors
 const char svm_keyword[] = "t2-svm-example";
 const char *thumbnail_dir = NULL;
 u_int pid, accid, aggid;
 if (svm_inquire_functions(fe, svm_keyword, thumbnail_dir, pid, accid, aggid)
     == false)				// inquiry fails
     return -1;

 cout	<< "done inquiring functions, pid=" << pid
	<< " accid= " << accid << " aggid=" << aggid << endl;

     // inquire T2 front-end about datasets
 svm_ImageInfoCollection allimages;
 if (svm_inquire_datasets(fe, svm_keyword, pid, accid, aggid, thumbnail_dir,
                          allimages)
     == false)
     return -1;

 cout << "done inquiring datasets:" << endl << allimages << endl;

 if (qfile == NULL)
     {
     if (interactive_query == true)
         use_interactive_query(fe);
     else if (app_front_end == true)
         AppFrontEnd(fe, allimages, client_port, dump_qfile);
     else
         {
         for (int j=0; j<sec; j++)
             {
             bool r = fe.pingT2FrontEnd();
             if (r == true)
                 cout << "ping " << j << " returns true" << endl;
             else
                 {
                 cout << "ping " << j << " returns false" << endl;
                 cout << "fe error: " << fe.errorValToString() << endl;
                 cout << "aborting ....." << endl;
                 return -1;
                 }
             sleep(1);
             }
         }
     }
 else
     {
     T2_IBFile qibf(qfile);
     if (!qibf)
         {
         t2_err_ret(__FILE__, __LINE__,
                    "Cannot open query batch file %s", qfile);
         }
     else
         {
         T2_QBatch q;
         qibf >> q;
         qibf.close();

         if (q.isValid() == false)
             {		// qbatch read error
             t2_err_ret(__FILE__, __LINE__, "Query batch read error");
             }
         else
             {
             for (int k=0; k<t2_nqry; k++)
                 {
                 cout << "sending query batch " << k << " times" << endl;
                 if (fe.submitQBatch(q) == false)
                     {
                     t2_err_ret(__FILE__, __LINE__,
                                "submitQBatch error %d", (int) fe.getErrorVal());
                     break;
                     }
                 sleep(1);
                 }
             }				// if (q.isValid()==false) else
         }					// if (!qibf) else
     }					// if (qfile == NULL) else




 cout << endl << "now disconnect from T2 front-end" << endl;
 fe.disconnectT2FrontEnd();

 cout << "program " << argv[0] << " exits gracefully" << endl;
 return 0;
}



/////////////////////////////////////////////////////////////////////////////
//
// help functions
//
/////////////////////////////////////////////////////////////////////////////
void
printusage(const char* pname)
{
  cout << "Usage: " << pname
       << " [-h] |"
       << " [-f <T2 front-end hostname> <port number>]" << endl
       << " [-a <client-connection-port-number>]" << endl
       << " [-d <query batch file>]" << endl
       << " [-i]"
       << " [-q <query batch file>]"
       << " [-n <# query batches>]" << endl
       << " [-s <# times to ping>]" << endl
       << endl;

  cout << "Where:" << endl;
  cout << "-h prints this message" << endl;
  cout << "-f specifies the hostname and port number that T2 front-end listens to" << endl
       << "   (default = " << default_t2fe_host << ", "
       << default_t2fe_port  << ")" 
       << endl;
  cout << "-a specifies the port number that this program listens to" << endl
       << "   (default = " << default_client_port << ")" << endl;
  cout << "-d if specified, dumps the query into the given output file" << endl
       << "   (usually used for debugging purpose)" << endl;
  cout << "-q specifies the query batch file" << endl
       << "   (this option disables -i option before it, default = NULL)"
       << endl;
  cout << "-i interactively creates queries within the given image boundary"
       << endl
       << "   (this option disables the -q option before it)"
       << endl;
  cout << "-n when -q is used to specify the query batch file, -n specifies the"
       << endl
       << "   number of times the query batch file is read and the query batches"
       << endl
       << "   stored in the file are sent to the back-end before shutting down;"
       << endl
       << "   when -q is not specified, -n specifies the number of times an empty"
       << endl
       << "   query batch file is sent to the back-end before shutting down"
       << endl
       << "   (default = " << default_t2_nqry  << ")" << endl
       << endl;
  cout << "-s when -q is not set, -s specifies the number of times to ping the"
       << endl
       << "   the T2 front-end (default = " << default_sec << ")" << endl;
}


void
badusage(usageError e, const char* p, char a)
{
  switch (e) {
    case eUNKNOWN:
        cerr << "Unknown parameter for -" << a << " in command line" << endl;
        break;
    case eBADARG:
        cerr << "Missing argument for parameter -" << a << " in command line"
             << endl;
        break;
    default:
        cerr << "Some kind of error in the parameter -" << a
             << " in command line" << endl;
  }
  printusage(p);
}



/////////////////////////////////////////////////////////////////////////////
//
// get_choice
//
// Read an integer from cin between min and max.
//
/////////////////////////////////////////////////////////////////////////////
int
get_choice(const char* msg, int min, int max)
{
  int choice;
  do {
     if (msg != NULL)
     cout << msg << "  ";
     cout.flush();
     cin >> choice;
     if (choice >=min && choice <=max)		// a valid choice
		break;
     cout << "invalid choice, please try again ....." << endl;
  } while (1);
  return choice;
}


/////////////////////////////////////////////////////////////////////////////
//
// use_interactive_query
//
// This function would allow the user to interactively generate a query
// to the T2 front-end. Only one query per query batch, and only one
// dataset per query.
//
/////////////////////////////////////////////////////////////////////////////
void
use_interactive_query(T2_FrontEnd& fe)
{
  const u_int buflen = 128;
  char buf[buflen];			// just a buffer

  const u_int nqueries = 1;
  T2_QBatch qbatch(nqueries);		// only one query per batch
  T2_QSpec& qspec = qbatch.getQuerySpec(0);
  qspec.setNumberDatasets(1);		// only one input dataset

  T2_UsrArgWriter write_arg;		// user argument writer

  cout  << endl
	<< "Interactive query creation: use the constructor catalog,"
	<< endl
	<< "the dataset catalog, and the index catalog to make sure you have all the id's"
	<< endl
	<< "correct."
	<< endl
	<< "You can setup the basic query information and just use different output query"
	<< endl
	<< "windows to generate different queries into the same dataset. You can also"
	<< endl
	<< "reset the basic query information when you want to."
	<< endl << endl;

  qspec.useBigEndianUserArg();		// use big endian for now

  u_int packet_number, packet_type = 6;
  u_int menu_selection = 1;
  do {
     // get a reference to the dataset specification part of the query
     T2_QSpecDataset& dataset_spec = qspec.getDatasetSpec(0);

     // reset basic information if necessary
     if (menu_selection == 1) {

	// get information about aggregation function:
	// 	aggregation function id (from constructor catalog)
	//	constructor user arg: packet_number and packet_type
	cout << "aggregation function id (from constructor catalog) = ";
	cout.flush();
	cin >> qspec.getAggrID();
	cout << "packet number = "; cout.flush();
	cin >> packet_number;
	cout << "packet type = " << packet_type << endl;
	qspec.getAggrConstructorArg().allocBuffer(sizeof(u_int)*2);
	write_arg.open(qspec.getAggrConstructorArg());
	if (fe.getMyEndianess() != fe.getT2BackEndEndianess())
	   write_arg.setConvertEndian(true);
	write_arg << packet_number << packet_type;
	write_arg.close();

	// get accumulator information
	cout << "accumulator id (from constructor catalog) = "; cout.flush();
	cin >> qspec.getAccID();
	qspec.getAccNavigatorID() = 0;	// assuming only 1 acc iterator

	// get dataset information
	cout << "dataset id (from dataset catalog) = "; cout.flush();
	cin >> dataset_spec.getDatasetID();
	dataset_spec.getIteratorID() = 0;  // assume only 1 dataset iterator

	// get index information
	cout << "index id (from index catalog) = " ; cout.flush();
	cin >> dataset_spec.getIndexID();

	// get projection function information
	cout << "projection function id (from constructor catalog) = ";
	cout.flush();
	cin >> dataset_spec.getProjID();

	// get output-handling information (ie what to do with the output);
	// currently, allows two options:
	//   write to a file on the back-end disks, or write to a socket:
	u_int choice = (u_int) get_choice("output-handling choice: 1 - write into a file, 2 - write to socket",
					  1, 2);
	T2_QSpecOutput& outspec = qspec.getOutputSpec();
	switch (choice) {
	  case 1:
		outspec.setOutputHandleType(t2_oBFile);   // output to file
		cout << "output file name (less than " << buflen << " bytes) = ";
		cout.flush();
		cin >> buf;
		outspec.setFilePrefix(buf);	// store the filename
		break;
	  case 2:
		outspec.setOutputHandleType(t2_oSocket);
		cout << "client host name (less than 64 bytes):" << endl;
		cout.flush();
		cin >> buf;
		outspec.setHostName(buf);	// store the hostname

		cout << "client port number:" << endl; cout.flush();
		cin >> outspec.getPortNumber();	// store the port number

		// do not use T2 output-handling protocol, 'cause svm uses
		// its own protocol
		outspec.disableT2Protocol();
		break;
	}

	choice = (u_int) get_choice("Use (1) Big Endian or (2) Little Endian",
				    1, 2);
	switch (choice) {
	   case 1:
		outspec.useBigEndianOutput();		// use big endian
		break;
	   case 2:
		outspec.useLittleEndianOutput();	// use little endian
		break;
	}

     }			// if (reset_basic_information == true)

     // get the output query window
     u_int x_pos, y_pos,		// top-left pixel of the output image
					// in the input image attribute space
	   width, height,		// width and height of the output image
	   x_zoom, y_zoom,		// zoom-out factors
	   pixelsize = 3;		// number of bytes per pixel
					// (set to 3 for rgb for now)
     cout << "x-coord for the top-left corner of the output image = ";
     cout.flush();
     cin >> x_pos;
     cout << "y-coord for the top-left corner of the output image = ";
     cout.flush();
     cin >> y_pos;
     cout << "width for the output image = ";
     cout.flush();
     cin >> width;
     cout << "height for the output image = ";
     cout.flush();
     cin >> height;
     cout << "x zoom-out factor the output image = ";
     cout.flush();
     cin >> x_zoom;
     cout << "y zoom-out factor the output image = ";
     cout.flush();
     cin >> y_zoom;

     if (width == 0 || height == 0)
	cout << "An empty output image !! Query ignore." << endl;
     else {
	// set the input query box
	T2_Box& inbox = dataset_spec.getQueryBox();
	inbox.setNumberDimensions(2);
	inbox.getLow()[0] = x_pos;
	inbox.getLow()[1] = y_pos;
	inbox.getHigh()[0] = x_pos + width * x_zoom - 1;
	inbox.getHigh()[1] = y_pos + height * y_zoom - 1;

	// set the output box
	T2_Box& output_box = qspec.getOutputBox();
	output_box.setNumberDimensions(2);
	output_box.getLow()[0] = 0;
	output_box.getLow()[1] = 0;
	output_box.getHigh()[0] = width - 1;
	output_box.getHigh()[1] = height - 1;

	// set projection constructor user arg, which needs the following:
	//	x_pos y_pos x_zoom y_zoom
	dataset_spec.getProjConstructorArg().allocBuffer(sizeof(u_int)*4);
	write_arg.open(dataset_spec.getProjConstructorArg());
	if (fe.getMyEndianess() != fe.getT2BackEndEndianess())
	   write_arg.setConvertEndian(true);
	write_arg << x_pos << y_pos << x_zoom << y_zoom;
	write_arg.close();

	// set accumulator constructor user arg, which needs the following:
	//	width height pixelsize
	qspec.getAccConstructorArg().allocBuffer(sizeof(u_int)*3);
	write_arg.open(qspec.getAccConstructorArg());
	if (fe.getMyEndianess() != fe.getT2BackEndEndianess())
	   write_arg.setConvertEndian(true);
	write_arg << width << height << pixelsize;
	write_arg.close();
	cout << endl;
     }

     // now submit the query to T2 front-end
     if (fe.submitQBatch(qbatch) == false) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: submitQBatch error %d", (int) fe.getErrorVal());
	break;
     }
     cout << "Query has been submitted to the T2 front-end" << endl;

     // print selection menu
     cout << endl
	  << "Choice:" << endl
	  << "	0. submit another query with the same input dataset and functions" << endl
	  << "	   (that is, only change the output query window and zoom factors)" << endl
	  << "	1. change the input dataset and functions for a new query" << endl
	  << "	2. exit" << endl
	  << "Your selection = ";
     cout.flush();
     cin >> menu_selection;

  } while (menu_selection < 2);		// do
}



/////////////////////////////////////////////////////////////////////////////
//
// svm_inquire_functions
//
// Inquire T2 front-end and download information about all functions
// with the given keyword in its name.
//
// svm_inquire_datasets
//
// Inquire T2 front-end and download information about all datasets
// with the given keyword in its name.
//
/////////////////////////////////////////////////////////////////////////////
bool
svm_inquire_functions(T2_FrontEnd& fe, const char* keyword,
		      const char* thumbnail_dir,
		      u_int& pid, u_int& accid, u_int& aggid)
{
  // id for projection, accumulator meta, and aggregation functions
  pid = 0, accid = 0, aggid = 0;		// id 0 is not used

  // ask for information about all constructors with the given keyword in their
  // names; just need two fields: name and id
  T2_FEFunctionInquiryResults func_results;
  const int func_fields = T2_FEInquiry::function_id_field |
			  T2_FEInquiry::function_name_field;
  if (fe.inquireFunctionRegExp(keyword, T2_UDF_Unknown, func_fields,
			       func_results) == false)
  {
     cout << "ERROR: T2 front-end function inquiry fails" << endl;
     return false;
  }

  bool	no_accid = true,
	no_pid = true,
	no_aggid = true;
  for (u_int i=0; i<func_results.getNumberEntries(); i++) {
     T2_FEFunctionEntry& entry = func_results[i];
     switch (entry.getFunctionType()) {
	case T2_UDF_AccMeta:		// accumulator meta object
			if (no_accid == true) {	// not assigned yet
			   t2_say(cout, "using accmeta object constructor: id="
					<< entry.getFunctionID()
					<< ", name=" << entry.getFunctionName()
					<< endl);
			   accid = entry.getFunctionID();
			   no_accid = false;
			}
			break;
	case T2_UDF_Projection:		// projection function
			if (no_pid == true) {	// not assigned yet
			   t2_say(cout, "using projection constructor: id="
					<< entry.getFunctionID()
					<< ", name=" << entry.getFunctionName()
					<< endl);
			   pid = entry.getFunctionID();
			   no_pid = false;
			}
			break;
	case T2_UDF_Aggregation:	// aggregation function
			if (no_aggid == true) {	// not assigned yet
			   t2_say(cout, "using aggregation constructor: id="
					<< entry.getFunctionID()
					<< ", name=" << entry.getFunctionName()
					<< endl);
			   aggid = entry.getFunctionID();
			   no_aggid = false;
			}
			break;
	default:			// ignore
			break;
     }
  }
  if (no_pid == true || no_accid == true || no_aggid == true) {
     if (pid == 0)
        cout << "ERROR: there is no projection function for keyword "
	     << keyword << endl;
     if (accid == 0)
        cout << "ERROR: there is no accumulator function for keyword "
	     << keyword << endl;
     if (aggid == 0)
        cout << "ERROR: there is no aggregation function for keyword "
	     << keyword << endl;
     return false;
  }
  return true;
}


bool
svm_inquire_datasets(T2_FrontEnd& fe, const char* keyword,
		     u_int pid, u_int accid, u_int aggid,
		     const char* thumbnail_dir,
		     svm_ImageInfoCollection& allimages)
{
  // ask for information about all constructors with the given keyword
  T2_FEDatasetInquiryResults dataset_results;

  // only choose to return dataset name, id, dataset description, blobsize,
  // blobobj, indexid (assuming only 1 iterator, so just use it);
  // note that the dataset description should contain the width and height
  // of the dataset image in ascii format
  const int fields = T2_FEInquiry::dataset_datasetid_field |
		     T2_FEInquiry::dataset_datasetname_field |
		     T2_FEInquiry::dataset_datasetdescription_field |
		     T2_FEInquiry::dataset_blobsize_field |
		     T2_FEInquiry::dataset_blobobj_field |
		     T2_FEInquiry::dataset_indexid_field;

  t2_say4(cout, "****** gonna call inquireDatasetRegExp" << endl);

  // let's say we are willing to have at most 10MB to hold the thumbnail
  // in memory
  u_int max_total_thumbnail_memory_limit = 10 * 1024 * 1024;
  if (fe.inquireDatasetRegExp(keyword, fields, max_total_thumbnail_memory_limit,
			      dataset_results) == false)
  {
     cout << "ERROR: T2 front-end dataset inquiry fails" << endl;
     return false;
  }

  t2_say4(cout, "****** return from inquireDatasetRegExp with " 
		<< dataset_results.getNumberEntries() << " results" << endl);

  // need to add more entries to allimages to accommodate for the new datasets
  u_int j = allimages.getNumberEntries();
  allimages.growSize(dataset_results.getNumberEntries());

  for (u_int i=0; i<dataset_results.getNumberEntries(); i++)
  {
     T2_FEDatasetEntry& inq_entry = dataset_results[i];
     if (inq_entry.getDatasetName() == NULL) {
	cout << "ERROR: dataset result entry " << i
	     << " has no dataset name, ignore it." << endl;
	delete [] inq_entry.getBlobObject();
	inq_entry.getBlobObject() = NULL;
	continue;
     }
     if (inq_entry.getNumberIndexes() == 0) {
	cout << "ERROR: dataset " << inq_entry.getDatasetName()
	     << " has no index, ignore it." << endl;
	delete [] inq_entry.getBlobObject();
	inq_entry.getBlobObject() = NULL;
	continue;
     }
     if (inq_entry.getDatasetDescription() == NULL) {
	cout << "ERROR: dataset " << inq_entry.getDatasetName()
	     << " has no description of width and height, ignore it." << endl;
	delete [] inq_entry.getBlobObject();
	inq_entry.getBlobObject() = NULL;
	continue;
     }
     if (inq_entry.getBlobDataSize() == 0) {
	cout << "ERROR: dataset " << inq_entry.getDatasetName()
	     << " has no thumbnail, ignore it." << endl;
	delete [] inq_entry.getBlobObject();
	inq_entry.getBlobObject() = NULL;
	continue;
     }

     if (inq_entry.getBlobObject() == NULL) {
	// blob object didn't make it, probably due to memory limit,
	// so ask for it again
	T2_FEDatasetInquiryResults res;
	if (fe.inquireDatasetExactMatch(inq_entry.getDatasetName(),
					T2_FEInquiry::dataset_blobobj_field,
					0, 		// no memory limit
					res) == false)
	{
	   cout << "ERROR: fail to download thumbnail for dataset "
	        << inq_entry.getDatasetName() << ", ignore dataset." << endl;
	   delete [] inq_entry.getBlobObject();
	   inq_entry.getBlobObject() = NULL;
	   continue;
	}
	if (res.getNumberEntries() != 1) {
	   cout << "ERROR: got " << res.getNumberEntries()
		<< " entries back when trying to download thumbnail again for dataset "
		<< inq_entry.getDatasetName() << ", ignore dataset."
		<< endl;
	   delete [] inq_entry.getBlobObject();
	   inq_entry.getBlobObject() = NULL;
	   continue;
	}

	// move res.getBlobObject() to inq_entry.getBlobObject()
	inq_entry.getBlobObject() = res[0].getBlobObject();
	res[0].getBlobObject() = NULL;
     }


     svm_ImageInfoEntry*& info_entry = allimages[j];
     info_entry = new svm_ImageInfoEntry;

     // move filename from inq_entry to info_entry
     info_entry->getDatasetName() = inq_entry.getDatasetName();
     inq_entry.getDatasetName() = NULL;

     info_entry->getDatasetID() = inq_entry.getDatasetID();
     info_entry->getIndexID() = inq_entry.getIndex(0).getIndexID();
     info_entry->getAccMetaID() = accid;
     info_entry->getAggregationID() = aggid;
     info_entry->getProjectionID() = pid;  // by UVC
     

     // extract width and height from the dataset description
     sscanf(inq_entry.getDatasetDescription(), "%d %d",
	    &(info_entry->getImageWidth()),
	    &(info_entry->getImageHeight()));
     if (info_entry->getImageWidth() == 0 || info_entry->getImageHeight() == 0)
     {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: dataset '%s' is an empty image with width=%d and height=%d",
		info_entry->getDatasetName(), info_entry->getImageWidth(),
		info_entry->getImageHeight());
	delete [] inq_entry.getBlobObject();
	inq_entry.getBlobObject() = NULL;
	delete info_entry;		// set entry pointer back to NULL
	info_entry = NULL;
	continue;
     }

     // get thumbnail size
     info_entry->getThumbnailFileSize() = inq_entry.getBlobDataSize();

     // get thumbnail filename
     const char fname_suffix[] = ".thumbnail";
     size_t len = strlen(info_entry->getDatasetName()) +
		  strlen(".thumbnail") + 1;
     char*& thumbnail_fname = info_entry->getThumbnailFilename();
     if (thumbnail_dir == NULL) {
	thumbnail_fname = new char[len];
	sprintf(thumbnail_fname, "%s%s",
		info_entry->getDatasetName(), fname_suffix);
     }
     else {
	len += strlen(thumbnail_dir) + 1;
	thumbnail_fname = new char[len];
	sprintf(thumbnail_fname, "%s/%s%s",
		thumbnail_dir, info_entry->getDatasetName(), fname_suffix);
     }

     // write thumbnail
     int fd = open(thumbnail_fname, O_WRONLY|O_CREAT, 0644);
     if (fd < 0) {			// open fail
	cout << "ERROR: fail to write thumbnail file '" << thumbnail_fname
	     << "' for dataset " << info_entry->getDatasetName()
	     << ", ignore this dataset."
	     << endl;
	delete [] thumbnail_fname;
	thumbnail_fname = NULL;

	// delete blob object to free up some memory
	delete [] inq_entry.getBlobObject();
	inq_entry.getBlobObject() = NULL;

	delete info_entry;		// set entry pointer back to NULL
	info_entry = NULL;
     }
     else {
	write(fd, inq_entry.getBlobObject(), inq_entry.getBlobFileSize());
	close(fd);

	// delete blob object to free up some memory
	delete [] inq_entry.getBlobObject();
	inq_entry.getBlobObject() = NULL;

        j++;
     }
  }

  return true;
}


int CreateListenSocket(u_int client_port)
{
 struct sockaddr_in addr;
 int listensd;
 
 printf("Waiting for clients at port %d ...\n", client_port);

 if ((listensd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0)
     {     
     cout << "Can't open socket." << endl;
     return -1;
     }
 
     /* Bind local address. */
 bzero((char *)&addr, sizeof(addr));
 addr.sin_family = AF_INET;
 addr.sin_addr.s_addr = htonl(INADDR_ANY);
 addr.sin_port = htons(client_port);
 if (bind(listensd, (struct sockaddr *)&addr, sizeof(addr)) < 0)
     {
     cout << "Can't bind local address." << endl;
     close(listensd);     
     return -1;
     }

     /* Now start listening to connection requests. */
 listen(listensd, 5);  /* the backlog is limited to 5 in the OS. */

 return listensd;
}


int WaitSocketsForRead(int soc1, int soc2, fd_set *fdset)
{
 FD_ZERO(fdset);
 if (soc1!=-1)
     FD_SET(soc1, fdset);
 if (soc2!=-1)
     FD_SET(soc2, fdset);
 return select(1+((soc1>soc2) ? soc1:soc2), fdset, NULL, NULL, NULL);
}


int Read_Raw(int sd, void *buf, int size)
{
 int i, j;
 unsigned char *b;

 b = (unsigned char *)buf;
 i = size;
 while (i)
     {
     j = read(sd, b, i); /* blocks until there is something <= i to return */
     b += j;
     i -= j;
     if (j <= 0)
         {
         size = -1;   /* Connection broken (or some other error). */
         break;
         }
     }
 return size;
}

int Write_Raw(int sd, void *buf, int size)
{
 int i, j;
 unsigned char *b;

 b = (unsigned char *)buf;
 i = size;
 while (i)
     {
     j = write(sd, b, i); /* blocks until there is something <= i to return */
     b += j;
     i -= j;
     if (j <= 0)
         {
         size = -1;   /* Connection broken (or some other error). */
         break;
         }
     }
 return size;
}


#ifdef _CONVERT_ENDIAN
void Convert_Endian(int nelem, int selem, void *vbuf)
{
  int i, j;
  char *buf=(char *)vbuf;

  for (i = 0; i < nelem; i++)
    for (j = 0; j < selem/2; j++)
    {
      char temp;

      temp = buf[(i * selem) + j];
      buf[(i * selem) + j] = buf[(i * selem) + selem - j - 1];
      buf[(i * selem) + selem - j - 1] = temp;
    }
}
#endif

int ReadInt(int sd)
{
 int v;

 if (Read_Raw(sd, &v, 4)!=4)
     return -1;

#ifdef _CONVERT_ENDIAN
 Convert_Endian(1, 4, &v);
#endif
 return v;
}

int WriteInt(int sd, int v)
{
#ifdef _CONVERT_ENDIAN
 Convert_Endian(1, 4, &v);
#endif
 if (Write_Raw(sd, &v, 4)!=4)
     return -1;
 
 return 0;
}


int WriteStr(int sd, char *st)
{
 int  l=strlen(st);
 char c='\n';

 if (Write_Raw(sd, st, l)==l)
     if (Write_Raw(sd, &c, 1)==1)
         return 0;
 return -1;
}


char *Read_Thumb_PPM(char *fname, int *width, int *height)
{
 FILE *in;
 char str[80], c;
 int  cnt = 0, vals[3];
 char *rgb;
    
 if ((in = fopen(fname, "r")) == NULL)
     {
     cout << "Can't open file" << fname << endl;
     return NULL;
     }
 
     /* Parse image header. */
 fscanf(in, "%s", str);
 if (strcmp(str, "P6"))
     {
     cout << "Not a PPM(raw) file." << endl;
     fclose(in);
     return NULL;
     }
 while (cnt < 3)
     {
     fscanf(in, "%s", str);
     if (str[0] == '#')
         {         
         do /* Skip until end of line. */
             {
             fscanf(in, "%c", &c);
             }
         while (c != '\n');
         }
     else
         {
         sscanf(str, "%d", &(vals[cnt]));
         cnt++;
         }
     }
 fgetc(in);  /* skip the last new line before the image. */
 
     /* read the image. */
 rgb = (char *)malloc(3 * vals[0] * vals[1]);
 if (rgb)
     fread(rgb, 3, vals[0]*vals[1], in);
 else
     cout << "Cannot allocate memory for rgb data. Requested size = " << 3*vals[0]*vals[1] << " ( "<<vals[0] << "x" << vals[1] << endl;
 
 fclose(in);
 *width = vals[0];
 *height = vals[1];
 return rgb;
}

int GenerateQuery(T2_FrontEnd& fe, int packno, int packtype, int x, int y, int w, int h, int z, char *hostname, int backendport, svm_ImageInfoEntry *ie, const char* dump_qfile)
{
 T2_QBatch qbatch(1);		        // only one query per batch
 T2_QSpec& qspec = qbatch.getQuerySpec(0);
 qspec.setNumberDatasets(1);		// only one input dataset

 qspec.useBigEndianUserArg();		// use big endian ???
  
 T2_UsrArgWriter write_arg;		// user argument writer

 T2_QSpecDataset& dataset_spec = qspec.getDatasetSpec(0);
  

 qspec.getAggrID() = ie->getAggregationID(); // ppm aggregation
 qspec.getAggrConstructorArg().allocBuffer(sizeof(u_int)*2);
 write_arg.open(qspec.getAggrConstructorArg());
 if (fe.getMyEndianess() != fe.getT2BackEndEndianess())
   write_arg.setConvertEndian(true);
 write_arg << packno << packtype;
 write_arg.close();


 qspec.getAccID() = ie->getAccMetaID();
 qspec.getAccNavigatorID() = 0;	// assuming only 1 acc iterator

 dataset_spec.getDatasetID() = ie->getDatasetID();
 dataset_spec.getIteratorID() = 0;  // assume only 1 dataset iterator

 dataset_spec.getIndexID() = ie->getIndexID();

 dataset_spec.getProjID() = ie->getProjectionID();

 T2_QSpecOutput& outspec = qspec.getOutputSpec();
 outspec.useBigEndianOutput();		// use big endian since client is written in Java
 outspec.setOutputHandleType(t2_oSocket);
  
 outspec.setHostName(hostname);	// store the hostname

 outspec.getPortNumber() = backendport;	// store the port number
 outspec.disableT2Protocol();

     // set the input query box
 T2_Box& inbox = dataset_spec.getQueryBox();
 inbox.setNumberDimensions(2);
 inbox.getLow()[0] = x;
 inbox.getLow()[1] = y;
 inbox.getHigh()[0] = x+w*z;
 inbox.getHigh()[1] = y+h*z;

     // set the output box
 T2_Box& output_box = qspec.getOutputBox();
 output_box.setNumberDimensions(2);
 output_box.getLow()[0] = 0;
 output_box.getLow()[1] = 0;

 output_box.getHigh()[0] = w-1;
 output_box.getHigh()[1] = h-1;

     // set projection constructor user arg, which needs the following:
     //	x_pos y_pos x_zoom y_zoom
 dataset_spec.getProjConstructorArg().allocBuffer(sizeof(u_int)*4);
 write_arg.open(dataset_spec.getProjConstructorArg());
 if (fe.getMyEndianess() != fe.getT2BackEndEndianess())
   write_arg.setConvertEndian(true);
 write_arg << x << y << z << z;
 write_arg.close();

     // set accumulator constructor user arg, which needs the following:
     //	width height pixelsize
 int pixelsize=3;
 
 qspec.getAccConstructorArg().allocBuffer(sizeof(u_int)*3);
 write_arg.open(qspec.getAccConstructorArg());
 if (fe.getMyEndianess() != fe.getT2BackEndEndianess())
   write_arg.setConvertEndian(true);
 write_arg << w << h << pixelsize;
 write_arg.close();

     // now submit the query to T2 front-end
 if (fe.submitQBatch(qbatch) == false)
     {
     t2_err_ret(__FILE__, __LINE__, "ERROR: submitQBatch error %d", (int) fe.getErrorVal());
     return -1;
     }

  // dump query into given file for debugging purpose
  if (dump_qfile != NULL)
     {
	T2_OBFile obf(dump_qfile);
	if (!obf)
	   t2_err_ret(__FILE__, __LINE__,
	      "WARNING: fail to open output query dump file '%s'", dump_qfile);
	else
	{
	   obf << qbatch;
	   obf.close();
	}
     }

 return 0;
}



int ReadAndProcessReq(int clientsd, T2_FrontEnd& fe, svm_ImageInfoCollection& allimages, char* hostname, int &backendport, u_int &openimage, const char* dump_qfile)
{
 const int PCK_HELLO       = 1;
 const int PCK_REP_HELLO   = 2;
 const int PCK_OPEN        = 3;
 const int PCK_REP_OPEN    = 4;
 const int PCK_QUERY       = 5;
 const int PCK_REP_QUERY   = 6;  
 int  packno, packtype, x, y, w, h, z; 
 char *rgb;

 packno = ReadInt(clientsd);
 packtype = ReadInt(clientsd);
 if (packno==-1 || packtype==-1)
     return -1;
 switch (packtype)
     {
     case PCK_HELLO:
         cout << "Hello Packet recieved from client " << hostname << endl;
         backendport = ReadInt(clientsd);
         if (backendport<0)
             return -1;
         
         WriteInt(clientsd, packno);
         WriteInt(clientsd, PCK_REP_HELLO);
         WriteInt(clientsd, allimages.getNumberEntries());
	 {
           for (u_int i=0; i<allimages.getNumberEntries(); ++i)
             {
             cout << "Dataset name " << allimages[i]->getDatasetName() << endl;
             WriteStr(clientsd, allimages[i]->getDatasetName());
             }
	 }

         break;
     case PCK_OPEN:
         cout << "Open Packet recieved from client " << hostname << endl;
         x = ReadInt(clientsd);
         if (x<0 || x>=(int)allimages.getNumberEntries())
             {
             cout << "Invalid image index = " << x;             
             return -1;
             }
         openimage = (u_int) x;         
             
         WriteInt(clientsd, packno);
         WriteInt(clientsd, PCK_REP_OPEN);
         WriteInt(clientsd, allimages[openimage]->getImageWidth());
         WriteInt(clientsd, allimages[openimage]->getImageHeight());
         
         rgb = Read_Thumb_PPM(allimages[openimage]->getThumbnailFilename(), &w, &h);
         if (rgb)
             {             
             WriteInt(clientsd, w);
             WriteInt(clientsd, h);
             Write_Raw(clientsd, rgb, 3*w*h);
             free(rgb);
             }
         else // if not enough memory somehow lets decide the size of thumbnail
             {
             WriteInt(clientsd, 100); 
             WriteInt(clientsd, 100);
             }
         
         break;
     case PCK_QUERY:
         cout << "Query Packet recieved from client " << hostname << endl;
         x=ReadInt(clientsd);
         y=ReadInt(clientsd);
         w=ReadInt(clientsd);
         h=ReadInt(clientsd);
         z=ReadInt(clientsd);

         if (GenerateQuery(fe, packno, PCK_REP_QUERY, x, y, w, h, z, hostname, backendport, allimages[openimage], dump_qfile)==-1)
             {
             cout << "Fail in GenerateQuery" << endl;             
             return -1;
             }
         break;         
     default:
         cout << "Unexpected packet type=" << packtype << endl;
         
     }
 cout << "Request Processed!" << endl;

 return 0; 
}



void AppFrontEnd(T2_FrontEnd& fe, svm_ImageInfoCollection& allimages, u_int client_port, const char* dump_qfile)
{
 fd_set	fdset;
 int    listensd, clientsd=-1, client_host=0, client_backendport=0;
 u_int  openimage=0xffff;
 char   *clientname = NULL;

 if ((listensd = CreateListenSocket(client_port))==-1)
     return;

 while (WaitSocketsForRead(listensd, clientsd, &fdset)>0)
     {
     if (clientsd!=-1 && FD_ISSET(clientsd, &fdset))
         if (ReadAndProcessReq(clientsd, fe, allimages, clientname, client_backendport, openimage, dump_qfile)==-1)
             {
             close(clientsd);
             clientsd = -1;
             }     
     if (FD_ISSET(listensd, &fdset))
         {
         struct sockaddr_in addr;
         int    newsd;
	 //int	len= (int) sizeof(addr);
	 socklen_t len= (socklen_t) sizeof(addr);

         newsd = accept(listensd, (struct sockaddr *)&addr, &len);
         if (clientsd!=-1)
             {             
             cout << "Sorry only one client is allowed." << endl;
             close(newsd);             
             }
         else
             {
             clientsd = newsd;
             client_host = addr.sin_addr.s_addr;

             struct hostent *hp = gethostbyaddr((char *)&addr.sin_addr,sizeof(addr.sin_addr),AF_INET);
             if(hp == NULL)
                 clientname = strdup(inet_ntoa(addr.sin_addr)); 
             else
                 clientname = strdup(hp->h_name);
             cout << "Connection request from client " << clientname << endl;
             }         
         }
     }
 if (clientname)
     free(clientname);
 close(listensd);
}





