
/* $Id: svm_aggr.C,v 1.3 2000/03/15 00:18:38 chialin Exp $ */

/* svm_aggr.C */

#include	<iostream.h>
#include	"svm_acc.h"
#include	"svm_aggr.h"
#include	"svm_output.h"


////////////////////////////////////////////////////////////////////////
//
// define APP_ACCMSG_NOCOPY to avoid memory copy in fillAccMsgBuffer();
// otherwise, pixels are copied into out-going mesage buffer
//
////////////////////////////////////////////////////////////////////////
#define APP_ACCMSG_NOCOPY

////////////////////////////////////////////////////////////////////////
//
// svm_aggregation methods
//
////////////////////////////////////////////////////////////////////////
bool
svm_aggregation::needGlobalCombine(const T2_QueryInfo& qinfo,
				   T2_Accumulator& acc)
{
  if (qinfo.getNumberNodes() > 1)
     return true;

  // only one processor, so no need to do global combine,
  // however, need to initialize the variables that describe the sub-image
  // since they were supposed to be set during fillAccMsgBuffer() if a global
  // combine phase was needed
  svm_ImageAcc *image = (svm_ImageAcc *) & acc;
  my_bytes = image->getBytes();
  my_nbytes = image->getNumberPixels() * image->getPixelSize();
  my_x_pos = image->getXPosition();
  my_y_pos = image->getYPosition();
  my_width = image->getWidth();
  my_height = image->getHeight();

  return false;
}
 

// for global combine phase, I'm just gonna use a simple partitioning along
// the rows to partition the work load, with the remaining rows assigned
// to the last processor; that is, each processor is assigned to work
// on a number of rows; the following implementation would ask for
// a message buffer for each remote processor, and 
// if APP_ACCMSG_NOCOPY is defined, 
//	just return a pointer into the pixel array and the appropriate size;
//	(this should be the more efficient implementation)
// or if APP_ACCMSG_NOCOPY is not defined, 
//	copy all the pixels from svm_ImageAcc into the message buffer
T2_UDFRet
svm_aggregation::fillAccMsgBuffer(const T2_QueryInfo& qinfo,
				  T2_Accumulator& acc,
				  T2_Array<T2_AccMsgBufferWriter>& msgbuf)
{
  u_int numnodes = qinfo.getNumberNodes();
  T2_ProcID mynode = qinfo.getMyNode();
  t2_Assert (numnodes > 1);
  t2_Assert (numnodes == msgbuf.getNumberElements());

  svm_ImageAcc *image = (svm_ImageAcc *) & acc;
  u_int width = image->getWidth(),
        height = image->getHeight(),
	pixelsize  = image->getPixelSize();
  u_char* bytes = image->getBytes();

  u_int y_part = height / numnodes,		// # rows per partition
        y_excess = height % numnodes;		// # excess rows
  u_int npixels_part = y_part * width;		// # pixels per partition
  size_t nbytes_part = npixels_part * pixelsize;   // # bytes per partition

  u_char* cur_bytes = bytes;
  for (u_int p=0, y=0; p<numnodes; p++) {
     u_int cur_y_part = y_part,
     cur_nbytes_part = nbytes_part;
     if (p == numnodes-1) {		// last processor gets all y_access rows
        cur_y_part += y_excess;
        cur_nbytes_part += y_excess * width * pixelsize;
     }

     if (p == (u_int) mynode) {		// no msg to myself, but remember
					// my assigned bytes
        my_bytes = cur_bytes;
        my_nbytes = cur_nbytes_part;
	my_x_pos = image->getXPosition();
        my_y_pos = y + image->getYPosition();
        my_width = width;
        my_height = cur_y_part;
	cur_bytes += cur_nbytes_part;        
    }
    else { 				// send a msg to remote proc p
#       ifdef APP_ACCMSG_NOCOPY

	   msgbuf[p].setUserDataBuffer((const char*) cur_bytes,
				       cur_nbytes_part);
	   cur_bytes += cur_nbytes_part;

#       else

	   T2_AccMsgBufferWriter& mbuf = msgbuf[p];
	   mbuf.allocBuffer(cur_nbytes_part);
	   for (u_int k=0; k<cur_nbytes_part; k++) {
	      mbuf << *cur_bytes;
	      cur_bytes++;
     	   }

#       endif

    }				// if (p == mynode) else

    y += cur_y_part;
  }				// for (u_int p=0, y=0; p<numnodes; p++) 

  return T2_UDFRet_OK;
}


T2_UDFRet 
svm_aggregation::finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
			  T2_Output*& outp)
{
  svm_ImageAcc *acc_image = (svm_ImageAcc *) &acc;

  // create the output sub-image object
  outp = (T2_Output *) new
	svm_OutputSubImage(packet_no, packet_type,
			   qinfo.getNumberNodes(), qinfo.getNumberTiles(),
			   my_x_pos, my_y_pos, my_width, my_height, my_bytes,
			   acc_image->getPixelSize(), acc_image->getMaxValue(),
			   acc_image->getTotalOutputImageWidth(),
			   acc_image->getTotalOutputImageHeight());

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrMax methods
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggrMax::aifElem(const T2_QueryInfo& qinfo, void* accElem)
{
  t2_Assert (accElem != NULL);
  svm_ImagePixel& pixel = *((svm_ImagePixel *) accElem);
  for (u_int i=0; i<pixel.getNumberBytes(); i++)
     pixel[i] = 0;

  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggrMax::dafElem(const T2_QueryInfo& qinfo, T2_DSID i, const void* dataElem,
		     void* accElem)
{
  // ignore dataset id

  t2_Assert (dataElem != NULL);
  t2_Assert (accElem != NULL);
  const svm_ImagePixel& input_pixel = *((const svm_ImagePixel *) dataElem);
  svm_ImagePixel& output_pixel = *((svm_ImagePixel *) accElem);

  t2_Assert (input_pixel.getNumberBytes() == input_pixel.getNumberBytes());
  for (u_int j=0; j<input_pixel.getNumberBytes(); j++)
      if (input_pixel[j] > output_pixel[j])
	 output_pixel[j] = input_pixel[j];

  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggrMax::processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
			   T2_AccMsgBufferReader& msgbuf,
                           T2_Accumulator& acc)
{
  T2_ProcID mynode = qinfo.getMyNode();
  t2_Assert (mynode != sender);
  t2_Assert (msgbuf.getDataSize() == my_nbytes);

  for (u_int i=0; i<my_nbytes; i++) {
     u_char v;
     msgbuf >> v;
     if (my_bytes[i] < v)
	my_bytes[i] = v;
  }

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrMin methods
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggrMin::aifElem(const T2_QueryInfo& qinfo, void* accElem)
{
  t2_Assert (accElem != NULL);
  svm_ImagePixel& pixel = *((svm_ImagePixel *) accElem);
  for (u_int i=0; i<pixel.getNumberBytes(); i++)
      pixel[i] = 255;
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggrMin::dafElem(const T2_QueryInfo& qinfo, T2_DSID i, const void* dataElem,
		     void* accElem)
{
  // ignore dataset id

  t2_Assert (dataElem != NULL);
  t2_Assert (accElem != NULL);

  const svm_ImagePixel& input_pixel = *((const svm_ImagePixel *) dataElem);
  svm_ImagePixel& output_pixel = *((svm_ImagePixel *) accElem);
  t2_Assert (input_pixel.getNumberBytes() == input_pixel.getNumberBytes());
  for (u_int j=0; j<input_pixel.getNumberBytes(); j++)
     if (input_pixel[j] < output_pixel[j])
	output_pixel[j] = input_pixel[j];
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggrMin::processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
			   T2_AccMsgBufferReader& msgbuf,
                           T2_Accumulator& acc)
{
  T2_ProcID mynode = qinfo.getMyNode();
  t2_Assert (mynode != sender);
  t2_Assert (msgbuf.getDataSize() == my_nbytes);

  for (u_int i=0; i<my_nbytes; i++) {
     u_char v;
     msgbuf >> v;
     if (my_bytes[i] > v)
	my_bytes[i] = v;
  }

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrMax, svm_aggrMin constructors
//
// Expecting the following numbers from the user argument, which should be
// used in the header of the finalized output.
//	u_int packet_number
//	u_int packet_type
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggregationMaxConstructor(const T2_QueryInfo& qinfo, T2_UsrArgReader& arg,
			      T2_AggregateFuncObj*& afp)
{
  u_int packet_no, packet_type;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(packet_no)+sizeof(packet_type)) {
     // not enough arguments !!
    t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(packet_no)+sizeof(packet_type));
     return T2_UDFRet_ERROR;
  }

  arg >> packet_no >> packet_type;

  afp = (T2_AggregateFuncObj *) new svm_aggrMax(packet_no, packet_type);
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggregationMinConstructor(const T2_QueryInfo& qinfo, T2_UsrArgReader& arg,
			      T2_AggregateFuncObj*& afp)
{
  u_int packet_no, packet_type;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(packet_no)+sizeof(packet_type)) {
     // not enough arguments !!
    t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(packet_no)+sizeof(packet_type));
     return T2_UDFRet_ERROR;
  }

  arg >> packet_no >> packet_type;

  afp = (T2_AggregateFuncObj *) new svm_aggrMin(packet_no, packet_type);
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggregationMaxPPMConstructor(const T2_QueryInfo& qinfo,T2_UsrArgReader& arg,
				 T2_AggregateFuncObj*& afp)
{
  u_int packet_no, packet_type;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(packet_no)+sizeof(packet_type)) {
     // not enough arguments !!
    t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(packet_no)+sizeof(packet_type));
     return T2_UDFRet_ERROR;
  }

  arg >> packet_no >> packet_type;

  afp = (T2_AggregateFuncObj *) new svm_aggrMaxPPM(packet_no, packet_type);
  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrMaxPPM methods
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggrMaxPPM::finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
			 T2_Output*& outp)
{
  svm_ImageAcc *acc_image = (svm_ImageAcc *) &acc;

  // create the output sub-image object
  outp = (T2_Output *) new
	svm_OutputSubImagePPM(packet_no, packet_type,
			   qinfo.getNumberNodes(), qinfo.getNumberTiles(),
			   my_x_pos, my_y_pos, my_width, my_height, my_bytes,
			   acc_image->getPixelSize(), acc_image->getMaxValue(),
			   acc_image->getTotalOutputImageWidth(),
			   acc_image->getTotalOutputImageHeight());
  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrFast methods
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggrFast::computeIntersection(const T2_QueryInfo& qinfo,
			const T2_Cluster& input, const T2_Box& mbr,
			T2_ClusterAuxInfoReader& meta,
			T2_Dataset& dataset, T2_DSID dsid,
			T2_IteratorID dsiterid,
			const T2_Box& iqr, const T2_Region& tilereg,
			T2_ProjectFuncObj& proj,
			u_int acciterid, T2_Accumulator& acc,
			const u_char*& in_ptr, u_char*& out_ptr,
			u_int& in_x_start, u_int& in_y_start,
			u_int& in_x_end, u_int& in_y_end,
			u_int& out_x_start, u_int& out_y_start,
			u_int& out_x_end, u_int& out_y_end,
			u_int& x_start_size, u_int& y_start_size,
			u_int& x_mid_size, u_int& y_mid_size,
			u_int& x_end_size, u_int& y_end_size,
			u_int& in_skip_bytes, u_int& out_skip_bytes,
			u_int& pixelsize)
{
  // this code uses the same code that svm_dataset::genIterator() to
  // find out the range of pixels that it needs to iterate through,
  // and then just go through the byte array to do the aggregation

  // ignores dataset iterator id

  // only deals with 2-dim query window
  if (mbr.getNumberDimensions() != 2 || iqr.getNumberDimensions() != 2 ||
      tilereg.getNumberDimensions() != 2)
  {
      t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot handle attribute space other than 2-dim.");
      return T2_UDFRet_ERROR;
  }

  // only handles with 1 data block from the cluster
  if (input.getNumberBlocks() != 1) {
      t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting a cluster with 1 data block, but got one with %d blocks.",
                input.getNumberBlocks());
      return T2_UDFRet_ERROR;
  }
  assert (input.getDataPointer(0) != NULL);

  // for now, only knows how to handle a tile region with one single box
  switch (tilereg.getNumberBoxes()) {
     case 0:	// an empty region, so nothing to aggregate
		return T2_UDFRet_OK;

     case 1:	// yes, I can handle this, so continue execution
		break;

     default:	// ooops, cannot handle more boxes
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: cannot handle regions with %d boxes",
			tilereg.getNumberBoxes());
		return T2_UDFRet_ERROR;
  }

  const T2_Box& tilebox = tilereg[0];
  u_int tile_x_low = (u_int) tilebox.getLow()[0],
        tile_x_high = (u_int) tilebox.getHigh()[0];
//	tile_y_low = (u_int) tilebox.getLow()[1],
//	tile_y_high = (u_int) tilebox.getHigh()[1];

  // compute all the variables declared above
  if (findIntersection(qinfo, mbr, iqr, tilebox, proj,
		       in_x_start, in_y_start, in_x_end, in_y_end,
		       out_x_start, out_y_start, out_x_end, out_y_end,
		       x_start_size, y_start_size,
		       x_mid_size, y_mid_size,
		       x_end_size, y_end_size)
      != T2_UDFRet_OK)
     return T2_UDFRet_ERROR;		// error occurs

  // Need to compute in_skip_bytes and out_skip_bytes, which are the number
  // of bytes to skip in the input cluster and the accumulator, respectively,
  // when advancing from a row to the next one.
  svm_ImageDataset *image_dataset = (svm_ImageDataset *) & dataset;
  svm_ImageAcc *image_acc = (svm_ImageAcc *) & acc;
  t2_Assert (image_dataset->getPixelSize() == image_acc->getPixelSize());
  pixelsize = image_acc->getPixelSize();
  const u_int in_width = (u_int) mbr.getHigh()[0] - (u_int) mbr.getLow()[0] + 1,
	      out_width = tile_x_high - tile_x_low + 1;
  in_skip_bytes = (in_width - in_x_end + in_x_start - 1) * pixelsize,
  out_skip_bytes = out_width * pixelsize;

  // now iterate over the output and input pixels to do aggregation
  in_ptr = (const u_char *) input.getDataPointer(0);
  in_ptr += pixelsize * (in_x_start + in_y_start * in_width);
  out_ptr = image_acc->getBytes();
  out_ptr += pixelsize * (out_x_start + out_y_start * out_width);

  return T2_UDFRet_OK;
}



T2_UDFRet
svm_aggrMaxFast::dafAcc(const T2_QueryInfo& qinfo,
			const T2_Cluster& input, const T2_Box& mbr,
			T2_ClusterAuxInfoReader& meta,
			T2_Dataset& dataset, T2_DSID dsid,
			T2_IteratorID dsiterid,
			const T2_Box& iqr, const T2_Region& tilereg,
			T2_ProjectFuncObj& proj,
			u_int acciterid, T2_Accumulator& acc)
{
  // the starting and ending offsets for the input cluster pixels
  // and the accumulator pixels, relative to their top-left pixels,
  // in the intersecting box between mbr, iqr, and tilebox
  u_int in_x_start, in_y_start, in_x_end, in_y_end,
	out_x_start, out_y_start, out_x_end, out_y_end;

  // the number of input pixels that map to an output pixel on the starting
  // (top-left) boundary of the intersecting box between mbr, iqr and tilebox,
  // an output pixel between the two boundaries of the intersecting box,
  // and an output pixel on the ending (bottom-right) boundary of the 
  // intersecting box
  u_int x_start_size, y_start_size, x_mid_size, y_mid_size,
	x_end_size, y_end_size;

  // number of bytes per pixel for both the input cluster pixels and the
  // output accumulator pixels
  u_int pixelsize;

  // number of bytes to skip when moving from one row to the next row
  u_int in_skip_bytes, out_skip_bytes;

  // pointers to the byte array of the input cluster, and to the
  // byte array of the accumulator that input pixels of cluster map to
  const u_char *in_ptr;
  u_char *out_ptr;

  if (computeIntersection(qinfo, input, mbr, meta, dataset, dsid, dsiterid,
			  iqr, tilereg, proj, acciterid, acc,
			  in_ptr, out_ptr,
			  in_x_start, in_y_start, in_x_end, in_y_end,
			  out_x_start, out_y_start, out_x_end, out_y_end,
			  x_start_size, y_start_size,
			  x_mid_size, y_mid_size,
			  x_end_size, y_end_size,
			  in_skip_bytes, out_skip_bytes, pixelsize)
     == T2_UDFRet_ERROR)
     return T2_UDFRet_ERROR;		// error occurs

  for (u_int out_y=out_y_start; out_y<=out_y_end; out_y++) {
      u_int in_num_y;
      if (out_y == out_y_start)
          in_num_y = y_start_size;
      else if (out_y == out_y_end)
          in_num_y = y_end_size;
      else
          in_num_y = y_mid_size;

      for (u_int ny=0; ny<in_num_y; ny++) {
           u_char *cur_out_ptr = out_ptr;	// back to current output row

           for (u_int out_x=out_x_start; out_x<=out_x_end; out_x++) {
                u_int in_num_x;
                if (out_x == out_x_start)
                    in_num_x = x_start_size;
                else if (out_x == out_x_end)
                    in_num_x = x_end_size;
		else
                    in_num_x = x_mid_size;

                for (u_int nx=0; nx<in_num_x; nx++) {
                     for (u_int i=0; i<pixelsize; i++) {
			  if (*in_ptr > *cur_out_ptr)
			     *cur_out_ptr = *in_ptr;
                          in_ptr++;
                          cur_out_ptr++;
                     }
                 }
           }
           in_ptr += in_skip_bytes;
      }
      out_ptr += out_skip_bytes;
  }

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrMaxFast, svm_aggrMin constructors
//
// Expecting the following numbers from the user argument, which should be
// used in the header of the finalized output.
//	u_int packet_number
//	u_int packet_type
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggrMaxFastConstructor(const T2_QueryInfo& qinfo, T2_UsrArgReader& arg,
			   T2_AggregateFuncObj*& afp)
{
  u_int packet_no, packet_type;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(packet_no)+sizeof(packet_type)) {
     // not enough arguments !!
    t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(packet_no)+sizeof(packet_type));
     return T2_UDFRet_ERROR;
  }

  arg >> packet_no >> packet_type;

  afp = (T2_AggregateFuncObj *) new svm_aggrMaxFast(packet_no, packet_type);
  return T2_UDFRet_OK;
}

