
/* $Id: svm_aggravg.C,v 1.3 2000/03/15 00:27:00 chialin Exp $ */

/* svm_aggravg.C */

#include	<iostream.h>
#include	"svm_dataset.h"
#include	"svm_aggravg.h"
#include	"svm_accavg.h"
#include	"svm_output.h"
#include	"svm_prj.h"
#include	"svm_utility.h"


////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVG methods
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggrAVG::aifElem(const T2_QueryInfo& qinfo, void* accElem)
{
  // accumulator svm_accAVG should have initialized the sums and counts, 
  // so no need to re-initialize them again
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggrAVG::dafElem(const T2_QueryInfo& qinfo, T2_DSID i, const void* dataElem,
		     void* accElem)
{
  // ignore dataset id
  t2_Assert (dataElem != NULL);
  t2_Assert (accElem != NULL);

  const svm_ImagePixel& input_pixel = *((const svm_ImagePixel *) dataElem);
  svm_ImagePixel& output_pixel = *((svm_ImagePixel *) accElem);
  u_int *sum_count = (u_int *) output_pixel.getBytes();
  t2_Assert ((input_pixel.getNumberBytes() +1) * sizeof(u_int)
		== output_pixel.getNumberBytes());

  // recall that the layout for output pixel in the accumulator is
  // sum_0, sum_1, ..., count
  for (u_int j=0; j<input_pixel.getNumberBytes(); j++)
     sum_count[j] += (u_int) input_pixel[j];		// add to sum
  sum_count[input_pixel.getNumberBytes()]++;		// increment count

  return T2_UDFRet_OK;
}


bool
svm_aggrAVG::needGlobalCombine(const T2_QueryInfo& qinfo, T2_Accumulator& acc)
{
  if (qinfo.getNumberNodes() > 1)
     return true;

  // only one processor, so no need to do global combine,
  // however, need to initialize the variables that describe the sub-image
  // since they were supposed to be set during fillAccMsgBuffer() if a global
  // combine phase was needed
  svm_ImageAccAVG *accumulator = (svm_ImageAccAVG *) & acc;
  my_sumcount = accumulator->getSumCount();
  my_num_sumcount_ints = accumulator->getSumCountSize();
  my_x_pos = accumulator->getXPosition();
  my_y_pos = accumulator->getYPosition();
  my_width = accumulator->getWidth();
  my_height = accumulator->getHeight();

  return false;
}
 

// for global combine phase, I'm just gonna use a simple partitioning along
// the rows to partition the work load, with the remaining rows assigned
// to the last processor; that is, each processor is assigned to work
// on a number of rows; the following implementation would ask for
// a message buffer for each remote processor, and 
//	just return a pointer into the sum_count array and the appropriate size;
T2_UDFRet
svm_aggrAVG::fillAccMsgBuffer(const T2_QueryInfo& qinfo,
				  T2_Accumulator& acc,
				  T2_Array<T2_AccMsgBufferWriter>& msgbuf)
{
  u_int numnodes = qinfo.getNumberNodes();
  T2_ProcID mynode = qinfo.getMyNode();
  t2_Assert (numnodes > 1);
  t2_Assert (numnodes == msgbuf.getNumberElements());

  svm_ImageAccAVG *accumulator = (svm_ImageAccAVG *) & acc;
  u_int width = accumulator->getWidth(),
        height = accumulator->getHeight(),
	pixelsize  = accumulator->getPixelSize();
  u_int* sum_count = accumulator->getSumCount();

  u_int y_part = height / numnodes,		// # rows per partition
        y_excess = height % numnodes;		// # excess rows
  u_int npixels_part = y_part * width;		// # pixels per partition
  size_t nints_part = npixels_part * (pixelsize+1); // # sum/count ints per 
						    // partition

  u_int* cur_sumcount = sum_count;
  for (u_int p=0, y=0; p<numnodes; p++) {
     u_int cur_y_part = y_part,
     cur_nints_part = nints_part;
     if (p == numnodes-1) {		// last processor gets all y_access rows
        cur_y_part += y_excess;
        cur_nints_part += y_excess * width * (pixelsize+1);
     }

     if (p == (u_int) mynode) {		// no msg to myself, but remember
					// my assigned bytes
	my_sumcount = cur_sumcount;
	my_num_sumcount_ints = cur_nints_part;
	my_x_pos = accumulator->getXPosition();
	my_y_pos = y + accumulator->getYPosition();
	my_width = width;
	my_height = cur_y_part;
     }
     else { 				// send a msg to remote proc p
	msgbuf[p].setUserDataBuffer((const char*) cur_sumcount,
				    cur_nints_part * sizeof(u_int));
     }				// if (p == mynode) else

     cur_sumcount += cur_nints_part;        
     y += cur_y_part;
  }				// for (u_int p=0, y=0; p<numnodes; p++) 

  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggrAVG::processAccMsg(const T2_QueryInfo& qinfo, T2_ProcID sender,
			   T2_AccMsgBufferReader& msgbuf,
                           T2_Accumulator& acc)
{
  T2_ProcID mynode = qinfo.getMyNode();
  t2_Assert (mynode != sender);
  t2_Assert (msgbuf.getDataSize() == my_num_sumcount_ints * sizeof(u_int));

  // get the pointer that points to the msg
  size_t actual;
  u_int *msg_sum_count = (u_int *) msgbuf.read(msgbuf.getDataSize(), actual);

  u_int *sum_count = my_sumcount;
  for (u_int i=0; i<my_num_sumcount_ints; i++) {
     sum_count[i] += msg_sum_count[i];
  }

  return T2_UDFRet_OK;
}


void
svm_aggrAVG::computeAVGBytes(svm_ImageAccAVG* acc_image)
{
  t2_Assert (acc_image != NULL);

  // allocate byte array for final output sub-image
  if (output_bytes != NULL)
     delete [] output_bytes;
  u_int n_outputpixels = my_width * my_height;
  t2_Alloc(output_bytes, n_outputpixels*acc_image->getPixelSize(), u_char);

  // compute the final output pixel byte values for the output sub-image
  t2_Assert (my_num_sumcount_ints
			== n_outputpixels * (acc_image->getPixelSize()+1));
  u_char *cur_byte = output_bytes;
  u_int *cur_sumcount = my_sumcount;
  for (u_int i=0; i<n_outputpixels; i++) {
     u_int count = cur_sumcount[acc_image->getPixelSize()];
     if (count == 0) {
	for (u_int j=0; j<acc_image->getPixelSize(); j++) {
	   *cur_byte = 0;
	   cur_byte++;
	}
	cur_sumcount += acc_image->getPixelSize() + 1;
     }
     else {
	for (u_int j=0; j<acc_image->getPixelSize(); j++) {
	   *cur_byte = (u_char) (*cur_sumcount / count);
	   cur_byte++;
	   cur_sumcount++;
	}
	cur_sumcount++;
     }
  }
}


T2_UDFRet 
svm_aggrAVG::finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
			  T2_Output*& outp)
{
  svm_ImageAccAVG *acc_image = (svm_ImageAccAVG *) &acc;

  // allocate byte array for final output sub-image, and compute the
  // byte values for the output pixels in the output sub-image
  computeAVGBytes(acc_image);

  // create the output sub-image object
  outp = (T2_Output *) new
	svm_OutputSubImage(packet_no, packet_type,
			 qinfo.getNumberNodes(), qinfo.getNumberTiles(),
			 my_x_pos, my_y_pos, my_width, my_height, output_bytes,
			 acc_image->getPixelSize(), acc_image->getMaxValue(),
			 acc_image->getTotalOutputImageWidth(),
			 acc_image->getTotalOutputImageHeight());

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVG, svm_aggrAVGPPM constructors
//
// Expecting the following numbers from the user argument, which should be
// used in the header of the finalized output.
//	u_int packet_number
//	u_int packet_type
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggregationAVGConstructor(const T2_QueryInfo& qinfo, T2_UsrArgReader& arg,
			      T2_AggregateFuncObj*& afp)
{
  u_int packet_no, packet_type;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(packet_no)+sizeof(packet_type)) {
     // not enough arguments !!
    t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(packet_no)+sizeof(packet_type));
     return T2_UDFRet_ERROR;
  }

  arg >> packet_no >> packet_type;

  afp = (T2_AggregateFuncObj *) new svm_aggrAVG(packet_no, packet_type);
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggregationAVGPPMConstructor(const T2_QueryInfo& qinfo,T2_UsrArgReader& arg,
				 T2_AggregateFuncObj*& afp)
{
  u_int packet_no, packet_type;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(packet_no)+sizeof(packet_type)) {
     // not enough arguments !!
    t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(packet_no)+sizeof(packet_type));
     return T2_UDFRet_ERROR;
  }

  arg >> packet_no >> packet_type;

  afp = (T2_AggregateFuncObj *) new svm_aggrAVGPPM(packet_no, packet_type);
  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVGPPM methods
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet 
svm_aggrAVGPPM::finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
			  T2_Output*& outp)
{
  svm_ImageAccAVG *acc_image = (svm_ImageAccAVG *) &acc;

  // allocate byte array for final output sub-image, and compute the
  // byte values for the output pixels in the output sub-image
  computeAVGBytes(acc_image);

  // create the output sub-image object
  outp = (T2_Output *) new
	svm_OutputSubImagePPM(packet_no, packet_type,
			 qinfo.getNumberNodes(), qinfo.getNumberTiles(),
			 my_x_pos, my_y_pos, my_width, my_height, output_bytes,
			 acc_image->getPixelSize(), acc_image->getMaxValue(),
			 acc_image->getTotalOutputImageWidth(),
			 acc_image->getTotalOutputImageHeight());

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVGFast methods
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggrAVGFast::dafAcc(const T2_QueryInfo& qinfo,
			const T2_Cluster& input, const T2_Box& mbr,
			T2_ClusterAuxInfoReader& meta,
			T2_Dataset& dataset, T2_DSID dsid,
			T2_IteratorID dsiterid,
			const T2_Box& iqr, const T2_Region& tilereg,
			T2_ProjectFuncObj& proj,
			u_int acciterid, T2_Accumulator& acc)
{
  // this code uses the same code that svm_dataset::genIterator() to
  // find out the range of pixels that it needs to iterate through,
  // and then just go through the byte array to do the aggregation

  // ignores dataset iterator id

  // only deals with 2-dim query window
  if (mbr.getNumberDimensions() != 2 || iqr.getNumberDimensions() != 2 ||
      tilereg.getNumberDimensions() != 2)
  {
      t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot handle attribute space other than 2-dim.");
      return T2_UDFRet_ERROR;
  }

  // only handles with 1 data block from the cluster
  if (input.getNumberBlocks() != 1) {
      t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting a cluster with 1 data block, but got one with %d blocks.",
                input.getNumberBlocks());
      return T2_UDFRet_ERROR;
  }
  assert (input.getDataPointer(0) != NULL);

  // for now, only knows how to handle a tile region with one single box
  switch (tilereg.getNumberBoxes()) {
     case 0:	// an empty region, so nothing to aggregate
		return T2_UDFRet_OK;

     case 1:	// yes, I can handle this, so continue execution
		break;

     default:	// ooops, cannot handle more boxes
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: cannot handle regions with %d boxes",
			tilereg.getNumberBoxes());
		return T2_UDFRet_ERROR;
  }

  const T2_Box& tilebox = tilereg[0];
  u_int tile_x_low = (u_int) tilebox.getLow()[0],
        tile_y_low = (u_int) tilebox.getLow()[1],
        tile_x_high = (u_int) tilebox.getHigh()[0],
        tile_y_high = (u_int) tilebox.getHigh()[1];

  // the starting and ending offsets for the input cluster pixels
  // and the accumulator pixels, relative to their top-left pixels,
  // in the intersecting box between mbr, iqr, and tilebox
  u_int in_x_start, in_y_start, in_x_end, in_y_end,
	out_x_start, out_y_start, out_x_end, out_y_end;

  // the number of input pixels that map to an output pixel on the starting
  // (top-left) boundary of the intersecting box between mbr, iqr and tilebox,
  // an output pixel between the two boundaries of the intersecting box,
  // and an output pixel on the ending (bottom-right) boundary of the 
  // intersecting box
  u_int x_start_size, y_start_size, x_mid_size, y_mid_size,
	x_end_size, y_end_size;

  // compute all the variables declared above
  if (findIntersection(qinfo, mbr, iqr, tilebox, proj,
		       in_x_start, in_y_start, in_x_end, in_y_end,
		       out_x_start, out_y_start, out_x_end, out_y_end,
		       x_start_size, y_start_size,
		       x_mid_size, y_mid_size,
		       x_end_size, y_end_size)
      != T2_UDFRet_OK)
     return T2_UDFRet_ERROR;		// error occurs

  // Need to compute in_skip_bytes and out_skip_bytes, which are the number
  // of bytes to skip in the input cluster and the accumulator, respectively,
  // when advancing from a row to the next one.
  svm_ImageDataset *image_dataset = (svm_ImageDataset *) & dataset;
  svm_ImageAccAVG *image_acc = (svm_ImageAccAVG *) & acc;
  t2_Assert (image_dataset->getPixelSize() == image_acc->getPixelSize());
  const u_int pixelsize = image_acc->getPixelSize(),
	      in_width = (u_int) mbr.getHigh()[0] - (u_int) mbr.getLow()[0] + 1,
	      out_width = tile_x_high - tile_x_low + 1;
  u_int in_skip_bytes = (in_width - in_x_end + in_x_start - 1) * pixelsize,
	out_skip_bytes = out_width * (pixelsize+1);

  // now iterate over the output and input pixels to do aggregation
  const u_char *in_ptr = (const u_char *) input.getDataPointer(0);
  in_ptr += pixelsize * (in_x_start + in_y_start * in_width);
  u_int *out_ptr = image_acc->getSumCount();
  out_ptr += (pixelsize+1) * (out_x_start + out_y_start * out_width);
  for (u_int out_y=out_y_start; out_y<=out_y_end; out_y++) {
      u_int in_num_y;
      if (out_y == out_y_start)
          in_num_y = y_start_size;
      else if (out_y == out_y_end)
          in_num_y = y_end_size;
      else
          in_num_y = y_mid_size;

      for (u_int ny=0; ny<in_num_y; ny++) {
           u_int *cur_out_ptr = out_ptr;	// back to current output row

           for (u_int out_x=out_x_start; out_x<=out_x_end; out_x++) {
                u_int in_num_x;
                if (out_x == out_x_start)
                    in_num_x = x_start_size;
                else if (out_x == out_x_end)
                    in_num_x = x_end_size;
		else
                    in_num_x = x_mid_size;

                for (u_int nx=0; nx<in_num_x; nx++) {
                     for (u_int i=0; i<pixelsize; i++) {
                          cur_out_ptr[i] += (u_int) *in_ptr;
                          in_ptr++;
                     }
                     cur_out_ptr[pixelsize]++;	// increment count
                 }
                 cur_out_ptr += pixelsize + 1;
           }
           in_ptr += in_skip_bytes;
      }
      out_ptr += out_skip_bytes;
  }

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggrAVGPPMFast methods
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet 
svm_aggrAVGPPMFast::finalize(const T2_QueryInfo& qinfo, T2_Accumulator& acc,
			     T2_Output*& outp)
{
  // This method should be exactly the same as svm_aggrAVGPPM::finalize()

  svm_ImageAccAVG *acc_image = (svm_ImageAccAVG *) &acc;

  // allocate byte array for final output sub-image, and compute the
  // byte values for the output pixels in the output sub-image
  computeAVGBytes(acc_image);

  // create the output sub-image object
  outp = (T2_Output *) new
	svm_OutputSubImagePPM(packet_no, packet_type,
			 qinfo.getNumberNodes(), qinfo.getNumberTiles(),
			 my_x_pos, my_y_pos, my_width, my_height, output_bytes,
			 acc_image->getPixelSize(), acc_image->getMaxValue(),
			 acc_image->getTotalOutputImageWidth(),
			 acc_image->getTotalOutputImageHeight());

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////
//
// svm_aggregationAVGFastConstructor
// svm_aggregationAVGPPMFastConstructor
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_aggregationAVGFastConstructor(const T2_QueryInfo& qinfo,
				     T2_UsrArgReader& arg,
				     T2_AggregateFuncObj*& afp)
{
  u_int packet_no, packet_type;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(packet_no)+sizeof(packet_type)) {
     // not enough arguments !!
    t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(packet_no)+sizeof(packet_type));
     return T2_UDFRet_ERROR;
  }

  arg >> packet_no >> packet_type;

  afp = (T2_AggregateFuncObj *) new svm_aggrAVGFast(packet_no, packet_type);
  return T2_UDFRet_OK;
}


T2_UDFRet
svm_aggregationAVGPPMFastConstructor(const T2_QueryInfo& qinfo,
				     T2_UsrArgReader& arg,
				     T2_AggregateFuncObj*& afp)
{
  u_int packet_no, packet_type;
  const size_t argsize = arg.getDataSize();
  if (argsize < sizeof(packet_no)+sizeof(packet_type)) {
     // not enough arguments !!
    t2_err_ret(__FILE__, __LINE__,
		"ERROR: argument size = %d bytes, but expect %d bytes.",
		argsize, sizeof(packet_no)+sizeof(packet_type));
     return T2_UDFRet_ERROR;
  }

  arg >> packet_no >> packet_type;

  afp = (T2_AggregateFuncObj *) new svm_aggrAVGPPMFast(packet_no, packet_type);
  return T2_UDFRet_OK;
}

