
/* $Id: svm_dataset.C,v 1.3 2000/02/14 19:37:39 chialin Exp $ */

/* svm_dataset.C */


#include	<iostream.h>
#include	<unistd.h>
#include	<ctype.h>
#include	"svm_utility.h"
#include	"svm_dataset.h"


////////////////////////////////////////////////////////////////////////////
//
// app_ImageDataset methods
//
////////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageDataset::genIterator(const T2_System& system, T2_IteratorID i,
			      const T2_Cluster& c,
			      const T2_Box& mbr, T2_ClusterAuxInfoReader& meta,
			      const T2_Box& iqr, const T2_Region& tilereg,
			      T2_ProjectFuncObj& prj, T2_Iterator*& iterp)
{
  // only generates one kind of iterator, so ignore i

  // only deals with 2-dim query window
  if (mbr.getNumberDimensions() != 2 || iqr.getNumberDimensions() != 2 ||
      tilereg.getNumberDimensions() != 2)
  {
      t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot handle attribute space other than 2-dim.");
      return T2_UDFRet_ERROR;
  }

  // only handles with 1 data block from the cluster
  if (c.getNumberBlocks() != 1) {
      t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting a cluster with 1 data block, but got one with %d blocks.",
		c.getNumberBlocks());
      return T2_UDFRet_ERROR;
  }
  assert (c.getDataPointer(0) != NULL);

  // for now, only knows how to handle a tile region with one single box
  switch (tilereg.getNumberBoxes()) {
     case 0:	// an empty region, so nothing to iteratre over
		// use range (1,1) - (0,0) for nothing to iterate
		iterp = (T2_Iterator *)
			new svm_ImageDatasetIterator(0, 0, 0, 0, 0, NULL,
						     1, 1, 0, 0);
		return T2_UDFRet_OK;

     case 1:	// yes, I can handle this, so continue execution
		break;

     default:	// ooops, cannot handle more boxes
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: cannot handle regions with %d boxes",
			tilereg.getNumberBoxes());
		return T2_UDFRet_ERROR;
  }

  // get the coordinates of mbr first
  u_int mbr_x_low = (u_int) mbr.getLow()[0],
	mbr_y_low = (u_int) mbr.getLow()[1],
	mbr_x_high = (u_int) mbr.getHigh()[0],
	mbr_y_high = (u_int) mbr.getHigh()[1];

  // find out the intersection between the mbr of the cluster, mbr, and the
  // input query window iqr to find out pixels of the cluster c that fall 
  // inside the query region, iqr; (I am ignoring the tile region, tilereg);
  // intersection is easy in this case, as both the bounding box of the
  // block (mbr) and the query region (iqr) are just simple rectangles
  // that are parallel to the two axis;
  // then need to find out the x and y offset into the chunk, relative
  // to the top-left pixel of the chunk, to quickly locate the pixels;
  // that is, all pixels with local offset between (x_start, y_start) upto
  // (x_end, y_end) need to be iterated over by the iterator;
  // this is all done by calling findIntersectionInLocalOffsets()
  u_int x_start, y_start, x_end, y_end;
  if (findIntersectionInLocalOffsets(
		mbr_x_low, mbr_y_low, mbr_x_high, mbr_y_high,
		(u_int) iqr.getLow()[0], (u_int) iqr.getLow()[1],
		(u_int) iqr.getHigh()[0], (u_int) iqr.getHigh()[1],
		x_start, y_start, x_end, y_end)
     == false)
  {						// no intersection
     iterp = (T2_Iterator *)
		new svm_ImageDatasetIterator(0, 0, 0, 0, 0, NULL, 1, 1, 0, 0);
  }
  else {
     iterp = (T2_Iterator *)
		new svm_ImageDatasetIterator(mbr_x_low, mbr_y_low,
					     mbr_x_high - mbr_x_low + 1,
					     mbr_y_high - mbr_y_low + 1,
					     pixelsize,
					     (u_char *) c.getDataPointer(0),
				             x_start, y_start, x_end, y_end);
  }

  // just read the aux information for demonstration purpose;
  // recall that x_pos and y_pos are stored in meta
  u_int x_pos, y_pos;
  meta >> x_pos >> y_pos;

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////////
//
// app_ImageDataset constructor
//
////////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageDatasetConstructor(const T2_System& system,
			    const T2_Array<int>& dfd, const T2_Array<int>& mdfd,
			    T2_Dataset*& dsp)
{
  if (mdfd.getNumberElements() != 1) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: only expecting one aux file.");
     return T2_UDFRet_ERROR;
  }

  // use streams instead of file descriptors
  FILE *auxfp = fdopen(mdfd[0], "r");
  if (auxfp == NULL) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: fdopen on aux file fails.");
     return T2_UDFRet_ERROR;
  }

  // read a line into line buffer
  const size_t max_line_size = 512;
  char line[max_line_size];
  if (get_uncomment_line(line, max_line_size, auxfp, '#') == 0) {
     // unexpected eof
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: unexpected eof when reading aux file.");
     fclose(auxfp);
     return T2_UDFRet_ERROR;
  }

  // assuming that the aux file has the following fields in ascii format on
  // the same line:
  //  <image width> <image height> <number of bytes per pixel> <max byte value>
  int width, height, pixelsize;
  u_int maxval;
  sscanf(line, "%d %d %d %d", &width, &height, &pixelsize, &maxval);
  if (width < 0  || height < 0 || pixelsize < 0) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: invalid values in aux file.");
     fclose(auxfp);
     return T2_UDFRet_ERROR;
  }

  dsp = (T2_Dataset *) new svm_ImageDataset(width, height, pixelsize, maxval);
  fclose(auxfp);

  return T2_UDFRet_OK;
}



////////////////////////////////////////////////////////////////////////////
//
// svm_ImageDatasetIterator methods
//
////////////////////////////////////////////////////////////////////////////
T2_UDFRet
svm_ImageDatasetIterator::
getNextElement(const T2_System& system, const T2_Cluster& c,
	       bool& eod, const void*& e, T2_Point& coord)
{
  u_int x_off, y_off;			// offsets for pixel to return

  if (svm_ImageIterator::getNextElement(x_off, y_off, pixel) == false) {
     // end-of-data is true
     eod = true;
     return T2_UDFRet_OK;
  }

  // compute coord
  coord.setNumberDimensions(2);		// 2-dim point
  coord[0] = x_pos + x_off;
  coord[1] = y_pos + y_off;

  e = (const void *) &pixel;
  eod = false;
  return T2_UDFRet_OK;
}

