
/* $Id: svm_image.h,v 1.3 2000/03/09 01:32:20 chialin Exp $ */

/* svm_image.h */

#ifndef _SVM_IMAGE_H_
#define _SVM_IMAGE_H_

#include	<sys/types.h>
#include	"t2_header.h"


// forward declarration
class svm_ImageIterator;

//////////////////////////////////////////////////////////////////////////
//
// a class for an image
//
//////////////////////////////////////////////////////////////////////////
class svm_Image {
public:
  svm_Image(): width(0), height(0), pixelsize(0), maxval(0), bytes(NULL) {};
  ~svm_Image() { delete [] bytes; };

  // access methods
  int getNumberPixels() const { return width*height; };
  int& getWidth() { return width; };
  int& getHeight() { return height; };
  u_int& getPixelSize() { return pixelsize; };
  u_int& getMaxValue() { return maxval; };
  u_char*& getBytes() { return bytes; };

  // access the z-th component of a pixel with position x and y
  u_char getPixelByte(int x, int y, u_int z) const;

  // set the z-th component of a pixel with position x and y;
  // return 0 if succeeds, 1 otherwise
  int setPixelByte(int x, int y, u_int z, u_char v);

  // wipe out the image and remove all pixels;
  void clear() { width = height = pixelsize = 0; maxval = 0;
		 delete [] bytes; bytes=NULL; };

  // create a sub-image starting from position x_pos, y_pos with width=w
  // and height=h; returns 0 if succeeds, 0 otherwise
  int copySubImage(svm_Image& sub_image, int x_pos, int y_pos, int w, int h);
  
protected:
  int width, height;		// number of pixels for the image
  u_int pixelsize,		// number of bytes per pixel
      maxval;			// maximum color-component value
  u_char* bytes;		// an array of size width*height*bytes

private:
};



///////////////////////////////////////////////////////////////////////
//
// svm_ImagePixel
//
// A class that points to a pixel inside an image.
//
///////////////////////////////////////////////////////////////////////
class svm_ImagePixel {
friend class svm_ImageIterator;
public:
  svm_ImagePixel(): nbytes(0), bytes(NULL) {};
  ~svm_ImagePixel() {};

  // access methods
  u_int getNumberBytes() const { return nbytes; };
  const u_char* getBytes() const { return bytes; };
  u_char operator [] (u_int i) const { return bytes[i]; };
  u_char& operator [] (u_int i) { return bytes[i]; };

protected:
  u_int nbytes;
  u_char* bytes;		// always point to char[] owned by some other 
				// object, so never need to delete it

  // set the bytes in the pixel
  void setBytes(u_int nb, u_char* b) { nbytes = nb; bytes = b; };

private:
};



///////////////////////////////////////////////////////////////////////
//
// svm_ImageIterator
//
// This is a class that would iterator through a given range of pixels
// in an smv_Image.
// The range is given in terms of relative offsets from the first pixel
// (ie the top-left pixel). This class assumes that a row-major order
// is used for the pixels.
//
///////////////////////////////////////////////////////////////////////
class svm_ImageIterator {
public:
  svm_ImageIterator(u_int w, u_int h, u_int psize, u_char* bytes,
		    u_int x_s, u_int y_s, u_int x_e, u_int y_e);
  virtual ~svm_ImageIterator() {};

  // get the next element; returns true if more to return, false if end-of-data;
  // when return true, x_off and y_off gives the relative offset from the
  // 1st pixel of the pixel returned;
  bool getNextElement(u_int& x_off, u_int& y_off, svm_ImagePixel& pixel);

protected:
  // width, height and pixelsize describe how the given byte array should
  // be interpretated
  u_int width, height,		// width and height of the cluster
	pixelsize;		// number of bytes per pixel
  u_char *bytes;		// byte array for pixels

  // x_start, y_start and x_end, y_end gives the range of pixels that the
  // iterator should iterate over; these are all relative to the 1st pixel
  // in the byte array (ie the 1st pixel is considered with coordinates (0,0) )
  u_int x_start, y_start,	// starting position, relative to the top-left
				// corner of the chunk, for the pixels to be
				// iterated over
	x_end, y_end;		// end position, relative to the top-left
				// corner of the chunk, for the pixels to be
				// iterated over
				// (ie all pixels between x_start and x_end,
				// between y_start and y_end need to be
				// iterated over)

  // these are the variables used to keep track of the next pixel to return
  u_int x_next, y_next;		// position of next pixel to return
  u_char *pixel_next;		// start of the next pixel to return
  size_t skip_bytes;		// this is the number of bytes to skip when
				// advancing a char pointer from pixel
				// (x_end, i) to (x_start, i+1)

private:
};


inline
svm_ImageIterator::
svm_ImageIterator(u_int w, u_int h, u_int psize, u_char* b,
		  u_int x_s, u_int y_s, u_int x_e, u_int y_e)
		: width(w), height(h), pixelsize(psize), bytes(b),
		  x_start(x_s), y_start(y_s), x_end(x_e), y_end(y_e),
		  x_next(x_s), y_next(y_s)
{
  if (x_end >= w)		// too wide, cut it back
     x_end = w-1;
  if (y_end >= h)		// too high, cut it back
     y_end = h-1;
  pixel_next = bytes + (x_start + width * y_start) * pixelsize;
  skip_bytes = (width - x_end + x_start) * pixelsize;
}





#endif	/* ! _SVM_IMAGE_H_ */
