
/* $Id: svm_output.C,v 1.4 2000/03/09 01:35:10 chialin Exp $ */

/* svm_output.C */

#include	"svm_output.h"


////////////////////////////////////////////////////////////////////////
//
// svm_OutputImage methods
//
// This class implements the protocol between the back-end and the
// display client. The protocol defines the following sequence of
// data:
//	u_int packet_no		-- packet number from aggregation function
//	u_int packet_type	-- packet type from aggregation function
//	u_int nprocs		-- total number of back-end nodes
//	u_int ntiles		-- total number of tiles
//	u_int x_res		-- width of the entire output image
//	u_int y_res		-- height of the entire output image
//	u_int x_pos		-- x position of sub-image in the entire
//				   output image
//	u_int y_pos		-- y position of sub-image in the entire
//				   output image
//	u_int sub_x_res		-- width of the sub-image
//	u_int sub_y_res		-- height of the sub-image
//	u_char[] bytes		-- byte array of sub_x_res*sub_y_res*3 bytes
//
// And all ints are assumed to be in big-endian.
//
////////////////////////////////////////////////////////////////////////
size_t
svm_OutputSubImage::getOutputBufferSize(const T2_QueryInfo& qinfo)
{
  assert (pixelsize == 3);

  // need to allocate buffer for the header
  return sizeof(packet_no) + sizeof(packet_type)
	 + sizeof(nprocs) + sizeof(ntiles)
	 + sizeof(x_res) + sizeof(y_res)
	 + sizeof(x_pos) + sizeof(y_pos)
	 + sizeof(sub_x_res) + sizeof(sub_y_res);
}


T2_UDFRet
svm_OutputSubImage::flushOutput(const T2_QueryInfo& qinfo, T2_OutputBuffer& buf,
				T2_OutputDataPtrList& ptr)
{
  const char *p1 = buf.getCurrentPosition(),	// pointer to header
             *p2;				// a temp pointer
  buf << packet_no << packet_type
      << nprocs << ntiles << x_res << y_res << x_pos << y_pos
      << sub_x_res << sub_y_res;
  p2 = buf.getCurrentPosition();
  ptr.insertDataPointer(p1, p2 - p1);		// insert pointer to header

  u_int npixels = sub_x_res*sub_y_res;
  if (npixels > 0)
     ptr.insertDataPointer((const char*) bytes, npixels * pixelsize);

  if (qinfo.getBackEndEndian() != qinfo.getExpectedQueryOutputEndian()) {
     // client receiving output expects output in different endian,
     // so swap the bytes of all data to be written;

     // first swap the bytes of all variables written into buf;
     // since all these variables are of the same type, u_int, I can
     // easily swap the bytes for the entire u_int array
     t2_swapBytesArray(p1, 10, sizeof(u_int));

     // well, since the pixels are really an array of u_char, so no need
     // to swap bytes
  }

  return T2_UDFRet_OK;
}


////////////////////////////////////////////////////////////////////////
//
// svm_OutputImagePPM methods
//
////////////////////////////////////////////////////////////////////////
size_t
svm_OutputSubImagePPM::getOutputBufferSize(const T2_QueryInfo& qinfo)
{
  t2_AssertOrDo (pixelsize == 3, cout << "pixelsize=" << pixelsize << endl);
  assert (pixelsize == 3);

  // assume we have a bound on how many digits any number in the PPM file
  // header can have, including the white space before it
  const size_t max_number_digits = 8;

  // need to allocate buffer for the header
  return sizeof(char) * 2		// for "P6"
	+ max_number_digits * 3		// for width, height, maxval
	+ sizeof(char) * 2		// for comment marker "# "
	+ (max_number_digits+8) * 8;	// for comments: packet_no, packet_type,
					// nprocs, ntiles, x_res, y_res, x_pos,
					// y_pos
}


T2_UDFRet
svm_OutputSubImagePPM::flushOutput(const T2_QueryInfo& qinfo,
				   T2_OutputBuffer& buf,
				   T2_OutputDataPtrList& ptr)
{
  // write header
  char header[512];
  sprintf(header, "P6\n# packet: %d %d\n# nprocs=%d ntiles=%d x_res=%d y_res=%d x_pos=%d y_pos=%d\n%d %d %d\n",
	  packet_no, packet_type, nprocs, ntiles, x_res, y_res, x_pos, y_pos,
	  sub_x_res, sub_y_res, maxval);
  const char *p1 = buf.getCurrentPosition(),	// pointer to header
             *p2;				// a temp pointer
  buf.writeString(header);
  p2 = buf.getCurrentPosition();
  ptr.insertDataPointer(p1, p2 - p1 - 1);	// insert pointer to header
						// (-1 to avoid '\0' at the
						/// end of header)
  u_int npixels = sub_x_res*sub_y_res;
  if (npixels > 0)
     ptr.insertDataPointer((const char*) bytes, npixels * pixelsize);

  return T2_UDFRet_OK;
}

