
/* $Id: svm_utility.C,v 1.4 2000/03/09 01:35:48 chialin Exp $ */

/* svm_utility.C */

#include 	<iostream.h>
#include 	<strings.h>
#include 	"svm_utility.h"


///////////////////////////////////////////////////////////////////////////
//
// get_uncomment_line
//
// Get the next uncomment-ed line, where comment is all texts in a line
// after the comment-marker.
// Returns 1 if such a line is found, 0 otherwise.
//
///////////////////////////////////////////////////////////////////////////
int
get_uncomment_line(char* line, size_t max_line_size, FILE* fp, const char mark)
{
  while (1) {
     if (fgets(line, max_line_size, fp) == NULL) 	// eof
        return 0;

     size_t len = strlen(line),
	    i;

     for (i=0; i<len; i++) 		// look for first non-space char
	if (!isspace(line[i]))
	   break;

     if (line[i] != '\0' && line[i] != mark)
	break;

     // this is either an empty line or a comment line, so read next line
  }
  return 1;
}



///////////////////////////////////////////////////////////////////////////
//
// findIntersectionInLocalOffsets
//
// Find out the intersection between two boxes described by arguments
// mbr_* and iqr_*, and describe the intersecting box in terms of 
// x and y offsets into the box described by mbr_*, relative to the 
// top-left corner of the box described by mbr_*.
// That is, all pixels with local offset between (x_start, y_start) upto
// (x_end, y_end) in the box described by mbr_* are inside the 
// intersecting box.
// Returns true if an intersection is found, false otherwise, in which
// case x_start, y_start, x_end and y_end are undefined.
//
///////////////////////////////////////////////////////////////////////////
bool
findIntersectionInLocalOffsets(const u_int mbr_x_low, const u_int mbr_y_low,
			       const u_int mbr_x_high, const u_int mbr_y_high,
			       const u_int iqr_x_low, const u_int iqr_y_low,
			       const u_int iqr_x_high, const u_int iqr_y_high,
			       u_int& x_start, u_int& y_start,
			       u_int& x_end, u_int& y_end)
{
  if (iqr_x_low > mbr_x_high || iqr_x_high < mbr_x_low ||
      iqr_y_low > mbr_y_high || iqr_y_high < mbr_y_low)	   // no intersection
     return false;

  // intersection is easy in this case, as both the bounding box of the
  // block (mbr) and the query region (iqr) are just simple rectangles
  // that are parallel to the two axis;
  // then need to find out the x and y offset into the mbr, relative
  // to the top-left pixel of the mbr, to quickly locate the pixels;
  // that is, all pixels with local offset between (x_start, y_start) upto
  // (x_end, y_end) are inside the intersecting box
  x_start = (iqr_x_low > mbr_x_low)? (iqr_x_low - mbr_x_low) : 0;
  y_start = (iqr_y_low > mbr_y_low)? (iqr_y_low - mbr_y_low) : 0;
  x_end = (iqr_x_high > mbr_x_high)?
			(mbr_x_high - mbr_x_low) : (iqr_x_high - mbr_x_low);
  y_end = (iqr_y_high > mbr_y_high)?
			(mbr_y_high - mbr_y_low) : (iqr_y_high - mbr_y_low);
  return true;
}



///////////////////////////////////////////////////////////////////////////
//
// adjust_start_end
//
// This method computes along some given dimension, the information 
// about a box that is the intersection of a box out_box, represented by
// out_start_coord and out_end_coord, and a box tile_box, represented by
// tile_start_coord, tile_end_coord. These two boxes are both in the
// output attribute space, while thhe box in_box, represented by
// in_start_coord and in_end_coord, is the same box as out_box but in
// the input attribute space. The projection from in_box to out_box
// is done through a coordinate translation with origin orig and scale
// size zoom_size. The range specified by tile_start and tile_end
// marks the offsets, relative to top-left corner of tile_box, of
// the intersecting box, while the range specified by in_start
// and in_end marks a box that might be slightly larger than the
// intersecting box. This function would update in_start and in_end
// to accurately reflect the intersecting box. This function also
// computes the number of input pixels that map to the output pixels
// falling on the boundaries of the intersecting boxes, in-start_size
// and in_end_size, as these output pixels may be mapped to by fewer than 
// zoom_size input pixels. All the output pixels not on the boundaries
// of the intersecting box are mapped to by exactly zoom_size input pixels.
//
///////////////////////////////////////////////////////////////////////////
void
adjust_start_end(u_int orig, u_int zoom_size,
		const u_int in_start_coord, const u_int in_end_coord,
		const u_int out_start_coord, const u_int out_end_coord,
		const u_int tile_start_coord,const u_int tile_end_coord,
		const u_int tile_start, const u_int tile_end,
		u_int& in_start, u_int& in_end, 
		u_int& in_start_size, u_int& in_end_size)
{
  // deal with the starting side first
  if (tile_start == 0) {
     if (out_start_coord < tile_start_coord) {
	u_int in_diff = (tile_start_coord - out_start_coord - 1) * zoom_size;
	if (in_start_coord < orig)
	   in_diff += (orig - in_start_coord) % zoom_size;
	else
	   in_diff += zoom_size - ((in_start_coord - orig) % zoom_size);
	in_start += in_diff;
	in_start_size = zoom_size;
     }
     else {		// if (out_start_coord < tile_start_coord)
	t2_Assert (out_start_coord == tile_start_coord);
	t2_Assert (in_start_coord >= orig);
	in_start_size = zoom_size - ((in_start_coord - orig) % zoom_size);
     }				// if (out_start_coord < tile_start_coord) else
  }
  else {			// if (tile_start == 0)
	t2_Assert (in_start_coord > orig);
	in_start_size = zoom_size - ((in_start_coord - orig) % zoom_size);
  }					// if (tile_start == 0) else

  // deal with the ending side
  if (tile_end == (tile_end_coord - tile_start_coord)) {
     if (out_end_coord > tile_end_coord) {
	u_int in_diff = (out_end_coord - tile_end_coord - 1) * zoom_size
				+ ((in_end_coord - orig) % zoom_size) + 1;
	t2_Assert (in_end_coord >= in_diff);

	in_end -= in_diff;
	in_end_size = zoom_size;
     }
     else {	// if (out_end_coord > tile_end_coord)
	t2_Assert (out_end_coord == tile_end_coord);
	t2_Assert (in_end_coord >= orig);
	in_end_size = ((in_end_coord - orig) % zoom_size) + 1;
     }		// if (out_end_coord > tile_end_coord) else
  }
  else {	// if (tile_end == (tile_end_coord - tile_start_coord))
	t2_Assert (in_end_coord >= orig);
	in_end_size = ((in_end_coord - orig) % zoom_size) + 1;
  }		// if (tile_end == (tile_end_coord - tile_start_coord + 1)) else
}



////////////////////////////////////////////////////////////////////////
//
// findIntersection
//
////////////////////////////////////////////////////////////////////////
T2_UDFRet
findIntersection(const T2_QueryInfo& qinfo,
		 const T2_Box& mbr, const T2_Box& iqr, const T2_Box& tilebox,
		 T2_ProjectFuncObj& proj,
		 u_int& in_x_start, u_int& in_y_start, 
		 u_int& in_x_end, u_int& in_y_end,
		 u_int& out_x_start, u_int& out_y_start,
		 u_int& out_x_end, u_int& out_y_end,
		 u_int& x_start_size, u_int& y_start_size, 
		 u_int& x_mid_size, u_int& y_mid_size, 
		 u_int& x_end_size, u_int& y_end_size)
{
  if (findIntersectionInLocalOffsets(
		(u_int) mbr.getLow()[0], (u_int) mbr.getLow()[1],
		(u_int) mbr.getHigh()[0], (u_int) mbr.getHigh()[1],
		(u_int) iqr.getLow()[0], (u_int) iqr.getLow()[1],
		(u_int) iqr.getHigh()[0], (u_int) iqr.getHigh()[1],
		in_x_start, in_y_start, in_x_end, in_y_end)
     == false)			// no intersection, so nothing to aggregate
     return T2_UDFRet_OK;

  // compute coordinates of the intersecting box in the input attribute space
  T2_Box in_ibox(2);
  in_ibox.getLow()[0] = mbr.getLow()[0] + in_x_start;
  in_ibox.getLow()[1] = mbr.getLow()[1] + in_y_start;
  in_ibox.getHigh()[0] = mbr.getLow()[0] + in_x_end;
  in_ibox.getHigh()[1] = mbr.getLow()[1] + in_y_end;

  // project the intersecting box into the output attribute space
  T2_Box out_ibox(2);
  bool intersects;
  if (proj.projectBox(qinfo.getSystem(), in_ibox, intersects, out_ibox)
       != T2_UDFRet_OK)		// error occurs
       return T2_UDFRet_ERROR;
  t2_Assert (intersects == true);

  // use the projected intersecting box to find out the range of output pixels
  // that fall inside the projected intersecting box
  if (findIntersectionInLocalOffsets(
		(u_int) tilebox.getLow()[0], (u_int) tilebox.getLow()[1],
		(u_int) tilebox.getHigh()[0], (u_int) tilebox.getHigh()[1],
		(u_int) out_ibox.getLow()[0], (u_int) out_ibox.getLow()[1],
		(u_int) out_ibox.getHigh()[0], (u_int) out_ibox.getHigh()[1],
		out_x_start, out_y_start, out_x_end, out_y_end)
     == false)			// no intersection, so nothing to aggregate
     return T2_UDFRet_OK;

  // the following implementation relies on the fact that the projeection
  // function is svm_ImageProjection
  svm_ImageProjection* svm_prj = (svm_ImageProjection *) &proj;
  int x_orig = svm_prj->getXOrigin(),
      y_orig = svm_prj->getYOrigin();
  u_int x_size = svm_prj->getXSize(),
	y_size = svm_prj->getYSize();

  // in the case where the intersecting box is not entirely contained within
  // the tilebox, we may have to adjust in_x_start, in_y_start, in_x_end 
  // and in_y_end to conform to the tilebox;
  //
  // For example:
  // Here is the tricky point. Let's say that the top-left input pixel pi
  // in the intersecting box, (x_start, y_start) -- (x_end, y_end),
  // maps to an output pixel po. The pixel on the right of pi may map to
  // output pixel po, or to the output pixel on the right of po.
  // Here is the strategy. Get from the projection function x_orig, y_orig,
  // x_size and y_size, and find out x_start_size and y_start_size, which 
  // are the number of input pixles that map to output pixel po.
  x_mid_size = x_size;
  y_mid_size = y_size;
  adjust_start_end(x_orig, x_size,
		(u_int) in_ibox.getLow()[0], (u_int) in_ibox.getHigh()[0],
		(u_int) out_ibox.getLow()[0], (u_int) out_ibox.getHigh()[0],
		(u_int) tilebox.getLow()[0], (u_int) tilebox.getHigh()[0],
		out_x_start, out_x_end,
		in_x_start, in_x_end, x_start_size, x_end_size);
  adjust_start_end(y_orig, y_size,
		(u_int) in_ibox.getLow()[1], (u_int) in_ibox.getHigh()[1],
		(u_int) out_ibox.getLow()[1], (u_int) out_ibox.getHigh()[1],
		(u_int) tilebox.getLow()[1], (u_int) tilebox.getHigh()[1],
		out_y_start, out_y_end,
		in_y_start, in_y_end, y_start_size, y_end_size);

  return T2_UDFRet_OK;
}

