
/* $Id: split-ppm.C,v 1.1 2000/03/01 06:00:14 chialin Exp $ */

/* split-ppm.C 
 *
 * This utility program splits the ppm file into multiple smaller files,
 * which could be just bytes of sub-images or smaller ppm files.
 */


#include	<iostream.h>
#include	<fstream.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<assert.h>
#include	<strings.h>

#ifndef NULL
#  define NULL		0
#endif


//////////////////////////////////////////////////////////////////////////
//
// a class for an image
//
//////////////////////////////////////////////////////////////////////////
class Image {
public:
  Image(): width(0), height(0), pixelsize(0), maxval(0), bytes(NULL) {};
  ~Image() { delete [] bytes; };

  // access methods
  int getNumberPixels() const { return width*height; };
  int& getWidth() { return width; };
  int& getHeight() { return height; };
  int& getPixelSize() { return pixelsize; };
  int& getMaxValue() { return maxval; };
  char*& getBytes() { return bytes; };

  // access the z-th component of a pixel with position x and y
  char getPixelByte(int x, int y, int z) const;

  // set the z-th component of a pixel with position x and y;
  // return 0 if succeeds, 1 otherwise
  int setPixelByte(int x, int y, int z, char v);

  // wipe out the image and remove all pixels;
  void clear() { width = height = pixelsize = 0; maxval = 0;
		 delete [] bytes; bytes=NULL; };

  // create a sub-image starting from position x_pos, y_pos with width=w
  // and height=h; returns 0 if succeeds, 0 otherwise
  int copySubImage(Image& sub_image, int x_pos, int y_pos, int w, int h);
  
protected:
  int width, height,		// number of pixels for the image
      pixelsize,		// number of bytes per pixel
      maxval;			// maximum color-component value
  char* bytes;			// an array of size width*height*bytes

private:
};


//////////////////////////////////////////////////////////////////////////
//
// prototypes
//
//////////////////////////////////////////////////////////////////////////
void
printusage(const char* prog);
int
read_ppm_file(ifstream& file, Image& image);
int
get_next_uncomment_string(ifstream& file, char* buf, int buf_len, char comment);
int
read_raw_bytes(ifstream& file, int nbytes, char* bytes);
void
write_raw_ppm_header(Image& image, ofstream& os, const char *descrition);
int
split_image(Image& image, int x_part, int y_part, char output_type,
	    ofstream* outfs, ofstream& summary);

//////////////////////////////////////////////////////////////////////////
//
// main program
//
//////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  if (argc < 6) {
     printusage(argv[0]);
     return -1;
  }

  // variables
  const char* inp_imagefile;
  const char* out_imagefileprefix;
  char  output_type;
  int   x_part, y_part,
	i = 1;

  // get arguments
  inp_imagefile = argv[i++];
  x_part = atoi(argv[i++]);
  y_part = atoi(argv[i++]);
  output_type = argv[i++][0];
  out_imagefileprefix = argv[i++];
  if (x_part <=0 || y_part <=0) {
     cerr << "ERROR: invalid x/y partitioning '" << argv[2] << " " << argv[3]
	  << "'" << endl;
     printusage(argv[0]);
     return -1;
  }
  if (output_type != 'p' && output_type != 'b') {
     cerr << "ERROR: invalid output type '" << argv[4] << "'" << endl;
     printusage(argv[0]);
     return -1;
  }

  ifstream inp(inp_imagefile, ios::in);
  if (!inp) {
     cerr << "ERROR: cannot open input image file '" << inp_imagefile
	  << "'" << endl;
     return -1;
  }

  Image inp_image;
  if (read_ppm_file(inp, inp_image) != 0) {	// some thing went wrong
     inp.close();
     return -1;
  }
  inp.close();

  // get ready to open output files
  const nparts = x_part * y_part;
  ofstream *outfs = new ofstream[nparts];
  ofstream summary;
  assert (outfs != NULL);

  // output summary file
  char outfilename[256];
  sprintf(outfilename, "%s.summary", out_imagefileprefix);
  summary.open(outfilename, ios::out);
  if (!summary) {
     cerr << "ERROR: cannot open summary file '" << outfilename << "'" << endl;
     return -1;
  }
  // write summary file header
  summary << "# generated by: " << argv[0]
	  << " " << argv[1] << " " << argv[2]
	  << " " << argv[3] << " " << argv[4]
	  << " " << argv[5]
	  << endl
	  << "# format: <x_pos> <y_pos> <width> <height> <number bytes per pixel>"
	  << endl
	  << "#"
	  << endl;

  // open output files
  i = 0;
  for (int y=0; y<y_part; y++) {
     for (int x=0; x<x_part; x++) {
	sprintf(outfilename, "%s.%d-%d", out_imagefileprefix, x, y);
	outfs[i].open(outfilename, ios::out);
	if (!outfs[i]) {			// open fails
	   cerr << "ERROR: cannot open output file '" << outfilename
		<< "'" << endl;
	   for (int k=0; k<i; k++)
	      outfs[k].close();
	    return -1;
	}
	i++;
     }
  }

  int r = split_image(inp_image, x_part, y_part, output_type, outfs, summary);
  for (i=0; i<nparts; i++)
     outfs[i].close();

  delete [] outfs;

  return r;
}


//////////////////////////////////////////////////////////////////////////
//
// printusage
//
//////////////////////////////////////////////////////////////////////////
void
printusage(const char* prog)
{
  cerr << "Usage: " << prog << "\\" << endl
       << "       <ppm/pgm file> \\"
       << endl
       << "       <number of x partitions> <number of y partitions> \\"
       << endl
       << "       <output type (p/b)> <output file prefix>"
       << endl
       << "where <number of x partitions> and <number of y partitions> must be"
       << endl
       << "positive numbers."
       << endl
       << endl
       << "This program splits a ppm/pgm file into "
       << endl
       << "<number of x partitions> X <number of y partitions> sub-images."
       << endl
       << endl
       << "There are two types of output."
       << endl
       << "If <output type> is 'p', then the output files would be ppgm/pgm raw files."
       << endl
       << "If <output type> is 'b', then only the pixels of each sub-image would"
       << endl
       << "be stored in the output files, listed in row-major order."
       << endl
       << "In both cases, the filename for the n-th partition along the x-dimension "
       << endl
       << "and the m-th partition along the y-dimension would be"
       << endl
       << "	<output file prefix>.m-n"
       << endl
       << "In the case where <output type> is 'b', an additional file,"
       << endl
       << "<output file prefix>.summary is generated, and that file stores for"
       << endl
       << "each sub-image its top-left position in the original image, its number"
       << endl
       << "of pixels along x and y dimension."
       << endl;
}



//////////////////////////////////////////////////////////////////////////
//
// read_ppm_file
//
// Read a ppm/pgm file. Return 0 if succeeds, 1 otherwise
//
//////////////////////////////////////////////////////////////////////////
int
read_ppm_file(ifstream& file, Image& image)
{
  image.clear();

  const int max_line_length = 71;
  char line[max_line_length];
  const char comment = '#';

  // get magic number for file type
  if (get_next_uncomment_string(file, line, max_line_length, comment) != 0) {
     // read fails
     cerr << "ERROR: cannot read magic number for file type." << endl;
     image.clear();
     return 1;
  }
  if (line[0] != 'P' || (line[1] != '3' && line[1] != '6')) {
     cerr << "ERROR: invalid magic number '" << line << "'" << endl;
     image.clear();
     return 1;
  }
  char ppm_type = line[1];

  // get width
  if (get_next_uncomment_string(file, line, max_line_length, comment) != 0) {
     // read fails
     cerr << "ERROR: cannot read width for image." << endl;
     image.clear();
     return 1;
  }
  if ((image.getWidth() = atoi(line)) <= 0) {
     cerr << "ERROR: invalid width = " << image.getWidth() << endl;
     image.clear();
     return 1;
  }

  // get height
  if (get_next_uncomment_string(file, line, max_line_length, comment) != 0) {
     // read fails
     cerr << "ERROR: cannot read height for image." << endl;
     image.clear();
     return 1;
  }
  if ((image.getHeight() = atoi(line)) <= 0) {
     cerr << "ERROR: invalid width = " << image.getHeight() << endl;
     image.clear();
     return 1;
  }

  // get maxval
  if (get_next_uncomment_string(file, line, max_line_length, comment) != 0) {
     // read fails
     cerr << "ERROR: cannot read max value for image." << endl;
     image.clear();
     return 1;
  }
  if ((image.getMaxValue() = atoi(line)) <= 0) {
     cerr << "ERROR: invalid width = " << image.getMaxValue() << endl;
     image.clear();
     return 1;
  }

  // skip the white space immediately after maxval
  char ch;
  file.get(ch);

  switch (ppm_type) {
    case '3':	// ascii format, not implelmented yet ...
		image.getPixelSize() = 3;	// 3 bytes per pixel
		cerr << "ERROR: sorry, P3 (ascii) format is not implemented yet"		     << endl;
		image.clear();
		return 1;

    case '6':	// raw format
		image.getPixelSize() = 3;	// 3 bytes per pixel
		{
		   const int image_size_bytes =
			image.getNumberPixels() * image.getPixelSize();
		   image.getBytes() = new char[image_size_bytes];
		   if (image.getBytes() == NULL) {		// new fails
			cerr << "ERROR: fail to new " << image_size_bytes
			     << " bytes for image." << endl;
			image.clear();
			return 1;
		   }
		   if (read_raw_bytes(file, image_size_bytes,
				      image.getBytes()) != 0)
		   {	// error
			image.clear();
			return 1;
		   }
		}
		break;
  }
  return 0;
}


//////////////////////////////////////////////////////////////////////////
//
// some I/O help functions
//
// return 0 if succeeds, 1 otherwise
//
//////////////////////////////////////////////////////////////////////////
int
get_next_uncomment_string(ifstream& file, char* buf, int buf_len, char comment)
{
  while (!file.eof()) {
    file >> buf;
    if (file.eof())			// eof
       return 1;

    // search for comments
    int string_len = strlen(buf);
    for (int i=0; i<string_len; i++) {
	if (buf[i] == comment) {
	   buf[i] = '\0';

	   // skip the rest of the line
	   char ch;
	   while (file.get(ch) && ch != '\n');

	   break;
	}
    }
    if (buf[0] != '\0')			// found a string
	return 0;
  }
  return 1;				// eof
}


int
read_raw_bytes(ifstream& file, int nbytes, char* bytes)
{
  char *p = bytes,
       *endp = bytes + nbytes;
  while (p != endp && file.get(*p))
	p++;

  if (p == endp)			// all bytes are read
     return 0;
  return 1;
}


int
split_image(Image& image, int x_part, int y_part, char output_type,
	    ofstream* outfs, ofstream& summary)
{
  assert (output_type == 'p' || output_type == 'b');

  // compute how many pixels per sub-image in each dimension;
  // when cannot divide, the excess ones always to the last sub-image along
  // each dimension
  // (eg partitioning 5 pixels into 3 sub-images, first two sub-images get
  // 1 pixel each and last one gets 3 pixels)
  // partitioning fails when image is too small to be partitioned

  int sub_width = image.getWidth() / x_part,	// width of sub-image
      sub_height = image.getHeight() / y_part;	// height of sub-image

  if (sub_width == 0 || sub_height == 0) {
     cerr << "ERROR: image is too small to be partitioned"
	  << endl;
     return 1;
  }
  int excess_width = image.getWidth() - sub_width * x_part,
      excess_height = image.getHeight() - sub_height * y_part;

  // just so that I don't have to keep all the sub-images in the memory
  // at any time, I am simply going to create each sub-image at a time
  // and write them to disk as soon as they are created
  int x_pos = 0, y_pos = 0;	// top-left position of current sub-image
  for (int i=0, y=0; y<y_part; y++) {
      int h = sub_height;
      if (y == y_part - 1)
	  h += excess_height;

      x_pos = 0;
      for (int x=0; x<x_part; x++) {
	  int w = sub_width;
	  if (x == x_part - 1)
	      w += excess_width;

	  Image sub_image;
	  if (image.copySubImage(sub_image, x_pos, y_pos, w, h) != 0) {
	     // copy fails
	     return 1;
	  }

	  // write header if necessary
	  switch (output_type) {
	    case 'p':	// output ppm file 
			{
			   char pos_info[512];
			   sprintf(pos_info, "# position x=%d y=%d",
				   x_pos, y_pos);
			   write_raw_ppm_header(sub_image, outfs[i], pos_info);

			}
			break;

	    case 'b':	// output binary file, no header is needed
			break;
	  }


	  // output bytes
	  char *pixel = sub_image.getBytes();
	  for (int k=0; k<sub_image.getNumberPixels(); k++) {
	     for (int n=0; n<sub_image.getPixelSize(); n++) {
		outfs[i].put(*pixel);
		pixel++;
	     }
	  }

	  summary << x_pos << " " << y_pos << " " << w << " " << h
		  << " " << sub_image.getPixelSize()
		  << "		# sub-image." << x << "-" << y
		  << endl;

	  x_pos += w;			// advance x position
	  i++;				// get to next outfs
      }

      y_pos += h;			// advance y position
  }

  return 0;
}


void
write_raw_ppm_header(Image& image, ofstream& os, const char *description)
{
  assert (image.getPixelSize() == 3);
  os << "P6" << endl;
//  os << "# generated by split-image" << endl;
  if (description != NULL)
     os << description << endl;
  os << image.getWidth() << " " << image.getHeight()
     << endl << image.getMaxValue() << endl;
}


//////////////////////////////////////////////////////////////////////////
//
// some I/O help functions
//
// return 0 if succeeds, 1 otherwise
//
//////////////////////////////////////////////////////////////////////////
int
Image::copySubImage(Image& sub_image, int x_pos, int y_pos, int w, int h)
{
  sub_image.clear();

  int x_end = x_pos + w,
      y_end = y_pos + h;

  // check ranges
  if (x_pos < 0 || x_pos >= width || y_pos < 0 || y_pos >= height ||
      x_end < 0 || x_end > width  || y_end < 0 || y_end > height)
  {
     cerr << "ERROR: invalid range for Image::copySubImage" << endl;
     cerr << "width=" << width << " height=" << height
	  << " x_pos=" << x_pos << " y_pos=" << y_pos
	  << " w=" << w << " h=" << h
	  << endl;
     return 1;
  }

  sub_image.getWidth() = w;
  sub_image.getHeight() = h;
  sub_image.getPixelSize() = getPixelSize();
  sub_image.getMaxValue() = getMaxValue();

  char*& b = sub_image.getBytes();
  b = new char[sub_image.getNumberPixels() * sub_image.getPixelSize()];
  if (b == NULL) {
     cerr << "ERROR: cannot allocate bytes for sub-image in Image::copySubImage"
	  << endl;
     sub_image.clear();
     return 1;
  }

  for (int y=y_pos; y<y_end; y++)
     for (int x=x_pos; x<x_end; x++)
	for (int z=0; z<getPixelSize(); z++) {
	   int r = 
	      sub_image.setPixelByte(x-x_pos, y-y_pos, z, getPixelByte(x,y,z));
	   assert (r == 0);
	}

  return 0;
}


char
Image::getPixelByte(int x, int y, int z) const
{
  assert (x >= 0 && x < width);
  assert (y >= 0 && y < height);
  assert (z >= 0 && z < pixelsize);

  int i = z + pixelsize * (x + width * y);
  return bytes[i];
}


int
Image::setPixelByte(int x, int y, int z, char v)
{
  if (x >= 0 && x < width && y >= 0 && y < height &&
      z >= 0 && z < pixelsize)
  {
     int i = z + pixelsize * (x + width * y);
     bytes[i] = v;
     return 0;
  }
  return 1;
}
