
/* $Id: t2_fe.h,v 1.6 2000/03/16 00:50:02 chialin Exp $ */

/* t2_fe.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/25/98
 * T2 front-end
 */


#ifndef _T2_FE_H_
#define _T2_FE_H_


#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_socket.h"
#include	"t2_bfile.h"
#include	"t2_qspec.h"
#include	"t2_packets.h"
#include	"t2_feqpool.h"


//////////////////////////////////////////////////////////////////////////////
//
// T2_FE
//
// The constructor takes an unsigned integer, which is used to specify
// the maximum number of pending queries that T2_FE can hold at any
// time.
//
//////////////////////////////////////////////////////////////////////////////
class T2_FE {
public:
  T2_FE(u_int l, u_int qwin=3);
  ~T2_FE();

  // starting up or shutting down the front-end
  bool startup(u_int be_port);
  void shutdown();

  //
  // connecting with the back-end nodes:
  //
  // connectAllBackEnds		waits until all the back-end nodes have
  //				connected to the front-end, and returns the
  // 				number of back-end nodes connected if
  //				succeeds, 0 if time out (even if some, but not
  //				all back-end nodes may have connected, use
  //				getNumberConnectedBackEnds() to check the
  //				number of back-end nodes connected),
  //				-1 otherwise
  // pollBackEndConnection	polls for a back-end connection in a
  //				non-blocking fashion; returns 1 if a connection
  //				was made, 0 if time out, -1 if error
  //
  // getTotalNumberBackEnds	returns the total number of back-end nodes
  //				expected to connect
  //
  // getNumberConnectedBackEnds returns the number of back-end nodes already
  //				connected
  //
  int connectAllBackEnds(struct timeval* timeout=NULL);
  int connectAllBackEnds(int nsec, int nusec=0);
  int pollBackEndConnection(struct timeval* timeout=NULL);
  int pollBackEndConnection(int nsec, int nusec=0);
  int getTotalNumberBackEnds() const { return beconnects.getNumberElements(); };
  int getNumberConnectedBackEnds() const { return nconnected; };
  u_char getFrontEndEndian() const { return my_endian; };
  u_char getBackEndsEndian() const { return be_endian; };

  // submit a query specification to the T2 front-end; returns true if
  // submission succeeds, false otherwise
  bool submitQuerySpec(T2_QSpec* q, T2_QSpecStatus* s);

  // check whether a submitted qspec has been forwarded to the back-end;
  // returns 1 if it has, 0 if it has not, -1 if error
  int testQuerySpec(T2_QSpecStatus& s) { do_work(); return s.test(); };

  // cancel a submitted query if possible;
  // returns true if cancel succeeds, false otherwise
  bool cancelQuerySpec(T2_QSpecStatus& s) { return pool.cancelQuerySpec(&s); };

  // send a qbatch to back-end; users are not supposed to call these methods
  bool sendQBatch(T2_QBatch& q);
  bool sendQBatch(T2_QPtrBatch& q);


  //////////
  // ordinary usrs should not use any methods below
  //////////

  // send a qbatch without involving qtokens at all
  bool sendQBatch_NoWaitForBEAck(T2_QPtrBatch& q);

  // accessing qtokens (ordinary users should not use any of these methods)
  u_int getMax_QTokens() const { return max_qtokens; };
  u_int getQTokens() const { return qtokens; };
  u_int setQTokens(u_int i) { return (qtokens = i); };
  u_int incQTokens() { return ++qtokens; };

  // accessing back-end connection socket (ordinary users should not use any
  // of these methods)
  T2_Socket& getBackEndConnectionSocket(u_int i) 
			{ return beconnects[i].getSocket(); };

  // recv an ack from be for a submitted query
  bool recvQueryAck();

protected:
  T2_SocketServer sockserver;
  T2_Array<T2_BEConnection> beconnects;
  u_int nconnected;			// # of back-end nodes already connected
  u_char be_endian,			// endian-ness of all back-end nodes
	 my_endian;			// my endian-ness (ie T2 front-end)

  T2_FEQPool pool;			// query pool

  const u_int max_qtokens;		// how many qbatches to be sent to
  u_int qtokens;			// the back-end before starting to
					// read acks from the back-end

  void shutdownBEConnections();		// shut down all back-end connections

  int connectBackEnd(T2_BEConnection& s, u_char& e);
					// connected by a back-end

  void do_work();			// send query batches to back-end

private:
};


inline int
T2_FE::connectAllBackEnds(int nsec, int nusec)
{
  struct timeval tval;
  tval.tv_sec = nsec;
  tval.tv_usec = nusec;
  return connectAllBackEnds(&tval);
}

inline int
T2_FE::pollBackEndConnection(int nsec, int nusec)
{
  struct timeval tval;
  tval.tv_sec = nsec;
  tval.tv_usec = nusec;
  return pollBackEndConnection(&tval);
}

inline bool
T2_FE::submitQuerySpec(T2_QSpec* q, T2_QSpecStatus* s)
{
  bool r = pool.submitQuerySpec(q, s);
  do_work();
  return r;
}

inline void
T2_FE::shutdownBEConnections()
{
  if (nconnected > 0) {
     const u_int nbe = beconnects.getNumberElements();
     for (u_int i=0; i<nbe; i++)
	beconnects[i].shutdown();
     nconnected = 0;
     be_endian = T2_UnknownEndian;
  }
}



// more stuff for T2_FrontEnd
#include	"t2_frontend.h"

#endif	/* ! _T2_FE_H_ */
