
/* $Id: t2_feinquiry.h,v 1.3 2000/02/24 17:06:47 chialin Exp $ */

/* t2_feinquiry.h */

#ifndef _T2_FEINQUIRY_H_
#define _T2_FEINQUIRY_H_

#include	<iostream.h>
#include	<string.h>
#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_socket.h"
#include	"t2_defs.h"
#include	"t2_bstream.h"
#include	"t2_conscat.h"
#include	"t2_dsinfo.h"


// forward declaration
class T2_FEInquiryResultsReader;

//////////////////////////////////////////////////////////////////////////
//
// T2_FEInquiry
//
// A class that represents a single inquiry from the application to 
// inquire the meta data stored with the T2 front-end.
//
//////////////////////////////////////////////////////////////////////////
class T2_FEInquiry {
private:
  // deleting name if necessary
  void delete_name() { if (del_name == true) delete [] name; name = NULL;
		       del_name = true; };

public:
  // inquiry type for datasets or constructor functions
  enum inquiry_type {
	inquiry_error = 0,		// error
	inquiry_dataset,		// inquiry for dataset info
	inquiry_function,		// inquiry for constructor functions
	inquiry_accmeta_function,	// inquiry for acc meta constructor func
	inquiry_aggregation_function,	// inquiry for aggr constructor func
	inquiry_projection_function,	// inquiry for proj constructor func
	inquiry_metachaos_function	// inquiry for metachaos constructorfunc
  };

  // fields of the schema for dataset information
  enum dataset_schema_field {
	dataset_datasetid_field			= T2_USHORT_BIT_0,
	dataset_datasetname_field		= T2_USHORT_BIT_1,
	dataset_datasetdescription_field	= T2_USHORT_BIT_2,
	dataset_blobsize_field			= T2_USHORT_BIT_3,
	dataset_blobobj_field			= T2_USHORT_BIT_4,
	dataset_iteratorname_field		= T2_USHORT_BIT_5,
	dataset_indexid_field			= T2_USHORT_BIT_6,
	dataset_indexname_field			= T2_USHORT_BIT_7,
	dataset_indexdescription_field		= T2_USHORT_BIT_8
  };

  // fields of the schema for constructor function information
  enum function_schema_field {
	function_id_field		= T2_USHORT_BIT_0,
	function_name_field		= T2_USHORT_BIT_1,
	function_description_field	= T2_USHORT_BIT_2
  };

  // possible status of inquiry results
  enum inquiry_result_tag {
	results_ok = 0,			// results are fine
	bad_socket_error,		// some thing wrong with the socket
	invalid_inquiry,		// some thing wrong with the inquiery
	t2_frontend_not_ready		// some thing wrong with the front-end
  };

  T2_FEInquiry();
  ~T2_FEInquiry() { delete_name(); };

  // const access methods
  inquiry_type getInquiryType() const { return qtype; };
  T2_UDFType getFunctionType() const { return ftype; };
  const char* getName() const { return name; };
  bool getExactMatchFlag() const { return exact_match; };
  u_short getFieldFlag() const { return field_flag; };
  size_t getTotalBlobDataSizeLimit() const { return blobsize_limit; };
  
  // non-const access methods
  inquiry_type& getInquiryType() { return qtype; };
  T2_UDFType& getFunctionType() { return ftype; };
  bool& getExactMatchFlag() { return exact_match; };
  u_short& getFieldFlag() { return field_flag; };
  size_t& getTotalBlobDataSizeLimit() { return blobsize_limit; };

  // two methods that handles name:
  // setNameNoCopy simply sets the pointer without copying it, in which case
  // no delete should be done to name at the end;
  void setNameNoCopy(const char* s)
			{ delete_name(); name = (char*) s; del_name = false; };
  // copyName makes a copy of the given name, in which case
  // delete should be done to name at the end;
  void copyName(const char* s)
		{ delete_name(); del_name = true; size_t len = strlen(s)+1;
		  t2_Alloc(name, len, char); memcpy(name, s, len);
		};
  
  // write into an output binary stream
  bool bwrite(T2_OBStream& obs) const;

  // read from an input binary stream
  bool bread(T2_IBStream& ibs);

  // write to ostream for debugging purpose
  ostream& write(ostream& os) const;

  // static functions to check the bits in the field flag for dataset inquiries
  static bool checkDatasetIDField(const u_short f);
  static bool checkDatasetNameField(const u_short f);
  static bool checkDatasetDescriptionField(const u_short f);
  static bool checkDatasetBlobDataSizeField(const u_short f);
  static bool checkDatasetBlobDataField(const u_short f);
  static bool checkDatasetIteratorNameField(const u_short f);
  static bool checkDatasetIndexIDField(const u_short f);
  static bool checkDatasetIndexNameField(const u_short f);
  static bool checkDatasetIndexDescriptionField(const u_short f);

  // static functions to check the bits in the field flag for function inquiries
  static bool checkFunctionIDField(const u_short f);
  static bool checkFunctionNameField(const u_short f);
  static bool checkFunctionDescriptionField(const u_short f);

protected:
  inquiry_type qtype;			// inquiry type
  T2_UDFType ftype;			// function type (only valid when
					// qtype == inquiry_function)

  char* name;				// target string for name
  bool exact_match;			// whether name should be 
					// exact match (true) or
					// name is a regular expression (false)

  u_short field_flag;			// mark which fields should be returned

  bool del_name;			// a flag to tell whether I need to
					// delete the name or not

  size_t blobsize_limit;		// a limit on how many bytes in total
					// the caller wants the T2 front-end
					// to send for blob objects for all
					// results satisfying this inquiry
					// (zero means no limit)
					// -- only makes sense for dataset
					// inquiries

private:
};


inline
T2_FEInquiry::T2_FEInquiry(): qtype(inquiry_error), ftype(T2_UDF_Unknown),
			      name(NULL), exact_match(false), field_flag(0),
			      del_name(false), blobsize_limit(0)
{
}

inline T2_OBStream&
operator << (T2_OBStream& obs, const T2_FEInquiry& q)
{
  q.bwrite(obs);
  return obs;
}

inline T2_IBStream&
operator >> (T2_IBStream& ibs, T2_FEInquiry& q)
{
  q.bread(ibs);
  return ibs;
}

inline bool
T2_FEInquiry::checkFunctionIDField(const u_short f)
{
  if (f ^ (u_short) function_id_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkFunctionNameField(const u_short f)
{
  if (f ^ (u_short) function_name_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkFunctionDescriptionField(const u_short f)
{
  if (f ^ (u_short) function_description_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetIDField(const u_short f)
{
  if (f ^ (u_short) dataset_datasetid_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetNameField(const u_short f)
{
  if (f ^ (u_short) dataset_datasetname_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetDescriptionField(const u_short f)
{
  if (f ^ (u_short) dataset_datasetdescription_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetBlobDataSizeField(const u_short f)
{
  if (f ^ (u_short) dataset_blobsize_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetBlobDataField(const u_short f)
{
  if (f ^ (u_short) dataset_blobobj_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetIteratorNameField(const u_short f)
{
  if (f ^ (u_short) dataset_iteratorname_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetIndexIDField(const u_short f)
{
  if (f ^ (u_short) dataset_indexid_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetIndexNameField(const u_short f)
{
  if (f ^ (u_short) dataset_indexname_field)
     return true;
  return false;
}

inline bool
T2_FEInquiry::checkDatasetIndexDescriptionField(const u_short f)
{
  if (f ^ (u_short) dataset_indexdescription_field)
     return true;
  return false;
}

inline ostream&
operator << (ostream& os, const T2_FEInquiry& iqr)
{
  return iqr.write(os);
}



//////////////////////////////////////////////////////////////////////////
//
// T2_FEDatasetEntry
//
// A class that stores an entry of the results for a single inquiry
// about the information that the T2 front-end stores for datasets.
//
//////////////////////////////////////////////////////////////////////////
class T2_FEDatasetEntry: protected T2_DSInfoEntry {
friend class T2_FEInquiryResultsReader;
public:
  T2_FEDatasetEntry(): T2_DSInfoEntry() {};
  ~T2_FEDatasetEntry() {};

  // query methods
  u_int getNumberIterators() const
			{ return T2_DSInfoEntry::getNumberIterators(); };
  u_int getNumberIndexes() const
			{ return T2_DSInfoEntry::getNumberIndexes(); };

  // const access methods
  T2_DSID getDatasetID() const { return T2_DSInfoEntry::getDatasetID(); };
  const char* getDatasetName() const
			{ return T2_DSInfoEntry::getDatasetName(); };
  const char* getDatasetDescription() const
			{ return T2_DSInfoEntry::getDatasetDescription(); };
  size_t getBlobDataSize() const { return T2_DSInfoEntry::getBlobFileSize(); };
  const char* getBlobObject() const
			{ return T2_DSInfoEntry::getBlobFileName(); };
  const char* getIteratorName(u_int i) const
			{ return T2_DSInfoEntry::getIteratorName(i); };
  const T2_IndexInfoEntry& getIndex(u_int i) const
			{ return T2_DSInfoEntry::getIndex(i); };
  
  // non-const access methods
  T2_DSID& getDatasetID() { return T2_DSInfoEntry::getDatasetID(); };
  char*& getDatasetName() { return T2_DSInfoEntry::getDatasetName(); };
  char*& getDatasetDescription() 
			{ return T2_DSInfoEntry::getDatasetDescription(); };
  size_t& getBlobFileSize() { return T2_DSInfoEntry::getBlobFileSize(); };
  char*& getBlobObject() { return T2_DSInfoEntry::getBlobFileName(); };
  char*& getIteratorName(u_int i) 
			{ return T2_DSInfoEntry::getIteratorName(i); };
  T2_IndexInfoEntry& getIndex(u_int i) { return T2_DSInfoEntry::getIndex(i); };
  
  // copy constructor and operator =
  T2_FEDatasetEntry(const T2_FEDatasetEntry& e): T2_DSInfoEntry(e) {};
  T2_FEDatasetEntry& operator =(const T2_FEDatasetEntry& e)
			{ T2_FEDatasetEntry::operator =(e); return *this; };

  // output to ostream for debugging purpose
  ostream& write(ostream& os) const { return T2_DSInfoEntry::write(os); };

protected:
  // note that I am using the data member blob_fname as the member that
  // stores the actual content of the blob filename

private:
};


inline ostream&
operator << (ostream& os, const T2_FEDatasetEntry& e)
{
  return e.write(os);
}



//////////////////////////////////////////////////////////////////////////
//
// T2_FEDatasetInquiryResults
//
// A class that stores the results for a single inquiry about the info
// that the T2 front-end stores for datasets.
//
//////////////////////////////////////////////////////////////////////////
class T2_FEDatasetInquiryResults {
friend class T2_FEInquiryResultsReader;
public:
  T2_FEDatasetInquiryResults(): status(T2_FEInquiry::results_ok), entries(10){};
  ~T2_FEDatasetInquiryResults() {};

  // const access method
  T2_FEInquiry::inquiry_result_tag getStatus() const { return status; };
  u_int getNumberEntries() const { return entries.getNumberElements(); };
  const T2_FEDatasetEntry& operator [] (u_int i) const { return entries[i]; };
  T2_FEDatasetEntry& operator [] (u_int i) { return entries[i]; };

  // non-const access methods
  T2_FEInquiry::inquiry_result_tag& getStatus() { return status; };

  // remove all entries
  void clear() { entries.clear(); status = T2_FEInquiry::results_ok; };

protected:
  T2_FEInquiry::inquiry_result_tag status;
  T2_VArray<T2_FEDatasetEntry> entries;

  // append a new blank result; return the new blank entry
  T2_FEDatasetEntry& appendElement() { return entries.appendElement(); };

private:
};



//////////////////////////////////////////////////////////////////////////
//
// T2_FEFunctionEntry
//
// A class that stores an entry of the results for a single inquiry
// about the information that the T2 front-end stores for constructor
// functions.
//
//////////////////////////////////////////////////////////////////////////
class T2_FEFunctionEntry: protected T2_ConsCatalogEntry {
friend class T2_FEInquiryResultsReader;
public:
  T2_FEFunctionEntry(): T2_ConsCatalogEntry() {};
  ~T2_FEFunctionEntry() {};

  // access methods
  T2_UDFType getFunctionType() const { return T2_ConsCatalogEntry::type; };
  u_int getFunctionID() const { return T2_ConsCatalogEntry::id; };
  const char* getFunctionName() const { return T2_ConsCatalogEntry::name; };
  const char* getFunctionDescription() const
			{ return T2_ConsCatalogEntry::description; };

  // output to ostream for debugging purpose
  ostream& write(ostream& os) const { return T2_ConsCatalogEntry::write(os); };

protected:
private:
};


inline ostream&
operator << (ostream& os, const T2_FEFunctionEntry& e)
{
  return e.write(os);
}



//////////////////////////////////////////////////////////////////////////
//
// T2_FEFunctionInquiryResults
//
// A class that stores a set of results for a single inquiry about the info
// that the T2 front-end stores for constructor functions.
//
//////////////////////////////////////////////////////////////////////////
class T2_FEFunctionInquiryResults {
friend class T2_FEInquiryResultsReader;
public:
  T2_FEFunctionInquiryResults(): status(T2_FEInquiry::results_ok),entries(10){};
  ~T2_FEFunctionInquiryResults() {};

  // const access method
  T2_FEInquiry::inquiry_result_tag getStatus() const { return status; };
  u_int getNumberEntries() const { return entries.getNumberElements(); };
  const T2_FEFunctionEntry& operator [] (u_int i) const { return entries[i]; };
  T2_FEFunctionEntry& operator [] (u_int i) { return entries[i]; };

  // non-const access methods
  T2_FEInquiry::inquiry_result_tag& getStatus() { return status; };

  // remove all entries in the object
  void clear() { entries.clear(); status = T2_FEInquiry::results_ok; };

  // copy constructor and operator =
//  T2_FEFunctionInquiryResult(const T2_FEFunctionInquiryResults& r);
//  T2_FEFunctionInquiryResult& operator =(const T2_FEFunctionInquiryResults& r);

protected:
  T2_FEInquiry::inquiry_result_tag status;
  T2_VArray<T2_FEFunctionEntry> entries;

  // append a new blank result; return the new blank entry
  T2_FEFunctionEntry& appendElement() { return entries.appendElement(); };

private:
};



//////////////////////////////////////////////////////////////////////////
//
// T2_FEInquiryResultsWriter
//
// A class that the T2 front-end uses to sends back to the application
// the results for a single inquiry.
// This class works with class T2_FEInquiryResultsReader to exchange
// the results correctly.
//
//////////////////////////////////////////////////////////////////////////
class T2_FEInquiryResultsWriter {
public:
  // tage for communication between T2_FEInquiryResultsWriter and
  // T2_FEInquiryResultsReader
  enum result_tag {
	no_more_results = 0,
	one_more_result,		// yet another result
	invalid_inquiry,
	t2_frontend_not_ready
  };

  T2_FEInquiryResultsWriter(T2_Socket& s, u_short ff, size_t blimit);
  ~T2_FEInquiryResultsWriter() {};

  // initiate the transfer
  bool initiate();

  // send a result
  bool writeResult(T2_ConsCatalogEntry& e);
  bool writeResult(T2_DSInfoEntry& e);

  // terminate the transfer with the given tag
  bool terminate(result_tag t);

protected:
  T2_Socket& sock;
  const u_short field_flag;
  size_t blobsize_limit,
	 blobsize_cur;			// how many bytes of blob data has been
					// transferred so far

private:
};


inline
T2_FEInquiryResultsWriter::
T2_FEInquiryResultsWriter(T2_Socket& s, u_short ff, size_t blimit)
			: sock(s), field_flag(ff), blobsize_limit(blimit),
			  blobsize_cur(0)
{
}



//////////////////////////////////////////////////////////////////////////
//
// T2_FEInquiryResultsReader
//
// A class that the application uses to recv the inquiry results sent
// from the T2 front-end.
// This class works with class T2_FEInquiryResultsWriter to exchange
// the results correctly.
//
//////////////////////////////////////////////////////////////////////////
class T2_FEInquiryResultsReader {
public:
  T2_FEInquiryResultsReader(T2_Socket& s): sock(s) {};

  // initiate exchange
  bool initiate();

  // recv a bunch of results
  bool readResults(T2_FEFunctionInquiryResults& r);
  bool readResults(T2_FEDatasetInquiryResults& r);

protected:
  T2_Socket& sock;
  u_short field_flag;

private:
};



#endif	/* ! _T2_FEINQUIRY_H_ */
