
/* $Id: t2_frontend.h,v 1.5 2000/03/16 00:50:06 chialin Exp $ */

/* t2_frontend.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 10/12/98
 * A class to be used by the application front-end. 
 */

#ifndef _T2_FRONTEND_H_
#define _T2_FRONTEND_H_


#include	"t2_header.h"
#include	"t2_socket.h"
#include	"t2_qspec.h"
#include	"t2_feinquiry.h"


// forward declaration
class T2_AppFrontEnd;
class T2_QBatchList;

//////////////////////////////////////////////////////////////////////////
//
// T2_AppToFEProtocol
//
// An object that specifies the protocol between the application
// front-end and the T2 front-end
//
//////////////////////////////////////////////////////////////////////////
enum T2_AppToFECommand {		// command sent by app fe to T2 fe
	t2_acmdSubmit = 0,		// submit a query spec
	t2_acmdDisconnect,		// disconnect from T2 front-end
	t2_acmdPing,			// ping the T2 front-end
	t2_acmdUnexpectedDisconnect,	// connection is unexpectedly closed
	t2_acmdInquiry,			// inquiry about T2 meta data
	t2_acmdError			// error, must be last in enum
};

// dispite the name, T2_AppToFEAck are really ack's from FE back to App
enum T2_AppToFEAck {			// ack sent by T2 fe to app fe
	t2_ackConnect = 0,		// ack connection accepted
	t2_ackBusy,			// ack connection rejected due to busy
	t2_ackPing,			// ack for a ping request
	t2_ackUnexpectedDisconnect,	// connection is unexpectedly closed
	t2_ackInquiry,			// results for inquiry from app
	t2_ackError			// error, must be last in enum
};

class T2_AppToFEProtocol {
public:
  T2_AppToFEProtocol() {};
  ~T2_AppToFEProtocol() {};

  // read/write a command from/to socket
  T2_AppToFECommand readAppCommand(T2_Socket& sock);
  bool writeAppCommand(T2_Socket& sock, T2_AppToFECommand cmd);

  // read/write an ack from/to socket
  T2_AppToFEAck readFEAck(T2_Socket& sock);
  bool writeFEAck(T2_Socket& sock, T2_AppToFEAck ack);

  //
  // the following methods are higher level protocols
  //

  // application front-end connects to T2 front-end
  T2_AppToFEAck app_connect(T2_Socket& s, const u_char my_endian,
			    u_char& t2fe_endian, u_char& t2be_endian,
			    u_int& numnodes);

  // T2 front-end acks for connection; 
  // the ack could be busy (by calling fe_busyAck), in which case the T2 
  // front-end is busy and cannot accept more connections, 
  // or ok (by calling fe_connectAck), in which case the connection is accepted
  void fe_busyAck(T2_Socket& s);
  bool fe_connectAck(T2_Socket& s, const u_char t2fe_endian,
		     const u_char t2be_endian, const u_int numnodes,
		     u_char& appfe_endian);

  // application fron-end asks to disconnect from T2 front-end
  void app_disconnect(T2_Socket& s);
  // T2 front-end ack for disconnection request from application
  void fe_disconnectAck(T2_Socket& s);

  // application pings T2 front-end and T2 front-end acks
  bool app_ping(T2_Socket& s);
  void fe_pingAck(T2_Socket& s);

  // application submit a query batch
  bool app_submitQBatch(T2_Socket& s, T2_QBatch& qbatch);

  // application sends an inquiry to the T2 front-end
  bool app_sendInquiry(T2_Socket& s, T2_FEInquiry& iqr);

protected:

private:
};

inline T2_AppToFECommand
T2_AppToFEProtocol::readAppCommand(T2_Socket& sock)
{
  if (sock.connected() == false)
     return t2_acmdError;
  u_char i = '\0';
  sock >> i;
  if (sock.getNumberOfBytesLastRead() == 0)
     return t2_acmdUnexpectedDisconnect;
  if ((u_int) i >= (u_int) t2_acmdError)
     return t2_acmdError;
  return (T2_AppToFECommand) i;
}

inline bool
T2_AppToFEProtocol::writeAppCommand(T2_Socket& sock, T2_AppToFECommand cmd)
{
  if (sock.connected() == false)
     return false;
  sock << (u_char) cmd;
  return true;
}

inline T2_AppToFEAck
T2_AppToFEProtocol::readFEAck(T2_Socket& sock)
{
  if (sock.connected() == false)
     return t2_ackError;
  u_char i = 0;
  sock >> i;
  if (sock.getNumberOfBytesLastRead() == 0)
     return t2_ackUnexpectedDisconnect;
  if ((u_int) i >= (u_int) t2_ackError)
     return t2_ackError;
  return (T2_AppToFEAck) i;
}

inline bool
T2_AppToFEProtocol::writeFEAck(T2_Socket& sock, T2_AppToFEAck ack)
{
  if (sock.connected() == false)
     return false;
  sock << (u_char) ack;
  return true;
}



//////////////////////////////////////////////////////////////////////////
//
// T2_TDBList
//
// A template class for a doubly-linked list. The data class
// stored with each node must have the following methods.
//	a copy constructor
//	a destructor
//
//////////////////////////////////////////////////////////////////////////
template<class Data>
class T2_TDBList {
public:
  T2_TDBList() { head = tail = cur = NULL; numnodes = 0; };
  ~T2_TDBList() { cur = head;
		while (cur != NULL)
		{ T2_TList_Node* p = cur; cur = cur->next; delete p; }
  };

  // query methods
  bool isEmpty() const { return (numnodes == 0)? true:false; };
  u_int getListSize() const { return numnodes; };

  // moving the current pointer
  bool curAtHead() const { return (cur == head)? true:false; };
  bool curAtTail() const { return (cur == tail)? true:false; };
  bool moveHead() { if (isEmpty() == true) return false;
		    cur = head; return true;
  };
  bool moveNext() { if (cur == NULL || cur->next == NULL) return false;
		    cur = cur->next; return true;
  };
  bool moveNextWrap() { if (cur == NULL) return false;
			cur = cur->next;
		        if (cur == NULL) moveHead(); return true;
  };
  bool movePrev() { if (cur == NULL || cur->prev == NULL) return false;
		    cur = cur->prev; return true;
  };

  // accessing the current record
  const Data& getCurData() const { t2_Assert (cur != NULL); return cur->data; };
  Data& getCurData() { t2_Assert (cur != NULL); return cur->data; };
  Data& getHeadData() { t2_Assert (head != NULL); return head->data; };
  Data& getTailData() { t2_Assert (tail != NULL); return tail->data; };

  // list append
  u_int appendList(const Data& d) {
		T2_TList_Node *node = new T2_TList_Node(d);
		if (isEmpty() == true) { head = tail = cur = node; }
		else { t2_Assert(tail!=NULL);
		       tail->next = node; node->prev = tail; tail = node; }
		return ++numnodes;
  };

  // delete current record
  u_int deleteCurWrap() {
		if (isEmpty() == true) return 0;
		t2_Assert (cur != NULL);
		T2_TList_Node *p = cur; cur = cur->next;
		if (p == head) { head = p->next;
				 if (head) head->prev = NULL; else tail=NULL; }
		else if (p == tail)
		     { t2_Assert(cur == NULL);
		       tail = tail->prev; tail->next = NULL; cur = head; }
		else { t2_Assert (cur != NULL);
		       cur->prev = p->prev; p->prev->next = cur; }
		delete p;
		return --numnodes;
  };

protected:
  struct T2_TList_Node {
	 T2_TList_Node *next, *prev;
	 Data data;

	 T2_TList_Node(const Data& d): data(d) { next = prev = NULL; };
  };

  T2_TList_Node *head, *tail, *cur;
  u_int numnodes;

private:
  // do not use copy constructor and operator =
  T2_TDBList(const T2_TDBList<Data>& l);
  T2_TDBList<Data>& operator =(const T2_TDBList<Data>& l);
};


template<class Data>
inline
T2_TDBList<Data>::T2_TDBList(const T2_TDBList<Data>& l)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_TDBList() copy constructor");
}

template<class Data>
inline T2_TDBList<Data>&
T2_TDBList<Data>::operator =(const T2_TDBList<Data>& l)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_TDBList() operator =");
  return *this;
}



//////////////////////////////////////////////////////////////////////////
//
// T2_QBatchList
//
//////////////////////////////////////////////////////////////////////////
class T2_QBatchList {
public:
  T2_QBatchList(): qlist(), app(NULL) {};
  ~T2_QBatchList();

  // query methods
  bool isEmpty() const { return qlist.isEmpty(); };
  u_int getNumberQBatches() const { return qlist.getListSize(); };

  // append an entry for an empty qbatch
  T2_QPtrBatch& appendEmptyQBatch();

  // set and get app
  T2_AppFrontEnd* getAppFrontEnd() { return app; };
  T2_AppFrontEnd* setAppFrontEnd(T2_AppFrontEnd* q);
  void clearAppFrontEnd();

  // get query batch from qlist
  T2_QPtrBatch* getAndRemoveQBatch();

protected:
  T2_TDBList<T2_QPtrBatch*> qlist;
  T2_AppFrontEnd* app;

private:
};


inline
T2_QBatchList::~T2_QBatchList()
{
  t2_Assert (qlist.isEmpty() == true);
  t2_Assert (app == NULL);
}

inline T2_AppFrontEnd*
T2_QBatchList::setAppFrontEnd(T2_AppFrontEnd* a)
{
  t2_Assert (a != NULL);
  t2_Assert (app == NULL);
  return (app = a); 
}

inline void
T2_QBatchList::clearAppFrontEnd()
{
  app = NULL;
}



//////////////////////////////////////////////////////////////////////////
//
// T2_AppFrontEnd
//
// A class to represent an application front-end connected to T2 front-end
// by the T2 front-end process.
//
//////////////////////////////////////////////////////////////////////////
class T2_AppFrontEnd {
public:
  T2_AppFrontEnd(): sock(), qlist(NULL) {};
  ~T2_AppFrontEnd() { t2_Assert (qlist == NULL); };

  // access methods
  T2_Socket& getSocket() { return sock; };
  T2_QBatchList* setQueryList(T2_QBatchList *q);
  T2_QBatchList* getQueryList() { return qlist; };
  T2_QBatchList* clearQueryList() { return (qlist = NULL); };
  u_char getAppFrontEndEndian() const { return endian; };
  void setAppFrontEndEndian(u_char e) { endian = e; };

protected:
  T2_Socket sock;
  T2_QBatchList *qlist;
  u_char endian;		// endian of the application front-end

private:
};


inline T2_QBatchList*
T2_AppFrontEnd::setQueryList(T2_QBatchList* q)
{
  t2_Assert (q != NULL);
  t2_Assert (qlist == NULL);
  return (qlist = q);
}




//////////////////////////////////////////////////////////////////////////
//
// T2_FrontEnd
//
//////////////////////////////////////////////////////////////////////////
enum T2_FrontEndError {
	t2_feNoError = 0,		// no error
	t2_feSysError,			// serious error !!
	t2_feNotConnected,		// not connected to T2 front-end
	t2_feConnectFailure,		// cannot connect to T2 front-end
					// (check errno for detailed reason)
	t2_feUnexpectedDisconnect,	// T2 front-end unexpectedly close conn.
	t2_feT2Busy,			// T2 front-end is busy
	t2_fePingAckError,		// T2 front-end returns unrecognizable
					// ack for ping
	t2_feEInvalid,			// invalid argument
	t2_feWriteError,		// error occurs during write
	t2_feReadError			// error occurs during read
};


class T2_FrontEnd {
public:
  T2_FrontEnd() { nbackends = 0; errval = t2_feNoError; 
		  t2fe_endian = t2be_endian = T2_UnknownEndian;
		  my_endian = t2_checkEndianess(); };
  ~T2_FrontEnd();

  // get error value
  T2_FrontEndError getErrorVal() const { return errval; };

  // connect to/disconnect from T2 front-end process
  // return true if succeed, false otherwise;
  // when return false, sets errno to the errval returned by sock
  bool connectT2FrontEndByHostname(const char* hostname, short port);
  bool connectT2FrontEndByAddress(const char* ipaddr, short port);
  void disconnectT2FrontEnd();

  // number of back-end nodes
  u_int getNumberBackEndNodes() const { return nbackends; };

  // get endianess of the T2 front-end machine, the T2 back-end machine,
  // and the local machine
//  u_char getT2FrontEndEndianess() const { return t2fe_endian; };
  u_char getT2BackEndEndianess() const { return t2be_endian; };
  u_char getMyEndianess() const { return my_endian; };

  // ping the T2 front-end and wait for an ack back
  bool pingT2FrontEnd();

  // submit a query batch
  bool submitQBatch(T2_QBatch& qbatch);

  // convert error value to string, usually for printout
  static const char* errorValToString(T2_FrontEndError e);
  const char* errorValToString() const
		{ return errorValToString(getErrorVal()); };

  // inquiry methods for datasets and functions for exact matching
  // or regular expression matching;
  // the argument max_totblobsize for inquiring datasets only makes
  // sense when content's of blob object is retrieved, in which case
  // at most max_totblobsize would be transferred over and stored
  // in the results; the sizes of those blob objects that didn't make 
  // its way over to the caller due to the limit can be found in
  // the field of blob data size; further inquries can be made to
  // get the blob objects over the socket
  bool inquireDatasetExactMatch(const char* pattern, const int fields,
				const size_t max_totblobsize,
				T2_FEDatasetInquiryResults& results);
  bool inquireDatasetRegExp(const char* pattern, const int fields,
				const size_t max_totblobsize,
				T2_FEDatasetInquiryResults& results);
  bool inquireFunctionExactMatch(const char* pattern, const T2_UDFType ftype,
				const int fields,
				T2_FEFunctionInquiryResults& results);
  bool inquireFunctionRegExp(const char* pattern, const T2_UDFType ftype,
				const int fields,
				T2_FEFunctionInquiryResults& results);

  // a general inquiry method
  bool inquireDataset(const char* pattern, bool exact_match,
		      const int fields, const size_t max_totblobsize,
		      T2_FEDatasetInquiryResults& results);
  bool inquireFunction(const char* pattern, bool exact_match,
		       const T2_UDFType func_type, const int fields,
		       T2_FEFunctionInquiryResults& results);

protected:
  T2_Socket sock;
  u_int nbackends;
  u_char t2fe_endian,			// endianess of the t2-frontend
	 t2be_endian,			// endianess of the t2-backend
	 my_endian;			// endianess of the local process
  T2_AppToFEProtocol protocol;

  T2_FrontEndError errval;

private:
  bool waitforConnectAck();

};


inline bool
T2_FrontEnd::inquireDatasetExactMatch(const char* pattern, const int fields,
				      const size_t max_totblobsize,
				      T2_FEDatasetInquiryResults& results)
{
  return inquireDataset(pattern, true, fields, max_totblobsize, results);
}

inline bool
T2_FrontEnd::inquireDatasetRegExp(const char* pattern, const int fields,
				  const size_t max_totblobsize,
				  T2_FEDatasetInquiryResults& results)
{
  return inquireDataset(pattern, false, fields, max_totblobsize, results);
}

inline bool
T2_FrontEnd::inquireFunctionExactMatch(const char* pattern,
				       const T2_UDFType ftype,
				       const int fields,
				       T2_FEFunctionInquiryResults& results)
{
  return inquireFunction(pattern, true, ftype, fields, results);
}

inline bool
T2_FrontEnd::inquireFunctionRegExp(const char* pattern, const T2_UDFType ftype,
				   const int fields,
				   T2_FEFunctionInquiryResults& results)
{
  return inquireFunction(pattern, false, ftype, fields, results);
}


#endif	/* ! _T2_FRONTEND_H_ */
