
/* client.C */

/*
 * Chialin Chang, chialin@cs.umd.edu, 7/22/98
 * An simple client that reads queries from files, and uses T2_FrontEnd
 * interface to talk to a T2 Front-End process.
 */


#include	<iostream.h>
#include	<unistd.h>
#include	"t2_bfile.h"
#include	"t2_frontend.h"


/////////////////////////////////////////////////////////////////////////////
//
// some help function prototypes and variables
//
/////////////////////////////////////////////////////////////////////////////
enum usageError { eUNKNOWN, eBADARG };
static void printusage(const char*);
static void badusage(usageError, const char*, char);

const char default_t2fe_host[] = "condor.cs.umd.edu";
const u_int default_t2fe_port = 9987,
	    default_t2_nqry	= 10;


/////////////////////////////////////////////////////////////////////////////
//
// main
//
/////////////////////////////////////////////////////////////////////////////
main(int argc, char** argv)
{
  int t2fe_port				// port that T2 front-end listens to
		= default_t2fe_port,
      t2_nqry 				// # query batches to send to T2
		= default_t2_nqry;
  const char *hostname = default_t2fe_host,
             *qfile = NULL;

  int sec = 1;

  // deal with the command-line arguments
  int i;
  for (i=0; i<argc; i++) {
     if (argv[i][0] == '-') {
	switch(argv[i][1]) {
	case 'h':
		printusage(argv[0]); return 0; break;

	case 'f':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		hostname = argv[i];
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-2][1]); return -1; }
		t2fe_port = atoi(argv[i]);
		if (t2fe_port < 0)
		   { badusage(eBADARG, argv[0], argv[i-2][1]); return -1; }
		break;

	case 'n':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		t2_nqry = atoi(argv[i]);
		if (t2_nqry < 0)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		break;

	case 'q':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		qfile = argv[i];
		break;

	case 's':
		i++;
		if (i >= argc)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		sec = atoi(argv[i]);
		if (t2_nqry < 0)
		   { badusage(eBADARG, argv[0], argv[i-1][1]); return -1; }
		break;

	default:
		badusage(eUNKNOWN, argv[0], argv[i][1]); return -1; break;
        }
     }
  }

  cout << "T2 front-end host = " << hostname
       << ", port = " << t2fe_port << endl;
  if (qfile == NULL)
     cout << "Will issue " << t2_nqry << " empty query batches to back-end"
          << endl;
  else {
     cout << "Will read query batch file " << qfile << " and send the query"
	  << endl
          << "batch in the file to the back-end " << t2_nqry << " times"
	  << endl;
  }


  T2_FrontEnd fe;
  if (fe.connectT2FrontEndByHostname(hostname, t2fe_port) == false) {
     // cannot connect to T2 front-end
     if (fe.getErrorVal() == t2_feT2Busy)
	cout << "T2 front-end is busy" << endl;
     else
	t2_err_ret(__FILE__, __LINE__, "fe.connectT2FrontEndByHostname error");
     return -1;
  }

  cout << "Connected to T2 front-end" << endl;
  cout << "T2 current has " << fe.getNumberBackEndNodes()
       << " back-end nodes" << endl;


  if (qfile == NULL) {
     for (u_int j=0; j<sec; j++) {
	bool r = fe.pingT2FrontEnd();
	t2_Assert (r == true);
	cout << "ping " << j << " returns true" << endl;
	sleep(1);
     }
  }
  else {
     T2_IBFile qibf(qfile);
     if (!qibf) {
	t2_err_ret(__FILE__, __LINE__,
		  "Cannot open query batch file %s", qfile);
     }
     else {

        T2_QBatch q;
        qibf >> q;
        qibf.close();

        if (q.isValid() == false) {		// qbatch read error
	   t2_err_ret(__FILE__, __LINE__, "Query batch read error");
        }
        else {
	   cout << "qbatch file:" << endl
		<< q << endl;
	   for (u_int k=0; k<t2_nqry; k++) {
	      cout << "sending query batch " << k << " times" << endl;
	      if (fe.submitQBatch(q) == false) {
	         t2_err_ret(__FILE__, __LINE__,
		           "submitQBatch error %d", (int) fe.getErrorVal());
	         break;
	      }
	      sleep(1);
	   }
        }				// if (q.isValid()==false) else
     }					// if (!qibf) else
  }					// if (qfile == NULL) else




  cout << endl << "now disconnect from T2 front-end" << endl;
  fe.disconnectT2FrontEnd();

  cout << "program " << argv[0] << " exits gracefully" << endl;
  return 0;
}



/////////////////////////////////////////////////////////////////////////////
//
// help functions
//
/////////////////////////////////////////////////////////////////////////////
void
printusage(const char* pname)
{
  cout << "Usage: " << pname
    << " [-h] |"
    << " [-f <T2 front-end hostname> <port number>]" << endl
    << " [-q <query batch file>]"
    << " [-n <# query batches>]" << endl
    << endl;

  cout << "Where:" << endl;
  cout << "-h prints this message" << endl;
  cout << "-f specifies the hostname and port number that T2 front-end listens to" << endl
       << "   (default = " << default_t2fe_host << ", "
       << default_t2fe_port  << ")" 
       << endl;
  cout << "-q specifies the query batch file" << endl
       << "   (default = NULL)" << endl;
  cout << "-n when -q is used to specify the query batch file, -n specifies the"
       << endl
       << "   number of times the query batch file is read and the query batches"
       << endl
       << "   stored in the file are sent to the back-end before shutting down;"
       << endl
       << "   when -q is not specified, -n specifies the number of times an empty"
       << endl
       << "   query batch file is sent to the back-end before shutting down"
       << endl
       << "   (default = " << default_t2_nqry  << ")" << endl
       << endl;
}


void
badusage(usageError e, const char* p, char a)
{
  switch (e) {
    case eUNKNOWN:
        cerr << "Unknown parameter for -" << a << " in command line" << endl;
        break;
    case eBADARG:
        cerr << "Missing argument for parameter -" << a << " in command line"
             << endl;
        break;
    default:
        cerr << "Some kind of error in the parameter -" << a
             << " in command line" << endl;
  }
  printusage(p);
}
