/*
  File name: hilbert.c 

  Mapping to 1D from 2D and 3D using Hilbert Curve. 
*/
#ifndef HILBERT_C
#define HILBERT_C

#include <stdlib.h>
#include <stdio.h>
#include "hilbert.h"

#if 0  /* Implementations moved into three files (hilbert[2,3,4]d.c) */

/* 2D Hilbert Curve */
static int _RotationTable[4]   = {3,0,0,1};
static int _SenseTable[4]      = {-1,1,1,-1};
static int _QuadTable[4][2][2] = {{{0,1},{3,2}},
				  {{1,2},{0,3}},
				  {{2,3},{1,0}},
				  {{3,0},{2,1}}};
 
/* Returns the index of a point (x,y) in hilbert curve ordering */
unsigned int hilbert2D(unsigned short x, unsigned short y) 
{
  int rotation = 0;
  int sense    = 1;
  unsigned int side, k, num = 0;
  int quad, xbit, ybit;
 
  /* IMPORTANT: side should be a power of 2 */
  side = (unsigned int) 65536;
  for (k=side; k > 0; k >>= 1) { 
    xbit = x/k;
    ybit = y/k;
    x -= k*xbit;
    y -= k*ybit;
    quad = _QuadTable[rotation][xbit][ybit];
    if (sense == -1)
      num += k*k*(3-quad);
    else
      num += k*k*quad;
    rotation += _RotationTable[quad];
    if (rotation >= 4)
      rotation -= 4;
    sense *=_SenseTable[quad];
  }
  return num;
}

/* 3D Hilbert Space-filling Curve */
/*
  Source: 
     "Fractals for Secondary Key Retrieval", C. Faloutsos, S. Roseman
     University of Maryland, Computer Science Dept., Tech. Rept. CS-TR-2242.
*/
static unsigned int first_column[8]       = {0,1,2,3,4,5,6,7};
static unsigned int second_column[8]      = {0,1,3,2,6,7,5,4};
static unsigned int inv_second[8]         = {0,1,3,2,7,6,4,5};
static unsigned int third_column[8][2]    = {
  {0,1},{0,2},{0,4},{5,4},
  {0,2},{3,2},{0,4},{5,4}
};
static unsigned int fourth_column[8]      = {1,2,4,1,2,1,4,1};
static int fifth_column[8][3][3] = {
  {{0,0,1},{0,1,0},{1,0,0}},
  {{0,1,0},{0,0,1},{1,0,0}},
  {{1,0,0},{0,1,0},{0,0,1}},
  {{0,0,-1},{0,1,0},{-1,0,0}},
  {{0,1,0},{1,0,0},{0,0,1}},
  {{0,0,-1},{0,-1,0},{1,0,0}},
  {{1,0,0},{0,1,0},{0,0,1}},
  {{0,0,-1},{0,1,0},{-1,0,0}}
};

unsigned int hilbert3D(unsigned int x, unsigned int y, unsigned int z, 
		       unsigned int hn[])
{
  unsigned char input[3], output[3], rotated[32][3];
  int          i;
  unsigned int j,n,k,l;
  int          cmat[3][3] = {{0,0,1},{0,1,0},{1,0,0}};
  int          cnew[3][3];

  for (i=31;i>=0;i--) {
    input[0] = (x&(1<<i)) ? 1 : 0;
    input[1] = (y&(1<<i)) ? 1 : 0;
    input[2] = (z&(1<<i)) ? 1 : 0;

    for (j=0;j<3;j++) {
      for (k=0;k<3;k++) 
	if (cmat[j][k]!=0) break;

      if (j==3 || k==3) {
	fprintf(stderr, "There is an error in the matrix.\n");
	exit(0);
      }
      
      if (cmat[j][k]==1)
	output[j] = input[k];
      else
	output[j] = 1-input[k];
    }

    k = (unsigned int) (((int)output[0])*4+
			((int)output[1])*2+
			((int)output[2]));
    j = inv_second[k];
    rotated[i][2] = (first_column[j]&(1<<2)) ? 1 : 0;
    rotated[i][1] = (first_column[j]&(1<<1)) ? 1 : 0;
    rotated[i][0] = (first_column[j]&(1<<0)) ? 1 : 0;

    cnew[0][0] = (fifth_column[j][0][0] * cmat[0][0] +
		  fifth_column[j][0][1] * cmat[1][0] +
		  fifth_column[j][0][2] * cmat[2][0]);
    cnew[0][1] = (fifth_column[j][0][0] * cmat[0][1] +
		  fifth_column[j][0][1] * cmat[1][1] +
		  fifth_column[j][0][2] * cmat[2][1]);
    cnew[0][2] = (fifth_column[j][0][0] * cmat[0][2] +
		  fifth_column[j][0][1] * cmat[1][2] +
		  fifth_column[j][0][2] * cmat[2][2]);

    cnew[1][0] = (fifth_column[j][1][0] * cmat[0][0] +
		  fifth_column[j][1][1] * cmat[1][0] +
		  fifth_column[j][1][2] * cmat[2][0]);
    cnew[1][1] = (fifth_column[j][1][0] * cmat[0][1] +
		  fifth_column[j][1][1] * cmat[1][1] +
		  fifth_column[j][1][2] * cmat[2][1]);
    cnew[1][2] = (fifth_column[j][1][0] * cmat[0][2] +
		  fifth_column[j][1][1] * cmat[1][2] +
		  fifth_column[j][1][2] * cmat[2][2]);

    cnew[2][0] = (fifth_column[j][2][0] * cmat[0][0] +
		  fifth_column[j][2][1] * cmat[1][0] +
		  fifth_column[j][2][2] * cmat[2][0]);
    cnew[2][1] = (fifth_column[j][2][0] * cmat[0][1] +
		  fifth_column[j][2][1] * cmat[1][1] +
		  fifth_column[j][2][2] * cmat[2][1]);
    cnew[2][2] = (fifth_column[j][2][0] * cmat[0][2] +
		  fifth_column[j][2][1] * cmat[1][2] +
		  fifth_column[j][2][2] * cmat[2][2]);

    
    for (j=0;j<3;j++)
      for (k=0;k<3;k++)
	cmat[j][k] = cnew[j][k];
  }

  for (i=0;i<3;i++) hn[i] = 0;

  k = 0; n = 2; 
  for (i=0;i<32;i++) {
    for (j=0;j<3;j++) {
      hn[n] |= (((unsigned int)rotated[i][j])<<k);
      k++;
      if (k==32) {
	n--; k = 0;
      }
    }
  }

  return 1;
}

/* Comparison routines for quick sort */
int hsf2d_compare(const void *a, const void *b)
{
#define H2DCAST(x) ((_thsf2d_order*)x)
  if (H2DCAST(a)->hsf_idx>H2DCAST(b)->hsf_idx)
    return 1;
  if (H2DCAST(a)->hsf_idx<H2DCAST(b)->hsf_idx)
    return -1;
  return 0;
}

int hsf3d_compare(const void *a, const void *b)
{
#define H3DCAST(x) ((_thsf3d_order*)x)
  if (H3DCAST(a)->v[0]>H3DCAST(b)->v[0]) return  1;
  if (H3DCAST(a)->v[0]<H3DCAST(b)->v[0]) return -1;
  if (H3DCAST(a)->v[1]>H3DCAST(b)->v[1]) return  1;
  if (H3DCAST(a)->v[1]<H3DCAST(b)->v[1]) return -1;
  if (H3DCAST(a)->v[2]>H3DCAST(b)->v[2]) return  1;
  if (H3DCAST(a)->v[2]<H3DCAST(b)->v[2]) return -1;
  return 0;
}

#endif /* 0 */ 

#endif /* HILBERT_MAP_C */

