/*
  File name: hilbert4d.c 

  Used to find the Hilbert mapping of a point in 4-dimensional space.

  Source: 
     "Fractals for Secondary Key Retrieval", C. Faloutsos, S. Roseman
     University of Maryland, Computer Science Dept., Tech. Rept. CS-TR-2242.

  $Id: hilbert4d.c,v 1.1 2000/02/23 17:11:42 chialin Exp $
*/
#ifndef HILBERT4D_C
#define HILBERT4D_C

#include <stdlib.h>
#include <stdio.h>
#include "hilbert.h"

static unsigned int first_column[16] = {0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15};
static unsigned int second_column[16] = {0,1,3,2,6,7,5,4,12,13,15,14,10,11,9,8};
static unsigned int inv_second[16] = {0,1,3,2,7,6,4,5,15,14,12,13,8,9,11,10};
static unsigned int third_column[16][2] = {{0,1},{0,2},{0,4},{5,13},
					   {9,11},{10,8},{10,8},{9,11},
					   {3,11},{10,2},{0,8},{9,11},
					   {15,13},{12,8},{10,8},{9,8}};
static unsigned int fourth_column[16] = {1,2,4,8,2,2,2,2,8,8,8,2,2,4,2,1};
static unsigned int fifth_column[16][4][4] = {
  {{0,0,0,1},{0,0,1,0},{0,1,0,0},{1,0,0,0}},
  {{0,0,1,0},{0,0,0,1},{0,1,0,0},{1,0,0,0}},
  {{0,1,0,0},{0,0,0,1},{0,0,1,0},{1,0,0,0}},
  {{1,0,0,0},{0,-1,0,0},{0,0,1,0},{0,0,0,-1}},
  {{0,0,1,0},{0,0,0,-1},{0,1,0,0},{-1,0,0,0}},
  {{0,0,-1,0},{0,0,0,1},{0,1,0,0},{-1,0,0,0}},
  {{0,0,1,0},{0,0,0,1},{0,1,0,0},{-1,0,0,0}},
  {{0,0,1,0},{0,0,0,-1},{0,1,0,0},{-1,0,0,0}},
  {{1,0,0,0},{0,1,0,0},{0,0,-1,0},{0,0,0,-1}},
  {{-1,0,0,0},{0,1,0,0},{0,0,-1,0},{0,0,0,1}},
  {{1,0,0,0},{0,1,0,0},{0,0,1,0},{0,0,0,1}},
  {{0,0,1,0},{0,0,0,-1},{0,1,0,0},{-1,0,0,0}},
  {{0,0,-1,0},{0,0,0,-1},{0,-1,0,0},{-1,0,0,0}},
  {{0,-1,0,0},{0,0,0,1},{0,0,1,0},{-1,0,0,0}},
  {{0,0,-1,0},{0,0,0,1},{0,1,0,0},{-1,0,0,0}},
  {{0,0,0,-1},{0,0,1,0},{0,1,0,0},{-1,0,0,0}}
};

int hilbert4D(unsigned int  x, unsigned int  y, unsigned int  z, 
	      unsigned int  w, unsigned int  hn[])
{
  unsigned char input[4], output[4], rotated[32][4];
  int           i;
  unsigned int  j,n,k,l;
  unsigned int  o,p,r;
  unsigned int  cmat[4][4] = {{0,0,0,1},{0,0,1,0},{0,1,0,0},{1,0,0,0}};
  unsigned int  cnew[4][4];

  for (i=31;i>=0;i--) {
    input[0] = (x&(1<<i)) ? 1 : 0;
    input[1] = (y&(1<<i)) ? 1 : 0;
    input[2] = (z&(1<<i)) ? 1 : 0;
    input[3] = (w&(1<<i)) ? 1 : 0;

    for (j=0;j<4;j++) {
      for (k=0;k<4;k++) 
	if (cmat[j][k]!=0) break;

      if (j==4 || k==4) {
	fprintf(stderr, "There is an error in the matrix.\n");
	return 0;
      }
      
      if (cmat[j][k]==1)
	output[j] = input[k];
      else
	output[j] = 1-input[k];
    }

    k = (unsigned int) (((int)output[0])*8+
			((int)output[1])*4+
			((int)output[2])*2+
			((int)output[3]));

    j = inv_second[k];

    rotated[i][3] = (first_column[j]&(1<<3)) ? 1 : 0;
    rotated[i][2] = (first_column[j]&(1<<2)) ? 1 : 0;
    rotated[i][1] = (first_column[j]&(1<<1)) ? 1 : 0;
    rotated[i][0] = (first_column[j]&(1<<0)) ? 1 : 0;

    for (o=0;o<4;o++)
      for (p=0;p<4;p++)
	cnew[o][p] = 0;

    for (o=0;o<4;o++)
      for (p=0;p<4;p++)
	for (r=0;r<4;r++) 
	  cnew[o][p] += (fifth_column[j][o][r]*cmat[r][p]);

    for (j=0;j<4;j++)
      for (k=0;k<4;k++)
	cmat[j][k] = cnew[j][k];
  }

  for (i=0;i<4;i++) hn[i] = 0;

  n = 3; k = 0;
  for (i=0;i<32;i++) {
    for (j=0;j<4;j++) {
      hn[n] |= (((unsigned int)rotated[i][j])<<k);
      
      k++;
      if (k==32) {
	n--; k = 0;
      }
    }
  }

  return 1;
}

int hsf4d_compare(const void *a, const void *b)
{
#define H4DCAST(x) ((_thsf4d_order*)x)
  if (H4DCAST(a)->v[0]>H4DCAST(b)->v[0]) return  1;
  if (H4DCAST(a)->v[0]<H4DCAST(b)->v[0]) return -1;
  if (H4DCAST(a)->v[1]>H4DCAST(b)->v[1]) return  1;
  if (H4DCAST(a)->v[1]<H4DCAST(b)->v[1]) return -1;
  if (H4DCAST(a)->v[2]>H4DCAST(b)->v[2]) return  1;
  if (H4DCAST(a)->v[2]<H4DCAST(b)->v[2]) return -1;
  if (H4DCAST(a)->v[3]>H4DCAST(b)->v[3]) return  1;
  if (H4DCAST(a)->v[3]<H4DCAST(b)->v[3]) return -1;
  return 0;
}

#endif /* HILBERT4D_C */
