/*
  File name: t2dld_hilbert.C

  Declustering and clustering using 3D hilbert curve for T2 data loading 
  service.

  $Id: t2dld_hilbert.C,v 1.1 2000/02/23 16:56:54 chialin Exp $
*/
#ifndef T2DLD_HILBERT_C
#define T2DLD_HILBERT_C

#include "t2dld_hilbert.h"

#define  MAX_VALUE      (1<<30)

bool 
T2DLD_Hilbert3D::placeClusters(const T2DLD_DataClusterArray& tclusters,
			       T2_Array<T2DLD_MoverInstruction>& mover_instr)
{
  int    i, j;
  double pntMin[MAX_DIMENSIONS_ALLOWED];
  double pntMax[MAX_DIMENSIONS_ALLOWED];
  double pntMid;
  u_int  pntCrv[MAX_DIMENSIONS_ALLOWED];
  u_int  nblocks; 

  nblocks = tclusters.getNumberDataClusters();

# if (T2_VERBOSE > 2)
     fprintf(stdout, "T2DLD_Hilbert3D::placeClusters, nblocks: %d\n",
	     nblocks);
     fflush(stdout);
# endif

  if (nDims>MAX_DIMENSIONS_ALLOWED) {
     fprintf(stderr, "number of dimensions is larger than allowed.\n");
     fflush(stderr);
     return false;
  }

  /* 1. Normalization */
  // Find min and max values ... 
  T2_Box mbr      = tclusters[0]->getMBR();
  T2_Point lowpnt = mbr.getLow();
  T2_Point hghpnt = mbr.getHigh();
  for (j=0;j<nDims;j++) {
    pntMid    = (lowpnt[j]+hghpnt[j])/2.0;
    pntMin[j] = pntMid;
    pntMax[j] = pntMid;
  }
  for (i=1;i<nblocks;i++) {
    mbr    = tclusters[i]->getMBR();
    lowpnt = mbr.getLow();
    hghpnt = mbr.getHigh();
    for (j=0;j<nDims;j++) {
      pntMid = (lowpnt[j]+hghpnt[j])/2.0;
      if (pntMin[j]>pntMid) pntMin[j] = pntMid;
      if (pntMax[j]<pntMid) pntMax[j] = pntMid;
    }
  }

  if (nDims<=2) {  /* 2-dimensional hilbert curve */

#   if (T2_VERBOSE > 1)
       fprintf(stdout, "USING 2D HILBERT .... (%d,%d) \n",nblocks,nDims);
       fflush(stdout);

       fprintf(stdout, "T2DLD_Hilbert3D::placeClusters, ");
       fprintf(stdout, "min (%f,%f) max (%f,%f)\n",
	       pntMin[0],pntMin[1],
	       pntMax[0],pntMax[1]);
       fflush(stdout);
#   endif

    _thsf2d_order *hsfc;
    hsfc = (_thsf2d_order *) malloc(sizeof(_thsf2d_order)*nblocks);
    if (hsfc==NULL) {
      fprintf(stderr, "Cannot allocate space for hilbert curve.\n");
      fflush(stderr);
      return false;
    }
  
    /* 2. Hilbert traversal */
    for (i=0;i<2;i++) pntCrv[i] = 0;
    for (i=0;i<nblocks;i++) {
      mbr    = tclusters[i]->getMBR();
      lowpnt = mbr.getLow();
      hghpnt = mbr.getHigh();

      for (j=0;j<nDims;j++) {
	pntMid    = (lowpnt[j]+hghpnt[j])/2.0;
	pntCrv[j] = (u_int) 
	  (((pntMid-pntMin[j])/(pntMax[j]-pntMin[j]))*MAX_VALUE);
      }

#     if (T2_VERBOSE > 2)
         fprintf(stdout, "T2DLD_Hilbert3D::placeClusters, points (");
         for (j=0;j<nDims;j++) 
	   fprintf(stdout, "%f ",lowpnt[j]);
         fprintf(stdout, ")x(");
         for (j=0;j<nDims;j++) 
	   fprintf(stdout, "%f ",hghpnt[j]);
         fprintf(stdout, ") ");
         fprintf(stdout,
		 "normalized (%u,%u,%u)\n",pntCrv[0],pntCrv[1],pntCrv[2]);
         fflush(stdout);
#     endif

      hsfc[i].pnt_idx = i;
      if (hilbert2D(pntCrv[0],pntCrv[1],hsfc[i].v)==0) {
	fprintf(stderr, "Something is wrong ... \n");
	fflush(stderr);
	return false;
      }

#     if (T2_VERBOSE > 2)
         fprintf(stdout, "T2DLD_Hilbert3D::placeClusters, ");
         fprintf(stdout, "Hilbert value (%u, %u)\n",
	         hsfc[i].v[0],hsfc[i].v[1]);
         fflush(stdout);
#     endif

    }

    /* 3. Placement -- declustering/clustering */
    qsort((void*)hsfc,(size_t)nblocks,(size_t)sizeof(_thsf2d_order),
	  hsf2d_compare);

#   if (T2_VERBOSE > 2)
       fprintf(stdout, "Done with mapping... doing assignment...\n");
       fflush(stdout);
#   endif

    int curr_disk = 0;
    for (i=0;i<nblocks;i++) {
      j = hsfc[i].pnt_idx;

#     if (T2_VERBOSE > 2)
         fprintf(stdout, "T2DLD_Hilbert3D::placeClusters, ");
         fprintf(stdout, "value: %u %u id: %d disk: %d\n",
	         hsfc[i].v[0],hsfc[i].v[1],j,curr_disk);
         fflush(stdout);
#     endif

      mover_instr[curr_disk] << tclusters[j];
      curr_disk = (curr_disk+1)%nDisks;
    }
    free(hsfc);

  } else if (nDims==3) { /* 3-dimensional hilbert curve */

#   if (T2_VERBOSE > 1)
       fprintf(stdout, "USING 3D HILBERT .... \n");
       fflush(stdout);
#   endif

    _thsf3d_order *hsfc;
    hsfc = (_thsf3d_order *) malloc(sizeof(_thsf3d_order)*nblocks);
    if (hsfc==NULL) {
      fprintf(stderr, "Cannot allocate space for hilbert curve.\n");
      fflush(stderr);
      return false;
    }
  
    /* 2. Hilbert traversal */
    for (i=0;i<3;i++) pntCrv[i] = 0;
    for (i=0;i<nblocks;i++) {
      mbr    = tclusters[i]->getMBR();
      lowpnt = mbr.getLow();
      hghpnt = mbr.getHigh();

      for (j=0;j<nDims;j++) {
	pntMid    = (lowpnt[j]+hghpnt[j])/2.0;
	pntCrv[j] = (u_int) 
	  (((pntMid-pntMin[j])/(pntMax[j]-pntMin[j]))*MAX_VALUE);
      }

      hsfc[i].pnt_idx = i;
      if (hilbert3D(pntCrv[0],pntCrv[1],pntCrv[2],hsfc[i].v)==0) 
	return false;
    }

    /* 3. Placement -- declustering/clustering */
    qsort((void*)hsfc,(size_t)nblocks,(size_t)sizeof(_thsf3d_order),
	  hsf3d_compare);

    int curr_disk = 0;
    for (i=0;i<nblocks;i++) {
      j = hsfc[i].pnt_idx;
      mover_instr[curr_disk] << tclusters[j];
      curr_disk = (curr_disk+1)%nDisks;
    }
    free(hsfc);

  } else if (nDims==4) { /* 4-dimensional hilbert curve */

#   if (T2_VERBOSE > 1)
       fprintf(stdout, "USING 4D HILBERT .... \n");
       fflush(stdout);
#   endif

    _thsf4d_order *hsfc;
    hsfc = (_thsf4d_order *) malloc(sizeof(_thsf4d_order)*nblocks);
    if (hsfc==NULL) {
      fprintf(stderr, "Cannot allocate space for hilbert curve.\n");
      fflush(stderr);
      return false;
    }
  
    /* 2. Hilbert traversal */
    for (i=0;i<4;i++) pntCrv[i] = 0;
    for (i=0;i<nblocks;i++) {
      mbr    = tclusters[i]->getMBR();
      lowpnt = mbr.getLow();
      hghpnt = mbr.getHigh();

      for (j=0;j<nDims;j++) {
	pntMid    = (lowpnt[j]+hghpnt[j])/2.0;
	pntCrv[j] = (u_int) 
	  (((pntMid-pntMin[j])/(pntMax[j]-pntMin[j]))*MAX_VALUE);
      }

      hsfc[i].pnt_idx = i;
      if (hilbert4D(pntCrv[0],pntCrv[1],pntCrv[2],pntCrv[3],hsfc[i].v)==0) 
	return false;
    }

    /* 3. Placement -- declustering/clustering */
    qsort((void*)hsfc,(size_t)nblocks,(size_t)sizeof(_thsf4d_order),
	  hsf4d_compare);

    int curr_disk = 0;
    for (i=0;i<nblocks;i++) {
      j = hsfc[i].pnt_idx;
      mover_instr[curr_disk] << tclusters[j];
      curr_disk = (curr_disk+1)%nDisks;
    }
    free(hsfc);

  } else {

#   if (T2_VERBOSE > 1)
       fprintf(stdout, "USING 4D HILBERT for a larger dimension .... \n");
       fflush(stdout);
#   endif

    nDims = 4; /* for the moment use the first four points */

    _thsf4d_order *hsfc;
    hsfc = (_thsf4d_order *) malloc(sizeof(_thsf4d_order)*nblocks);
    if (hsfc==NULL) {
      fprintf(stderr, "Cannot allocate space for hilbert curve.\n");
      fflush(stderr);
      return false;
    }
  
    /* 2. Hilbert traversal */
    for (i=0;i<4;i++) pntCrv[i] = 0;
    for (i=0;i<nblocks;i++) {
      mbr    = tclusters[i]->getMBR();
      lowpnt = mbr.getLow();
      hghpnt = mbr.getHigh();

      for (j=0;j<nDims;j++) {
	pntMid    = (lowpnt[j]+hghpnt[j])/2.0;
	pntCrv[j] = (u_int) 
	  (((pntMid-pntMin[j])/(pntMax[j]-pntMin[j]))*MAX_VALUE);
      }

      hsfc[i].pnt_idx = i;
      if (hilbert4D(pntCrv[0],pntCrv[1],pntCrv[2],pntCrv[3],hsfc[i].v)==0) 
	return false;
    }

    /* 3. Placement -- declustering/clustering */
    qsort((void*)hsfc,(size_t)nblocks,(size_t)sizeof(_thsf4d_order),
	  hsf4d_compare);

    int curr_disk = 0;
    for (i=0;i<nblocks;i++) {
      j = hsfc[i].pnt_idx;
      mover_instr[curr_disk] << tclusters[j];
      curr_disk = (curr_disk+1)%nDisks;
    }
    free(hsfc);

  }

  return true;
}

#endif /* T2DLD_HILBERT_C */
