
/* $Id: t2dld_base.h,v 1.3 1999/03/30 21:45:46 chialin Exp $ */

/* t2dld_base.h */


#ifndef _T2DLD_BASE_H_
#define _T2DLD_BASE_H_

///////////////////////////////////////////////////////////////////////////
//
// This file contains the basic classes needed by the T2 loader. These
// classes define a physical tape data block, a data cluster which consists
// of a number of data blocks, and an array of data clusters.
//
///////////////////////////////////////////////////////////////////////////

#include	<iostream.h>
#include	"t2dld_header.h"
#include	"t2_tarray.h"
#include	"t2_bfile.h"
#include	"t2_usrarg.h"
#include	"t2_shape.h"


// forward declaration
class T2DLD_TapeClusterArray;
class T2DLD_MoverInstruction;

///////////////////////////////////////////////////////////////////////////
//
// T2DLD_DataBlock
//
///////////////////////////////////////////////////////////////////////////
class T2DLD_DataBlock {
public:
  T2DLD_DataBlock() { fid = 0; offset = 0; nbytes = 0; data = NULL; };
  ~T2DLD_DataBlock() { delete [] data; };

  // const access methods
  u_int getDataFileID() const { return fid; };
  off_t getOffset() const { return offset; };
  size_t getSize() const { return nbytes; };
  const char* getData() const { return data; };

  // set data members
  u_int& getDataFileID() { return fid; };
  off_t& getOffset() { return offset; };
  size_t& getSize() { return nbytes; };
  char*& getData() { return data; };
  void setDataFileID(u_int i) { fid = i; };
  void setOffset(off_t f) { offset = f; };
  void setSize(size_t s) { nbytes = s; };
  void setData(char* c, size_t s) { data = c; nbytes = s; };

  // output to ostream
  ostream& write(ostream& os) const;

protected:
  u_int fid;			// data file id
  off_t offset;			// tape offset
  size_t nbytes;		// block size
  char* data;			// actual data

private:
};


inline ostream&
operator << (ostream& os, const T2DLD_DataBlock& b)
{
  return b.write(os);
}



/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_DataCluster
//
// An array of data blocks.
//
/////////////////////////////////////////////////////////////////////////////
class T2DLD_DataCluster {
public:
  T2DLD_DataCluster();
  ~T2DLD_DataCluster() {};

  // query methods
  u_int getNumberDataBlocks() const { return tblocks.getNumberElements(); };

  // const access methods
  const T2_Box& getMBR() const { return mbr; };
  const T2DLD_DataBlock& getDataBlock(u_int i) const { return tblocks[i]; };
  const T2_UsrArg& getUsrData() const { return user_data; };

  // non-const access methods
  T2_Box& getMBR() { return mbr; };
  T2DLD_DataBlock& getDataBlock(u_int i) { return tblocks[i]; };
  T2_UsrArg& getUsrData() { return user_data; };

  // set number of physical blocks on tapes
  void setNumberDataBlocks(u_int n) { tblocks.clear(n); };

  // set number of dimensions for the bounding box
  void setMBRNumberDimensions(u_int n) { mbr.setNumberDimensions(n); };

  // output to ostream
  ostream& write(ostream& os) const;

  // write information into an index file using the binary format
  void bwrite(T2_OBFile& obf) const;

protected:
  T2_Box mbr;				// bounding box
  T2_Array<T2DLD_DataBlock> tblocks;	// physical blocks on tapes
  T2_UsrArg user_data;			// user-defined data

private:
};


inline ostream&
operator << (ostream& os, const T2DLD_DataCluster& instr)
{
  return instr.write(os);
}



/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_DataClusterArray
//
// An array of data clusters.
//
/////////////////////////////////////////////////////////////////////////////
class T2DLD_DataClusterArray {
friend class T2DLD_TapeClusterArray;
friend class T2DLD_MoverInstruction;
public:
  T2DLD_DataClusterArray(u_int l=200);
  ~T2DLD_DataClusterArray();

  // query methods
  u_int getNumberDataFiles() const
				{ return tape_filenames.getNumberElements(); };
  u_int getNumberDataClusters() const
				{ return clusters.getNumberElements(); };

  // access methods
  const T2DLD_DataCluster* operator[] (u_int i) const;
  const char* getDataFileName(u_int i) const;

protected:
  T2_Array<char*> tape_filenames;		// a list of filenames on tape
  T2_VArray<T2DLD_DataCluster*> clusters;

  // access cluster array
  T2_VArray<T2DLD_DataCluster*>& getClusterArray() { return clusters; };

  // append a new cluster pointer into the array
  T2DLD_DataClusterArray& operator << (T2DLD_DataCluster* i);

private:
};


inline const T2DLD_DataCluster*
T2DLD_DataClusterArray::operator[] (u_int i) const
{
  if (i < getNumberDataClusters())
     return clusters[i];
  return NULL;
}

inline T2DLD_DataClusterArray&
T2DLD_DataClusterArray::operator << (T2DLD_DataCluster* i)
{
  t2_Assert (i != NULL);
  clusters << i;
  return *this;
}

inline const char* 
T2DLD_DataClusterArray::getDataFileName(u_int i) const
{
  if (i < getNumberDataFiles())
     return tape_filenames[i];
  return NULL;
}



#endif	/* ! _T2DLD_BASE_H_ */
