
/* $Id: t2dld_loader.h,v 1.6 2000/02/21 07:27:41 chialin Exp $ */

/* t2dld_loader.h */

#ifndef _T2DLD_LOADER_H_
#define _T2DLD_LOADER_H_

#include	"t2dld_header.h"
#include	"t2dld_mover.h"
#include	"t2dld_getclusters.h"
#include	"t2dld_placement.h"
#include	"t2_beconfig.h"


///////////////////////////////////////////////////////////////////////////
//
// T2DLD_Loader
//
///////////////////////////////////////////////////////////////////////////
static const size_t T2DLD_MAX_COMMAND_LINE_LEN = 1024;

class T2DLD_Loader {
public:
  T2DLD_Loader(u_int nprocs, u_int me, u_int m);
  ~T2DLD_Loader();

  // initialization by hostname or by rank
  bool initByHostname(const char* hostname, T2_BEConfigFile& beconfig)
		{ if (hostname == NULL) return false;
		  return init(hostname, beconfig); };
  bool initByRank(T2_BEConfigFile& beconfig) { return init(NULL, beconfig); };

  // am I the master node?
  bool master_is_me() const { return (mynode == master)? true:false; };

  // const access methods
  u_int getNumberMovers() const { return movers.getNumberElements(); };

  // create getCluster function object and placement function object
  T2DLD_GetClustersFuncObj* getGetClustersFuncObj(const char* index_file);
  T2DLD_PlacementFuncObj* getPlacementFuncObj(u_int ndims, u_int ndisks);

  // load data according to command file, and write log into logfs
  bool loadData(T2_TxtFile& cmdfile, ostream& logfs);

protected:
  u_int numnodes, mynode,		// # of back-end nodes and my proc rank
	master;				// rank of master node

  // data members decided by the back-end configuration file in init()
  u_int total_num_disks;		// total number of logical disks
  T2_Array<T2DLD_Mover> movers;		// data mover, one mover per assigned
					// logical disk id

  // temporary variables used during data loading
  T2_BitVector my_data_disks;		// stores the movers involved for
					// loading the current dataset

  T2_UsrArg msgbuf;			// buffer used to send/recv msgs
					// to/from master and slave loaders
  u_int ndims;				// number of dimensions of the
					// attribute space for the current
					// dataset

  // the following are only used by the master node
  T2_BitVector data_disks;		// logical disk id's chosen to 
					// decluster dataset clusters over
  u_int num_data_disks;			// total number of data disks for 
					// the current dataset being loaded
  T2_Array<T2_VArray<u_int>*> assigned_diskids;
					// an array of numnodes elements, with
					// one T2_VArray<u_int>* per node,
					// to store the list of logical disk
					// id's assigned to each node
  T2_Array<int> disk_to_instr;		// an array of tot_num_disk_ids 
					// elements, to store for each disk id
					// the array index of the corresponding
					// instruction for that disk in
					// the instruction array;
					// if -1, means this disk id is not
					// used as a data disk for loading
					// this dataset
  T2_VArray<char*> datafile_prefix,	// an array of dataset file prefixes
		   indexfile_prefix;	// an array of index file prefixes
  T2_Array<u_int> file_prefix_i;	// an array of tot_num_disk_ids
					// elements, to store for each disk id
					// the data file prefix stored in
					// datafile_prefix[] to use, and
					// the index file prefix stored in
					// indexfile_prefix[] to use
					// (that is, data file on disk id k
					// should use name stored in
					// datafile_prefix[file_prefix_i[k]]
					// as data file prefix, and
					// indexfile_prefix[file_prefix_i[k]]
					// as index file prefix)

  // initialization either by hostname when hostname != NULL, or
  // by processor rank
  bool init(const char* hostname, T2_BEConfigFile& beconfig);

  // check if reached marker for end of a command entry
  bool check_end_marker(const char* string) const;

  // load data according to given command file
  bool master_loadData(T2_TxtFile& cmdfile, ostream& logfs);
  bool slave_loadData();

  // prepare to load a dataset
  bool master_prepareLoadOneDataset(const T2_BitVector& data_disks,
				    ostream& logfs);
  bool slave_prepareLoadOneDataset();

  // load a data sub-set
  void master_loadOneDataSubset(const u_int ndims,
				T2DLD_GetClustersFuncObj& getClustersFunc,
				T2DLD_PlacementFuncObj& placementFunc,
				const T2_VArray<char*>& tapefiles);

  // move data
  bool master_moveData(const u_int ndims,
		       T2_Array<T2DLD_MoverInstruction>& instr,
		       const T2_VArray<char*>& tapefiles);
  bool slave_moveData();

  // finish loading one dataset
  bool master_endLoadOneDataset();
  bool slave_endLoadOneDataset();

  // done loading all datasets
  bool master_allDataLoadingDone();
  bool slave_allDataLoadingDone();

private:
};


inline
T2DLD_Loader::T2DLD_Loader(u_int nprocs, u_int me, u_int m)
		: numnodes(nprocs), mynode(me), master(m),
		  total_num_disks(0), movers(0),
		  my_data_disks(), msgbuf(), ndims(0),
		  data_disks(), num_data_disks(0),
		  assigned_diskids(0), disk_to_instr(0),
		  datafile_prefix(10), indexfile_prefix(10),
		  file_prefix_i()
{
}

inline
T2DLD_Loader::~T2DLD_Loader()
{
  for (u_int i=0; i<assigned_diskids.getNumberElements(); i++)
    delete assigned_diskids[i];
  t2_Assert (datafile_prefix.getNumberElements() == 0);
  t2_Assert (indexfile_prefix.getNumberElements() == 0);
}

inline bool
T2DLD_Loader::check_end_marker(const char* string) const
{
  const char marker[] = "===";
  if (strcmp(string, marker) == 0)
      return true;
  return false;
}

inline bool
T2DLD_Loader::loadData(T2_TxtFile& cmdfile, ostream& logfs)
{
  if (master_is_me() == true)
     return master_loadData(cmdfile, logfs);
  return slave_loadData();
}



#endif	/* ! _T2DLD_LOADER_H_ */
