
/* $Id: dld-index.C,v 1.6 2000/02/23 17:10:18 chialin Exp $ */

/* dld-index.C */


#include	<string.h>
#include	<sys/uio.h>
#include	<unistd.h>
#include	<ctype.h>
#include	"t2dld_index.h"
#include	"t2dld_getclusters.h"


// max length of a line in an ascii index file
static const size_t MAXLINELEN = 512;

/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_STDIndexFile expects an index file, which has the following format:
//   <header info>
//   <info for entry 1>
//   <info for entry 2>
//	...
// where <header info> consists of the following:
//	char format: format of index file ('a' - ascii, 'b' - binary)
//	u_int ndims: number of dimensions of the bounding boxes
//	u_int nentries: number of entries (ie logical chunks) in the file
//
// The <info for entry x> field of an index file in the binary format
// consists of the following;
//	float low[ndims]: coordinates for the low point of the bounding box
//	float high[ndims]: coordinates for the high point of the bounding box
//	u_int nblocks: number of physical blocks in the logical chunk
//	<info for block 1>
//	<info for block 2>
//	  ...
//	<user data>: additional user-defined data
//
// and <info for block y> consists of the following:
//	u_int file_id: the file id of the data file listed in a
//		       corresponding name file
//	off_t offset: the offset into the data file for the physical block
//	size_t size: the number of bytes for the physical block
//
// and <user data> consists of the following:
//	size_t nbytes: number of bytes for the user-defined data
//	char usrdata[nbytes]: user-defined data
//
// The <info for entry x> field of an index in the ascii format consists
// of the same fields, with the exception that <user data> consists of a
// single line, which would appear right after the line that contains the
// information for the last physical block (ie <info for block n>).
// An empty line or a line that starts with '#' in this case indicates no 
// user data. 
// Also note that for an index file in the ascii format:
//   (1) Each line cannot contain more than MAXLINELEN-1 bytes. 
//   (2) If multiple fields appear on the same line, they must be separated
//	 by while spaces (ie blanks or tabs or newlines).
//	 In particular, note that for <header info>, the field format must
//	 be separated from the field ndims.
//   (3) Empty lines between <info for entry x>'s would be ignored. 
//   (4) Text after '#' on each line is treated as comment and will be ignored.
//
/////////////////////////////////////////////////////////////////////////////


/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_STDIndexFile methods
//
/////////////////////////////////////////////////////////////////////////////
T2DLD_STDIndexFile::T2DLD_STDIndexFile()
	: index_a('#',MAXLINELEN), index_b(),
	  index_format(0), ndims(0), nentries(0)
{
}


bool
T2DLD_STDIndexFile::open(const char* index_file)
{
  reset();

  if (index_file == NULL)
     return false;

  const u_int MAX_OPEN_TRIES = 10;
  u_int tt;

  // open index file
  for (tt=0; tt<MAX_OPEN_TRIES; tt++) {
     index_a.open(index_file);
     if (index_a)
	break;
  }
  if (!index_a) {				// open failed
     t2_err_ret(__FILE__, __LINE__, 
		"ERROR: failed to open index file %s", index_file);
     return false;
  }

  // read meta data from indexfile
  t2_Assert (index_format == 0);
  index_a >> index_format;

  if (index_format == 0) {			// read failed
     reset();
     return false;
  }
  switch (index_format) {
     case 0:					// read failed
		reset();
		t2_err_ret(__FILE__, __LINE__,
			  "ERROR: read to index file failed");
		return false;
     case 'a':					// ascii format
		index_a >> ndims >> nentries;
		break;
     case 'b':					// binary format
		index_a.close();
		for (tt=0; tt<MAX_OPEN_TRIES; tt++) {
		   index_b.open(index_file);
		   if (index_b)
		      break;
		}
		if (!index_b) {			// open failed
		   reset();
		   t2_err_ret(__FILE__, __LINE__, 
			   "ERROR: failed to open index file %s", index_file);
		   return false;
		}
		index_b.bseek(sizeof(char));
		index_b >> ndims >> nentries;
		break;
     default :					// unrecognized format
		reset();
		t2_err_ret(__FILE__, __LINE__,
		   "ERROR: unrecognized index file format %c (ascii code= %d)",
		   index_format, (int) index_format);
		return false;
  }

  return true;
}


void
T2DLD_STDIndexFile::reset()
{
  if (index_a)
     index_a.close();
  if (index_b)
     index_b.close();

  index_format = 0;
  ndims = nentries = 0;
}


bool
T2DLD_STDIndexFile::read(T2DLD_DataCluster& leaf)
{
  T2_Box& mbr = leaf.getMBR();
  mbr.setNumberDimensions(getMBRNumberDimensions());
  T2_Point& low = mbr.getLow();
  T2_Point& high = mbr.getHigh();

  T2_UsrArg& usr_data = leaf.getUsrData();
  t2_Assert (usr_data.getBufferSize() == 0);

  u_int nblocks;			// # physical blocks for leaf

  switch (index_format) {
     case 'a':				// ascii format
		if (index_a.eof() == true)
		   return false;
		{
		   u_int d, i;

		   // read coordinates for low and high points of bounding box
		   for (d=0; d<getMBRNumberDimensions(); d++) {
		   	t2_Assert (index_a.eof() == false);
			index_a >> low[d];

		   }
		   for (d=0; d<getMBRNumberDimensions(); d++) {
		   	t2_Assert (index_a.eof() == false);
			index_a >> high[d];
		   }

		   // read # of physical blocks
		   t2_Assert (index_a.eof() == false);
		   index_a >> nblocks;
		   t2_Assert (nblocks > 0);
		   leaf.setNumberDataBlocks(nblocks);

		   // read each physical block
		   for (i=0; i<nblocks; i++) {
			T2DLD_DataBlock& datablock = leaf.getDataBlock(i);
		   	t2_Assert (index_a.eof() == false);
			index_a >> datablock.getDataFileID()
			        >> datablock.getOffset()
			        >> datablock.getSize();
		   }
		   {
			// read user data
			u_int nn;
			index_a.getNextLine();
			const char *line = index_a.getLine();
			const char *s = searchString((char *) line, 0, &nn);
			if (s == NULL || s[0] == '#') {	// empty line or comment
			}
			else {				// user data exists
			   t2_Assert (nn > 1);
			   for (; nn<index_a.getLineLength(); nn++) {
			      // search for '#'
			      if (line[nn] == '#' && isspace(line[nn-1]))
				 break;
			   }
			   usr_data.allocBuffer(nn+1);
			   T2_UsrArgWriter writer(usr_data);
			   writer.write(line, nn);
			   writer << '\0';
			}
			index_a.skipLine();
		   }
		}
		break;
     case 'b':				// binary format
		if (index_b.eof() == true)
		   return false;
		{
		   u_int d, i;

		   // read coordinates for low and high points of bounding box
		   for (d=0; d<getMBRNumberDimensions(); d++) {
		   	t2_Assert (index_b.eof() == false);
			index_b >> low[d];
		   }
		   for (d=0; d<getMBRNumberDimensions(); d++) {
		   	t2_Assert (index_b.eof() == false);
			index_b >> high[d];
		   }

		   // read # of physical blocks
		   t2_Assert (index_b.eof() == false);
		   index_b >> nblocks;
		   t2_Assert (nblocks > 0);
		   leaf.setNumberDataBlocks(nblocks);

		   // read each physical block
		   for (i=0; i<nblocks; i++) {
			T2DLD_DataBlock& datablock = leaf.getDataBlock(i);
		   	t2_Assert (index_b.eof() == false);
			index_b >> datablock.getDataFileID()
			        >> datablock.getOffset()
			        >> datablock.getSize();
		   }
		   {
			// read user data
			size_t udata_len;
		   	t2_Assert (index_b.eof() == false);
			index_b >> udata_len;
			if (udata_len > 0) {		// user data exists
			   char* udata = new char[udata_len];
		   	   t2_Assert (index_b.eof() == false);
			   ssize_t s =
				index_b.bread(udata, udata_len, sizeof(char));
			   t2_Assert (s > 0);
			   usr_data.setUsrArg(udata, s);
			}
		   }
		}
		break;
     default:				// unrecognized format
		t2_err_quit(__FILE__, __LINE__,
			"ERROR: unrecoginzed index format %c (%d)",
			index_format, (int) index_format);
		return false;
  }

  return true;
}


bool
T2DLD_STDIndexFile::read(T2DLD_TapeClusterArray& c)
{
  if (! (*this)) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no index file is open for T2DLD_STDIndexFile");
     return false;
  }

  for (u_int n=0; n<getNumberEntries(); n++) {
     T2DLD_DataCluster *cluster = new T2DLD_DataCluster;
     if (cluster == NULL) {
        t2_err_ret(__FILE__, __LINE__,
		   "ERROR: cannot allocate memory for T2DLD_DataCluster %d",
		   n);
        return false;
     }

     read(*cluster);
     c << cluster;
  }
  return true;
}


ostream&
T2DLD_STDIndexFile::write(ostream& os) const
{
  os << "index fomat = " << getIndexFormat() << endl
     << "# dims for bounding boxes = " << getMBRNumberDimensions()
     << ", # entries = " << getNumberEntries() << endl;

  return os;
}


void
T2DLD_STDIndexFile::bwrite(T2_OBFile& obf) const
{
  if (!obf)
     return;

  obf << 'b'			// binary format
      << ndims			// # of dimensions for bounding box
      << nentries;		// # of entries
}



/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_STDIndexLeafNode methods
//
/////////////////////////////////////////////////////////////////////////////
//T2DLD_STDIndexLeafNode::T2DLD_STDIndexLeafNode()
//		: mbr(), tblocks(0), user_data()
//{
//}
//
//
//ostream&
//T2DLD_STDIndexLeafNode::write(ostream& os) const
//{
//  os << "low=" << mbr.getLow() << ", high=" << mbr.getHigh() << endl;
//  for (u_int i=0; i<getNumberDataBlocks(); i++) {
//     const T2DLD_TapeBlock& block = getDataBlock(i);
//     os << " <" << block.getDataFileID() << "," << block.getOffset()
//	<< "," << block.getSize() << ">";
//  }
//  os << endl;
//
//  os << "user data:[" << user_data.getStartOfData() << "]" << endl;
//}
//
//
//void
//T2DLD_STDIndexLeafNode::bwrite(T2_OBFile& obf) const
//{
//  if (!obf)
//     return;
//
//  u_int d;
//  const T2_Point& low = mbr.getLow();
//  const T2_Point& high = mbr.getHigh();
//
//  // write coordinates of low and high points for bounding box
//  for (d=0; d<mbr.getNumberDimensions(); d++)
//     obf << low[d];
//  for (d=0; d<mbr.getNumberDimensions(); d++)
//     obf << high[d];
//
//  // write # of physical blocks and each block
//  obf << getNumberDataBlocks();
//  for (u_int i=0; i<getNumberDataBlocks(); i++) {
//     const T2DLD_TapeBlock& block = getDataBlock(i);
//     obf << block.getDataFileID()
//	 << block.getOffset()
//	 << block.getSize();
//  }
//
//  // write user data
//  obf << user_data.getDataSize();
//  if (user_data.getDataSize() > 0)
//     obf.bwrite(user_data.getStartOfData(), user_data.getDataSize(),
//		sizeof(char));
//}
