
/* $Id: dld-loader.C,v 1.16 2000/09/29 16:37:58 chialin Exp $ */

/* dld-loader.C */

#include	<string.h>
#include	<strings.h>
#include	"t2dld_comm.h"
#include	"t2dld_loader.h"
#include	"t2_idrange.h"


extern int _g_mynode;

// prototypes
T2DLD_GetClustersFuncObj* t2dld_stdGetClustersFunc();
T2DLD_PlacementFuncObj* t2dld_stdPlacementFunc();

////////////////////////////////////////////////////////////////////////////
//
// a help function to increase the buffer size in a msgbuf if its smaller
// than the expected msg size
//
////////////////////////////////////////////////////////////////////////////
inline void
resetMsgBuf(T2_UsrArg& msgbuf, size_t msgsize)
{
  msgbuf.setDataSize(0);
  if (msgbuf.getBufferSize() < msgsize)
     msgbuf.allocBuffer(msgsize);
}


////////////////////////////////////////////////////////////////////////////
//
// a help function to write a bitvector into a message buffer and to
// read it back from a message buffer;
// getBitVectorMsgSize estimates the number of bytes that the flattened
// bitvector would take
//
////////////////////////////////////////////////////////////////////////////
inline size_t
getBitVectorMsgSize(const T2_BitVector& bits)
{
  return sizeof(u_int) * 2 + bits.getNumberLongs() * sizeof(u_long);
}


inline void
writeBitVector(const T2_BitVector& bits, T2_UsrArgWriter& msgbuf)
{
  t2_Assert (getBitVectorMsgSize(bits) <=
	     msgbuf.getBufferSize() - msgbuf.getNumberBytesWritten());
  const u_int nlongs = bits.getNumberLongs();
  msgbuf << bits.getNumberBits() << nlongs;
  for (u_int i=0; i<nlongs; i++)
     msgbuf << bits.getLong(i);
}


inline void
readBitVector(T2_BitVector& bits, T2_UsrArgReader& msgbuf)
{
  t2_Assert (msgbuf.getDataSizeYetToRead() > 0);

  u_int nbits, nlongs;
  msgbuf >> nbits >> nlongs;
  t2_Assert (msgbuf.getDataSizeYetToRead() > 0);

  bits.setNumberBits(nbits);
  t2_Assert (nlongs == bits.getNumberLongs());
  T2_Bits* bs = bits.getBitsArray();
  for (u_int i=0; i<nlongs; i++) {
     t2_Assert (msgbuf.getDataSizeYetToRead() >= sizeof(u_long));
     msgbuf >> bs[i].getVec();
  }
}


////////////////////////////////////////////////////////////////////////////
//
// sync_error_status
//
// This function does a global-sum of argument my_status across all
// processors, and compares the sum against the expected sum expect_sum.
// If the sum equals expect_sum, it returns true; otherwise, it returns
// false. This is mainly used for processors to proceed only if all
// processors succeeded in their earlier operations.
// Usually, the processor that encounters an error would call the
// function with my_status = 0 but without the second argument, 
// since its main goal is to inform the other processors of its error,
// and therefore does not care about the return value. 
// On the other hand, the processor that encounters no error would call
// this function with my_status = 1 and expect_sum = <number of processors>,
// and check if this function returns true (in which case no processor
// encounters any error) or false (in which case at least one processor
// encounters error).
//
////////////////////////////////////////////////////////////////////////////
bool
sync_error_status(const u_int my_status, const u_int expect_sum = 0)
{
  if (t2_be_gsum(my_status) == expect_sum)
     return true;
  return false;
}


////////////////////////////////////////////////////////////////////////////
//
// a helper function to read a name file and store the filenames into
// tapefiles[]
//
////////////////////////////////////////////////////////////////////////////
bool
readNameFile(const char* name_file, T2_VArray<char*>& tapefiles)
{
  T2_TxtFile namef('#', T2DLD_MAX_COMMAND_LINE_LEN);
  for (u_int tt=0; tt<10; tt++) {
     namef.open(name_file);
     if (namef)
        break;
  }
  if (!namef) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open name file %s", name_file);
     return false;
  }

  const char* path;
  while ((path = namef.getNextUncommentString()) != NULL) {
     char *fn = new char[strlen(path)+1];
     strcpy(fn, path);
     tapefiles << fn;
  }
  t2_Assert (namef.eof() == true);

  namef.close();
  return true;
}



////////////////////////////////////////////////////////////////////////////
//
// message tags used for communication between parallel mover managers
//
////////////////////////////////////////////////////////////////////////////
static
const int T2DLD_MYDISKS_TAG	= 6500,
	  T2DLD_COMMAND_TAG	= 6501,
	  T2DLD_PREPARE_TAG	= 6502,
	  T2DLD_MOVEDATA_TAG	= 6503,
	  T2DLD_BEENTRIES_TAG	= 6504;


////////////////////////////////////////////////////////////////////////////
//
// methods for the master to broadcast a command to the slaves
//
////////////////////////////////////////////////////////////////////////////
enum T2DLD_LoaderOP {
	mop_error = 0,
	mop_prepare, mop_movedata, mop_endloadonedataset, mop_alldone };


void
master_send_command(T2DLD_LoaderOP op, int numnodes, int mynode)
{
  int c;
  switch (op) {
    case mop_error:	c = 0; break;
    case mop_prepare:	c = 1; break;
    case mop_movedata:	c = 2; break;
    case mop_endloadonedataset:	c = 3; break;
    case mop_alldone:	c = 4; break;
    default:		t2_err_quit(__FILE__, __LINE__,
				"ERROR: unrecognized op %d", (int) op);
			return;
  }

  t2_say5(cout, "master sending op = " << c << endl);

  for (int p=0; p<numnodes; p++) {
    if (p == mynode)
	continue;
    bool r = t2_be_send(c, p, T2DLD_COMMAND_TAG);
    t2_Assert (r == true);
  }
}


T2DLD_LoaderOP
slave_recv_command(int src)
{
  int c,
      source = src,
      tag = T2DLD_COMMAND_TAG;
  bool r = t2_be_recv(c, source, tag);
  t2_Assert (r == true);
  t2_Assert (source == src);
  t2_Assert (tag == T2DLD_COMMAND_TAG);

  T2DLD_LoaderOP op;
  switch (c) {
    case 0:		op = mop_error; break;
    case 1:		op = mop_prepare; break;
    case 2:		op = mop_movedata; break;
    case 3:		op = mop_endloadonedataset; break;
    case 4:		op = mop_alldone; break;
    default:		t2_err_quit(__FILE__, __LINE__,
				"ERROR: unrecognized op %d", c);
			return mop_error;
  }

  t2_say5(cout, "slave recving op = " << c << endl);

  return op;
}


inline void
master_inform_slave_error(const u_int numnodes, const u_int mynode)
{
  master_send_command(mop_error, (int) numnodes, (int) mynode);
}


////////////////////////////////////////////////////////////////////////////
//
// communication methods between master and slave loaders
//
////////////////////////////////////////////////////////////////////////////
bool
collect_mydisks(T2_UsrArg& msgbuf, const u_int ndisks, const u_int numnodes,
		const u_int mynode, const T2_BitVector& mydisks,
		T2_Array<T2_VArray<u_int>*>& assigned_diskids)
{
  // this function collects the bit vectors mydisks from all processors
  // and use them to initialize assigned_diskids[], an array with numnodes 
  // elements;
  // after initialized, assigned_diskids[p] points to an var-sized array
  // which stores the logical disk id's assigned to processor p;
  u_int p;
  T2_BitVector all_disks(ndisks);	// remember all assigned disk id's
  all_disks.clrAll();
  T2_Array<u_int> all_disk_owner(ndisks);	// remember owners of disk id's

  // initialize assigned_diskids[] with bitvector mydisks
  for (p=0; p<assigned_diskids.getNumberElements(); p++)
     delete assigned_diskids[p];
  assigned_diskids.clear(numnodes);
  for (p=0; p<numnodes; p++)
     assigned_diskids[p] = new T2_VArray<u_int>(5);
  t2_Assert (mynode < numnodes);
  for (u_int d=0; d<ndisks; d++) {
     if (mydisks.isSet(d) == true) {
	*(assigned_diskids[mynode]) << d;
	all_disks.set(d);
	all_disk_owner[d] = mynode;
     }
  }

  // collect remote_mydisks from other nodes
  T2_BitVector remote_mydisks(ndisks);
  for (p=1; p<numnodes; p++) {		// recv numnodes-1 msgs
     int source = T2_AnySource,
	 msgtag = T2DLD_MYDISKS_TAG;
     size_t msgsize = 0;
     if (t2_be_probe(source, msgtag, msgsize) == false) {
	t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_probe fails");
	return false;
     }
     t2_Assert (source >= 0);
     t2_Assert ((u_int)source < numnodes);

     resetMsgBuf(msgbuf, msgsize);
     if (t2_be_recv(msgbuf.getStartOfData(), msgbuf.getBufferSize(), source,
		    msgtag, msgsize) == false)
     {
	t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_recv fails");
	return false;
     }
     msgbuf.setDataSize(msgsize);

     // read bit vector from msgbuf
     T2_UsrArgReader reader(msgbuf);
     readBitVector(remote_mydisks, reader);

     t2_Assert (assigned_diskids[source] != NULL);
     T2_VArray<u_int>& dlist = *(assigned_diskids[source]);

     // scan through remote_mydisks to set array assigned_diskids[]
     for (u_int dd=0; dd<ndisks; dd++) {
	if (remote_mydisks.isClr(dd) == true)
	   continue;
	if (all_disks.isSet(dd) == true) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: logical disk id %d was assigned to both node %d and node %d",
		dd, all_disk_owner[dd], source);
	   return false;
	}
	else {
	   dlist << dd;
	   all_disks.set(dd);
	   all_disk_owner[dd] = source;
	}
     }			// for (u_int dd=0; dd<ndisks; dd++)
  }			// for (p=1; p<numnodes; p++)

  return true;
}


bool
send_mydisks(T2_UsrArg& msgbuf, const T2_BitVector& mydisks, u_int dest)
{
  // write the bit vector, mydisks, into msgbuf
  resetMsgBuf(msgbuf, getBitVectorMsgSize(mydisks));
  T2_UsrArgWriter writer(msgbuf);
  writeBitVector(mydisks, writer);
  writer.close();

  // send msgbuf to dest
  return t2_be_send(msgbuf.getStartOfData(), msgbuf.getDataSize(), dest,
		    T2DLD_MYDISKS_TAG);
}


bool
collect_be_entries(T2_UsrArg& msgbuf, const u_int numnodes, const u_int mynode,
		   const T2_BitVector& be_entries, T2_BEConfigFile& beconfig)
{
  u_int p;
  const u_int num_entries = be_entries.getNumberBits();
  const int not_assigned = -1;

  // return value
  bool retv = true;

  // record the entries assigned to processor so far, 
  // initialized with be_entries
  T2_Array<int> entries_assigned_so_far(num_entries);
  for (p=0; p<num_entries; p++) {
     if (be_entries.isSet(p) == true)
	entries_assigned_so_far[p] = mynode;
     else
	entries_assigned_so_far[p] = not_assigned;
  }

  // collect remote_be_entries from other nodes
  T2_BitVector remote_be_entries(num_entries);
  for (p=1; p<numnodes; p++) {		// recv numnodes-1 msgs
     int source = T2_AnySource,
	 msgtag = T2DLD_BEENTRIES_TAG;
     size_t msgsize = 0;
     if (t2_be_probe(source, msgtag, msgsize) == false) {
	t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_probe fails");
	return false;
     }
     t2_Assert (source >= 0);
     t2_Assert ((u_int)source < numnodes);

     resetMsgBuf(msgbuf, msgsize);
     if (t2_be_recv(msgbuf.getStartOfData(), msgbuf.getBufferSize(), source,
		    msgtag, msgsize) == false)
     {
	t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_recv fails");
	return false;
     }
     msgbuf.setDataSize(msgsize);

     // read bit vector from msgbuf
     T2_UsrArgReader reader(msgbuf);
     readBitVector(remote_be_entries, reader);

     if (remote_be_entries.allClr() == true) {
	 t2_err_ret(NULL, 0,
		"\nWARNING: processor with MPI rank %d has NOT been assigned any entry in the back-end configuration file, and therefore it will not move any data.",
		source);
     }
     else {
       for (u_int b=0; b<num_entries; b++) {
	 if (remote_be_entries.isSet(b) == false)
	   continue;
	 if (entries_assigned_so_far[b] == not_assigned) {
	   entries_assigned_so_far[b] = source;
	   continue;
	 }

	 // hmm, this entry was assigned to other processor before
	 const T2_BEConfigEntry* e = beconfig[b];
	 t2_Assert (e != NULL);
	 t2_err_ret(NULL, 0,
		"ERROR: the entry for host '%s' in back-end configuration file is repeatedly assigned to processors with MPI rank %d and %d; each entry can only be assigned to one single processor.",
		e->getHostName(), entries_assigned_so_far[b], source);
	 retv = false;
       }		// for (u_int b=0; b<num_entries; b++)
     }
  }			// for (p=1; p<numnodes; p++)

  // check if no entry is assigned to any processor
  bool all_assigned = true,
       some_assigned = false;
  for (p=0; p<num_entries; p++) {
      if (entries_assigned_so_far[p] == not_assigned) {
	 const T2_BEConfigEntry* e = beconfig[p];
	 if (e == NULL)	// no entry !!
	    continue;

	 t2_err_ret(NULL, 0,
		"\nWARNING: the entry for host '%s' in back-end configuration file is not assigned to any processor, and therefore no data will be moved to the disks of that entry.",
		e->getHostName());
	 all_assigned = false;
      }
      else
	  some_assigned = true;
  }
  if (some_assigned == false) {	// no processor has any work to do !!
     t2_err_ret(NULL, 0,
		"\nWARNING: no processor is assigned any entry in the back-end configuration file at all, so the execution of this program would only generate a log file without actually moving any data at all.\nTo assign entries in the back-end configuration file to processors, make sure host names listed in the back-end configuration file match names of at least some of the processors running, or use the optional argument 'r' when starting the loader program.\n");
  }

  return retv;
}


bool
send_be_entries(T2_UsrArg& msgbuf, const T2_BitVector& be_entries, u_int dest)
{
  // write the bit vector, be_entries, into msgbuf
  resetMsgBuf(msgbuf, getBitVectorMsgSize(be_entries));
  T2_UsrArgWriter writer(msgbuf);
  writeBitVector(be_entries, writer);
  writer.close();

  // send msgbuf to dest
  return t2_be_send(msgbuf.getStartOfData(), msgbuf.getDataSize(), dest,
		    T2DLD_BEENTRIES_TAG);
}


bool
send_prepareLoadOneDataset(T2_UsrArg& msgbuf, u_int dest,
			   u_int ndims, const T2_BitVector& data_disks,
			   const T2_Array<char*>& datafilename,
			   const T2_Array<char*>& indexfilename)
{
  const u_int ndisks = datafilename.getNumberElements();
  t2_Assert (ndisks == indexfilename.getNumberElements());

  // compute msg size
  size_t msgbuf_size = sizeof(ndims) + getBitVectorMsgSize(data_disks);
# ifdef T2_ASSERTION
     msgbuf_size += sizeof(ndisks);
# endif
  for (u_int i=0; i<ndisks; i++)
    if (datafilename[i] != NULL) {
	t2_Assert (indexfilename[i] != NULL);
	msgbuf_size += strlen(datafilename[i]) + 1
		    + strlen(indexfilename[i]) + 1;
    }
    else {
	t2_Assert (indexfilename[i] == NULL);
	msgbuf_size += 1 + 1;
    }

  // allocate buffer, if necessary
  resetMsgBuf(msgbuf, msgbuf_size);
  {
     T2_UsrArgWriter writer(msgbuf);
     writer << ndims;
     writeBitVector(data_disks, writer);

#    ifdef T2_ASSERTION
	writer << ndisks;
#    endif

     for (u_int n=0; n<datafilename.getNumberElements(); n++) {
	writer.writeString(datafilename[n]);
	writer.writeString(indexfilename[n]);
     }
  }

  if (t2_be_send(msgbuf.getStartOfData(), msgbuf.getDataSize(), dest,
		 T2DLD_PREPARE_TAG) == false)
  {
     t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_send to %d fails", dest);
     return false;
  }

  return true;
}


bool
recv_prepareLoadOneDataset(T2_UsrArg& msgbuf, int src,
			   u_int& ndims, T2_BitVector& data_disks,
			   T2_Array<const char*>& datafilename,
			   T2_Array<const char*>& indexfilename)
{
  int source = src,
      msgtag = T2DLD_PREPARE_TAG;
  size_t msgsize = 0;
  if (t2_be_probe(source, msgtag, msgsize) == false) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_probe fails");
     return false;
  }
  t2_Assert (msgtag == T2DLD_PREPARE_TAG);

  resetMsgBuf(msgbuf, msgsize);
  if (t2_be_recv(msgbuf.getStartOfData(), msgbuf.getBufferSize(), source,
		 msgtag, msgsize) == false)
  {
     t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_recv fails");
     return false;
  }
  msgbuf.setDataSize(msgsize);

  T2_UsrArgReader reader(msgbuf);

  u_int ndisks;

  reader >> ndims;
  t2_Assert (reader.getDataSizeYetToRead() > 0);
  readBitVector(data_disks, reader);
  t2_Assert (reader.getDataSizeYetToRead() > 0);

# ifdef T2_ASSERTION
     reader >> ndisks;
     t2_Assert (ndisks == datafilename.getNumberElements());
# endif

  // read data file prefixes and index file prefixes;
  // no copy, just set the pointers !!
  t2_Assert (datafilename.getNumberElements() 
	     == indexfilename.getNumberElements());
  for (u_int i=0; i<datafilename.getNumberElements(); i++) {
     t2_Assert (reader.getDataSizeYetToRead() > 0);
     datafilename[i] = reader.readString();
     indexfilename[i] = reader.readString();
  }

  return true;
}


bool
send_moveDataInstruction(T2_UsrArg& msgbuf, int dest, u_int ndims,
			 const T2_VArray<char*>& tapefiles,
			 const T2_VArray<u_int>& assigned_disks,
			 const T2_Array<int>& disk_to_instr,
			 const T2_Array<T2DLD_MoverInstruction>& instr)
{
  // send to dest a message that contains all filenames in tapefiles[] and
  // all entries in instr[] that are found through assigned_disks[] and
  // disk_to_instr[];
  // that is, all data clusters of instr[disk_to_instr[assigned_disks[i]]]
  // will be included in the message if disk_to_instr[assigned_disks[i]] >= 0;
  // intuitively, assigned_disks[] stores all the logical disk id's assigned
  // to processor dest, and disk_to_instr[assigned_disks[i]], if >= 0,
  // stores the array index of the corresponding instruction in instr[];
  // disk_to_instr[assigned_disks[i]] < 0 means that this logical disk 
  // is not used for the current dataset, in which case 0 clusters will
  // be sent to dest

  u_int i;

  size_t msgsize = 0;

  // scan through tapefiles[] and disk_to_instr[] to find the message size
  msgsize += sizeof(u_int);		// for # of tapefiles
  for (i=0; i<tapefiles.getNumberElements(); i++) {
     t2_Assert (tapefiles[i] != NULL);
     msgsize += strlen(tapefiles[i]) + 1;
  }
  msgsize += sizeof(u_int);		// for # of disks
  for (i=0; i<assigned_disks.getNumberElements(); i++) {
     msgsize += sizeof(u_int) +		// for logical disk id
		sizeof(u_int);		// for # data clusters in
					// disk_to_instr[assigned_disks[i]]
     t2_Assert (assigned_disks[i] < disk_to_instr.getNumberElements());
     int di_i = disk_to_instr[assigned_disks[i]];
     if (di_i < 0)			// disk not used for current dataset
	continue;
     t2_Assert ((u_int) di_i < instr.getNumberElements());
     const T2DLD_MoverInstruction& instruction = instr[di_i];
     for (u_int k=0; k<instruction.getNumberDataClusters(); k++) {
	t2_Assert (instruction[k] != NULL);
	const T2DLD_DataCluster& dcluster = *(instruction[k]);
	msgsize += 2 * ndims * sizeof(T2_ShapeCoord_t)	// for mbr
		 + sizeof(u_int);			// for # data blocks
	for (u_int t=0; t<dcluster.getNumberDataBlocks(); t++) {
	   const T2DLD_DataBlock& dblock = dcluster.getDataBlock(t);
	   msgsize += sizeof(char);		// 0 for no data, 1 for has data
	   if (dblock.getData() == NULL)	// no data
	      msgsize += sizeof(u_int)		// for fid
		       + sizeof(off_t)		// for offset
		       + sizeof(size_t);	// for nbytes
	   else
	      msgsize += sizeof(size_t)		// for nbytes
		       + dblock.getSize();	// for data

	}		// for (t=0; t<dcluster.getNumberDataBlocks(); t++)
     }			// for (k=0; k<instruction.getNumberDataClusters(); k++)
  }			// for (i=0; i<assigned_disks.getNumberElements(); i++)
	
  resetMsgBuf(msgbuf, msgsize);

  T2_UsrArgWriter writer(msgbuf);

  // write all tapefiles[]
  writer << tapefiles.getNumberElements();
  for (i=0; i<tapefiles.getNumberElements(); i++) {
     t2_Assert (writer.isFull() == false);
     writer.writeString(tapefiles[i]);
  }

  // write instructions
  writer << assigned_disks.getNumberElements();
  for (i=0; i<assigned_disks.getNumberElements(); i++) {
	writer << assigned_disks[i];		// logical disk id

	int di_i = disk_to_instr[assigned_disks[i]];
	if (di_i < 0) {			// disk not used for current dataset
	   writer << (u_int) 0;
	   continue;
	}
	t2_Assert ((u_int) di_i < instr.getNumberElements());
	const T2DLD_MoverInstruction& instruction = instr[di_i];
	writer << instruction.getNumberDataClusters();

	for (u_int k=0; k<instruction.getNumberDataClusters(); k++) {
	   t2_Assert (writer.isFull() == false);
	   const T2DLD_DataCluster& dcluster = *(instruction[k]);

	   // write mbr
	   u_int dim;
	   const T2_Point& low = dcluster.getMBR().getLow();
	   const T2_Point& high = dcluster.getMBR().getHigh();
	   for (dim=0; dim<ndims; dim++)
	      writer << low[dim];
	   for (dim=0; dim<ndims; dim++)
	      writer << high[dim];

	   // write clusters
	   writer << dcluster.getNumberDataBlocks();
	   for (u_int t=0; t<dcluster.getNumberDataBlocks(); t++) {
	      t2_Assert (writer.isFull() == false);
	      const T2DLD_DataBlock& dblock = dcluster.getDataBlock(t);
	      if (dblock.getData() == NULL)	// no data
		 writer << '\0'			// flag for no data
			<< dblock.getDataFileID()
			<< dblock.getOffset()
			<< dblock.getSize();
	     else {
		 writer << '\1'			// flag for has data
			<< dblock.getSize();
		 writer.write(dblock.getData(), dblock.getSize());
	     }
	   }		// for (t<dcluster.getNumberDataBlocks();)

	}		// for (k=0; k<instruction.getNumberDataClusters(); k++)
  } 			// for (i=0; i<assigned_disks.getNumberElements(); i++)

  writer.close();

  return t2_be_send(msgbuf.getStartOfData(), msgbuf.getDataSize(), dest,
		    T2DLD_MOVEDATA_TAG);
}


bool
recv_moveDataInstruction(T2_UsrArg& msgbuf, int src, const u_int ndims,
			 const T2_Array<T2DLD_Mover>& movers,
			 const T2_BitVector& my_data_disks,
			 T2_VArray<char*>& tapefiles,
			 T2_Array<T2DLD_MoverInstruction>& instr)
{
  int source = src,
      msgtag = T2DLD_MOVEDATA_TAG;
  size_t msgsize = 0;
  if (t2_be_probe(source, msgtag, msgsize) == false) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_probe fails");
     return false;
  }
  t2_Assert (msgtag == T2DLD_MOVEDATA_TAG);

  resetMsgBuf(msgbuf, msgsize);
  if (t2_be_recv(msgbuf.getStartOfData(), msgbuf.getBufferSize(), source,
		 msgtag, msgsize) == false)
  {
     t2_err_ret(__FILE__, __LINE__, "ERROR: t2_be_recv fails");
     return false;
  }
  msgbuf.setDataSize(msgsize);

  T2_UsrArgReader reader(msgbuf);

  // read all tapefiles[];
  // note that I am not copying the filenames out of msgbuf, therefore
  // (1) msgbuf must not be over-written until tapefiles are all open
  // (2) do not delete memory in tapefiles when done
  u_int n_tapefiles;
  reader >> n_tapefiles;
  for (u_int i=0; i<n_tapefiles; i++) {
     char*& filename = tapefiles.appendElement();
     filename = (char *) reader.readString();
  }

  u_int n_disks;
  reader >> n_disks;			// number of my disks
  t2_Assert (n_disks == movers.getNumberElements());
  t2_Assert (n_disks == instr.getNumberElements());
  for (u_int d=0; d<n_disks; d++) {
     t2_Assert (reader.getDataSizeYetToRead() >= 2*sizeof(u_int));
     T2DLD_MoverInstruction& instruction = instr[d];

     u_int did, n_dclusters;
     reader >> did >> n_dclusters;
     t2_Assert (did == movers[d].getLogicalDiskID());
     if (n_dclusters == 0)		// no clusters for this disk
	continue;

     t2_Assert (my_data_disks.isSet(did) == true);
      for (u_int k=0; k<n_dclusters; k++) {
	 t2_Assert (reader.getDataSizeYetToRead() >= sizeof(u_int));
	 T2DLD_DataCluster* dcluster_p = new T2DLD_DataCluster;
	 assert (dcluster_p != NULL);
	 T2DLD_DataCluster& dcluster = *dcluster_p;

	 T2_Box& mbr = dcluster.getMBR();
	 mbr.setNumberDimensions(ndims);

	 // read mbr
	 u_int dim;
	 T2_Point& low = mbr.getLow();
	 T2_Point& high = mbr.getHigh();
	 for (dim=0; dim<ndims; dim++)
	    reader >> low[dim];
	 for (dim=0; dim<ndims; dim++)
	    reader >> high[dim];

	 t2_Assert (reader.getDataSizeYetToRead() > 0);

	 u_int n_dblocks;
	 reader >> n_dblocks;
	 dcluster.setNumberDataBlocks(n_dblocks);
	 for (u_int t=0; t<n_dblocks; t++) {
	    t2_Assert (reader.getDataSizeYetToRead() > 0);
	    T2DLD_DataBlock& dblock = dcluster.getDataBlock(t);
	    char block_data_type;
	    reader >> block_data_type;
	    switch (block_data_type) {
		case '\0':	// no data
				reader >> dblock.getDataFileID()
				       >> dblock.getOffset()
				       >> dblock.getSize();
				t2_Assert (dblock.getData() == NULL);
				break;
		case '\1': {	// has data
				size_t data_size;
				reader >> data_size;
				char* data;
				t2_Alloc (data, data_size, char);
				reader.read(data, data_size);
				dblock.setData(data, data_size);
				break;
			   }
		default:	// unrecognized type, error !!
				t2_err_quit(__FILE__, __LINE__,
					"ERROR: unrecognized type %d",
					(int) block_data_type);
				delete dcluster_p;
				return false;
	    }			// switch (block_data_type)
	 }			// for (u_int t=0; t<n_dblocks; t++)

	 // insert dcluster into instruction
	 instruction << dcluster_p;

      }				// for (u_int k=0; k<n_dclusters; k++)
  }				// for (u_int d=0; d<n_disks; d++)

  return true;
}


/////////////////////////////////////////////////////////////////////////////
//
// helper function to delete an array of strings
//
/////////////////////////////////////////////////////////////////////////////
static inline void
deleteVArrayStrings(T2_VArray<char*>& va)
{
  for (u_int i=0; i<va.getNumberElements(); i++)
      delete [] va[i];
  va.clear();
}


/////////////////////////////////////////////////////////////////////////////
//
// T2DLD_Loader methods
//
/////////////////////////////////////////////////////////////////////////////
bool
T2DLD_Loader::init(const char* hostname, T2_BEConfigFile& beconfig)
{
  // intialize loader;
  // if hostname != NULL, use hostname to assign logical disk id's to nodes;
  // if hostname == NULL, use ((logical proc id)%numnodes == mynode) to assign
  // logical processor id's to nodes, which are assigned logical disk id's
  // assigned to those logical processor id's

  movers.clear(0);

  if (!beconfig) {
     sync_error_status(0);
     return false;
  }

  // initialize data members
  total_num_disks = beconfig.getMinInvalidLogicalDiskID();
  if (master_is_me() == true) {
     data_disks.setNumberBits(total_num_disks);
     num_data_disks = 0;
     disk_to_instr.clear(total_num_disks);
     file_prefix_i.clear(total_num_disks);
  }

  // initialize variables
  u_int num_mydisks = 0;
  T2_BitVector mydisks(total_num_disks);
  mydisks.clrAll();
  T2_BitVector be_entries(beconfig.getNumberEntries());
  be_entries.clrAll();

  // look for entries assigned to the local processor
  for (u_int i=0; i<beconfig.getNumberEntries(); i++) {
     const T2_BEConfigEntry* e = beconfig[i];
     if (e == NULL)
	continue;

     if (hostname != NULL) {		// check hostname
	if (strcmp(hostname, e->getHostName()) != 0)
	   continue;
     }
     else {				// check rank
	if ((e->getLogicalProcessorID() % numnodes) != mynode)
	   continue;
     }

     // record that this entry (ie entry i) is assigned to local processor
     be_entries.set(i);

     const T2_VArray<T2_BEConfigLocalDisk>& logdid = e->getLocalDiskEntries();
     for (u_int n=0; n<logdid.getNumberElements(); n++) {
	const T2_BEConfigLocalDisk& disk = logdid[n];
	for (u_int d=0; d<disk.logdid.getNumberElements(); d++) {
	   u_int k = disk.logdid[d];
	   if (mydisks.isSet(k) == true)
		t2_err_ret(__FILE__, __LINE__,
			"WARNING: logical disk id %d is repeatedly assigned to processor %d",
			k, mynode);
	   else {
		mydisks.set(k);
		num_mydisks++;
	   }
	}	// for (u_int d=0; d<disk.logdid.getNumberElements(); d++)
     }		// for (u_int n=0; n<logdid.getNumberElements(); n++)
  }		// for (u_int i=0; i<beconfig.getNumberEntries(); i++)

  // we now need to collect mydisk from all processors to master node
  if (master_is_me() == true) {
     if (collect_mydisks(msgbuf, total_num_disks, numnodes, mynode, mydisks, 
			 assigned_diskids)
	 == false)
     {
	sync_error_status(0);
	return false;			// error
     }
     // collect be_entries to master node to alert the users which entries
     // in the backend configuration file are assigned to no processors
     if (collect_be_entries(msgbuf, numnodes, mynode, be_entries, beconfig)
	 == false)
     {
	sync_error_status(0);
	return false;			// error
     }
  }
  else {
     if (send_mydisks(msgbuf, mydisks, master) == false) {
	sync_error_status(0);
	return false;			// error
     }
     if (send_be_entries(msgbuf, be_entries, master) == false) {
	sync_error_status(0);
	return false;			// error
     }
  }

  // we need one mover per assigned logical disk id
  movers.clear(num_mydisks);
  for (u_int d=0, m=0; m<num_mydisks; d++) {
     t2_Assert (d < total_num_disks);
     if (mydisks.isSet(d) == false)
	continue;

     // mover m is responsible for logical disk id d
     if (movers[m].init(d) == false) {
	sync_error_status(0);
	return false;
     }
     m++;
  }

  // return true if all processors return true
  return sync_error_status(1, numnodes);
}


bool
T2DLD_Loader::master_loadData(T2_TxtFile& cmdfile, ostream& logfs)
{
  if (!cmdfile) {
      master_inform_slave_error(numnodes, mynode);
      return false;
  }

  // A data-loader command file is a text file that consists of a number of
  // command entries, one per dataset, where each command is in the 
  // following format:
  //	char[] datasetname		- name of dataset
  //	<disk and data file prefix assignment 1>
  //	<disk and data file prefix assignment 2>
  //	   :
  //	-1				- no more disk and prefix assignment
  //	<name-index-file-pair 1>
  //	<name-index-file-pair 2>
  //	   :
  //	===				- end of command entry
  //
  // where <disk and data file prefix assignment> has the following format:
  //	<logical disk id range>		-- one range of logical disk id's
  //	<data file prefix>		-- prefix for the dataset files
  //					   assigned to the logical disk id's
  //					   listed in the logical disk id range
  //	<index file prefix>		-- prefix for the index files
  //					   assigned to the logical disk id's
  //					   listed in the logical disk id range
  //
  // and <logical disk id range> is in one of the following formats:
  //	n			- a single logical disk id
  //	n-m			- a set of logical disk id's from n upto m
  //
  // and <name-index-file-pair> is in the following format:
  //	char[] name_file	- filename of a name file
  //	char[] index_file	- filename of an index file

  char datasetname[T2DLD_MAX_COMMAND_LINE_LEN];
  while (1) {
     // get dataset name
     const char* string = cmdfile.getNextUncommentString();
     if (string == NULL)	// eof
	break;

     // copy dataset name
     strcpy(datasetname, string);

     // initialize data members
     data_disks.clrAll();
     num_data_disks = 0;
     t2_Assert (datafile_prefix.getNumberElements() == 0);
     t2_Assert (indexfile_prefix.getNumberElements() == 0);

     // create an entry in the log file
     logfs << string << "	# <dataset name>" << endl
	  << "	# <dataset constructor function name>" << endl
	  << "	# <blob object filename>" << endl
	  << "	# <description for dataset>" << endl;

     const u_int lineno = cmdfile.getCurrentLineNumber();

     // get logical disk id's and their datafile/indexfile prefixes
     while (1) {
	// read a logical disk id range
	if ((string = cmdfile.getNextUncommentString()) == NULL) {
	    t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting a logical disk id range for dataset '%s' at line %d in the command file, but reads eof",
		datasetname, lineno);
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    return false;
	}
	u_int idrange_lineno = cmdfile.getCurrentLineNumber();
	int n, m;
	if (T2_IDRange::stringToRange((char *)string, n, m) == false) {
	    t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid logical disk id range '%s' at line %d in the command file",
		string, idrange_lineno);
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    return false;
	}

	if (n < 0 || m < 0)	// done
	    break;

	if ((u_int) m >= data_disks.getNumberBits()) {
	    // invalid logical disk id range
	    t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid logical disk id range '%s' at line %d in the command file, as logical disk id must be < %d",
		string, idrange_lineno, data_disks.getNumberBits());
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    return false;
	}
	num_data_disks += m - n + 1;
	t2_Assert (datafile_prefix.getNumberElements() ==
		   indexfile_prefix.getNumberElements());
	u_int new_prefix_i = datafile_prefix.getNumberElements();
	for (u_int did=n; did<=(u_int)m; did++) {
	   if (data_disks.isSet(did) == true) {
	      t2_err_ret(__FILE__, __LINE__,
		"ERROR: logical disk id %d, listed at line %d in the command file, has already been assigned the data file prefix '%s' earlier; each logical disk id can only be assigned one data file prefix.",
		did, idrange_lineno, datafile_prefix[file_prefix_i[did]]);
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    return false;
	   }
	   data_disks.set(did);
	   file_prefix_i[did] = new_prefix_i;
	}

	// read dataset file prefix
	if ((string = cmdfile.getNextUncommentString()) == NULL) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting a datafile prefix for logical disk id range given at line %d in the command file, but reads eof",
		idrange_lineno);
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    return false;
	}
	size_t prefix_len = strlen(string) + 1;
	char*& new_datafile_prefix = datafile_prefix.appendElement();
	t2_Alloc (new_datafile_prefix, prefix_len, char);
	memcpy(new_datafile_prefix, string, prefix_len);

	// read index file prefix
	if ((string = cmdfile.getNextUncommentString()) == NULL) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting a index file prefix for logical disk id range given at line %d in the command file, but reads eof",
		idrange_lineno);
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    return false;
	}
	prefix_len = strlen(string) + 1;
	char*& new_indexfile_prefix = indexfile_prefix.appendElement();
	t2_Alloc (new_indexfile_prefix, prefix_len, char);
	memcpy(new_indexfile_prefix, string, prefix_len);

     }			// while (1)

     if (num_data_disks == 0) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: no logical disk id was specified for dataset '%s' defined at line %d in the command file.",
		datasetname, lineno);
	master_inform_slave_error(numnodes, mynode);
	return false;
     }

     // initialize disk_to_instr
     t2_Assert (disk_to_instr.getNumberElements() == total_num_disks);
     for (u_int di=0, dc=0; di<total_num_disks; di++) {
	if (data_disks.isSet(di) == true) {
	   t2_Assert(dc < num_data_disks);
	   disk_to_instr[di] = dc++;
	}
	else
	   disk_to_instr[di] = -1;		// not use for this dataset
     }

     T2DLD_PlacementFuncObj* placementfunc = NULL;

     // read name-index file pairs
    while ((string = cmdfile.getNextUncommentString()) != NULL) {
	if (check_end_marker(string) == true)
	    break;

	char name_file[T2DLD_MAX_COMMAND_LINE_LEN],
	     index_file[T2DLD_MAX_COMMAND_LINE_LEN];

	// get filenames for name file and index file
	strcpy(name_file, string);
	if ((string = cmdfile.getNextUncommentString()) == NULL) {
	    t2_err_ret(__FILE__, __LINE__,
		"ERROR: expects filename for an index file for dataset '%s' defined at line %d in the command file, but sees eof.",
		datasetname, lineno);
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    // have each mover close dataset file and index
	    for (u_int mm=0; mm<movers.getNumberElements(); mm++)
		 movers[mm].endLoadOneDataset();
	    return false;
	}
	strcpy(index_file, string);

	// read filenames from the name file and store them into tapefiles[]
	T2_VArray<char*> tapefiles(32);
	if (readNameFile(name_file, tapefiles) == false) {	// error
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    // have each mover close dataset file and index
	    for (u_int mm=0; mm<movers.getNumberElements(); mm++)
		 movers[mm].endLoadOneDataset();
	    return false;
	}

	// create getCluster function object
	T2DLD_GetClustersFuncObj* getclustersfunc =
					getGetClustersFuncObj(index_file);
	if (getclustersfunc == NULL) {		// error
	    deleteVArrayStrings(tapefiles);
	    deleteVArrayStrings(datafile_prefix);
	    deleteVArrayStrings(indexfile_prefix);
	    master_inform_slave_error(numnodes, mynode);
	    // have each mover close dataset file and index
	    for (u_int mm=0; mm<movers.getNumberElements(); mm++)
		 movers[mm].endLoadOneDataset();
	    return false;
	}

	if (placementfunc == NULL) {	// processing 1st name-index files
	    ndims = getclustersfunc->getNumberDimensions();

	    // create placement function object
	    placementfunc = getPlacementFuncObj(ndims, num_data_disks);
	    if (placementfunc == NULL) {	// error
	        deleteVArrayStrings(tapefiles);
	        delete getclustersfunc;
	        deleteVArrayStrings(datafile_prefix);
	        deleteVArrayStrings(indexfile_prefix);
	        master_inform_slave_error(numnodes, mynode);
	        // have each mover close dataset file and index
	        for (u_int mm=0; mm<movers.getNumberElements(); mm++)
		     movers[mm].endLoadOneDataset();
	        return false;
	    }

	    // tell the mover that the loader is about to initiate
	    // the loading process of a dataset
	    if (master_prepareLoadOneDataset(data_disks, logfs) == false) {
		// error
	        deleteVArrayStrings(tapefiles);
	        delete getclustersfunc;
	        delete placementfunc;
	        deleteVArrayStrings(datafile_prefix);
	        deleteVArrayStrings(indexfile_prefix);
	        master_inform_slave_error(numnodes, mynode);
	        // have each mover close dataset file and index
	        for (u_int mm=0; mm<movers.getNumberElements(); mm++)
		     movers[mm].endLoadOneDataset();
	        return false;
	    }
	}			// if (placementfunc == NULL)
	else {
	   // make sure that number of dimensions is consistent
	   if (ndims != getclustersfunc->getNumberDimensions()) {
		t2_err_ret(__FILE__, __LINE__,
			"ERROR: attribute space for dataset '%s' was assumed to be %d, but now changes to %d.",
			datasetname, ndims,
			getclustersfunc->getNumberDimensions());
		deleteVArrayStrings(tapefiles);
		delete getclustersfunc;
		delete placementfunc;
		deleteVArrayStrings(datafile_prefix);
		deleteVArrayStrings(indexfile_prefix);
		master_inform_slave_error(numnodes, mynode);
		// have each mover close dataset file and index
		for (u_int mm=0; mm<movers.getNumberElements(); mm++)
		   movers[mm].endLoadOneDataset();
		return false;
	   }
	}			// if (placementfunc == NULL) else

	// start to load data subset
	master_loadOneDataSubset(ndims, *getclustersfunc, *placementfunc,
				 tapefiles);

	// release memory
	deleteVArrayStrings(tapefiles);
	delete getclustersfunc;

    }

    delete placementfunc;
    deleteVArrayStrings(datafile_prefix);
    deleteVArrayStrings(indexfile_prefix);

    // tell the mover that data loading for the current dataset
    // is complete
    if (master_endLoadOneDataset() == false) {		// error
	return false;
    }

    // complete rest of entry in the log file
    logfs << "===		# end of entry for dataset " 
	  << datasetname << endl << endl;		// end of record

    t2_say(cout, "Done loading dataset '" << datasetname << "'" << endl);

  }			// while (1) -- load one dataset per iteration


  master_allDataLoadingDone();			// all done

  t2_say2(cout, "master node leaving master_loadData" << endl);

  return true;
}


T2DLD_GetClustersFuncObj*
T2DLD_Loader::getGetClustersFuncObj(const char* index_file)
{
  // create the getCluster function object
  // ok, so we only have one choice for now, so I'll just hardwire it ...
  T2DLD_GetClustersFuncObj* getclustersfunc = t2dld_stdGetClustersFunc();
  if (getclustersfunc == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: t2dld_stdGetClustersFunc returns NULL");
     return NULL;
  }

  // initialize getclustersfunc
  if (getclustersfunc->init(index_file) == false) {
     t2_err_ret(__FILE__, __LINE__, 
		"ERROR: GetClustersFuncObj::init(%s) failed", index_file);
     delete getclustersfunc;
     return NULL;
  }

  t2_Assert (getclustersfunc->getNumberDimensions() != 0);
  return getclustersfunc;
}


T2DLD_PlacementFuncObj*
T2DLD_Loader::getPlacementFuncObj(u_int ndims, u_int ndisks)
{
  // create the placement function object
  // ok, so we only have one choice for now, so I'll just hardwire it ...
  T2DLD_PlacementFuncObj* placementfunc = t2dld_stdPlacementFunc();
  if (placementfunc == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: t2dld_stdPlacementFunc returns NULL");
     return NULL;
  }

  // initialize placementfunc
  if (placementfunc->init(ndims, ndisks) == false) {
     delete placementfunc;
     t2_err_ret(__FILE__, __LINE__, "ERROR: PlacementFuncObj::init failed");
     return NULL;
  }

  return placementfunc;
}


bool
T2DLD_Loader::
master_prepareLoadOneDataset(const T2_BitVector& data_disks, ostream& logfs)
{
  t2_Assert (master_is_me() == true);

  // send stuff to other processors
  master_send_command(mop_prepare, numnodes, mynode);  // send op first
  for (u_int p=0; p<numnodes; p++) {
     if (p == mynode)
	continue;

     // find out the logical disk id's assigned to node p
     t2_Assert (assigned_diskids[p] != NULL);
     T2_VArray<u_int>& assigned_diskids_p = *(assigned_diskids[p]);
     u_int num_assigned_diskids_p = assigned_diskids_p.getNumberElements();

     // find out the dataset file prefixes and index file prefixes for node p
     T2_Array<char*> datafile_prefix_p(num_assigned_diskids_p),
		     indexfile_prefix_p(num_assigned_diskids_p);
     for (u_int i=0; i<num_assigned_diskids_p; i++) {
	if (data_disks.isSet(assigned_diskids_p[i]) == true) {
	   datafile_prefix_p[i] =
		datafile_prefix[file_prefix_i[assigned_diskids_p[i]]];
	   indexfile_prefix_p[i] =
		indexfile_prefix[file_prefix_i[assigned_diskids_p[i]]];
	   t2_Assert (datafile_prefix_p[i] != NULL);
	   t2_Assert (indexfile_prefix_p[i] != NULL);
	}
	else {
	   datafile_prefix_p[i] = NULL;
	   indexfile_prefix_p[i] = NULL;
	}
     }

     if (send_prepareLoadOneDataset(msgbuf, p, ndims, data_disks,
				    datafile_prefix_p, indexfile_prefix_p)
	 == false)
       return false;				// error
  }

  my_data_disks.clrAll();

  // have movers open output files
  // in the loop below, i is the current logical disk id, m is the mover
  // that is responsible for i, n is the number of data disks seen so far;
  // assuming that there is always one data file per data disk, n also
  // gives the current dataset file id;
  // we will also write the dataset filenames to logfile in this loop,
  // and write the index filenames to logfile later
  t2_Assert (data_disks.getNumberBits() == total_num_disks);
  u_int i, m, n;
  char datafname[T2DLD_MAX_COMMAND_LINE_LEN],
       indexfname[T2DLD_MAX_COMMAND_LINE_LEN];
  for (i=0, m=0, n=0; i<data_disks.getNumberBits(); i++) {
     if (data_disks.isSet(i) == true) {
	// append logical disk id to dataset filename and index filename
	sprintf(datafname, "%s.%d", datafile_prefix[file_prefix_i[i]], i);
	sprintf(indexfname, "%s.%d", indexfile_prefix[file_prefix_i[i]], i);

	// write data filename into log file, and record its data file id as n
	logfs << "d " << datafname << " " << n << "	# data file" 
	      << endl;

	if (m < getNumberMovers() && movers[m].getLogicalDiskID() == i) {
	   // mover m is responsible for logical disk id i
	   if (movers[m].prepareLoadOneDataset(ndims, n, datafname,
					       indexfname) == false)
	      return false;		// error

	   my_data_disks.set(i);	// mover m is used to load this dataset
	   m++;
	}
	n++;
     }			// if (data_disks.isSet(i) == true)
     else {
	if (m < getNumberMovers() && movers[m].getLogicalDiskID() == i)
	   m++;
     }			// if (data_disks.isSet(i) == true) else
  }			// for (i=0,m=0,n=0; i<data_disks.getNumberBits(); i++)

  // write index filename into log file, and the data file id it supports
  logfs << "	# index name" << endl
	<< "	# index constructor name" << endl
	<< "	# index description" << endl;
  for (i=0, n=0; i<data_disks.getNumberBits(); i++) {
     if (data_disks.isSet(i) == true) {
	// append logical disk id to index filename
	sprintf(indexfname, "%s.%d", indexfile_prefix[file_prefix_i[i]], i);

	// write index filename into log file, and it supports data file id = n
	// NOTE: the current R-tree implementation uses two files per node:
	// one index file and one aux file, in that order
	logfs << "i " << indexfname << " " << n << " -1	# index file"
	      << endl
	      << "i " << indexfname << ".aux " << n << " -1	# index aux file"
	      << endl;
	n++;
     }
  }

  return true;
}


bool
T2DLD_Loader::
slave_prepareLoadOneDataset()
{
  t2_Assert (master_is_me() == false);

  // array to point to dataset/index filename prefixes;
  // note that recv_prepareLoadOneDataset() would not copy the file prefixes
  // into the following two arrays; it only sets the pointers to point into
  // msgbuf, therefore no need free memory at the end;
  // also note that msgbuf cannot be changed before these filename prefixes
  // are used
  T2_Array<const char*> datafile_prefix(movers.getNumberElements()),
		  indexfile_prefix(movers.getNumberElements());

  // recv stuff from master
  T2_BitVector data_disks;
  if (recv_prepareLoadOneDataset(msgbuf, master, ndims, data_disks,
				 datafile_prefix, indexfile_prefix) == false)
     return false;				// error;

  my_data_disks.clrAll();

  // have movers open output files
  // in the loop below, i is the current logical disk id, m is the mover
  // that is responsible for i, n is the number of data disks seen so far;
  // assuming that there is always one data file per data disk, n also
  // gives the current dataset file id
  t2_Assert (data_disks.getNumberBits() == total_num_disks);
  char datafname[T2DLD_MAX_COMMAND_LINE_LEN],
       indexfname[T2DLD_MAX_COMMAND_LINE_LEN];
  for (u_int i=0, m=0, n=0; i<data_disks.getNumberBits(); i++) {
     if (data_disks.isSet(i) == true) {
	if (m < getNumberMovers() && movers[m].getLogicalDiskID() == i) {
	   // mover m is responsible for logical disk id i

	   // append logical disk id to dataset filename and index filename
	   sprintf(datafname, "%s.%d", datafile_prefix[m], i);
	   sprintf(indexfname, "%s.%d", indexfile_prefix[m], i);

	   if (movers[m].prepareLoadOneDataset(ndims, n, datafname,
					       indexfname) == false)
	      return false;		// error

	   my_data_disks.set(i);	// mover m is used to load this dataset
	   m++;
	}
	n++;
     }			// if (data_disks.isSet(i) == true)
     else {
	if (m < getNumberMovers() && movers[m].getLogicalDiskID() == i)
	   m++;
     }			// if (data_disks.isSet(i) == true) else
  }			// for (i=0,m=0,n=0; i<data_disks.getNumberBits(); i++)

  return true;
}


void
T2DLD_Loader::
master_loadOneDataSubset(const u_int ndims,
			 T2DLD_GetClustersFuncObj& getClustersFunc,
			 T2DLD_PlacementFuncObj& placementFunc,
			 const T2_VArray<char*>& tapefiles)
{
  t2_Assert (getClustersFunc.getNumberDimensions() != 0);

  while (1) {
     T2DLD_DataClusterArray clusters;

     {
	T2DLD_TapeClusterArray tclusters(clusters);

	// get data clusters from getclusterfunc,
	getClustersFunc.getClusters(tclusters);
     }

     if (clusters.getNumberDataClusters() == 0)   // no more data clusters
	break;

     // create an array of mover instructions, one entry per data disk
     // that has a bit set in data_disks
     T2_Array<T2DLD_MoverInstruction> mover_instr(num_data_disks);

     // assign data clusters to disks
     placementFunc.placeClusters(clusters, mover_instr);

     // move data
     master_moveData(ndims, mover_instr, tapefiles);

#    if (T2_VERBOSE > 2)
        // print out the instruction list for now
        cout << endl;
        for (u_int d=0; d<num_data_disks; d++) {
	   cout << "Instruction for disk " << d << ":" << endl;
	   cout << mover_instr[d];
	   cout << "-----------" << endl;
        }
        cout << endl;
#    endif

  }		// while (1)

}


bool
T2DLD_Loader::
master_moveData(const u_int ndims, T2_Array<T2DLD_MoverInstruction>& instr,
		const T2_VArray<char*>& tapefiles)
{
  t2_Assert (master_is_me() == true);

  // distribute the instructions to the appropriate processors
  master_send_command(mop_movedata, numnodes, mynode);	// send op first
  for (u_int p=0; p<numnodes; p++) {
     if (p == mynode)
	continue;

     t2_Assert (assigned_diskids[p] != NULL);
     if (send_moveDataInstruction(msgbuf, p, ndims, tapefiles,
			         *(assigned_diskids[p]), disk_to_instr, instr)
	== false)
	return false;
  }

  // call local movers to actually move the data according to instructions
  for (u_int i=0; i<movers.getNumberElements(); i++) {
     u_int did = movers[i].getLogicalDiskID();
     t2_Assert (did < disk_to_instr.getNumberElements());
     int instr_offset = disk_to_instr[did];
     if (instr_offset < 0)	// mover not used for current dataset
	continue;
     t2_Assert ((u_int) instr_offset < instr.getNumberElements());
     if (instr[instr_offset].getNumberDataClusters() == 0)
	continue;		// no cluster for this mover
     t2_Assert (my_data_disks.isSet(did) == true);
     if (movers[i].moveData(tapefiles, instr[instr_offset]) == false)
	return false;
  }
  return true;
}


bool
T2DLD_Loader::master_endLoadOneDataset()
{
  t2_Assert (master_is_me() == true);

  master_send_command(mop_endloadonedataset, numnodes, mynode);

  // have each mover close dataset file and index
  for (u_int i=0; i<movers.getNumberElements(); i++)
     movers[i].endLoadOneDataset();

  return true;
}


bool
T2DLD_Loader::master_allDataLoadingDone()
{
  t2_Assert (master_is_me() == true);

  master_send_command(mop_alldone, numnodes, mynode);

  return true;
}


bool
T2DLD_Loader::slave_loadData()
{
  // enter an infiite loop, getting commands from the master,
  // return when allDataLoadingDone is recv'ed, or when error occurred

  while (1) {
     T2DLD_LoaderOP op = slave_recv_command(master);
     switch (op) {
	case mop_prepare:	// prepare to load a dataset
				t2_say5(cout, 
					"about to call slave_prepareLoadOneDataset" << endl);
				if (slave_prepareLoadOneDataset() == false)
				   return false;
				break;

	case mop_movedata:	// move data
				t2_say5(cout, "about to call slave_moveData" 	
						<< endl);
				slave_moveData();
				break;

	case mop_endloadonedataset:
				t2_say5(cout, "about to call slave_endLoadOneDataset" << endl);
				if (slave_endLoadOneDataset() == false)
				   return false;
				break;

	case mop_alldone:	// done
				t2_say5(cout, "about to call slave_allDataLoadingDone" << endl);
				slave_allDataLoadingDone();
				return true;

	default:		// error
				slave_endLoadOneDataset();
				return false;
     }
  }
  t2_err_quit(__FILE__, __LINE__, "ERROR: should never reach this statement");
  return false;
}


bool
T2DLD_Loader::slave_moveData()
{
  T2_VArray<char*> tapefiles(10);
  T2_Array<T2DLD_MoverInstruction> instr(getNumberMovers());

  // recv instructions from master
  if (recv_moveDataInstruction(msgbuf, master, ndims, movers, my_data_disks,
			       tapefiles, instr)
      == false)
     return false;		// error

  // call local movers to actually move the data according to instructions
  for (u_int i=0; i<movers.getNumberElements(); i++) {
     T2DLD_MoverInstruction& instruction = instr[i];
     if (instruction.getNumberDataClusters() == 0)
	continue;		// no data to move

     u_int did = movers[i].getLogicalDiskID();
     t2_Assert (my_data_disks.isSet(did) == true);
     if (movers[i].moveData(tapefiles, instr[i]) == false)
	return false;		// error
  }

  return true;
}


bool
T2DLD_Loader::slave_endLoadOneDataset()
{
  t2_Assert (master_is_me() == false);

  // have each mover close dataset file and index
  for (u_int i=0; i<movers.getNumberElements(); i++)
     movers[i].endLoadOneDataset();

  return true;
}


bool
T2DLD_Loader::slave_allDataLoadingDone()
{
  return true;
}

