
/* $Id: t2loader.C,v 1.13 2000/09/29 16:24:21 chialin Exp $ */

/* t2loader.C */


#include	<iostream.h>
#include	<fstream.h>
#include	<stdlib.h>
#include	<unistd.h>
#include	"t2_header.h"
#include	"t2dld_mover.h"
#include	"t2dld_loader.h"


//////////////////////////////////////////////////////////////
//
// prototypes
//
//////////////////////////////////////////////////////////////
enum usageError { eUNKNOWN, eBADARG };
static void printusage(const char*);
static void badusage(usageError, const char*, const char*);

#ifndef T2_SYS_SUPPORT_GETHOSTNAME
  extern "C" int gethostname(char* name, int namelen);
#endif

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN  256
#endif

bool sync_error_status(const u_int my_status, const u_int expect_sum = 0);

//////////////////////////////////////////////////////////////
//
// main program
//
//////////////////////////////////////////////////////////////
main(int argc, char** argv, char** envp)
{
  int numnodes = 1, mynode = 0;

  // initializing environment
  t2dld_setup(argc, argv, envp, numnodes, mynode);

  // variables
  const char *cmdfile = NULL,	// dataload command file
	     *beconfile = NULL,	// back-end configuration file
	     *logfile = NULL;	// dataset loader file to generate
  bool init_by_hostname = true;	// init loader by hostname (true) or by rank

  if (argc < 4) {
     if (mynode == 0)
	printusage(argv[0]);
     t2dld_finalize();
     return -1;
  }

  // deal with the command-line arguments
  int i = 1;
  beconfile = argv[i++];
  cmdfile = argv[i++];
  logfile = argv[i++];		// output file

  if (i < argc) {
     if (argv[i][0] != 'r' && argv[i][1] != 'r') {
	if (mynode == 0)
	   printusage(argv[0]);
	t2dld_finalize();
	return -1;
     }

     init_by_hostname = false;
  }

  // reporting arguments
  if (mynode == 0) {
     cout << "numnodes = " << numnodes << endl;
     cout << "command file = " << cmdfile << endl;
     cout << "back-end configuration file = " << beconfile << endl;
     cout << "dataset loader log file to create = " << logfile << endl;
     cout << endl;
  }

  // the master node
  const int masternode = 0;

  // open log file and command file
  ofstream logfs;
  T2_TxtFile cmd('#', T2DLD_MAX_COMMAND_LINE_LEN);
  if (mynode == masternode) {
     // check if output log file exists before opening output log file
     ifstream ifs(logfile, ios::in);
     if (ifs) {
        ifs.close();
        extern bool rename_bak(const char*);
        if (rename_bak(logfile) == false) {
	   t2_err_ret(__FILE__, __LINE__,
	      "ERROR: log file '%s' already exists, cannot rename it to '%s.bak'.",
	      logfile, logfile);
	   sync_error_status(0);
	   return -1;
        }
        t2_say(cout, "WARNING: log file " << logfile << " exists, renamed to "
			<< logfile << ".bak" << endl);
     }

     logfs.open(logfile, ios::out);
     if (!logfs) {
        t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open output log file '%s'", logfile);
	sync_error_status(0);
        t2dld_finalize();
        return -1;
     }

     // open command file
     cmd.open(cmdfile);
     if (!cmd) {
        t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open loader command file '%s'.", cmdfile);
        logfs.close();
	sync_error_status(0);
        t2dld_finalize();
        return -1;
     }
  }
  
  // open back-end config file
  T2_BEConfigFile beconfig(beconfile);
  if (!beconfig) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open back-end configuration file '%s'.",
		beconfile);
     logfs.close();
     cmd.close();
     sync_error_status(0);
     t2dld_finalize();
     return -1;
  }

  // initialize loader
  T2DLD_Loader loader(numnodes, mynode, masternode);
  if (init_by_hostname == true) {
     // get hostname
     char hostname[MAXHOSTNAMELEN];
     if (gethostname(hostname, MAXHOSTNAMELEN) != 0) {	// gethostname failed
	t2_err_ret(__FILE__, __LINE__, "ERROR: gethostname failed");
	logfs.close();
	cmd.close();
        beconfig.close();
	sync_error_status(0);
	t2dld_finalize();
	return -1;
     }

     if (loader.initByHostname(hostname, beconfig) == false) {	// error
	logfs.close();
	cmd.close();
        beconfig.close();
	sync_error_status(0);
        t2dld_finalize();
        return -1;
     }
  }
  else {			// if (init_by_hostname == true)
     if (loader.initByRank(beconfig) == false) {		// error
	logfs.close();
	cmd.close();
        beconfig.close();
	sync_error_status(0);
        t2dld_finalize();
        return -1;
     }
  }				// if (init_by_hostname == true) else

  // close back-end configuration file
  beconfig.close();

  // make sure that loaders on all processors were initialized properly
  if (sync_error_status(1, numnodes) == false) {	// some proc failed
     logfs.close();
     cmd.close();
     t2dld_finalize();
     return -1;
  }

  // load datasets
  loader.loadData(cmd, logfs);

  // close output logfile and command file
  logfs.close();
  cmd.close();

  // cleaning up
  t2dld_finalize();

  return 0;
}



//////////////////////////////////////////////////////////////
//
// some help functions
//
//////////////////////////////////////////////////////////////
void printusage(const char *name)
{
  cout << "Usage: " << name
    << " <back-end configuration file> <cmd file> <loader log file> [r]"
    << endl;

  cout << "Where:" << endl
      << "<back-end configuration file> specifies T2 back-end configuration,"
      << endl 
      << "<cmd file> specifies the data-loader command file," << endl
      << "<loader log file> specifies the output file to store the results of"
      << endl
      << "	the data loading process."
      << " "
      << endl
      << endl
      << "If the optional argument 'r' is given, the entries in the back-end"
      << endl
      << "configuration file will be assigned to the processors in a round-robin"
      << endl
      << "order, according to their MPI processor ranks. Otherwise, the entries"
      << endl
      << "are assigned based on the host names listed in the back-end configuration"
      << endl
      << "file."
      << endl;
}


void badusage(usageError e, const char* p, const char* a)
{
  switch (e) {
    case eUNKNOWN:
	cerr << "Unknown parameter " << a << " in command line" << endl;
	break;
    case eBADARG:
	cerr << "Missing argument for parameter " << a << " in command line"
	     << endl;
	break;
    default:
	cerr << "Some kind of error in the parameter " << a
	     << " in command line" << endl;
  }
  printusage(p);
}


