/*
  File name: mc_system.h

  Server and client classes. Each program may be a server, and there 
  may be multiple clients it is communicating. 
*/
#ifndef MC_SYSTEM_H
#define MC_SYSTEM_H

// IBM RS6000 machines.
/* #if !defined(XLC) && (defined(__AIX__) || defined(__ALF__)) */
#ifndef MC_SYS_SUPPORT_BOOL
    #ifdef MC_USE_T2_HEADER_H
	#include "t2_header.h"		/* will define type bool if necessary */
    #else
	enum bool {false = 0, true = 1};
    #endif
#endif 

#include <string.h>
#include <stdlib.h>
#include "pgme.h"
#include "mc_wrapper.h"
#include "mc_assert.H"

extern "C" int pvm_mytid();

/*
 * Class definition for client. This class holds information about 
 * the client (receiver) side.
 */
class MC_Client {
  friend class MC_Server;
  friend class MC_DataSender;
  friend class MC_DataReceiver;
  friend class MC_DataSame;
  friend class MC_Schedule;
  friend class MC_Options;
  friend class MC_OffBandComm;
private:
  char   *prog_name; // name of the client 
  pgme_s *prog;      // program of the client
  int    numNodes;   // number of nodes in the client side

  pgme_s *program()   { MC_ASSERT(prog!=NULL); return prog; };
  void   set_prog_name(char *pname) {
    if (prog_name!=NULL) delete [] prog_name;
    prog_name = new char[strlen(pname)+1];
    memcpy(prog_name,pname,strlen(pname)+1);
  };
public:
  MC_Client() {
    prog_name = NULL;
    prog      = NULL;
    numNodes  = 0;
  };

  MC_Client(char *clientName, int numOfNodes) {
    prog_name = NULL;
    set_prog_name(clientName);
    numNodes  = numOfNodes;
    prog      = NULL;
  };

  ~MC_Client() {
    exitClient();
  };

  bool isInitialized() { return ((prog!=NULL) ? true : false); };

  bool checkClient() {
    MC_ASSERT(prog_name!=NULL);
    if (CheckPgme(prog_name,numNodes)) {
      prog = WaitPgme(prog_name,numNodes);
      MC_ASSERT(prog!=NULL);
      return true;
    } else {
      return false;
    }
  };

  bool checkClient(char *clientName, int numOfNodes) {
    if (CheckPgme(clientName,numOfNodes)) {
      set_prog_name(clientName);
      prog      = WaitPgme(clientName,numOfNodes);
      numNodes  = numOfNodes;
      MC_ASSERT(prog!=NULL);
      return true;
    } else {
      return false;
    }
    return (CheckPgme(clientName,numOfNodes) ? true : false);
  };

  void waitClient() {
    if (prog==NULL) {
      MC_ASSERT(prog_name!=NULL);
      prog = WaitPgme(prog_name,numNodes);
      MC_ASSERT(prog!=NULL);
    }
  }

  void waitClient(char *clientName, int numOfNodes) {
    if (prog==NULL) {
      set_prog_name(clientName);
      prog      = WaitPgme(clientName,numOfNodes);
      numNodes  = numOfNodes;
      MC_ASSERT(prog!=NULL);
    }
  };

  void exitClient() {
    if (prog_name!=NULL) {
      delete [] prog_name;
      prog_name = NULL;
    }
    if (prog==NULL) 
      return;
    MC_FreePgme(prog);
    prog      = NULL;
    numNodes  = 0;
  };

  int numOfNodes() { return numNodes; };
};

/*
 * The server class definition. This class holds information about 
 * server (sender) side.
 */
class MC_Server {
  friend class MC_DataSender;
  friend class MC_DataSame;
  friend class MC_DataReceiver;
  friend class MC_Schedule;
  friend class MC_Options;
private:
  pgme_s  *prog;     // program of the server
  int     numNodes;  // number of nodes in the server
  int     myID;      // id of this node assigned by Meta-Chaos 

  pgme_s *program()   { return prog; };
public:
  MC_Server() { 
    prog     = NULL; 
    numNodes = 0;
    myID     = -1;
  };

  MC_Server(char *serverName, int numOfNodes) {
    initServer(serverName,numOfNodes);
  };

  MC_Server(char *serverName, int numOfNodes, int me) {
    initServer(serverName,numOfNodes,me);
  };

  ~MC_Server() {
    exitServer();
  };

  operator const void* () const { return (prog == NULL)? NULL : this; };
  static bool pvm_is_running() { return (pvm_mytid() <= 0)? false : true; };

  void synchronize(MC_Client& clnt) {
    Sync2Pgme(prog,clnt.program());
  };

  bool initServer(char *serverName, int numOfNodes) {
    if (prog != NULL)
      return true; // exitServer();
    if (pvm_is_running()) {
      prog     = InitPgme(serverName,numOfNodes);
      numNodes = numOfNodes;
      myID     = MyPosMyPgme();
    }
    return (*this)? false : true;
  };

  bool initServer(char *serverName, int numOfNodes, int me) {
    if (prog != NULL)
      return true; // exitServer();
    if (pvm_is_running()) {
      prog     = InitPgmeMe(serverName,numOfNodes,me);
      numNodes = numOfNodes;
      myID     = MyPosMyPgme();
      MC_ASSERT(myID==me);
    }
    return (*this)? false : true;
  };

  void exitServer() {
    if (prog==NULL) return;
    MC_FreePgme(prog);
    prog     = NULL;
    numNodes =  0;
    myID     = -1;
  };    

  int    numOfNodes() { return numNodes; };
  int    myId()       { return myID; };
};


#endif /* MC_SYSTEM_H */
