
/* $Id: t2_rtree.h,v 1.1 2000/03/08 02:54:30 chialin Exp $ */

/* t2_rtree.h */

#ifndef _T2_RTREE_H_
#define _T2_RTREE_H_

#include	<fstream.h>
#include	"t2_header.h"
#include	"t2_shape.h"
#include	"t2_usrarg.h"
#include	"w_rtree.h"

//////////////////////////////////////////////////////////////////////////////
//
// T2_RTree
//
// A wrapper for the W_RTree class, an R-tree implementation using the Gist
// library. This wrapper provides the appropriate interface for T2. In
// particular, it supports the use of T2 shape classes (eg T2_Point, T2_Box,
// etc.). This wrapper function also uses an aux index file to store the
// max size of a leaf node in the R-tree.
//
//////////////////////////////////////////////////////////////////////////////
class T2_RTree {
public:
  T2_RTree(): rtree(), ndims(0), coord(NULL), qrycoord(NULL),
	      buflen(0), buf(NULL), 
	      leaf_size(sizeof(u_int) + sizeof(off_t) + sizeof(size_t)),
	      max_leaf_size(0), aux_ofs() {};
  T2_RTree(const u_int dim);
  ~T2_RTree() { close(); delete [] coord; delete [] qrycoord; delete [] buf; };

  // initialization
  // note: the initialization of ndims is only effective for creating a
  // new R-tree; for appending and reading an existing R-tree, the number
  // of dimensions would come from the aux index file.
  void initNumberDimensions(const unsigned int dim);

  // This function opens the R-tree file with a mode
  // fname: name of the R-tree file
  // mode : open mode for the R-tree file
  // mode can take one of the following
  //    W_CREATE_FILE: create an R-tree file. If there is a file with
  //                 the same name, it is deleted. Insertions to R-tree
  //                 are written to this file.
  //    W_APPEND_FILE: appends to an existing R-tree file. If the file
  //                 does not exist, it is created. Insertions to R-tree
  //                 are written to this file.
  //    W_READ_FILE: opens an existing R-tree file for performing
  //                 search on the R-tree.
  //
  // returns 1 on success, 0 otherwise.
  int open(const char *fname, int mode);

  // create an empty rtree;
  // this is equivalent to open(fname, W_CREATE_FILE)
  int create(const char *fname) { return open(fname, W_CREATE_FILE); };

  // Closes the file opened by a call to open
  // returns 1 on success, 0 otherwise.
  int close();

  // Flushes the R-tree nodes created so far to the R-tree file when
  // inserting new nodes into R-tree.
  // returns 1 on success, 0 otherwise.
  int flush() { return rtree.flush(); };

  // insert into a R-tree node the following information about a data cluster:
  // its mbr, its file id, its offset into the data file, its total size in
  // bytes, and its user data, if any
  // returns 1 on success, 0 otherwise.
  int insert(const T2_Box& mbr, u_int fid, off_t offset, size_t size,
	     const T2_UsrArg& arg);

  // initialize the iterator for a search on query box, qb
  // returns 1 on success, 0 otherwise.
  int initSearch(const T2_Box& qb);

  // Get the next node that intersects the query box.
  // returns 1  if it finds an R-tree node that intersects with query window.
  //            results are returned in out, dim, data, and size.
  // returns -1 if it cannot find any more nodes that intersect with query
  //            window. The values in variables out, dim, data, and size are
  //            undefined.
  // returns 0 on error.
  int getNext(T2_Box& mbr, u_int& fid, off_t& offset, size_t& size,
	      T2_UsrArg& arg);

  // check if the index is empty; return 1 if empty, 0 otherwise
  int isEmpty() { return rtree.isEmpty(); };

  // check if the index is open
  operator const void*() const { if (rtree) return this; return NULL; };

protected:
  W_RTree rtree;

  // data members used for conversion between T2_Box and double array
  // (W_RTree expects a double array)
  u_int ndims;				// # dimensions
  double *coord;			// coordinates of size ndims

  double *qrycoord;                     // coordinates for the query

  // data members used for storing data cluster information
  size_t buflen;			// size of buf
  char* buf;				// data buffer
  const size_t leaf_size;		// size for storing mbr,fid,offset,size
  size_t max_leaf_size;			// max size for a leaf node in rtree

  ofstream aux_ofs;			// output stream for aux index file

private:
};


inline
T2_RTree::T2_RTree(const u_int dim)
		: rtree(dim), ndims(dim), buflen(0), buf(NULL),
		  leaf_size(sizeof(u_int) + sizeof(off_t) + sizeof(size_t)),
		  max_leaf_size(0), aux_ofs()
{
  if (dim == 0) {
     coord = qrycoord = NULL;
  }
  else {
     t2_Alloc (coord, dim*2, double);
     t2_Alloc (qrycoord, dim*2, double);
  }
}



//////////////////////////////////////////////////////////////////////////////
//
// T2_StaticRTree
//
// A wrapper for the W_StaticRTree class, a static R-tree implementation using
// the Gist library. This wrapper provides the appropriate interface for T2.
// In particular, it supports the use of T2 shape classes (eg T2_Point, T2_Box,
// etc.).
//
//////////////////////////////////////////////////////////////////////////////
class T2_StaticRTree: public W_StaticRTree {
public:
  T2_StaticRTree(): W_StaticRTree(), ndims(0), coord(NULL), qrycoord(NULL),
		    buflen(0), buf(NULL),
		    leaf_size(sizeof(u_int)+sizeof(off_t)+sizeof(size_t)) {};
  T2_StaticRTree(const u_int dim);
  ~T2_StaticRTree() { close(); };

  // access methods
  u_int getNumberDimensions() const { return ndims; };

  // open an r-tree file; returns 1 on success, 0 otherwise.
  int open(const char *fname);

  // attach to an file descriptor fd with aux index file descriptor aux_fd;
  // returns 1 on success, 0 otherwise.
  int attach(int fd, int aux_fd);

  // close an rtree
  void close() {  W_StaticRTree::close(); ndims = 0;
		  delete [] coord; delete [] qrycoord; delete [] buf;
		  coord = NULL; qrycoord = NULL; buf = NULL; buflen = 0;
	       };

  // initialize the iterator for a search on query box, qb
  // returns 1 on success, 0 otherwise.
  int initSearch(const T2_Box& qb);

  // Get the next node that intersects the query box.
  // returns 1  if it finds an R-tree node that intersects with query window.
  //            results are returned in out, dim, data, and size.
  // returns -1 if it cannot find any more nodes that intersect with query
  //            window. The values in variables out, dim, data, and size are
  //            undefined.
  // returns 0 on error.
  int getNext(T2_Box& mbr, u_int& fid, off_t& offset, size_t& size,
	      T2_UsrArg& arg);

  // do not use constant constructor and operator = 
  T2_StaticRTree(const T2_StaticRTree&);
  T2_StaticRTree& operator = (const T2_StaticRTree&);

protected:
  // data members used for conversion between T2_Box and double array
  // (W_RTree expects a double array)
  u_int ndims;				// # dimensions
  double *coord;			// coordinates of size ndims

  double *qrycoord;                     // query coordinates ... 

  // data members used for storing data cluster information
  size_t buflen;			// size of buf
  char* buf;				// data buffer
  const size_t leaf_size;		// size for storing fid,offset,size

  // initialize number of dimensions
  void initNumberDimensions(const unsigned int dim);

private:
};


inline
T2_StaticRTree::T2_StaticRTree(const u_int dim)
		: W_StaticRTree(dim), ndims(dim), buflen(0), buf(NULL),
		  leaf_size(sizeof(u_int)+sizeof(off_t)+sizeof(size_t))
{
  if (dim == 0) {
     coord = qrycoord = NULL;
  }
  else {
     t2_Alloc (coord, dim*2, double);
     t2_Alloc (qrycoord, dim*2, double);
  }
}

inline
T2_StaticRTree::T2_StaticRTree(const T2_StaticRTree& t)
		: W_StaticRTree(0), ndims(0), buflen(0), buf(NULL),
		  leaf_size(sizeof(u_int)+sizeof(off_t)+sizeof(size_t))
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use copy constructor for class T2_StaticRTree");
}

inline T2_StaticRTree&
T2_StaticRTree::operator = (const T2_StaticRTree& t)
{
  t2_err_quit(__FILE__, __LINE__,
	     "do not use operator = for class T2_StaticRTree");
  return *this;
}



#endif	/* ! _T2_RTREE_H_ */
