
/* $Id: w_rtree.h,v 1.7 1999/06/30 19:54:43 chialin Exp $ */

/* w_rtree.h.h */

#ifndef _W_RTREE_H_
#define _W_RTREE_H_


#include	<sys/types.h>
#include	<sys/stat.h>
#include	<unistd.h>
#include	"gist_file.h"
#include	"gist.h"
#include	"gist_rtree.h"


// forward declaration
class W_StaticRTree;

////////////////////////////////////////////////////////////////////////////
//
// W_GistM
//
// Just a wrapper for the class gist_m from the Gist library with additional
// methods.
// 
////////////////////////////////////////////////////////////////////////////
class W_GistM: public gist_m {
public:
  W_GistM(): gist_m() {};
  ~W_GistM() {};

  // attach to an open file descriptor;
  // if updates are made to gist_m, then the file descriptor is supposed to
  // be open with O_RDWR;
  // if only lookup through the fetch methods are needed from gist_m, then
  // the file descriptor can be open with O_RDONLY;
  // return RCOK if succeeds, eFILEERROR otherwise
  rc_t attach(int fd, gist_ext_t* extension);

protected:
private:
};



////////////////////////////////////////////////////////////////////////////
//
// W_RTree
//
// This is a wrapper for the R-tree implementation from the Gist library.
//
////////////////////////////////////////////////////////////////////////////
#define  W_CREATE_FILE   100
#define  W_APPEND_FILE   200
#define  W_READ_FILE     300

class W_RTree {
friend class W_StaticRTree;
public:
  W_RTree();
  W_RTree(const unsigned int dim);
  ~W_RTree();

  // query methods
  int getOpenMode() const { return open_mode; };
  int getOpenFlag() const { return opened; };
  unsigned int getNumberDimensions() const { return dimension; };

  // This function sets the dimensions of the points forming 
  // the bounding rectangle 
  int init(const unsigned int dim) { dimension = dim; return 1; };

  // This function opens the R-tree file with a mode
  // fname: name of the R-tree file 
  // mode : open mode for the R-tree file 
  // mode can take one of the following 
  //    W_CREATE_FILE: create an R-tree file. If there is a file with 
  //                 the same name, it is deleted. Insertions to R-tree
  //                 are written to this file. 
  //    W_APPEND_FILE: appends to an existing R-tree file. If the file 
  //                 does not exist, it is created. Insertions to R-tree 
  //                 are written to this file. 
  //    W_READ_FILE: opens an existing R-tree file for performing 
  //                 search on the R-tree. 
  // 
  // returns 1 on success, 0 otherwise.
  int open(const char *fname, int mode);

  // create an empty rtree;
  // this is equivalent to open(fname, W_CREATE_FILE)
  int create(const char *fname) { return open(fname, W_CREATE_FILE); };

  // Closes the file opened by a call to open
  // returns 1 on success, 0 otherwise.
  int close();

  // Flushes the R-tree nodes created so far to the R-tree file when 
  // inserting new nodes into R-tree. 
  // returns 1 on success, 0 otherwise.
  int flush();

  // Insert a new node into the R-tree.
  // inp  : R-tree node. It represents a n-dimensional rectangle. Its size 
  //        should be 2*n. 
  //    If a node is a n-dimensional rectangle, then the format
  //    of the node is as follows
  //    (inp[0],inp[1]) -- minimum and maximum values of first dimension 
  //                       of the two points at the corners of the rectangle
  //    (inp[2],inp[3]) -- minimum and maximum values of second dimension 
  //                       of the two points at the corners of the rectangle
  //    (inp[2*n-2],inp[2*n-1]) -- minimum and maximum values of nth dimension 
  //                       of the two points at the corners of the rectangle
  // data : a pointer to the user defined data that we may want to store 
  //        with the R-tree node.
  // size : size (in bytes) of the user defined data 
  //
  // returns 1 on success, 0 otherwise.
  int insert(const double inp[], void *data, size_t size);

  // Initialize the iterator for a search on query window, qw.
  // qw: the query window. It has the same size of R-tree node. That is,
  //     for an n-dimensional R-tree, its size is 2*n. The format of the 
  //     query is the same as an R-tree node. 
  //
  // returns 1 on success, 0 otherwise.
  int initSearch(const double qw[]);

  // Get the next node that intersects the query window.
  // out  : the R-tree node that intersects with the query window.
  //        Its size should be at the size of R-tree node. 
  // dim  : The dimension of the R-tree node is returned in this variable. 
  // data : the user defined data stored with the R-tree node.
  // size : size of the user defined data 
  // 
  // returns 1  if it finds an R-tree node that intersects with query window.
  //            results are returned in out, dim, data, and size.
  // returns -1 if it cannot find any more nodes that intersect with query
  //            window. The values in variables out, dim, data, and size are 
  //            undefined. 
  // returns 0 on error. 
  int getNext(double out[], u_int& dim, void *data, size_t& size);

  // check if the index is empty; return 1 if empty, 0 otherwise
  int isEmpty() { return rt_index.is_empty(); };

  // check if the index is open
  operator const void*() const { if (opened) return this; return NULL; };

protected:
  W_GistM	 rt_index;
  int		 open_mode;
  gist_cursor_t  cursor;
  unsigned int   dimension;
  int            opened;
  bool           eof;
  smsize_t       datasz;
  rt_query_t     *qry;
  rt_rect        *rectQry;

private:
};



////////////////////////////////////////////////////////////////////////////
//
// W_StaticRTree
//
// This is a class that implements an R-tree that only allows for searching.
// That is, no insertion nor deletion is allowed. It uses internally the
// R-tree implementation from the gist library.
//
////////////////////////////////////////////////////////////////////////////
class W_StaticRTree {
public:
  W_StaticRTree(): rtree() {};
  W_StaticRTree(const unsigned int dim): rtree(dim) {};
  ~W_StaticRTree() {};

  // query methods
  int getOpenFlag() const { return rtree.getOpenFlag(); };
  unsigned int getNumberDimensions() const
			{ return rtree.getNumberDimensions(); };

  // initialization
  int init(const unsigned int dim) { return rtree.init(dim); };

  // open an r-tree file; returns 1 on success, 0 otherwise.
  int open(const char *fname) { return rtree.open(fname, W_READ_FILE); };

  // attach to an file descriptor; returns 1 on success, 0 otherwise.
  int attach(int fd);

  // Closes the file opened by a call to open
  // returns 1 on success, 0 otherwise.
  int close() { return rtree.close(); };

  // Initialize the iterator for a search on query window, qw.
  // qw: the query window. It has the same size of R-tree node. That is,
  //     for an n-dimensional R-tree, its size is 2*n. The format of the 
  //     query is the same as an R-tree node. 
  // returns 1 on success, 0 otherwise.
  int initSearch(const double qw[]) { return rtree.initSearch(qw); };

  // Get the next node that intersects the query window.
  // out  : the R-tree node that intersects with the query window.
  //        Its size should be at the size of R-tree node. 
  // dim  : The dimension of the R-tree node is returned in this variable. 
  // data : the user defined data stored with the R-tree node.
  // size : size of the user defined data 
  // 
  // returns 1  if it finds an R-tree node that intersects with query window.
  //            results are returned in out, dim, data, and size.
  // returns -1 if it cannot find any more nodes that intersect with query
  //            window. The values in variables out, dim, data, and size are 
  //            undefined. 
  // returns 0 on error. 
  int getNext(double out[], u_int& dim, void *data, size_t& size)
			{ return rtree.getNext(out, dim, data, size); };

  // check if the index is empty; return 1 if empty, 0 otherwise
  int isEmpty() { return rtree.isEmpty(); };

protected:
  W_RTree rtree;

private:
};



#endif	/* ! _W_RTREE_H_ */
