
/* $Id: t2_dscat.h,v 1.4 2000/05/09 15:21:19 chialin Exp $ */

/* t2_dscat.h */

#ifndef _T2_DSCAT_H_
#define _T2_DSCAT_H_


#include	"t2_header.h"
#include	"t2_template.h"
#include	"t2_txtfile.h"
#include	"t2_bits.h"
#include	"t2_beconfig.h"
#include	"t2_dsreg.h"


/////////////////////////////////////////////////////////////////////////
//
// T2_DSCatalogHeader
//
/////////////////////////////////////////////////////////////////////////
class T2_DSCatalogHeader {
public:
  // write the header information from beconfig to file
  static bool writeHeader(fstream& file, T2_BEConfigFile& beconfig);

  // read the header information from the header and set the bits in argument
  // diskmap for the logical disk ids assigned to the local processor;
  // assignment can be done by matching the hostname in the header (use
  // argument hostname that is non-null), by matching the logical processor
  // ids (use argument my_log_pids that is non-null), or simply ignore
  // the assignment given in the header (set arugment hostname and my_log_pids
  // both to null)
  static bool readHeader(T2_TxtFile& file, const char* hostname,
			 const T2_BitVector* my_log_pids,
			 u_int& num_pids, T2_BitVector& diskmap);

  // just peek into a dataset catalog file and get the number of logical
  // processor ids and number of logical disk ids;
  // note that this method would NOT reset the file pointer
  static bool readHeaderNumberIDs(T2_TxtFile& file,
				  u_int& nprocids, u_int& ndiskids);

protected:
private:
};



/////////////////////////////////////////////////////////////////////////
//
// T2_DSCatalogWriter
//
/////////////////////////////////////////////////////////////////////////
class T2_DSCatalogWriter {
public:
  T2_DSCatalogWriter();
  ~T2_DSCatalogWriter() { close(); };

  // open and close file
  void open(const char* fname, T2_BEConfigFile& beconfig);
  void close();

  // query methods
  operator const void*() const { return (valid == true)? this:NULL; };
  u_int getNumberDatasets() const { return ndatasets; };

  // add a dataset entry to the file, and assign id's to the dataset and
  // its associated indexes;
  // returns true if succeed, and set dsid to the assigned dataset id,
  // store in idx_ids the assigned index id's
  // if error, returns NULL and set eflag to 1
  bool addEntry(T2_DSRegistrationEntry& e, T2_DSID dsid, u_int cons_id);

protected:
  const u_int max_ndigits_ndatasets;	// max # digits for max dataset id
  fstream file;
  off_t datasets_start;			// start position for dataset info
  u_int ndatasets,			// # datasets added so far
	numnodes,			// total number of back-end nodes
	numdisks;			// total number of back-end disks
					// (numdisks is not really used now)
  bool valid;				// is a file open? true = yes

private:
};


inline
T2_DSCatalogWriter::T2_DSCatalogWriter()
			: max_ndigits_ndatasets(20), file(), datasets_start(0),
			  ndatasets(0), numnodes(0), numdisks(0), valid(false)
{
}



////////////////////////////////////////////////////////////////////////
//
// T2_DSCatEntry
//
// This is the class for an entry in a dataset catalog.
// Every T2 dataset should contain an entry in the dataset catalog.
// An entry consists of the following information.
//   . dataset id
//   . number of local data files
//   . <filename, disk id number> for each local data file
//   . number of meta data files
//   . a list filenames for the meta data files
// At the beginning of the file, there will be a number indicating
// the number of datasets stored in this catalog.
//
////////////////////////////////////////////////////////////////////////
class T2_DSCatEntry {
public:
  T2_DSCatEntry(): id(0), constructor_id(0), fnames(0), diskno(0), mfnames(0){};
  ~T2_DSCatEntry() { clear(); };

  // read from an T2_TxtFile;
  // if offset is not null, the offset of the start of this entry is stored
  // in *offset
  bool read(T2_TxtFile& ifs, off_t* offset = NULL);

  // access methods
  T2_DSID getDatasetID() const { return id; };
  u_int getConstructorID() const { return constructor_id; };
  u_int getNumberDataFiles() const { return fnames.getNumberElements(); };
  u_int getNumberMetaFiles() const { return mfnames.getNumberElements(); };
  const char* getDataFileName(u_int i) const;
  u_int getDataFileDiskNo(u_int i) const { return diskno[i]; };
  const char* getMetaFileName(u_int i) const;

  // clear up the entry
  void clear();

  // write to ostream
  ostream& write(ostream& os) const;

protected:
  T2_DSID id;
  u_int constructor_id;			// id for the dataset constructor
  T2_Array<char*> fnames;
  T2_Array<u_int> diskno;		// assigned logical disk ids
  T2_Array<char*> mfnames;

  // do not use copy constructor
  T2_DSCatEntry(const T2_DSCatEntry& e);
  T2_DSCatEntry& operator =(const T2_DSCatEntry& e);

private:
};


inline const char*
T2_DSCatEntry::getDataFileName(u_int i) const
{
  if (i >= getNumberDataFiles())
     return NULL;
  return fnames[i];
}

inline const char*
T2_DSCatEntry::getMetaFileName(u_int i) const
{
  if (i >= getNumberMetaFiles())
     return NULL;
  return mfnames[i];
}

inline
T2_DSCatEntry::T2_DSCatEntry(const T2_DSCatEntry& e)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_DSCatEntry copy constructor");
}

inline T2_DSCatEntry&
T2_DSCatEntry::operator =(const T2_DSCatEntry& e)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_DSCatEntry operator =");
  return *this;
}

inline ostream&
operator << (ostream& os, const T2_DSCatEntry& e)
{
  return e.write(os);
}



////////////////////////////////////////////////////////////////////////
//
// T2_DSCatalog
//
// A T2 Dataset catalog file reader. Used to access and query a
// dataset catalog.
//
////////////////////////////////////////////////////////////////////////
class T2_DSCatalog {
public:
  T2_DSCatalog();
  ~T2_DSCatalog() { close(); };

  // query methods
  u_int getNumberDatasets() const;
  bool getDSCatalogEntry(T2_DSID i, T2_DSCatEntry& e, T2_BitVector& dbits);
  const T2_BitVector& getDiskBitMap() const { return mydisks; };

  // 3 different methods to open a catalog file; these methods read the 
  // catalog header and check logical disk id assignment if
  // necessary, and find the offset for every dataset id:
  // openHostName() follows the instructions in the header and assign to local
  // procesor all logical disk ids assigned to the given hostname in the header
  void openHostName(const char* fname, const char* hostname);
  // openProcID() follows the instructions in the header and assign to local
  // procesor all logical disk ids assigned to the given logical processor id
  // in the header
  void openProcID(const char* fname, const T2_BitVector& my_log_pids);
  // openDiskID() totally ignores the header and assign to local
  // procesor all logical disk ids marked by the bits in the given bit vector
  void openDiskID(const char* fname, const T2_BitVector& my_log_dids);

  // close the catalog
  void close() { file.close(); offsets.clear(0); mydisks.setNumberBits(0);
		 min_invalid_log_pid = 0; };

  // check if a file is open
  operator const void*() const { return file; };

  // write to ostream for debugging purpose (not to generate a dataset catalog,
  // in which case T2_DSCatalogWriter should be used)
  ostream& write(ostream& os);

protected:
  T2_TxtFile file;
  T2_Array<off_t> offsets;
  T2_BitVector mydisks;
  u_int min_invalid_log_pid;		// min invalid logical proc id
					// (ie all valid logical proc id should
					// be less than this val)

  const off_t T2_DSCAT_NO_OFFSET;	// initial value for offsets[]

  // set the offsets for all dataset ids in the catalog
  bool find_offsets();

private:
};


inline
T2_DSCatalog::T2_DSCatalog(): file('#', 512), offsets(0), mydisks(0),
			      min_invalid_log_pid(0), T2_DSCAT_NO_OFFSET(-1)
{
}

inline u_int
T2_DSCatalog::getNumberDatasets() const
{
  if (!file)
     return 0;
  return offsets.getNumberElements();
}

inline ostream&
operator << (ostream& os, T2_DSCatalog& c)
{
  return c.write(os);
}



#endif	/* ! _T2_DSCAT_H_ */
