
/* $Id: t2_idxcat.h,v 1.4 2000/03/14 23:28:50 chialin Exp $ */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/5/98
 * Dataset catalog class
 */

#ifndef _T2_IDXCAT_H_
#define _T2_IDXCAT_H_


#include	"t2_header.h"
#include	"t2_tarray.h"
#include	"t2_txtfile.h"
#include	"t2_bits.h"
#include	"t2_dsreg.h"


////////////////////////////////////////////////////////////////////////
//
// T2_IndexCatEntry
//
// This is the class for an entry in an index catalog.
// Every T2 index should contain an entry in the index catalog.
// An entry consists of the following information.
//   . index id
//   . number of index files
//   . a list of filenames for each index file and the logical data
//     file id's that those files support
// At the beginning of the file, there will be a number indicating
// the number of index stored in this catalog.
//
////////////////////////////////////////////////////////////////////////
class T2_IndexCatEntry {
public:
  T2_IndexCatEntry(): id(0), constructor_id(0), idxfiles(0) {};
  ~T2_IndexCatEntry();

  // read from an T2_TxtFile;
  // if offset is not null, the offset of this record from the beginning
  // of the file is stored in *offset
  bool read(T2_TxtFile& ifs, off_t* offset = NULL);

  // access methods
  T2_IndexID getIndexID() const { return id; };
  u_int getConstructorID() const { return constructor_id; };
  u_int getNumberFiles() const { return idxfiles.getNumberElements(); };
  const char* getFileName(u_int i) const;

  // clear the records
  void clear();

  // find out the set of index files that support a given set of dataset
  // file id's represented by argument datafilemap; the index files
  // that support the data files are marked in the argument idxfilemap,
  // and the total number of index files marked in idxfilemap is returned
  u_int support (const T2_BitVector& datafilemap,
		 T2_BitVector& idxfilemap) const;

  // output to ostream
  ostream& write(ostream& os) const;

protected:
  T2_IndexID id;			// index id
  u_int constructor_id;			// constructor id
  T2_Array<T2_NameIDRangePair> idxfiles;

  // do not use copy constructor
  T2_IndexCatEntry(const T2_IndexCatEntry& e);
  T2_IndexCatEntry& operator =(const T2_IndexCatEntry& e);

private:
};


inline const char*
T2_IndexCatEntry::getFileName(u_int i) const
{
  if (i >= getNumberFiles())
     return NULL;
  return idxfiles[i].name;
}

inline void
T2_IndexCatEntry::clear()
{
  id = 0;
  constructor_id = 0;
  idxfiles.clear(0);
}

inline
T2_IndexCatEntry::~T2_IndexCatEntry()
{
  clear();
}

inline
T2_IndexCatEntry::T2_IndexCatEntry(const T2_IndexCatEntry& e)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_IndexCatEntry copy constructor");
}

inline T2_IndexCatEntry&
T2_IndexCatEntry::operator =(const T2_IndexCatEntry& e)
{
  t2_err_quit(__FILE__, __LINE__, "do not use T2_IndexCatEntry operator =");
  return *this;
}

inline ostream&
operator <<(ostream& os, const T2_IndexCatEntry& e)
{
  return e.write(os);
}



////////////////////////////////////////////////////////////////////////
//
// T2_IndexCatalog
//
// T2 index catalog.
//
////////////////////////////////////////////////////////////////////////
class T2_IndexCatalog {
public:
  T2_IndexCatalog(const char* fname=NULL);
  ~T2_IndexCatalog() {};

  // query methods
  u_int getNumberIndexes() const { return offsets.getNumberElements(); };
  bool getIndexCatalogEntry(u_int i, T2_IndexCatEntry& e);
  operator const void* () const { return file; };

  // open index file
  void open(const char* fname);
  void close();

  // write to ostream for debugging purpose (not to generate an index catalog,
  // in which case T2_IndexCatalogWriter should be used)
  ostream& write(ostream& os);

protected:
  T2_TxtFile file;
  T2_Array<off_t> offsets;

  const off_t T2_IDXCAT_NO_OFFSET;	// initial value for offsets[]

private:
};


inline
T2_IndexCatalog::T2_IndexCatalog(const char* fname)
			: file ('#', 2048), offsets(0),
			  T2_IDXCAT_NO_OFFSET(-1)
{
  if (fname != NULL)
     open(fname);
}

inline void
T2_IndexCatalog::close()
{
  file.close();
}

inline ostream&
operator << (ostream& os, T2_IndexCatalog& i)
{
  return i.write(os);
}



/////////////////////////////////////////////////////////////////////////
//
// T2_IndexCatalogWriter
//
/////////////////////////////////////////////////////////////////////////
class T2_IndexCatalogWriter {
public:
  T2_IndexCatalogWriter(const char* fname= NULL);
  ~T2_IndexCatalogWriter() { close(); };

  // open and close file
  void open(const char*);
  void close();

  // query methods
  operator const void*() const { return (valid == true)? this:NULL; };
  u_int getNumberIndexes() const { return nindexes; };

  // add a dataset entry to the file, and assign id's to the dataset and
  // its associated indexes;
  // returns true if succeed, and set dsid to the assigned dataset id,
  // store in idx_ids the assigned index id's
  // if error, returns NULL and set eflag to 1
  bool addEntry(T2_DSRegistrationEntry& e, T2_Array<T2_IndexID>& idx_id,
		T2_Array<u_int>& cons_id);

protected:
  const u_int max_ndigits_nindexes;	// max # digits for max index id
  fstream file;
  u_int nindexes;			// # indexes added so far
  bool valid;				// is a file open? true = yes

private:
};


inline
T2_IndexCatalogWriter::T2_IndexCatalogWriter(const char* fname)
			: max_ndigits_nindexes(20), file(),
			  nindexes(0), valid(false)
{
  if (fname != NULL)
     open(fname);
}



#endif	/* _T2_IDXCAT_H_ */

