
/* t2_map.h */

/*
 * Chialin Chang, chialin@cs.umd.edu, 5/8/98
 * A template class for an associative map.
 */


#ifndef _T2_MAP_H_
#define _T2_MAP_H_


#include	"t2_header.h"
#include	"t2_tset.h"


//////////////////////////////////////////////////////////////////////////
//
// T2_Map
//
// T2_Map is a template class with two data types, Key and Val.
// Each entry in T2_Map has a key of type Key and a value of type Val.
// T2_Map allows the retrieval of a value by specifying a key.
//
// The data type Key is expected to have the following methods.
//	Key()					// empty constructor
//	Key(const Key&)				// copy constructor
//	Key& operator =(const Key&)		// assignment operator
//	bool operator <(const Key& k) const;
//	bool operator ==(const Key& k) const;
// Furthermore, both the less-than relation and the equality relation
// must be transitive.
//
// The data type Val is expected to have the following methods.
//	Val()					// empty constructor
//	Val(const Val&)				// copy constructor
//	Val& operator =(const Val&)		// assignment operator
//
// Note that the removal of an entry in the map may change the position
// of other existing entries (ie the entries are not "pinned").
//
//////////////////////////////////////////////////////////////////////////
template <class Key, class Val>
class T2_Map {
public:
  T2_Map(u_int l = 100);
  T2_Map(const T2_Map<Key,Val>& m);
  virtual ~T2_Map() {};

  // operator =
  T2_Map<Key,Val>& operator =(const T2_Map<Key,Val>& m);

  // query methods
  u_int getNumberElements() const { return map.getNumberElements(); };
  bool isEmpty() const { return (getNumberElements() == 0)? true : false; };
  bool isFull() const
	 { return (getNumberElements() == getMaxMapSize())? true : false; };

  // access methods
  bool search(const Key& k, Val& v) const;
  Val& getValue(const Key& k, bool& found);
  Val& touch(const Key& k);

  // inserting/removing entries into/from the map
//  T2_Map<Key,Val>& insert(const Key& k, const Val& v);
  Val& insert(const Key& k, const Val& v);
  bool remove(const Key& k);
  void emptyMap();		// remove every thing

  // insert entry (k,v) if k is not found in the map
  // returns true if found, false otherwise
//  bool searchOrInsert(const Key& k, const Val& v);

  // due to the simple implementation, for now, let's allow the map grow
  // its size to a given s
  u_int getMaxMapSize() const { return map.getMaxSetSize(); };
  void growMaxMapSize(u_int newlen) { map.growMaxSetSize(newlen); };

protected:
  struct Entry {
	Key key; Val val;
//	Entry(): key(), val() {};
//	Entry(const Key& k, const Val& v): key(k), val(v) {};
	Entry& operator =(const Entry& e)
				{ key = e.key; val = e.val; return *this; };
  };

  // the current implementation simply uses a tset, instead of some tree ...
  T2_TSet<Entry> map;

  Val errorVal;		// return value when search fails to find the key

  // insert method that returns the position of the inserted entry
  u_int insertEntry(const Key& k, const Val& v);

  // returns the position of element with key k; returns -1 if not found
  int search(const Key& k) const;

private:
};


template <class Key, class Val>
inline
T2_Map<Key,Val>::T2_Map(u_int l): map(l)
{
}

template <class Key, class Val>
inline
T2_Map<Key,Val>::T2_Map(const T2_Map<Key,Val>& m):
				map(m.map.getMaxSetSize())
{
  for (u_int i=0; i<m.map.getNumberElements(); i++)
    map << m.map[i];
}

template <class Key, class Val>
inline T2_Map<Key,Val>&
T2_Map<Key,Val>::operator =(const T2_Map<Key,Val>& m)
{
  map.clear(m.map.getMaxSetSize());
  for (u_int i=0; i<m.map.getNumberElements(); i++)
    map << m.map[i];
  return *this;
}

template <class Key, class Val>
inline int
T2_Map<Key,Val>::search(const Key& k) const
{
  for (u_int i=0; i<getNumberElements(); i++) {
     if (k == map[i].key)
	return i;
  }
  return -1;
}

template <class Key, class Val>
inline bool
T2_Map<Key,Val>::search(const Key& k, Val& v) const
{
  int i = search(k);
  if (i < 0)
     return false;
  v = map[i].val;
  return true;
}
     
template <class Key, class Val>
inline Val&
T2_Map<Key,Val>::getValue(const Key& k, bool& found)
{
  int i = search(k);
  if (i < 0) {
     found = false;
     return errorVal;
  }
  found = true;
  return map[i].val;
}
     
template <class Key, class Val>
inline u_int
T2_Map<Key,Val>::insertEntry(const Key& k, const Val& v)
{
  int i = search(k);
  if (i < 0) {

     if (map.isFull()) {	// this is due to a simple implementation
        t2_err_quit(__FILE__,__LINE__,
		    "T2_Map is full, cannot insert another entry");
        return 0;
     }

     Entry e;
     e.key = k; e.val = v;
     map << e;
     return map.getNumberElements() - 1;
  }

  map[i].val = v;
  return i;
}

template <class Key, class Val>
//T2_Map<Key,Val>&
inline Val&
T2_Map<Key,Val>::insert(const Key& k, const Val& v)
{
  u_int i = insertEntry(k, v);
  return map[i].val;
//  return *this;
}

//template <class Key, class Val>
//nline bool
//T2_Map<Key,Val>::searchOrInsert(const Key& k, const Val& v)
//{
//  int i = search(k);
//  if (i < 0) {
//
//     if (map.isFull()) {	// this is due to a simple implementation
//        t2_err_quit(__FILE__,__LINE__,
//		    "T2_Map is full, cannot insert another entry");
//        return false;
//     }
//
//     Entry e;
//     e.key = k; e.val = v;
//     map << e;
//     return false;
//  }
//  return true;
//}

template <class Key, class Val>
inline Val&
T2_Map<Key,Val>::touch(const Key& k)
{
  int i = search(k);
  if (i < 0) {
     u_int j = insertEntry(k, Val());
     return map[j].val;
  }
  return map[i].val;
}

template <class Key, class Val>
inline bool
T2_Map<Key,Val>::remove(const Key& k)
{
  int i = search(k);
  if (i < 0)			// cannot find key
     return false;		// remove fails
  map.out(i);
  return true;			// remove succeeds
}

template <class Key, class Val>
inline void
T2_Map<Key,Val>::emptyMap()
{
  map.emptySet();
}



//////////////////////////////////////////////////////////////////////////
//
// T2_PMap
//
// T2_PMap is a map that also allows its elements being accessed by their
// positions (ie array index). This allows one to, say, iterator through
// all entries in the map.
//
//////////////////////////////////////////////////////////////////////////
template <class Key, class Val>
class T2_PMap: public T2_Map<Key,Val> {
public:
  T2_PMap(u_int l = 100): T2_Map<Key,Val>(l) {};
  ~T2_PMap() {};

  // access methods
  int getPosition(const Key& k) const { return search(k); };
  const Key& getKeyByPosition(u_int i) const { return map[i].key; };
  const Val& getValueByPosition(u_int i) const { return map[i].val; };
  Key& getKeyByPosition(u_int i) { return map[i].key; }; 
  Val& getValueByPosition(u_int i) { return map[i].val; };

  // insert an entry
  T2_PMap<Key,Val>& insert(const Key& k, const Val& v, u_int& i);
  T2_PMap<Key,Val>& insert(const Key& k, const Val& v);

protected:
private:
};


template <class Key, class Val>
inline T2_PMap<Key,Val>&
T2_PMap<Key,Val>::insert(const Key& k, const Val& v, u_int& i)
{
  i = T2_Map<Key,Val>::insertEntry(k, v);
  return *this;
}

template <class Key, class Val>
inline T2_PMap<Key,Val>&
T2_PMap<Key,Val>::insert(const Key& k, const Val& v)
{
  T2_Map<Key,Val>::insertEntry(k, v);
  return *this;
}



#endif	/* _T2_MAP_H_ */
