
/* $Id: beconfig.C,v 1.2 2000/02/04 00:15:04 chialin Exp $ */

/* beconfig.C */


#include	<stdlib.h>
#include	"t2_beconfig.h"
#include	"t2_idrange.h"


///////////////////////////////////////////////////////////////////////////
//
// T2_BEConfigLocalDisk methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_BEConfigLocalDisk::read(T2_TxtFile& file)
{
  clear();
  if (!file)
     return false;

  t2_Assert (file.getFileName() != NULL);

  const char* string = file.getNextUncommentString();
  if (string == NULL || T2_BEConfigFile::check_eor_mark(string) == true) {
     t2_err_ret(NULL, 0,
	   "ERROR: expecting local disk information at line %d in file '%s'",
	   file.getCurrentLineNumber(), file.getFileName());
     return false;
  }

  int atoi_err;
  phydid = t2_atoi(string, &atoi_err);
  if (atoi_err != 0) {
     t2_err_ret(NULL, 0, "ERROR: invalid number '%s' at line %d in file '%s'",
                string, file.getCurrentLineNumber(), file.getFileName());
     return false;
  }

  string = file.getNextUncommentString();
  while (string != NULL && T2_BEConfigFile::check_eor_mark(string) == false) {
     int mini, maxi;
     if (T2_IDRange::stringToRange((char *) string, mini, maxi) == false) {
	t2_err_ret(NULL, 0, "ERROR: invalid logical disk id range '%s' at line %d in file '%s'",
		  string, file.getCurrentLineNumber(), file.getFileName());
	clear();
	return false;
     }

     if (mini < 0) 				// end of record
	return true;

     for (int ri=mini; ri<=maxi; ri++)
        logdid << (u_int) ri;
     string = file.getNextUncommentString();
  }

  clear();
  if (string == NULL)
     t2_err_ret(NULL, 0,
		"ERROR: expecting local disk information but encounters eof in file '%s'",
		file.getFileName());
  else
     t2_err_ret(NULL, 0,
		"ERROR: expecting local disk information but encounters end-of-record at line %d in file '%s'",
		file.getCurrentLineNumber(), file.getFileName());
  return false;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_BEConfigEntry methods
//
///////////////////////////////////////////////////////////////////////////
u_int
T2_BEConfigEntry::read(T2_TxtFile& file)
{
  clear();
  if (!file)
     return 2;

  t2_Assert (file.getFileName() != NULL);
  const char* string;

  // Here is the format of an entry in the back-end configuration file.
  //	char[] hostname
  //	u_int logical_pid		-- logical processor id
  //	<disk_info_1>
  //	<disk_info_2>
  //		:
  //	===				-- end of entry mark
  // where <disk_info> can have one of the following formats:
  // 1) format for local disks
  //	l				-- a label for a local disk
  //	u_int physical_did		-- physical global disk id
  //	<logical_did_1>			-- logical global disk id
  //	<logical_did_2>			-- logical global disk id
  //		:
  //	-1				-- end of logical disk id list
  // where <logical_did> is an unsigned integer.
  //
  // 2) format for remote but locally accessible disks (ie cross-mounted)
  //	r				-- a label for a remote disk
  //	<physical_did_1_range>		-- physical global disk id range
  //	<physical_did_2_range>		-- physical global disk id range
  //		:
  //	-1				-- end of remote disk id list
  //
  // Note that a physical disk, which is associated with a unique global 
  // physical disk id, contains files that are assigned to the logical 
  // disk id's associated with its physical disk id. 

  // read dataset name
  while (1) {
     if ((string = file.getNextUncommentString()) == NULL)      // eof
	return 2;

     if (T2_BEConfigFile::check_eor_mark(string) == false)
	break;

     // so this is an end-of-record marker; skip it and read next string
  }
  str_alloc_cpy(hostname, string);

  const u_int lineno = file.getCurrentLineNumber();

  // get physical global disk id
  if ((string = file.getNextUncommentString()) == NULL) {
     clear();
     t2_err_ret(NULL, 0,
	"ERROR: expects physical global disk id but encounters eof in file '%s'",
	file.getFileName());
     return 0;
  }
  int atoi_err;
  int i = t2_atoi(string, &atoi_err);
  if (atoi_err != 0) {
     clear();
     t2_err_ret(NULL, 0,
	   "ERROR: invalid physical global disk id '%s' is invalid at line %d in file '%s'.",
	   string, file.getCurrentLineNumber(), file.getFileName());
     return 0;
  }
  if (i < 0) {
     clear();
     t2_err_ret(NULL, 0,
	    "ERROR: negative physical global disk id %d at line %d in file '%s' is not allowed.",
	    i, file.getCurrentLineNumber(), file.getFileName());
     return 0;
  }
  logpid = i;

  // get disk information
  while ((string = file.getNextUncommentString()) != NULL) {
     if (T2_BEConfigFile::check_eor_mark(string) == true)
	return 1;

     switch (string[0]) {
	case 'L':
	case 'l':	// local disk
		  {
			T2_BEConfigLocalDisk& lde = local_disks.appendElement();
			if (lde.read(file) == false) {	// error
			   clear();
			   return 0;
			}
		  }
		  break;

	case 'R':
	case 'r':	// remote disk id ranges
		  {
			u_int cnt = 0;
			while ((string = file.getNextUncommentString()) != NULL)
			{
			   int mini, maxi;
			   if (T2_IDRange::stringToRange((char *) string,
							 mini, maxi) == false)
			   {
			      t2_err_ret(NULL, 0,
					"ERROR: invalid physical disk id range '%s' at line %d in file '%s'.",
					string, file.getCurrentLineNumber(),
					file.getFileName());
			      clear();
			      return 0;
			   }
			   if (mini < 0)		// end of remote disks
				break;
			   for (int ri=mini; ri<=maxi; ri++)
			      remote_disks << (u_int) ri;
			   cnt += maxi - mini + 1;
			}
			if (cnt == 0) {
			   clear();
			   t2_err_ret(NULL, 0,
				"WARNING: no remote physical disk id given at line %d in file '%s'.",
				file.getCurrentLineNumber(),
				file.getFileName());
			}
		  }
		  break;

	default:	// invalid label
		  clear();
		  t2_err_ret(NULL, 0,
			"ERROR: invalid disk label '%c' at line %d in file '%s'",
			string[0], file.getCurrentLineNumber(),
			file.getFileName());
		  return 0;
     }			// switch (string[0])

  }			// while ((string=file.getNextUncommentString())!=NULL)

  t2_err_ret(NULL, 0,
	"WARNING: no end-of-record marker for the last entry at line %d in file '%s', assume one exists.",
	lineno, file.getFileName());
  return 1;
}


ostream&
T2_BEConfigEntry::write(ostream& os) const
{
  os << "hostname = " << t2_stringNullWrap(hostname)
     << "	" << logpid << endl;
  for (u_int i=0; i<local_disks.getNumberElements(); i++) {
     const T2_BEConfigLocalDisk& ld = local_disks[i];
     os << "   local disk physical id = " << ld.phydid
	<< " logical ids = ";
     for (u_int k=0; k<ld.logdid.getNumberElements(); k++)
	os << " " << ld.logdid[k];
     os << endl;
  }
  os << "   remote physical disk ids = ";
  for (u_int j=0; j<remote_disks.getNumberElements(); j++)
     os << " " << remote_disks[j];
  os << endl;

  return os;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_BEConfigFile methods
//
///////////////////////////////////////////////////////////////////////////
void
T2_BEConfigFile::open(const char* fname)
{
  close();
  if (fname == NULL)
     return;

  T2_TxtFile file('#', 512);
  file.open(fname, true);		// true makes file remember its filename
  if (!file) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: cannot open '%s'", fname);
     return;
  }

  T2_BEConfigEntry* newe = new T2_BEConfigEntry;
  t2_Assert (newe != NULL);
  do {
     switch (newe->read(file)) {
	case 0:		// error
		close();
		file.close();
		delete newe;
		return;

	case 1:		// succeed
		update_id(*newe);	// update max id's
		entries << newe;	// store entry
		newe = new T2_BEConfigEntry;	// get next entry
		break;
	case 2:				// empty entry, so reuse entry
		newe->clear();
		break;
     } 					// switch (newe.read(file))

  } while (file.eof() == false);

  file.close();
  delete newe;
  valid = true;
}


void
T2_BEConfigFile::close()
{
  min_invlogpid = 0;
  min_invphydid = 0;
  min_invlogdid = 0;

  for (u_int i=0; i<entries.getNumberElements(); i++)
     delete entries[i];
  entries.clear();

  valid = false;
}


void
T2_BEConfigFile::update_id(const T2_BEConfigEntry& e)
{
  if (e.getLogicalProcessorID() >= min_invlogpid)
     min_invlogpid = e.getLogicalProcessorID() + 1;

  const T2_VArray<T2_BEConfigLocalDisk>& local_disk = e.getLocalDiskEntries();
  for (u_int i=0; i<local_disk.getNumberElements(); i++) {
     if (local_disk[i].phydid >= min_invphydid)
	min_invphydid = local_disk[i].phydid + 1;
     const T2_VArray<u_int>& local_logdid = local_disk[i].logdid;
     for (u_int k=0; k<local_logdid.getNumberElements(); k++)
	if (local_logdid[k] >= min_invlogdid)
	   min_invlogdid = local_logdid[k] + 1;
  }

  const T2_VArray<u_int>& remote_disk = e.getRemotePhysicalDiskID();
  for (u_int j=0; j<remote_disk.getNumberElements(); j++)
     if (remote_disk[j] >= min_invphydid)
	min_invphydid = remote_disk[j] + 1;
}


bool
T2_BEConfigFile::check_content_correct() const
{
  if (valid == false) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: no back-end configuration file has been open");
     return false;
  }

  u_int i;
  bool r = true;

  // setup the bit maps first
  T2_BitVector  logpid(min_invlogpid),
		phydid(min_invphydid),
		logdid(min_invlogdid);
  logpid.clrAll();
  phydid.clrAll();
  logdid.clrAll();

  // remember which processor ids have the id's defined
  T2_Array<u_int> logpid_def(min_invlogpid),
		  phydid_def(min_invphydid),
		  logdid_def(min_invlogdid);

  for (i=0; i<entries.getNumberElements(); i++) {
     T2_BEConfigEntry* e = entries[i];
     if (e == NULL) {
	t2_err_quit(__FILE__, __LINE__, "ERROR: NULL pointer for entry[%d]", i);
	return false;
     }

     // make sure that a hostname is given
     if (e->getHostName() == NULL) {
	t2_err_ret(__FILE__, __LINE__,
		  "ERROR: no hostname is given for the %d-th entry", i);
	r = false;
     }

     t2_AssertOrDo (e->getLogicalProcessorID() < min_invlogpid,
		    cout << "e->pid = " << e->getLogicalProcessorID()
			 << ", min-invalid-pid = " << min_invlogpid << endl);

     // make sure that logical procesor id is unique
     if (logpid.isSet(e->getLogicalProcessorID()) == true) {
	t2_err_ret(__FILE__, __LINE__,
		   "ERROR: duplicate logical processor id %d used by host '%s' and '%s'",
		   e->getLogicalProcessorID(),
		   t2_stringNullWrap(entries[logpid_def[e->getLogicalProcessorID()]]->getHostName()),
		   t2_stringNullWrap(e->getHostName()));
	r = false;
     }
     else {
	logpid.set(e->getLogicalProcessorID());
	logpid_def[e->getLogicalProcessorID()] = i;
     }

     u_int d;

     const T2_VArray<T2_BEConfigLocalDisk>& ldisks = e->getLocalDiskEntries();
     for (d=0; d<ldisks.getNumberElements(); d++) {
	const T2_BEConfigLocalDisk& ld = ldisks[d];
	// make sure that physical disk id for local disks are unique
	if (phydid.isSet(ld.phydid) == true) {
	   t2_err_ret(__FILE__, __LINE__,
		      "ERROR: duplicate physical disk id %d used by host '%s' and '%s'",
		      ld.phydid,
		      t2_stringNullWrap(entries[phydid_def[ld.phydid]]->getHostName()),
		      t2_stringNullWrap(e->getHostName()));
	   r = false;
	}
	else {
	   phydid.set(ld.phydid);
	   phydid_def[ld.phydid] = i;
	}

	// make sure that logical disk ids for local disks are unique
	for (u_int j=0; j<ld.logdid.getNumberElements(); j++)
	   if (logdid.isSet(ld.logdid[j]) == true) {
	      t2_err_ret(__FILE__, __LINE__,
			"ERROR: duplicate logical disk id %d used by host '%s' and '%s'",
			ld.logdid[j],
			t2_stringNullWrap(entries[logdid_def[ld.logdid[j]]]->getHostName()),
			t2_stringNullWrap(e->getHostName()));
	      r = false;
	   }
	   else {
	      logdid.set(ld.logdid[j]);
	      logdid_def[ld.logdid[j]] = i;
	   }
     }
  }

  // now check if all used id's have been pre-defined
  for (i=0; i<entries.getNumberElements(); i++) {
     T2_BEConfigEntry* e = entries[i];
     if (e == NULL) 		// already checked before
	continue;

     const T2_VArray<u_int>& rdisks = e->getRemotePhysicalDiskID();
     for (u_int d=0; d<rdisks.getNumberElements(); d++) {
	if (logdid.isSet(rdisks[d]) == false) {
	   t2_err_ret(__FILE__, __LINE__,
		     "ERROR: logical disk id %d accessed by host '%s' is not defined",
	   rdisks[d], t2_stringNullWrap(e->getHostName()));
	   r = false;
	}
     }
  }

  // now check if all logical processor ids have been assumed by an entry
  for (i=0; i<min_invlogpid; i++) {
     if (logpid.isSet(i) == false) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: logical processor id %d is not assumed by any hostname",
		i);
	r = false;
     }
  }

  return r;
}


bool
T2_BEConfigFile::check_eor_mark(const char* s)
{
  if (s != NULL && s[0] == '=' && s[1] == '=' && s[2] == '=')
     return true;
  return false;
}


ostream&
T2_BEConfigFile::write(ostream& os) const
{
  if (valid == false)
     return os;

  os << "min invalid logical processor id = " << min_invlogpid << endl
     << "min invalid physical disk id = " << min_invphydid << endl
     << "min invalid logical disk id = " << min_invlogdid << endl;

  for (u_int i=0; i<entries.getNumberElements(); i++) {
     os << *entries[i];
  }

  return os;
}
