
/* $Id: dscat.C,v 1.3 2000/03/14 23:31:18 chialin Exp $ */

/* dscat.C */


#include	<limits.h>
#include	"t2_dscat.h"


///////////////////////////////////////////////////////////////////////////
//
// T2_DSCatalogHeader methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_DSCatalogHeader::writeHeader(fstream& file, T2_BEConfigFile& beconfig)
{
  t2_Assert (file);
  t2_Assert (beconfig);
  if (beconfig.check_content_correct() == false)
     return false;

  // the header has the following format:
  //	u_int max_number_logical_proc_ids
  //	u_int max_number_logical_disk_ids
  //	<host record 1>
  //	<host record 2>
  //		:
  // where <host record> has the following format:
  // 	char[] hostname
  //	u_int logical_proc_id
  //	u_int logical_disk_id_1
  //	u_int logical_disk_id_2
  //		:
  //	-1					-- end-of-record mark

  file << beconfig.getMinInvalidLogicalProcID() << "	"
       << beconfig.getMinInvalidLogicalDiskID()
       << endl;

  for (u_int i=0; i<beconfig.getNumberEntries(); i++) {
     const T2_BEConfigEntry* e = beconfig[i];
     t2_Assert (e != NULL);
     t2_Assert (e->getHostName() != NULL);
     file << e->getHostName() << "	"
	  << e->getLogicalProcessorID()
	  << "	";

     const T2_VArray<T2_BEConfigLocalDisk>& ldisks = e->getLocalDiskEntries();
     for (u_int k=0; k<ldisks.getNumberElements(); k++) {
	const T2_VArray<u_int>& logdids = ldisks[k].logdid;
	for (u_int j=0; j<logdids.getNumberElements(); j++)
	   file << logdids[j] << " ";
     }
     file << "-1" << endl;
  }

  return true;
}


bool
T2_DSCatalogHeader::readHeader(T2_TxtFile& file, const char* hostname,
			       const T2_BitVector* my_log_pids,
			       u_int& num_pids, T2_BitVector& diskmap)
{
  t2_Assert ((const void*) file != NULL);
  t2_Assert (file.getFileName() != NULL);

  // can't have both of hostname and my_log_pids set to non-null
  // (ie at least one has to be NULL)
  t2_Assert (hostname == NULL || my_log_pids == NULL);

  enum t2_dscathdr_match_t {
	t2_dscathdr_match_ignore = 0,		// just ignore this header
	t2_dscathdr_match_by_hostname,		// look for given hostname
	t2_dscathdr_match_by_logpid		// look for given logical pids
  };

  t2_dscathdr_match_t match_type;
  if (hostname != NULL)
     match_type = t2_dscathdr_match_by_hostname;
  else if (my_log_pids != NULL)
     match_type = t2_dscathdr_match_by_logpid;
  else
     match_type = t2_dscathdr_match_ignore;

  int atoi_err;
  u_int num_log_dids;
  const char *string;

  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: no number of logical processor ids is given in file '%s'",
	    file.getFileName());
     return false;
  }
  num_pids = t2_atoi(string, &atoi_err);
  if (atoi_err != 0) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: invalid number of logical processor ids '%s' is given in file '%s'",
	    string, file.getFileName());
     return false;
  }

  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: no number of logical disk ids is given in file '%s'",
	    file.getFileName());
     return false;
  }
  num_log_dids = t2_atoi(string, &atoi_err);
  if (atoi_err != 0) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: invalid number of logical disk ids '%s' is given in file '%s'",
	    string, file.getFileName());
     return false;
  }
  diskmap.setNumberBits(num_log_dids);
  diskmap.clrAll();

  // a bitvector to remember logical disk ids assigned to other processors
  // to check if any logical disk id is repeatedly assign to processors
  T2_BitVector others_diskmap(num_log_dids);
  others_diskmap.clrAll();

  for (u_int i=0; i<num_pids; i++) {
     if ((string =  file.getNextUncommentString()) == NULL) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting %d back-end nodes, but only saw %d of them in file %s.",
		num_pids, i);
	return false;
     }
     u_int lineno = file.getCurrentLineNumber();

     bool my_log_id = false;		// are these gonna be my disk ids?

     // check hostname
     if (match_type == t2_dscathdr_match_by_hostname &&
	 strcmp(hostname, string) == 0)
     {	// hostname match
	my_log_id = true;
     }

     // get logical processor id
     if ((string = file.getNextUncommentString()) == NULL) {
	t2_err_ret(__FILE__, __LINE__, 
		"ERROR: no logical processor id is assigned for the hostname at line %d in file '%s'.",
		lineno, file.getFileName());
	return false;
     }
     int log_pid = t2_atoi(string, &atoi_err);
     if (atoi_err != 0 || log_pid < 0) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid logical processor id '%s' at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), file.getFileName());
	return false;
     }

     // check logical pid
     if (match_type == t2_dscathdr_match_by_logpid && 
	 my_log_pids->isSet(log_pid) == true)
     {	// this is my assigned logical pid
	my_log_id = true;
     }

     // get logical disk ids
     while ((string = file.getNextUncommentString()) != NULL) {
	int log_did = t2_atoi(string, &atoi_err);
	if (atoi_err != 0) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid logical disk id '%s' at line %d in file '%s'.",
		 string, file.getCurrentLineNumber(), file.getFileName());
	   return false;
	}
	if (log_did < 0)			// end of record
	   break;

	if (diskmap.isSet(log_did)==true || others_diskmap.isSet(log_did)==true)
	{  // this disk already assigned
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: logical disk id %d at line %d in file '%s' has already appeared earlier; a logical disk id can only appear once in the dataset catalog file.",
		 log_did, file.getCurrentLineNumber(), file.getFileName());
	   return false;
	}
	if (my_log_id == true)
	   diskmap.set(log_did);
	else
	   others_diskmap.set(log_did);

     }		// while ((string = file.getNextUncommentString()) != NULL)
  }		// for (u_int i=0; i<num_pids; i++)

  return true;
}


bool
T2_DSCatalogHeader::readHeaderNumberIDs(T2_TxtFile& file,
					u_int& nprocids, u_int& ndiskids)
{
  t2_Assert ((const void*) file != NULL);
  t2_Assert (file.getFileName() != NULL);

  int atoi_err, k;
  const char *string;

  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: no number of logical processor ids is given in file '%s'",
	    file.getFileName());
     return false;
  }
  k = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || k < 0) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: invalid number of logical processor ids '%s' is given at line %d in file '%s'",
	    string, file.getCurrentLineNumber(), file.getFileName());
     return false;
  }
  nprocids = (u_int) k;

  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: no number of logical disk ids is given in file '%s'",
	    file.getFileName());
     return false;
  }
  k = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || k < 0) {
     t2_err_ret(__FILE__, __LINE__,
	   "ERROR: invalid number of logical disk ids '%s' is given at line %d in file '%s'",
	    string, file.getCurrentLineNumber(), file.getFileName());
     return false;
  }
  ndiskids = (u_int) k;
  return true;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_DSCatalogWriter methods
//
///////////////////////////////////////////////////////////////////////////
void
T2_DSCatalogWriter::open(const char* fname, T2_BEConfigFile& beconfig)
{
  close();
  if (fname == NULL || !beconfig)
     return;

  // check if file already exists
  ifstream ifs(fname, ios::in);
  if (ifs) {
     ifs.close();
     char bakfname[512];
     sprintf(bakfname, "%s.bak", fname);
     if (rename(fname, bakfname) != 0) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: file '%s' already exists; cannot rename to '%s'",
		fname, bakfname);
	return;
     }
     t2_say(cout, "WARNING: Rename " << fname << " to " << bakfname << endl);
  }

  file.open(fname, ios::in | ios::out);
  if (!file) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: cannot open file '%s'", fname);
     return;
  }

  // write the header into the catalog file
  if (T2_DSCatalogHeader::writeHeader(file, beconfig) == false) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: write error for dataset catalog header");
     close();
     return;
  }
  file << endl;

  // save first max_ndigits_ndatasets bytes to write total number of 
  // datasets later
  datasets_start = file.tellg();
  for (u_int i=0; i<max_ndigits_ndatasets; i++)
     file << " ";
  file << endl;

  // remember total number of processors and disks
  numnodes = beconfig.getMinInvalidLogicalProcID();
  numdisks = beconfig.getMinInvalidLogicalDiskID();

  valid = true;
}


void
T2_DSCatalogWriter::close()
{
  if (valid == true) {
     // go back to beginning of file to write number of datasets added
     file.clear();
     file.seekg(datasets_start, ios::beg);
     file << ndatasets;
     ndatasets = 0;
     valid = false;
     file.close();
  }
}


bool
T2_DSCatalogWriter::addEntry(T2_DSRegistrationEntry& e, T2_DSID ds_id,
			     u_int cons_id)
{
  if (!file)
     return false;

  t2_Assert (ds_id == ndatasets);
  ndatasets++;

  u_int i;

  file << ds_id << "	" << cons_id << endl
       << e.datafiles.getNumberElements() << endl;
  for (i=0; i<e.datafiles.getNumberElements(); i++) {
     t2_Assert (e.datafiles[i].name != NULL);
     file << e.datafiles[i].name << "	"
	  << e.datafiles[i].id
	  << endl;
  }
  file << e.auxfiles.getNumberElements() << endl;
  for (i=0; i<e.auxfiles.getNumberElements(); i++) {
     char*& p = e.auxfiles[i];
     t2_Assert (p != NULL);
     file << p << endl;
  }
  file << endl;
  return true;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_DSCatEntry methods
//
///////////////////////////////////////////////////////////////////////////
void
T2_DSCatEntry::clear()
{
  u_int i;
  for (i=0; i<fnames.getNumberElements(); i++) {
     delete [] fnames[i];
     fnames[i] = NULL;
  }
  for (i=0; i<mfnames.getNumberElements(); i++) {
     delete [] mfnames[i];
     mfnames[i] = NULL;
  }
  id = 0;
  constructor_id = 0;
  fnames.clear(0);
  diskno.clear(0);
  mfnames.clear(0);
}


bool
T2_DSCatEntry::read(T2_TxtFile& ifs, off_t* offset)
{
  t2_Assert ((const void*) ifs != NULL);
  t2_Assert (ifs.getFileName() != NULL);

  clear();
  if (ifs.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: T2_DSCatEntry sees unexpected eof before reading dataset id.");
     return false;
  }

  int i,				// loop index
      nds;				// number of data files
  const u_int maxNameLen = 256;

  // get dataset id using getNextUncommentString() so that we can get the offset
  const char* string = ifs.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: no dataset id is found in file '%s'.", ifs.getFileName());
     return false;
  }
  int atoi_err;
  i = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || i < 0) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: invalid dataset id at line %d in file '%s'.",
	ifs.getCurrentLineNumber(), ifs.getFileName());
     clear();
     return false;
  }
  id = (u_int) i;

  if (offset != NULL)
     *offset = ifs.getCurrentLineOffset() + (off_t) (string - ifs.getLine());

  // get constructor id
  ifs >> i;
  if (ifs.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: no constructor id is given in file '%s'.",
	ifs.getFileName());
     clear();
     return false;
  }
  if (i < 0) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: invalid constructor id %d at line %d in file '%s'.",
	i, ifs.getCurrentLineNumber(), ifs.getFileName());
     clear();
     return false;
  }
  constructor_id = i;

  // get number of data files
  ifs >> nds;
  if (ifs.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: no number of data files is given in file '%s'.",
	ifs.getFileName());
     clear();
     return false;
  }
  if (nds < 0) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: invalid number of data files %d is given at line %d in file '%s'.",
	nds, ifs.getCurrentLineNumber(), ifs.getFileName());
     clear();
     return false;
  }

  if ((u_int) nds != fnames.getNumberElements()) {
     for (i=0; (u_int) i<fnames.getNumberElements(); i++)
        delete [] fnames[i];
     fnames.clear(nds);
     for (i=0; i<nds; i++)
        t2_Alloc(fnames[i], maxNameLen, char);
     diskno.clear(nds);
  }
  for (i=0; i<nds; i++) {
     fnames[i][0] = '\0';
     ifs >> fnames[i] >> diskno[i];
  }
  if (ifs.eof()) {
     t2_err_quit(__FILE__, __LINE__,
	"ERROR: sees unexpected eof in file '%s' before reading all data files",
	ifs.getFileName());
  }

  // get number of aux files
  ifs >> nds;
  if (ifs.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: no number of aux files is given in file '%s'.",
	ifs.getFileName());
     clear();
     return false;
  }
  if (nds < 0) {
     t2_err_ret(__FILE__, __LINE__,
	"ERROR: invalid number of aux files %d is given at line %d in file '%s'.",
	nds, ifs.getCurrentLineNumber(), ifs.getFileName());
     clear();
     return false;
  }

  if ((u_int) nds != mfnames.getNumberElements()) {
     for(i=0; (u_int) i<mfnames.getNumberElements(); i++)
        delete [] mfnames[i];
     mfnames.clear(nds);
     for (i=0; i<nds; i++)
        t2_Alloc(mfnames[i], maxNameLen, char);
  }
  for (i=0; i<nds; i++) {
     mfnames[i][0] = '\0';
     ifs >> mfnames[i];
  }
  if (ifs.eof()) {
     t2_err_quit(__FILE__, __LINE__,
	"ERROR: sees unexpected eof in file '%s' before reading all aux files",
	ifs.getFileName());
  }

  return true;
}


ostream&
T2_DSCatEntry::write(ostream& os) const
{
  u_int i;
  os << "dataset id = " << id << "		"
     << "constructor id = " << constructor_id
     << endl
     << "datafiles:"
     << endl;
  for (i=0; i<fnames.getNumberElements(); i++)
     os << "	" << t2_stringNullWrap(fnames[i]) << "	" << diskno[i] << endl;
  os << "auxfiles:" << endl;
  for (i=0; i<mfnames.getNumberElements(); i++)
     os << "	" << t2_stringNullWrap(mfnames[i]) << endl;

  return os;
}



///////////////////////////////////////////////////////////////////////////
//
// T2_DSCatalog methods
//
///////////////////////////////////////////////////////////////////////////
void
T2_DSCatalog::openHostName(const char* fname, const char* hostname)
{
  close();
  if (fname == NULL || hostname == NULL)
     return;

  file.open(fname, true);		// remember file name
  if (!file) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open dataset catalog '%s'.", fname);
     return;
  }

  // now get all logical disk ids assigned in the header to the given hostname
  if (T2_DSCatalogHeader::readHeader(file, hostname, NULL,
				     min_invalid_log_pid, mydisks) == false
      || find_offsets() == false)
  {
     close();				// error
     return;
  }
}


void
T2_DSCatalog::openProcID(const char* fname, const T2_BitVector& my_log_pids)
{
  close();
  if (fname == NULL)
     return;

  file.open(fname, true);		// remember file name
  if (!file) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open dataset catalog '%s'.", fname);
     return;
  }

  // now get all logical disk ids assigned in the header to the given
  // logical processor ids
  if (T2_DSCatalogHeader::readHeader(file, NULL, &my_log_pids,
				     min_invalid_log_pid, mydisks) == false
      || find_offsets() == false)
  {
     close();				// error
     return;
  }
  if (min_invalid_log_pid != my_log_pids.getNumberBits()) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: the overriding logical processor id assignment vector has valid logical processor ids less than %d, but the dataset catalog file '%s' has in its header valid logical processor ids less than %d.",
		min_invalid_log_pid, fname, my_log_pids.getNumberBits());
     close();				// error
     return;
  }
}


void
T2_DSCatalog::openDiskID(const char* fname, const T2_BitVector& my_log_dids)
{
  close();
  if (fname == NULL)
     return;

  file.open(fname, true);		// remember file name
  if (!file) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open dataset catalog '%s'.", fname);
     return;
  }

  // just ignore the header and use the given logical disk id assignment 
  if (T2_DSCatalogHeader::readHeader(file, NULL, NULL,
				     min_invalid_log_pid, mydisks) == false
      || find_offsets() == false)
  {
     close();				// error
     return;
  }
  if (mydisks.getNumberBits() != my_log_dids.getNumberBits()) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: the overriding logical disk id assignment vector has valid logical disk ids less than %d, but the dataset catalog file '%s' has in its header valid logical disk ids less than %d.",
		my_log_dids.getNumberBits(),
		fname, mydisks.getNumberBits());
     close();				// error
     return;
  }
  mydisks = my_log_dids;
}


bool
T2_DSCatalog::find_offsets()
{
  t2_Assert ((const void *) file != NULL);
  t2_Assert (file.getFileName() != NULL);

  const char* string;

  // get number of datasets
  if ((string = file.getNextUncommentString()) == NULL) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot find number of datasets in file '%s'.",
		file.getFileName());
     return false;
  }
  int atoi_err;
  int n = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || n < 0) {			// error
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid number of datasets '%s' at line %d in file '%s'.",
		string, file.getCurrentLineNumber(), file.getFileName());
     return false;
  }
  
  u_int i;
  offsets.clear(n);				// initialize offsets
  for (i=0; i<(u_int)n; i++)
     offsets[i] = T2_DSCAT_NO_OFFSET;

  for (i=0; i<(u_int)n; i++) {
     T2_DSCatEntry entry;
     if (entry.read(file, &(offsets[i])) == false) 	// error
	return false;
  }

# ifdef T2_ASSERTION
     for (i=0; i<(u_int)n; i++) {
	if (offsets[i] == T2_DSCAT_NO_OFFSET)
	   t2_err_ret(__FILE__, __LINE__,
		"WARNING: no entry for dataset id %d is found in file '%s'.",
		i, file.getFileName());
     }
# endif

  return true;
}


bool
T2_DSCatalog::getDSCatalogEntry(u_int i, T2_DSCatEntry& e, T2_BitVector& dbits)
{
  if (!file || i >= getNumberDatasets() || offsets[i] == T2_DSCAT_NO_OFFSET)
     return false;

  file.seek(offsets[i]);
  if (e.read(file) == false) {		// error
     return false;
  }

  // set the bit vector dbits to mark the data files assigned to the local proc
  const u_int ndatafiles = e.getNumberDataFiles();
  dbits.setNumberBits(ndatafiles);
  dbits.clrAll();
  for (u_int j=0; j<ndatafiles; j++) {
     const u_int diskno = e.getDataFileDiskNo(j);
     if (mydisks.isSet(diskno) == true)
	dbits.set(j);
  }
  return true;
}


ostream&
T2_DSCatalog::write(ostream& os)
{
  // print out logical disk id assignment
  os << "my-logical-disk-id-map: ";
  for (u_int d=0; d<mydisks.getNumberBits(); d++) {
     if (mydisks.isSet(d) == true)
	os << " " << d;
  }
  os << " (out of " << mydisks.getNumberBits() << " disks)" << endl;

  u_int i;

  // print out offsets
  os << "offsets: ";
  for (i=0; i<offsets.getNumberElements(); i++)
     os << " " << offsets[i];
  os << endl;

  // print out all entries
  T2_DSCatEntry e;
  T2_BitVector mydatafiles;
  for (i=0; i<getNumberDatasets(); i++) {
     if (getDSCatalogEntry(i, e, mydatafiles) == false) {	// error
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot read dataset catalog entry for dataset id %d",
		i);
	continue;
     }
     os << e;
     os << "my data file ids =";
     for (u_int j=0; j<mydatafiles.getNumberBits(); j++)
	if (mydatafiles.isSet(j) == true)
	   os << " " << j;
     os << "   (out of " << mydatafiles.getNumberBits() << " files)" << endl;
  }

  return os;
}

