
/* $Id: idxcat.C,v 1.2 2000/01/25 20:54:43 chialin Exp $ */

/* idxcat.C */

#include	"t2_idxcat.h"


///////////////////////////////////////////////////////////////////////////
//
// T2_IndexCatEntry methods
//
///////////////////////////////////////////////////////////////////////////
bool
T2_IndexCatEntry::read(T2_TxtFile& ifs, off_t* offset)
{
  t2_Assert ((const void*) ifs != NULL);
  t2_Assert (ifs.getFileName() != NULL);

  clear();
  if (ifs.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
        "ERROR: T2_DSCatEntry sees unexpected eof before reading dataset id.");
     return false;
  }

  int i;
  int nfiles;				// number of index files
  const u_int maxNameLen = 256;

  // get index id using getNextUncommentString() so that we can get the offset
  const char* string = ifs.getNextUncommentString();
  if (string == NULL) {
     t2_err_ret(__FILE__, __LINE__,
        "ERROR: no index id is found in file '%s'.", ifs.getFileName());
     return false;
  }
  int atoi_err;
  i = t2_atoi(string, &atoi_err);
  if (atoi_err != 0 || i < 0) {
     t2_err_ret(__FILE__, __LINE__,
        "ERROR: invalid dataset id at line %d in file '%s'.",
        ifs.getCurrentLineNumber(), ifs.getFileName());
     clear();
     return false;
  }
  id = (u_int) i;

  if (offset != NULL)
     *offset = ifs.getCurrentLineOffset() + (off_t) (string - ifs.getLine());

  // get constructor id
  ifs >> i;
  if (ifs.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
        "ERROR: no constructor id is given in file '%s'.",
        ifs.getFileName());
     clear();
     return false;
  }
  if (i < 0) {
     t2_err_ret(__FILE__, __LINE__,
        "ERROR: invalid constructor id %d at line %d in file '%s'.",
        i, ifs.getCurrentLineNumber(), ifs.getFileName());
     clear();
     return false;
  }
  constructor_id = i;

  // get number of index files
  ifs >> nfiles;
  if (ifs.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
        "ERROR: no number of index files is given in file '%s'.",
        ifs.getFileName());
     clear();
     return false;
  }
  if (nfiles < 0) {
     t2_err_ret(__FILE__, __LINE__,
        "ERROR: invalid number of index files %d is given at line %d in file '%s'.",
        nfiles, ifs.getCurrentLineNumber(), ifs.getFileName());
     clear();
     return false;
  }
  idxfiles.clear(nfiles);

  for (i=0; i<nfiles; i++) {
     T2_NameIDRangePair& idr = idxfiles[i];

     // get index filename
     t2_Alloc(idr.name, maxNameLen, char);
     ifs >> idr.name;
     if (ifs.eof() == true) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting %d index files, but only saw %d in file '%s'",
		nfiles, i, ifs.getFileName());
	clear();
	return false;
     }

     // get number of ranges
     int nranges;
     ifs >> nranges;
     if (ifs.eof() == true) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting number of ranges for index %d but see eof in file '%s'.",
		i, ifs.getFileName());
	clear();
	return false;
     }
     if (nranges < 0) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid number of ranges %d for index %d but see eof in file '%s'.",
		nranges, i, ifs.getFileName());
	clear();
	return false;
     }
     for (int j=0; j<nranges; j++) {
	if ((string = ifs.getNextUncommentString()) == NULL) {
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: expecting %d logical dataset file id ranges but only saw %d before seeing eof in file '%s'.",
		nranges, j, ifs.getFileName());
	   clear();
	   return false;
	}

	T2_IDRange& rr = idr.ranges.appendElement();
	if (rr.stringToRange((char*)string) == false
	    || rr.id_low < 0 || rr.id_high < 0)
	{
	   t2_err_ret(__FILE__, __LINE__,
		"ERROR: invalid logical dataset file id '%s' at line %d in file '%s'.",
		string, ifs.getCurrentLineNumber(), ifs.getFileName());
	   clear();
	   return false;
	}
     }
     if (string == NULL) {		// end of file
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: unexpected eof while reading datafile ids in file '%s'.",
		ifs.getFileName());
	clear();
	return false;
     }

  }					// for (i=0; i<nfiles; i++) 
  return true;
}


u_int
T2_IndexCatEntry::support(const T2_BitVector& datafilemap,
			  T2_BitVector& idxfilemap) const
{
  const u_int ndatafiles = datafilemap.getNumberBits();
  idxfilemap.setNumberBits(getNumberFiles());
  idxfilemap.clrAll();
  u_int cnt = 0;
  for (u_int i=0; i<getNumberFiles(); i++) {
     bool support_file = false;
     const T2_NameIDRangePair& idr = idxfiles[i];
     for (u_int j=0; j<idr.ranges.getNumberElements(); j++) {
	const T2_IDRange& r = idr.ranges[j];
	t2_AssertOrDo ((u_int) r.id_high < ndatafiles,
			cout << "ndatafiles = " << ndatafiles
			     << ", but index file " << i
			     << " supports file id "
			     << r.id_low << "-" << r.id_high << endl);
	for (int k=r.id_low; k<=r.id_high; k++)
	   if (datafilemap.isSet(k) == true) {	// support !
	      support_file = true;
	      j = idr.ranges.getNumberElements();	// to terminate j-loop
	      break;
	   }
     }
     if (support_file == true) {
	idxfilemap.set(i);
	cnt++;
     }
  }

  return cnt;
}


ostream&
T2_IndexCatEntry::write(ostream& os) const
{
  os << "index id = " << id << "		"
     << "constructor id = " << constructor_id << endl;
  for (u_int i=0; i<getNumberFiles(); i++) {
     const T2_NameIDRangePair& ifile = idxfiles[i];
     os << "	" << t2_stringNullWrap(ifile.name) << "	";
     for (u_int j=0; j<ifile.ranges.getNumberElements(); j++) {
	const T2_IDRange& rr = ifile.ranges[j];
	if (rr.id_low == rr.id_high)
	   os << " " << rr.id_low;
	else
	   os << " " << rr.id_low << "-" << rr.id_high;
     }
      os << endl;
  }
  return os;
}


///////////////////////////////////////////////////////////////////////////
//
// T2_IndexCatalog methods
//
///////////////////////////////////////////////////////////////////////////
void
T2_IndexCatalog::open(const char* fname)
{
  close();
  if (fname == NULL)
     return;

  file.open(fname, true);
  if (!file) {
     t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot open index catalog '%s'.", fname);
     return;
  }

  int n = -1;
  file >> n;
  if (file.eof() == true) {
     t2_err_ret(__FILE__, __LINE__,
	  "ERROR: no number of indexes is given in the index catalog '%s'.",
	  fname);
     close();
     return;
  }
  if (n < 0) {
     t2_err_ret(__FILE__, __LINE__,
	  "ERROR: invalid number of indexes is given at line %d in file '%s'.",
	  file.getCurrentLineNumber(), fname);
     close();
     return;
  }

  int i;

  offsets.clear(n);
  for (i=0; i<n; i++)
     offsets[i] = T2_IDXCAT_NO_OFFSET;

  T2_IndexCatEntry e;
  for (i=0; i<n; i++) {
     if (e.read(file, & (offsets[i])) == false) {	// error
	close();
	return;
     }
  }
}


bool
T2_IndexCatalog::getIndexCatalogEntry(u_int i, T2_IndexCatEntry& e)
{
  if (!file || i >= getNumberIndexes() || offsets[i] == T2_IDXCAT_NO_OFFSET)
     return false;

  file.seek(offsets[i]);
  return e.read(file);
}


ostream&
T2_IndexCatalog::write(ostream& os)
{
  u_int i;
  os << "index offsets = " ;
  for (i=0; i<offsets.getNumberElements(); i++)
     os << " " << offsets[i];
  os << endl;

  T2_IndexCatEntry ie;
  for (i=0; i<offsets.getNumberElements(); i++) {
     if (getIndexCatalogEntry(i, ie) == false) {	// error
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: cannot read index catalog entry for index id %d",
		i);
	continue;
     }
     os << ie;
  }
  return os;
}


///////////////////////////////////////////////////////////////////////////
//
// T2_IndexCatalogWriter methods
//
///////////////////////////////////////////////////////////////////////////
void
T2_IndexCatalogWriter::open(const char* fname)
{
  close();
  if (fname == NULL)
     return;

  // check if file already exists
  ifstream ifs(fname, ios::in);
  if (ifs) {
     ifs.close();
     char bakfname[512];
     sprintf(bakfname, "%s.bak", fname);
     if (rename(fname, bakfname) != 0) {
	t2_err_ret(__FILE__, __LINE__,
		"ERROR: file '%s' already exists; cannot rename to '%s'",
		fname, bakfname);
	return;
     }
     t2_say(cout, "WARNING: Rename " << fname << " to " << bakfname << endl);
  }

  file.open(fname, ios::in | ios::out);
  if (!file) {
     t2_err_ret(__FILE__, __LINE__, "ERROR: cannot open file '%s'", fname);
     return;
  }

  // save first max_ndigits_nindexes bytes to write total number of 
  // datasets later
  for (u_int i=0; i<max_ndigits_nindexes; i++)
     file << " ";
  file << endl;

  valid = true;
}


void
T2_IndexCatalogWriter::close()
{
  if (valid == true) {
     // go back to beginning of file to write number of datasets added
     file.clear();
     file.seekg((streampos) 0, ios::beg);
     file << nindexes;
     nindexes = 0;
     valid = false;
     file.close();
  }
}


bool
T2_IndexCatalogWriter::addEntry(T2_DSRegistrationEntry& e,
				T2_Array<T2_IndexID>& idx_ids,
				T2_Array<u_int>& cons_id)
{
  if (!file)
     return false;

  T2_VArray<T2_IndexRegistrationEntry>& indexes = e.indexes;
  t2_Assert (idx_ids.getNumberElements() == indexes.getNumberElements());
  t2_Assert (cons_id.getNumberElements() == indexes.getNumberElements());

  for (u_int i=0; i<indexes.getNumberElements(); i++) {
     T2_IndexRegistrationEntry& idx_e = indexes[i];

     t2_Assert (idx_ids[i] == nindexes);
     nindexes++;

     file << idx_ids[i] << "	" << cons_id[i] << endl
	  << idx_e.idxfiles.getNumberElements() << endl;
     for (u_int j=0; j<idx_e.idxfiles.getNumberElements(); j++) {
	T2_NameIDRangePair& nir = idx_e.idxfiles[j];
	char*& p = nir.name;
	t2_Assert (p != NULL);
	file << p << "	" << nir.ranges.getNumberElements();
	for (u_int k=0; k<nir.ranges.getNumberElements(); k++) {
	   T2_IDRange& idr = nir.ranges[k];
	   if (idr.id_low == idr.id_high)
	      file << " " << idr.id_low;
	   else
	      file << " " << idr.id_low << "-" << idr.id_high;
	}
	file << endl;		// end of record
     }
  }

  return true;
}

